 /****************************************************************************
  *
  *  Copyright (c) 2001 by LSI Logic Corporation, all rights reserved.
  *
  *                         LSI Logic confidential
  *  The information contained herein is proprietary of LSI Logic Corporation.
  *
  *  This software is provided "as is" without any express or implied warranty.
  *  LSI Logic disclaims all representations and warranties with regard to
  *  this software, including all implied warranties of merchantability and
  *  fitness for a particular purpose. In no event shall LSI Logic be liable
  *  for any special, indirect or consequential damages or any damages
  *  whatsoever resulting from loss of use, data or profits, whether in an
  *  action of contract, negligence or other tortious act or omission,
  *  arising out of or in connection with the use or performance of this
  *  software.
  *
  *  File Description:
  *
  *  This is the first code run in the bootloader.  It does the following:
  *  (1) sets up the stack
  *  (2) clears the loader's zero-initialised (ZI) section
  *  (3) calls the bootloader_main() function
  *
  ****************************************************************************/

/* Where the bootloader builds its page tables up from... */
.set BOOTLOADER_PAGETABLES, 0x4000
/* ...and where the stack grow down from */
.set BOOTLOADER_STACK, BOOTLOADER_PAGETABLES
	

/* Definitions used in setting up the page tables */

/* Specify domain=0, write-through caching */
.set WRITETHRU, 0xc1a
/* Specify domain=0, write-back caching */
.set WRITEBACK, 0xc1e
/* Specify domain=0, write buffer, no caching */
.set NOCACHEWB, 0xc16
/* Specify domain=0, no caching, no buffering */
.set NOTCACHED, 0xc12

/* Control register bits */
.set D_CACHE_ENABLE, (1 << 2)
.set I_CACHE_ENABLE, (1 << 12)
.set MMU_ENABLE,(1 << 0)
.set ASYNC, (1 << 31)
.set NOTFASTBUS, (1 << 30)

/*
   Magic constants used in NoCachesNoMMU routine
   26 : bits 31-26 set the index of the D-cache entry being cleaned
   5  : bits 6-5 set the segment of the D-cache entry being cleaned
   0xfffffffa : clear D cache enable (bit 2) and MMU enable (bit 0)
   0xffffefff : clear I cache enable (bit 12) in control register
*/
.set DCACHE_INDEX_SHIFT, 26
.set DCACHE_SEG_SHIFT, 5
.set DISABLE_MMU_AND_DCACHE, 0xfffffffa
.set DISABLE_ICACHE, 0xffffefff 

.global BootloaderStart
.global SetupCachesAndMMU
.global NoCachesNoMMU
.global GetExecOffset

/* Define this to clock the core off FCLK.  This is a development
   test switch and must not be defined for production units
   as it can permanently damage your hardware.
*/
#undef SELECT_FCLK

/* BootloaderStart is the entry point and must be positioned at the
   start of all code and data because the bootloader load
   and execution addresses are mandated to be the same.
   This is enforced by the linker script.
*/
	
/*
  Make a note of where the bootloader is executing in FLASH - this is needed
  so that the kernel can be told where the callback functions can be found.
  This cannot be done at link time because the bootloader is relocatable
  and can be packed into the FLASH at any location.

  r3 = bootloader entry address.  Cannot save it here because the
  data region is cleared in the following code.
*/  

BootloaderStart:
/* r3 = address where bootloader started execution */
    sub   r3, pc, #8    
	  ldr		sp, =BOOTLOADER_STACK

/* Use linker generated symbols to get at the start and "just past the end"
   addresses of the zero-initialised section, then zero the intervening memory. */

    ldr   r0, =__bss_start__
    ldr   r1, =__bss_end__
    mov   r2, #0
ClearZiRegion:
    str   r2, [r0], #4
    cmp   r0, r1
    blo   ClearZiRegion

    ldr   r1, =BootloaderBaseAddress
    str   r3, [r1]

    ldr   r3, =__bootloader_start__
    ldr   r1, =BootloaderStartLocation
    str   r3, [r1]

    ldr   r2, =__bootloader_end__
    sub   r0, r2, r3
    ldr   r1, =BootloaderLength
    str   r0, [r1]

    b   BootloaderMain

/*****************************************************************************
 * Function Name  : NoCachesNoMMU
 * Arguments      : None
 * Returns        : None
 * Description    : Flushes caches then turns them off along with the MMU
 *
 * This function is called by C routines, hence it must preserve r4-r11 & r13
 * as mandated by the ARM calling convention.
 *****************************************************************************/

NoCachesNoMMU:
/* Clean D-cache.  Not necessary if the cache is write-back as
   intended, but included for safety so this function can be used
   in any situation.
      
   r0 = seg number, r1 = index number
*/
    mov   r0, #0
SegLoop:
    mov   r1, #0
IndexLoop:
    mov   r2, r1, lsl #DCACHE_INDEX_SHIFT
    orr   r2, r2, r0, lsl #DCACHE_SEG_SHIFT

/* Clean & invalidate D-cache entry */
    mcr   p15,0,r2,c7,c14,2
    add   r1, r1, #1
    cmp   r1, #64
    bne   IndexLoop
    add   r0, r0, #1
    cmp   r0, #4
    bne   SegLoop
      
/* Disable MMU & both caches */

    mov   r0, #0
    mrc   p15,0,r0,c1,c0,0
    and   r0, r0, #DISABLE_MMU_AND_DCACHE
    and   r0, r0, #DISABLE_ICACHE
    mcr   p15,0,r0,c1,c0,0

/* Ensure the instruction pipe is full of benign instructions */
/* when the MMU turns off. */
    nop
    nop
    nop

/* Invalidate caches */
    mov   r0, #0
    mcr   p15,0,r0,c7,c7,0

/* Invalidate I-TLB & D-TLB */
    mov   r0, #0
    mcr   p15,0,r0,c8,c7,0
    mov   pc, lr

/*****************************************************************************
 * Function Name  : SetupCachesAndMMU
 * Arguments      : None
 * Returns        : None
 * Description    : Creates linear physical:virtual mapping for the MMU and
 *                : turns on both I and D caches (this speeds up the linux
 *                : kernel decompression.  We are called from the bootrom
 *                : with the MMU & caches disabled.
 *
 * This function is called by C routines, hence it must preserve r4-r11 & r13
 * as mandated by the ARM calling convention.
 *****************************************************************************/

SetupCachesAndMMU:
/* Set page table base pointer */
    mov   r3, #BOOTLOADER_PAGETABLES
    mcr   p15,0,r3,c2,c0,0

/* Build the page tables */
    ldr   r0, =WRITEBACK
    mov   r2, #0
      
/* 0x00000000 - 0x7FFFFFFF is write back cacheable */
MmuTableLp1:
    mov   r1, r0
    orr   r1, r1, r2, lsl #20
    str   r1, [r3], #4
    add   r2, r2, #1
    cmp   r2, #2048
    bne   MmuTableLp1

/* 0x80000000 - 0xBFFFFFFF is non-cacheable */
    ldr   r0, =NOTCACHED
MmuTableLp2:
    mov   r1, r0
    orr   r1, r1, r2, lsl #20
    str   r1, [r3], #4
    add   r2, r2, #1
    cmp   r2, #3072
    bne   MmuTableLp2

/* 0xC0000000 - 0xFFFFFFFF is write back cacheable */
    ldr   r0, =WRITEBACK
MmuTableLp3:
    mov   r1, r0
    orr   r1, r1, r2, lsl #20
    str   r1, [r3], #4
    add   r2, r2, #1
    cmp   r2, #4096
    bne   MmuTableLp3

/* Set all domains to manager state (no permission faults generated) */
    mov   r0, #0xffffffff
    mcr   p15,0,r0,c3,c0,0

/* Enable the MMU, I-Cache and D-Cache.  Select async clocking mode */
    mrc   p15, 0, r0, c1,c0, 0
    orr   r0, r0, #(MMU_ENABLE | D_CACHE_ENABLE)
    orr   r0, r0, #(I_CACHE_ENABLE)
  
#ifdef SELECT_FCLK
/* Switch ARM into async clocking mode.  The assumption here is that
   we would only want to engage FCLK if it was faster than BCLK and
   which is the only restriction on using async mode.
   Synchronous mode would be more efficient but has restrictions
   on FCLK phase and frequency that cannot be guaranteed on current
   systems.

   ****************************************************************
   This is a development switch only and can permanently damage the
   hardware.  Do not define this switch for production units.
   ****************************************************************
*/
#warning "**WARNING** - Using FCLK core clock, this can permanently damage your system"
    orr   r0, r0, #(ASYNC | NOTFASTBUS)
#else
    bic   r0, r0, #(ASYNC | NOTFASTBUS)
#endif
    mcr   p15, 0, r0, c1,c0, 0
    
/* Stuff the instruction pipe full of benign instructions */
/* to excute after the MMU turns on... */
    nop
    nop
    nop
    mov   pc, lr

/*****************************************************************************
 * Function Name  : GetExecOffset
 * Arguments      : None
 * Returns        : Offset between where the code was linked to execute
 *                : and where it is actually executing (offset = actual - linked)
 * Description    : Works out the offset between where this code was linked
 *                : to run and where it is actually running (i.e. code is
 *                : linked to execute in the 0x800 region by the linker
 *                : script, but ends up executing out of FLASH in the
 *                : 0xec000000 region).  Used to fixup offsets into static
 *                : data tables staored in FLASH.
 *
 * This function is called by C routines, hence it must preserve r4-r11 & r13
 * as mandated by the ARM calling convention.
 *****************************************************************************/

GetExecOffset:
    /* r0 = current execution address */
    sub   r0, pc, #8
    /* r1 = linked address corresponding to r0's execution address */
    ldr   r1, =(GetExecOffset)
    sub   r0, r0, r1
    mov   pc, lr
