/****************************************************************************
 *
 *  Copyright (c) 2001 by LSI Logic Corporation, all rights reserved.
 *
 *                         LSI Logic confidential
 *  The information contained herein is proprietary of LSI Logic Corporation.
 *
 *  This software is provided "as is" without any express or implied warranty.
 *  LSI Logic disclaims all representations and warranties with regard to
 *  this software, including all implied warranties of merchantability and
 *  fitness for a particular purpose. In no event shall LSI Logic be liable
 *  for any special, indirect or consequential damages or any damages
 *  whatsoever resulting from loss of use, data or profits, whether in an
 *  action of contract, negligence or other tortious act or omission,
 *  arising out of or in connection with the use or performance of this
 *  software.
 *
 *  File Description:
 *
 *  This is the top level file for the Getafix bootloader.
 *
 ****************************************************************************/

#include <linux/linkage.h>

#include "bootloader.h"
#include "flash-specific.h"
#include "flash-definition.h"

/* The include path must contain the bootrom source directory */
#include "bootrom.h"
#include "crc32.c"

#define IS_RUNNING_IN_RAM (BootloaderBaseAddress < 0x80000000)

PUBLIC __u32 BootloaderBaseAddress;
PUBLIC __u32 BootloaderLength;
PUBLIC __u32 BootloaderStartLocation;
PUBLIC __u32 BootloaderStart;
PUBLIC asmlinkage __u32 GetExecOffset(void);
PUBLIC asmlinkage void NoCachesNoMMU(void);
PUBLIC asmlinkage void SetupCachesAndMMU(void);

PRIVATE __u32 CheckKernelStruct(struct KernelStructure *pAddr);
PRIVATE __u32 CopyKernelImage(struct KernelStructure *pKernelStruct);
PRIVATE void BuildKernelConfigurationTags(struct LinuxKernelConfigTags *pTags,
               struct KernelStructure *pAddr1, struct KernelStructure *pAddr2,
               struct MacStructure *pMacStruct, AuxMacStruct *pAuxMacs,
                                          unsigned long numAuxMacs);
PRIVATE void ExitBootloader(__u32 ErrorCode);
PRIVATE int GetMAC(unsigned long FLAddr, struct MacStructure *pMacStruct,
                   AuxMacStruct *pAux, unsigned long *pNum);
PRIVATE void EmitChar(char c);
PRIVATE void RelocSafePuts(char *s);
PRIVATE void Puts(char *s);
PRIVATE void EmitHexWord(unsigned long x);

#define STR(x) #x
#define XSTR(y) STR(y)

/*****************************************************************************
 *          BOOTLOADER VERSION NUMBER - UPDATE THIS WHEN CHANGING CODE       *
 * V1.0: Initial release with new style tags.                                *
 * V1.1: V1.0, but changed to compile under gcc rather than the ARM tools.   *
 * V1.2: V1.1, but added support for hw code and auxillary macs.             *
 * V1.3: V1.2, but with the ability to run from RAM for bootp images         *
 * V1.4: Adds ability to pick up a MAC placed after a mode 0 program         *
 * V1.5: Passes new PHY ID field in MAC structure up to linux kernel.        *
 *****************************************************************************/
#define BOOTLOADER_MAJOR 1
#define BOOTLOADER_MINOR 5

/* Controls where the bootloader searches for a MAC address that follows a
   mode 0 program.
*/
#define MODE0_START (0x420)
#define MODE0_STRIDE (0x400)
#define MODE0_END (0x20020)

/*****************************************************************************
 * Bootloader static data must be treated specially - see PtrFixup()
 * declaration for the reasons why.
 *****************************************************************************/

/* Callback functions can only rely on static constant data as they are
  called from the Linux driver (hence only data in FLASH is still available) */

PUBLIC const struct bootloader_callbacks CallBacks =
{
  0,
  {0xff, 0xff, 0xff},
  ReadFlash32,
  FlashLin2Virt,
};

PRIVATE const struct MacAddress DefaultMac = {DEFAULT_BE_MAC_ADDRESS_STRING};

/* Offset to add from the start of a low MAC address structure to get to the
   auxillary MACs.
*/
#define LO_MAC_TO_AUXMAC_OFFSET (64)

/* Offset to subtract from the start of a high MAC address structure to get to the
   auxillary MACs.
*/
#define HI_MAC_TO_AUXMAC_OFFSET (32 + sizeof(AuxMacStruct))

PRIVATE const unsigned long MacLocations[] = {
        0x00000020,	/* Primary location */
        0x000FFFC0,	/* Secondary location (1MB - 64) */
        0x001FFFC0,	/* Tertiary location (2MB - 64) */
        0x003FFFC0,	/* Quaternary location (4MB - 64) */
        0x007FFFC0,	/* Quinary location (8MB - 64) */
        0x00FFFFC0,	/* Senary location (16MB - 64) */
        0x01FFFFC0,	/* Septenary location (32MB - 64) */
        0x03FFFFC0	/* Octonary location (64MB - 64) */
  };

/************************ end of static data ********************************/

/* Pointer to Linux configuration tags in main memory - these tags are
   built by the bootloader. */
PRIVATE struct LinuxKernelConfigTags *pKernelTags;
  
/****************************************************************************
 * Function Name  : BootloaderMain
 * Arguments      : None.
 * Returns        : Never returns to caller
 * Description    : This function does not return to the caller in the normal
 *                  sense.  If the FLASH boot fails then the program jumps
 *                  directly to a re-entry vector stored at a well-known
 *                  location in the bootrom.  This restarts the bootrom as if
 *                  a reset had occurred, but with a note not to retry the
 *                  FLASH boot.
 *                  A result of this scheme is that the bootloader does not
 *                  need to preserve any of the bootrom's variables or stack
 *                  and hence can define its own with impunity.
 ****************************************************************************/
PUBLIC void BootloaderMain(void)
{
  struct KernelStructure *pImage;
  struct KernelStructure *pImage1;
  struct KernelStructure *pImage2;
  
  __u32 CheckAddr;
  int AddrOK;
  int i;
  struct MacStructure MacStruct;
  unsigned long MacLocation;
  char Version[6];
  AuxMacStruct auxMacs[OBELIX_MAX_AUXMACS];
  unsigned long numAuxMacs;
  int found;

  RelocSafePuts("Bootloader V");
  Version[0] = BOOTLOADER_MAJOR + '0';
  Version[1] = '.';
  Version[2] = BOOTLOADER_MINOR + '0';
  Version[3] = '\r';
  Version[4] = '\n';
  Version[5] = '\0';
  Puts(Version);

  pKernelTags = (struct LinuxKernelConfigTags *)(LINUX_CONFIG_PARAMS_BASE);
  
  SetupCachesAndMMU();          /* Engage MMU and caches to speed up download */
  setCRC32coefficientArray();   /* Calculate array needed for all CRC functions */  
  ConfigureEBB();               /* Setup EBB with regard to whatever FLASHs are used */
  
  /* ========================================================================= */
  /* =                        LOCATE MAC STRUCTURE                           = */
  /* ========================================================================= */

  /* 
     Copy the default MAC address to Mac variable (see PtrFixup for why this
     isn't as straighforward as you might expect)
  */
  for (i = 0; i < sizeof(DefaultMac); i++)
  {
      *(((char *)&(MacStruct.Mac)) + i) = *(((char *)PtrFixup(&DefaultMac)) + i);
  }

  /*
     We need to pass the MAC address as part of the kernel configuration data
     so scan for it now.  The MAC structure only occurs on certain predefined
     FLASH linear addresses.
   */
  found = 0;
  for (i = 0; i < sizeof(MacLocations)/sizeof(unsigned long); i++)
  {
    /* See PtrFixup for why this isn't simply MacLocations[i] */
    MacLocation = *((unsigned long *)PtrFixup(MacLocations) + i);
    numAuxMacs = 0;
    if (GetMAC(MacLocation, &MacStruct, &auxMacs[0], &numAuxMacs))
    {
	  found = (0 == 0); /* i.e. TRUE */
      break;
    }
  }

  if (!found)
  {
	  /*
		For the case where the MAC can follow an arbitrary mode 0 program, we check
		a range of low addresses.
	  */
	  for (MacLocation = MODE0_START; MacLocation <= MODE0_END; MacLocation += MODE0_STRIDE)
	  {
		  if (GetMAC(MacLocation, &MacStruct, &auxMacs[0], &numAuxMacs))
		  {
			  break;
		  }
	  }
  }

  /* ========================================================================= */
  /* =                        LOCATE KERNEL IMAGES                           = */
  /* ========================================================================= */

  /* Locate up to two images & work out which one to boot */
  pImage = (struct KernelStructure *)-1;
  pImage1 = (struct KernelStructure *)-1;
  pImage2 = (struct KernelStructure *)-1;

  if (IS_RUNNING_IN_RAM)
  {
      /* We are running as part of a bootp image, so the kernel has already
         been downloaded into RAM.
      */
      BuildKernelConfigurationTags(pKernelTags, pImage1, (struct KernelStructure *)-1,
                                   &MacStruct, &auxMacs[0], numAuxMacs);
      JUMP_TO_IMAGE;
  }
  else
  {
      for (CheckAddr = 0; CheckAddr < MaxFlashLinearAddress(); CheckAddr += KERNEL_STRUCT_SEARCH_STEP)
      {
          if (CheckKernelStruct((struct KernelStructure *)CheckAddr) != 0)
          {
              if (pImage1 == (struct KernelStructure *)-1)
              {
                  pImage1 = (struct KernelStructure *)FlashLin2Virt(CheckAddr, &AddrOK);
                  if (!AddrOK)
                  {
                      pImage1 = (struct KernelStructure *)-1;
                  }
                  else
                  {
                      RelocSafePuts("K1");
                  }
              }
              else
              {
                  /* Found two images, so stop looking for more */
                  pImage2 = (struct KernelStructure *)FlashLin2Virt(CheckAddr, &AddrOK);
                  if (!AddrOK)
                  {
                      pImage2 = (struct KernelStructure *)-1;
                  }
                  else
                  {
                      RelocSafePuts("K2");
                      /* Two images present - use the date code to work out which is the newest */
                      /* Set pImage1 as the default image to boot and then check if this was corect */
                      pImage = pImage1;
                      /* pImage2 is newer if its date code is higher than pImage1, and also if
                         the pImage2 code is 0 and the pImage code is 2 (the case where the code
                         has wrapped.
                      */
                      if ((pImage2->DateCode == (pImage1->DateCode + 1)) || 
                          ((pImage2->DateCode == 0) && (pImage1->DateCode == 2)) )
                      {
                          pImage = pImage2;
                          RelocSafePuts("(K2)");
                      }
                      else
                      {
                          RelocSafePuts("(K1)");
                      }
                      
                      /* It is a deliberate choice to boot pImage1 if both images have the
                         same date code.  This situation shouldn't happen, but given that
                         alledgedly we have two good images, we take a policy decision to boot
                         one of them.  This will at least bring up the system to a state where
                         the user can do something about the problem (like do a new update),
                         rather than failing the FLASH boot.
                      */
                      
                      /* Copy the first choice image and if the copy succeeds, run it */
                      if (CopyKernelImage(pImage) != 0)
                      {
                          BuildKernelConfigurationTags(pKernelTags, 
                                                       pImage, (pImage == pImage1) ? pImage2 : pImage1,
                                                       &MacStruct, &auxMacs[0], numAuxMacs);
                          RelocSafePuts("-OK\r\n");
                          JUMP_TO_IMAGE;
                      }
                      
                      /* The first choice image that looked OK has failed to copy.  Remember where
                         the second choice image is, then go back and look for any further images
                         before retaking the decision of which one to boot */
                      RelocSafePuts("(FAILED)");
                      if (pImage == pImage1)
                      {
                          pImage1 = pImage2;
                      }
                      pImage2 = (struct KernelStructure *)-1;
                  }
              }
          }
      }

      /* We only get here if there less than two valid images - if there are two then we decide
         which to use in the above loop and jump directly to it from there. */
      
      if (pImage1 == (struct KernelStructure *)-1)
      {
          /* No valid kernel images to boot */
          RelocSafePuts("-No Image\r\n");
          ExitBootloader(BOOTLOADER_NOIMAGES);
      }
      
      /* Exactly one image in FLASH - so boot it */
      RelocSafePuts("(K1)");
      if (CopyKernelImage(pImage1) != 0)
      {
          BuildKernelConfigurationTags(pKernelTags, pImage1, (struct KernelStructure *)-1,
                                       &MacStruct, &auxMacs[0], numAuxMacs);
          RelocSafePuts("-OK\r\n");
          JUMP_TO_IMAGE;
      }
      
      /*
        We are out of options.  No image described in the FLASH Linux kernel structures
        copies to main memory without ending up with a bad CRC (probably because the main
        DRAM is broken in some way).  The FLASH boot has failed so we exit the bootloader
        and jump back to the bootrom so that it can try something else.
      */
      RelocSafePuts("-Bad copy\r\n");
      ExitBootloader(BOOTLOADER_CANNOTCOPY);    
  }
  /* Execution never gets here */
}

/****************************************************************************
 * Function Name  : CheckKernelStruct
 * Arguments      : FLASH linear address location to check.
 * Returns        : Non-zero if address contains a valid kernel structure,
 *                  0 otherwise.
 * Description    : Finds valid Linux kernel structures in the FLASH.
 *                  Checks the structure's CRC and does a sanity check on the
 *                  values but does NOT check the CRC on the kernel
 *                  image pointed to by the structure.
 ****************************************************************************/
PRIVATE __u32 CheckKernelStruct(struct KernelStructure *pAddr)
{
  __u32 FsLinearAddr;
  __u8 Result;
  struct KernelStructure *pVirtAddr;
  int AddrOK;
  
  pVirtAddr = (struct KernelStructure *)FlashLin2Virt((__u32)pAddr, &AddrOK);
  /* Check AddrOK before dereferencing the possibly invalid pVirtAddr */
  if (!AddrOK)
  {
      return 0;
  }
  if (pVirtAddr->Telltale != KERNEL_TELLTALE)
  {
      return 0;
  }
  RelocSafePuts("(K");
  EmitHexWord((unsigned long)pVirtAddr);
  if ((pVirtAddr->KernelLength + pVirtAddr->FilingSystemLength) > pVirtAddr->MaximumLength)
  {
      RelocSafePuts("xL)");
      return 0;
  }
  if (pVirtAddr->DateCode > 2)
  {
      RelocSafePuts("xD)");
      return 0;
  }
  Result = checkCRC32edBlock((__u8 *)pVirtAddr, sizeof(struct KernelStructure));
  if (Result != CRC_RESULT_PASS)
  {
    RelocSafePuts("xCRC)");
    return 0;
  }
  
  /* Check the associated filing system that Linux will decompress
     directly out of the FLASH
  */
  FsLinearAddr = (__u32)pAddr + sizeof(struct KernelStructure) + pVirtAddr->KernelLength;
  Result = checkFlashCRC32edBlock(FsLinearAddr, pVirtAddr->FilingSystemLength);
  if (Result != CRC_RESULT_PASS)
  {
    RelocSafePuts("xFS)");
    return 0;
  }
  RelocSafePuts("-OK)");
  return 1;
}

/****************************************************************************
 * Function Name  : CopyKernelImage
 * Arguments      : Pointer to ARM virtual address of kernel structure
 * Returns        : Non-zero if the kernel copy succeeded, zero otherwise.
 * Description    : Copies a kernel image from FLASH into main memory based
 *                  on the contents of the kernel structure passed.
 ****************************************************************************/
PRIVATE __u32 CopyKernelImage(struct KernelStructure *pVirtKernelStruct)
{
  __u8 Result;
  __u32 KernelStartLinear;
  int AddressOK;

  /* Kernel starts immediately after the kernel structure */
  KernelStartLinear = FlashVirt2Lin(
                (__u32)(pVirtKernelStruct) + sizeof(struct KernelStructure), 
                &AddressOK);
                
  if (!AddressOK)
  {
    /* The virtual address passed is not a valid FLASH address */
    return 0;
  }
  CopyFlashLin2Virt(KernelStartLinear, 
                    pVirtKernelStruct->KernelLength, 
                    LINUX_KERNEL_EXECUTION_ADDRESS);
  Result = checkCRC32edBlock((__u8 *)LINUX_KERNEL_EXECUTION_ADDRESS,
                              pVirtKernelStruct->KernelLength);
  return (Result == CRC_RESULT_PASS) ? 1 : 0;
}

/****************************************************************************
 * Function Name  : BuildKernelConfigurationTags
 * Arguments      : Pointer to ARM virtual address of kernel structure,
 *                : pointer to kernel structure of kernel that is to be
 *                : booted, pointer to backup kernel structure (or -1 if
 *                : none), pointer to first r/w segment. 
 * Returns        : None.
 * Description    : Copies across parameters needed by the Linux device
 *                  drivers into the special parameter block.
 ****************************************************************************/
PRIVATE void BuildKernelConfigurationTags(struct LinuxKernelConfigTags *pTags,
               struct KernelStructure *pAddr1, struct KernelStructure *pAddr2,
               struct MacStructure *pMacStruct, AuxMacStruct *pAuxMacs,
               unsigned long numAuxMacs)
{
  /* This builds a list special structures that is read by the kernel during
     startup.  The structures contain all the configuration parameters
     (but these can be overridden by specifying a command line during kernel
     compile time).
     
     Not all of the tags are fully populated by this routine.  Some info
     is determined after scanning the FLASH chips in the system
  */
  const int sizeof_tag_header = sizeof(struct tag_header);
  __u32 KernelStartLinear;
  int AddressOK;
  unsigned long *pData;
  unsigned long *pStart;
  int i;

  if (IS_RUNNING_IN_RAM)
  {
      KernelStartLinear = 0;
  }
  else
  {
      /* Kernel starts immediately after the kernel structure */
      KernelStartLinear = FlashVirt2Lin((__u32)(pAddr1) + sizeof(struct KernelStructure), 
                                        &AddressOK);
      if (!AddressOK)
      {
          /* Bad address passed - can't do anything except give up */
          return;
      }
  }

  /* pKTags points to the well-known memory location where the kernel
     will pick up the parameters.  Normally this is 0x100 but is
     configurable in the kernel build
  */
  
  /* The first tag MUST be ATAG_CORE */
  pTags->CoreHeader.size = ((sizeof_tag_header + sizeof(struct tag_core)) >> 2);
  pTags->CoreHeader.tag = ATAG_CORE;
  pTags->Core.flags = 0; /* read-write not read-only */
  pTags->Core.pagesize = 0; /* unused in current kernel */
  pTags->Core.rootdev = 0x2500; /* boot device = flashfloppy driver */
  
  /* Note - tell Linux it has less memory than on the system.  This frees
     space for the ABM buffers at the top of memory. */
  pTags->MemHeader.size = ((sizeof_tag_header + sizeof(struct tag_mem32)) >> 2);
  pTags->MemHeader.tag = ATAG_MEM;
  pTags->Mem.size = (SYSTEM_MEMORY_SIZE - ABM_BUFFER_SIZE_IN_MB) * 1024 * 1024;
  pTags->Mem.start = 0; /* Physical start address of memory */
  
  pTags->RamdiskHeader.size = ((sizeof_tag_header + sizeof(struct tag_ramdisk)) >> 2);
  pTags->RamdiskHeader.tag = ATAG_RAMDISK;
  pTags->Ramdisk.flags = 2; /* Load ramdisk, don't prompt */
  pTags->Ramdisk.size = BOOT_RAMDISK_SIZE;
  pTags->Ramdisk.start = BOOT_RAMDISK_START;

  /******************** BUILD V1 TAG_OBELIX STRUCTURE *****************/

  /* Obelix driver specific information */
  pTags->ObelixHeader.size = ((sizeof_tag_header + sizeof(struct tag_obelix)) >> 2);
  pTags->ObelixHeader.tag = ATAG_OBELIX;

  if (IS_RUNNING_IN_RAM)
  {
      /* The filing system has already been loaded into memory as part of the
         bootp image.  Ensure that bit 0 = 0 which indicates that the address
         passed is a memory address and not a FLASH linear address.
      */
      pTags->Obelix.ffloppy.base_address = (ARMFS_BOOTP_ADDR & ~1);
  }
  else
  {
      /* The filing system always appears directly after the kernel.  Note that
         bit 0 is set in the address to indicate to the Linux flashfloppy driver
         that the address is a FLASH linear address and not a physical address
         (physical addresses are used in development) */
      pTags->Obelix.ffloppy.base_address = ((KernelStartLinear + pAddr1->KernelLength) | 1);
  }
  pTags->Obelix.ffloppy.flashsize_in_kbytes = (FLASH_TOTAL_SIZE / 1024);
  /* Don't override the UART clock (useful only on pre-production FPGA boards) */
  pTags->Obelix.uart_clock = 0;
  
  pTags->Obelix.mac.First = pMacStruct->Mac.First;
  pTags->Obelix.mac.Second = pMacStruct->Mac.Second;
  pTags->Obelix.mac.Third = pMacStruct->Mac.Third;
  pTags->Obelix.mac.Fourth = pMacStruct->Mac.Fourth;
  pTags->Obelix.mac.Fifth = pMacStruct->Mac.Fifth;
  pTags->Obelix.mac.Sixth = pMacStruct->Mac.Sixth;
  
  /* Pointer to the callbacks that the bootloader provides to the Linux kernel
    so that it can handle non-contiguous copies from FLASH.  The address has to be
    correct independent of both where the image has been linked to execute and
    where it has been placed in FLASH. */
  pTags->Obelix.BootloaderStart = BootloaderBaseAddress;
  pTags->Obelix.BootloaderOffset = BootloaderStartLocation;
  pTags->Obelix.BootloaderLength = BootloaderLength;
  /* Note that the callback structure starts with a version code so the
     reading code can recast this pointer to the correct revision 
  */
  pTags->Obelix.pcallbacks = (struct bootloader_callbacks *)PtrFixup(&CallBacks);
  
  /* Note : The finfo member was never read by kernels before the switch was made
     to the variable length tag_obelix_flashdata tag.  Hence the structure is
     obsolete and is not filled by V1.0+ bootloaders.
   */

  /******************** BUILD V4 TAG_OBELIX STRUCTURE *****************/
  /* Note that the V2 header is compatible with V3/V4 tags */
  pTags->Obelix2Header.size = ((sizeof_tag_header + sizeof(struct tag_obelix4)) >> 2);
  pTags->Obelix2Header.tag = ATAG_OBELIX2;
  pTags->Obelix4.tag_version = 4;
  pTags->Obelix4.bootloader_major = BOOTLOADER_MAJOR;
  pTags->Obelix4.bootloader_minor = BOOTLOADER_MINOR;
  pTags->Obelix4.hw_code = pMacStruct->hwCode;

  pTags->Obelix4.num_auxmacs = numAuxMacs;

  for (i = 0; i < OBELIX_MAX_AUXMACS; i++)
  {
      if (i < numAuxMacs)
      {
          pTags->Obelix4.auxmacs[i].tellTale = (__u32)(-1);
          pTags->Obelix4.auxmacs[i].first = pAuxMacs->first;
          pTags->Obelix4.auxmacs[i].second = pAuxMacs->second;
          pTags->Obelix4.auxmacs[i].third = pAuxMacs->third;
          pTags->Obelix4.auxmacs[i].fourth = pAuxMacs->fourth;
          pTags->Obelix4.auxmacs[i].fifth = pAuxMacs->fifth;
          pTags->Obelix4.auxmacs[i].sixth = pAuxMacs->sixth;
          pTags->Obelix4.auxmacs[i].index = pAuxMacs->index;
          pAuxMacs++;
      }
      else
      {
          pTags->Obelix4.auxmacs[i].tellTale = 0;
      }
  }

  pTags->Obelix4.phyid0 = pMacStruct->Mac.phyid0;
  
  /******************** BUILD TAG_OBELIX_FLASHDATA STRUCTURE *****************/

  pTags->FlashHeader.tag = ATAG_OBELIX_FLASHDATA;
  
  /* pData is used to write the variable length data tag - we fill in the 
     total size of it when we are finished.
  */
  pData = (unsigned long *)(&(pTags->FlashData));

  /***** Build basic flash data sub-tag *****/
  /* Come back to the tag length */
  pStart = pData++;
  /* ID */
  *pData++ = OBELIX_FLASH_BASICDATA;
  /* Tag data */
#if FLASH_0_PRESENT == 1
  *pData++ = 0;
  *pData++ = FLASH_0_SPEED_IN_NS;
  *pData++ = FLASH_0_WIDTH;
  *pData++ = FLASH_0_SIZE_IN_MBITS;
  *pData++ = FLASH0_LINEAR_OFFSET;
  *pData++ = CS0_VIRTUAL_ADDR;
#endif
#if FLASH_1_PRESENT == 1
  *pData++ = 1;
  *pData++ = FLASH_1_SPEED_IN_NS;
  *pData++ = FLASH_1_WIDTH;
  *pData++ = FLASH_1_SIZE_IN_MBITS;
  *pData++ = FLASH1_LINEAR_OFFSET;
  *pData++ = CS1_VIRTUAL_ADDR;
#endif
#if FLASH_2_PRESENT == 1
  *pData++ = 2;
  *pData++ = FLASH_2_SPEED_IN_NS;
  *pData++ = FLASH_2_WIDTH;
  *pData++ = FLASH_2_SIZE_IN_MBITS;
  *pData++ = FLASH2_LINEAR_OFFSET;
  *pData++ = CS2_VIRTUAL_ADDR;
#endif
#if FLASH_3_PRESENT == 1
  *pData++ = 3;
  *pData++ = FLASH_3_SPEED_IN_NS;
  *pData++ = FLASH_3_WIDTH;
  *pData++ = FLASH_3_SIZE_IN_MBITS;
  *pData++ = FLASH3_LINEAR_OFFSET;
  *pData++ = CS3_VIRTUAL_ADDR;
#endif
  /* Write the calculated sub-tag length back */
  *pStart = (unsigned long)(pData - pStart);

  /***** Build kernel data sub-tag *****/
  /* Come back to the tag length */
  pStart = pData++;
  /* ID */
  *pData++ = OBELIX_FLASH_KERNELDATA;
  /* Tag data */
  *pData++ = (unsigned long)pAddr1;
  *pData++ = KernelStartLinear;
  *pData++ = (unsigned long)pAddr2;
  if ((__u32)pAddr2 != (__u32)-1)
  {
      *pData++ = FlashVirt2Lin(((__u32)pAddr2 + sizeof(struct KernelStructure)), &AddressOK);
  }
  else
  {
      *pData++ = 0;
  }
  /* Write the calculated sub-tag length back */
  *pStart = (unsigned long)(pData - pStart);

  /* Now we know how long the sub-tags are, we can calculate how long
     the total tag is
  */
  pTags->FlashHeader.size = (unsigned long)
            (pData - (unsigned long *)(&(pTags->FlashHeader.size)));

  /* End of sub-tags */
  *pData++ = 0;
  *pData++ = OBELIX_FLASH_END;
}

/****************************************************************************
 * Function Name  : ExitBootloader
 * Arguments      : Return code passed back to bootrom.
 * Returns        : Never returns (re-enters bootrom)
 * Description    : Turns off MMU and caches, then jumps back into
 *                  drivers into the special parameter block.
 ****************************************************************************/
PRIVATE void ExitBootloader(__u32 ErrorCode)
{
  /*
    The MMU and caches must be disabled before returning to the bootrom.
  */
  NoCachesNoMMU();
  EXIT_BOOTLOADER(ErrorCode);
}

/****************************************************************************
 * Function Name  : GetMAC
 * Arguments      : FLASH linear address to check for valid MAC definition
 *                : structure.  Pointer to space reserved for auxillary MACs.
 * Returns        : Non-zero and MAC address in structure supplied or 0
 *                : if structure invalid.  Any auxillary MAC structures
 *                : found in the FLASH.
 * Description    : Searchs a single location to see if there is a MAC
 *                : address structure present.
 ****************************************************************************/
PRIVATE int GetMAC(unsigned long FLAddr, struct MacStructure *pMacStruct,
                   AuxMacStruct *pAux, unsigned long *pNum)
{
  __u8 Result;
  struct MacStructure *pVirtAddr;
  int AddrOK;
  AuxMacStruct *pAuxStruct;
  int i;

  *pNum = 0;
  pVirtAddr = (struct MacStructure *)FlashLin2Virt(FLAddr, &AddrOK);
  /* Check AddrOK before dereferencing the possibly invalid pVirtAddr */
  if (!AddrOK || (pVirtAddr->Telltale != MAC_TELLTALE) || (pMacStruct == 0))
  {
    return 0;
  }
  
  Result = checkCRC32edBlock((__u8 *)pVirtAddr, sizeof(struct MacStructure));
  if (Result != CRC_RESULT_PASS)
  {
    return 0;
  }
  
  /* Avoid using a library function to keep size down - hence labourious copy */
  pMacStruct->Mac.Sixth = pVirtAddr->Mac.Sixth;
  pMacStruct->Mac.Fifth = pVirtAddr->Mac.Fifth;
  pMacStruct->Mac.Fourth = pVirtAddr->Mac.Fourth;
  pMacStruct->Mac.Third = pVirtAddr->Mac.Third;
  pMacStruct->Mac.Second = pVirtAddr->Mac.Second;
  pMacStruct->Mac.First = pVirtAddr->Mac.First;
  pMacStruct->Mac.phyid0 = pVirtAddr->Mac.phyid0;

  pMacStruct->hwCode = pVirtAddr->hwCode;

  RelocSafePuts("M");

  /* Now we know where the MAC address is, we can look for auxillary MACs.
     These live just above the MAC/Bootloader structures if these structures
     are placed at the start of a sector, or just below them otherwise.
  */

  if ((FLAddr & 0x3ff) == 0x20)
  {
      /* Check after MAC structure */
      pAuxStruct = (AuxMacStruct *)((char *)pVirtAddr + LO_MAC_TO_AUXMAC_OFFSET);
      for (i = 0; i < OBELIX_MAX_AUXMACS; i++)
      {
          if (pAuxStruct->tellTale == AUXMAC_TELLTALE)
          {
              Result = checkCRC32edBlock((__u8 *)pAuxStruct, sizeof(AuxMacStruct));
              if (Result == CRC_RESULT_PASS)
              {
                  RelocSafePuts("a");
                  (*pNum)++;
                  pAux->tellTale = (__u32)(-1);
                  pAux->first = pAuxStruct->first;
                  pAux->second = pAuxStruct->second;
                  pAux->third = pAuxStruct->third;
                  pAux->fourth = pAuxStruct->fourth;
                  pAux->fifth = pAuxStruct->fifth;
                  pAux->sixth = pAuxStruct->sixth;
                  pAux->index = pAuxStruct->index;
                  pAux->crc = 0;
                  pAux++;
                  pAuxStruct++;
                  continue;
              }
          }
          break;
      }
  }
  else
  {
      /* Check before MAC structure */
      pAuxStruct = (AuxMacStruct *)((char *)pVirtAddr - HI_MAC_TO_AUXMAC_OFFSET);
      for (i = 0; i < OBELIX_MAX_AUXMACS; i++)
      {
          if (pAuxStruct->tellTale == AUXMAC_TELLTALE)
          {
              Result = checkCRC32edBlock((__u8 *)pAuxStruct, sizeof(AuxMacStruct));
              if (Result == CRC_RESULT_PASS)
              {
                  RelocSafePuts("A");
                  (*pNum)++;
                  /* Non-zero telltale indicates valid auxmac entry */
                  pAux->tellTale = (__u32)(-1);
                  pAux->first = pAuxStruct->first;
                  pAux->second = pAuxStruct->second;
                  pAux->third = pAuxStruct->third;
                  pAux->fourth = pAuxStruct->fourth;
                  pAux->fifth = pAuxStruct->fifth;
                  pAux->sixth = pAuxStruct->sixth;
                  pAux->index = pAuxStruct->index;
                  pAux->crc = 0;
                  pAux++;
                  pAuxStruct--;
                  continue;
              }
          }
          break;
      }
  }

  return 1;
}

/****************************************************************************
 * Function Name  : PtrFixup
 * Arguments      : Pointer to statically initialized data.
 * Returns        : Corrected pointer to data.
 * Description    : When the bootloader code is compiled it is impossible
 *                : to know where exactly in FLASH it will end up, so the
 *                : code must be relocatable.  The ARM tools produced an
 *                : image that could execute at any location, however whilst
 *                : the GNU tools will produce relocatable code, in their
 *                : case this means code that can be fixed up by an OS
 *                : dynamic linker, which we don't have.
 *                : The current solution is to use the non-relocatable code
 *                : mode of the compiler.  Branches are relative as are ldr
 *                : instructions (they compile to pc relative loads); stack
 *                : accesses and uninitialized data accesses use absolute
 *                : addresses in DRAM so these all work, the only things that
 *                : don't are initialized data accesses as this data gets
 *                : linked immediately above the code section and is accessed
 *                : by absolute addressing, not program counter relative
 *                : addressing.
 *                : This function does the runtime relocation of static,
 *                : initialized data pointers and must be used whenever
 *                : the read-only tables in FLASH are accessed (e.g the
 *                : FLASH layout tables).
 ****************************************************************************/
PUBLIC inline const void *PtrFixup(const void *p)
{
    return (const void *)((unsigned long)p + GetExecOffset());
}

/****************************************************************************
 * Function Name  : RelocSafePuts
 * Arguments      : Null terminated string to send to UART
 * Returns        : None
 * Description    : Send a string to the UART, but copes with the fact that
 *                : that the string is static data and may not be where
 *                : it was originally linked (see PtrFixup declaration).
 ****************************************************************************/
PRIVATE void RelocSafePuts(char *s)
{
    char *reloc_s;
    char c;
    reloc_s = (char *)PtrFixup(s);
    while ((c = *reloc_s++) != '\0')
    {
        EmitChar(c);
    }
}

/****************************************************************************
 * Function Name  : Puts
 * Arguments      : Null terminated string to send to UART
 * Returns        : None
 * Description    : Send a string to the UART, but the string must not be
 *                : stored in the static data area (see RelocSafePuts)
 ****************************************************************************/
PRIVATE void Puts(char *s)
{
    char c;
    while ((c = *s++) != '\0')
    {
        EmitChar(c);
    }
}

/****************************************************************************
 * Function Name  : EmitChar
 * Arguments      : Character to send
 * Returns        : None
 * Description    : Send a single char to the UART.  This routine relies on 
 *                : the fact that the bootrom has already initialized the
 *                : UART.
 ****************************************************************************/
#define UART_BASE (0x80000000)
#define UART_TX_REGISTER ((volatile unsigned long*)(UART_BASE + 0))
#define UART_LINE_STATUS_REG ((volatile unsigned long*)(UART_BASE + 5 * 4))
/* Wait for char to go completely - this ensures pauses between chars &
   allows a serial terminal to sync up to the stream
*/
#define TX_EMPTY_MASK (1 << 6)

PRIVATE void EmitChar(char c)
{
    while ((*UART_LINE_STATUS_REG & TX_EMPTY_MASK) == 0);
    *UART_TX_REGISTER = c;
}

PRIVATE void EmitHexWord(unsigned long x)
{
    int i;
    unsigned long v;

    for (i = 28; i >= 0; i -= 4)
    {
        v = (x >> i) & 0xf;
        if (v > 9)
        {
            EmitChar(v + 'A' - 10);
        }
        else
        {
            EmitChar(v + '0');
        }
    }
}
