/****************************************************************************
 *
 *  Copyright (c) 2001 by LSI Logic Corporation, all rights reserved.
 *
 *                         LSI Logic confidential
 *  The information contained herein is proprietary of LSI Logic Corporation.
 *
 *  This software is provided "as is" without any express or implied warranty.
 *  LSI Logic disclaims all representations and warranties with regard to
 *  this software, including all implied warranties of merchantability and
 *  fitness for a particular purpose. In no event shall LSI Logic be liable
 *  for any special, indirect or consequential damages or any damages
 *  whatsoever resulting from loss of use, data or profits, whether in an
 *  action of contract, negligence or other tortious act or omission,
 *  arising out of or in connection with the use or performance of this
 *  software.
 *
 *  File Description:
 *
 *  This file contains all of the routines that are FLASH specific.  This
 *  allows the Getafix bootloader to be used as a template for other FLASH
 *  configurations.
 *
 *  This file supports many different combinations of FLASH chips in a system,
 *  just alter the file flash-specific.h to build support 1-4 FLASH chips with
 *  varying timings and widths.
 *
 *  This file relies on definitions from the Getafix bootrom header file.
 *
 ****************************************************************************/

#include "bootloader.h"
#include "bootrom.h"
#include "flash-specific.h"
#include "crc32.h"
#include "flashmemcpy.h"

PRIVATE const struct FlashInfo
{
  __u32 VirtStart; /* Start virtual address of FLASH chip */
  __u32 Size;      /* Size of FLASH in bytes */
  __u32 WidthCode; /* Width of FLASH */
	__u32 LinearEnd; /* End location of FLASH chip as FLASH linear address */
	__u32 Offset;	   /* Virtual start address - FLASH linear start address */
} FlashChipLayout[] =
{
#ifdef FLASH_0_DEVICE
  {CS0_VIRTUAL_ADDR,
  FLASH_0_SIZE,
  FLASH_0_WIDTHCODE,
	FLASH0_LINEAR_OFFSET + FLASH_0_SIZE - 1,
	CS0_VIRTUAL_ADDR - FLASH0_LINEAR_OFFSET},
#endif
#ifdef FLASH_1_DEVICE
  {CS1_VIRTUAL_ADDR,
  FLASH_1_SIZE,
  FLASH_1_WIDTHCODE,
	FLASH1_LINEAR_OFFSET + FLASH_1_SIZE - 1,
	CS1_VIRTUAL_ADDR - FLASH1_LINEAR_OFFSET},
#endif
#ifdef FLASH_2_DEVICE
  {CS2_VIRTUAL_ADDR,
  FLASH_2_SIZE,
  FLASH_2_WIDTHCODE,
	FLASH2_LINEAR_OFFSET + FLASH_2_SIZE - 1,
	CS2_VIRTUAL_ADDR - FLASH2_LINEAR_OFFSET},
#endif
#ifdef FLASH_3_DEVICE
  {CS3_VIRTUAL_ADDR,
  FLASH_3_SIZE,
  FLASH_3_WIDTHCODE,
	FLASH3_LINEAR_OFFSET + FLASH_3_SIZE - 1,
	CS3_VIRTUAL_ADDR - FLASH3_LINEAR_OFFSET},
#endif
};

#define FlashChips (sizeof(FlashChipLayout) / sizeof(struct FlashInfo))

/* ============================ EBB register set macros ============================ */

/*
	MASK produces a value with bits cleared in the positions corresponding to the
	register field referred to by the macro passed as an argument.  This value
	is used as a AND mask to zero the field in the register value read.
	e.g.
	RDAD field occupies bits 27-30.
	EBB_CERW_RDAD(x) is the macro used to set this field in the CERWx register.
	MASK(EBB_CERW_RDAD) is a value with bits 27-30 cleared, all others set.
	Hence to set CERWx register field RDAD, use

		New CERWx value = ((Old CERWx Value) & MASK(EBB_CERW_RDAD)) 
													| EBB_CERW_RDAD(New field value)
	
*/
#define MASK(x) ~x(-1)

/* CERWx register field definitions */
#define EBB_CERW_RDAD(x) (((x) & 0xf) << 27)
#define EBB_CERW_RDDD(x) (((x) & 0x1f) << 22)
#define EBB_CERW_RDAF(x) (((x) & 0x1) << 21)
#define EBB_CERW_RDAN(x) (((x) & 0x1) << 20)
#define EBB_CERW_RDAL(x) (((x) & 0x1) << 19)
#define EBB_CERW_RDDF(x) (((x) & 0x1) << 18)
#define EBB_CERW_RDDN(x) (((x) & 0x1) << 17)
#define EBB_CERW_RDDL(x) (((x) & 0x1) << 16)

#define EBB_CERW_WRAD(x) (((x) & 0xf) << 11)
#define EBB_CERW_WRDD(x) (((x) & 0x1f) << 6)
#define EBB_CERW_WRAF(x) (((x) & 0x1) << 5)
#define EBB_CERW_WRAN(x) (((x) & 0x1) << 4)
#define EBB_CERW_WRAL(x) (((x) & 0x1) << 3)
#define EBB_CERW_WRDF(x) (((x) & 0x1) << 2)
#define EBB_CERW_WRDN(x) (((x) & 0x1) << 1)
#define EBB_CERW_WRDL(x) (((x) & 0x1) << 0)
/* 
	Create an AND mask for the CERWx registers that has bits clear corresponding to all
	the fields that will be updated in the register.
*/
#define EBB_CERW_MASK (0)

/* CERWx register field definitions */
#define EBB_CECS_RAD(x) (((x) & 0xf) << 27)
#define EBB_CECS_RDD(x) (((x) & 0x1f) << 22)
#define EBB_CECS_RAF(x) (((x) & 0x1) << 21)
#define EBB_CECS_RAN(x) (((x) & 0x1) << 20)
#define EBB_CECS_RAL(x) (((x) & 0x1) << 19)
#define EBB_CECS_RDF(x) (((x) & 0x1) << 18)
#define EBB_CECS_RDN(x) (((x) & 0x1) << 17)
#define EBB_CECS_RDL(x) (((x) & 0x1) << 16)

#define EBB_CECS_WAD(x) (((x) & 0xf) << 11)
#define EBB_CECS_WDD(x) (((x) & 0x1f) << 6)
#define EBB_CECS_WAF(x) (((x) & 0x1) << 5)
#define EBB_CECS_WAN(x) (((x) & 0x1) << 4)
#define EBB_CECS_WAL(x) (((x) & 0x1) << 3)
#define EBB_CECS_WDF(x) (((x) & 0x1) << 2)
#define EBB_CECS_WDN(x) (((x) & 0x1) << 1)
#define EBB_CECS_WDL(x) (((x) & 0x1) << 0)

/* 
	Create an AND mask for the CECSx registers that has bits clear corresponding to all
	the fields that will be updated in the register.
*/
#define EBB_CECS_MASK (0)

/* CECFGx register field definitions */
#define EBB_CECFG_AEN(x) (((x) & 0x1) << 31)
#define EBB_CECFG_BUSHR(x) (((x) & 0x1f) << 27)
#define EBB_CECFG_BUSHW(x) (((x) & 0x7) << 23)
#define EBB_CECFG_AHLDF(x) (((x) & 0x7) << 20)
#define EBB_CECFG_AHLDR(x) (((x) & 0xf) << 16)
#define EBB_CECFG_AHLDW(x) (((x) & 0xf) << 12)
#define EBB_CECFG_SAMPD(x) (((x) & 0x1f) << 7)
#define EBB_CECFG_XACK(x) (((x) & 0x1) << 6)
#define EBB_CECFG_SHS(x) (((x) & 0x1) << 3)
#define EBB_CECFG_MSTB(x) (((x) & 0x1) << 2)
#define EBB_CECFG_SIZE(x) (((x) & 0x3) << 0)

#define EBB_CECFG_MASK (0)

/* CEWAITx register field definitions */
#define EBB_CEWAIT_WAITF(x) (((x) & 0x3f) << 8)
#define EBB_CEWAIT_WAITN(x) (((x) & 0x3f) << 0)
#define EBB_CEWAIT_MASK (0)

/* CEBCx register field definitions */
#define EBB_CEBC_RBC(x) (((x) & 0x3) << 2)
#define EBB_CEBC_WBC(x) (((x) & 0x3) << 0)
#define EBB_CEBC_MASK (0)

/* CEMSTBx register field definitions */
#define EBB_CEMSTB_MSAD(x) (((x) & 0xf) << 11)
#define EBB_CEMSTB_MSDD(x) (((x) & 0x1f) << 6)
#define EBB_CEMSTB_MSAF(x) (((x) & 0x1) << 5)
#define EBB_CEMSTB_MSAN(x) (((x) & 0x1) << 4)
#define EBB_CEMSTB_MSAL(x) (((x) & 0x1) << 3)
#define EBB_CEMSTB_MSDF(x) (((x) & 0x1) << 2)
#define EBB_CEMSTB_MSDN(x) (((x) & 0x1) << 1)
#define EBB_CEMSTB_MSDL(x) (((x) & 0x1) << 0)
#define EBB_CEMSTB_MASK (0)

/* ========================= Initialisation structure ============================== */

/*
	This defines an arbitary set of modifications to arbitary registers.
	Each entry specifies the ARM virtual address of the register to change
	and the change to make
		New value = (old value & and_mask) | or_mask
	This array is used to setup the EBB so that it maps in all the FLASH
	chips in the system (so that they match the description in the
	flash-definition.h file).  All the entries in this structure are
	'executed'.
*/

PRIVATE const struct RegisterInitialization
{
	__u32 *pAddress;
	__u32 AndMask;
	__u32 OrMask;
} RegInit[] =
{
#ifdef FLASH_0_DEVICE
	/* At least one FLASH device - so do global setups required for all FLASHs */
	{EBB_CEMSTB0,
		EBB_CEMSTB_MASK,
		EBB_CEMSTB_MSAD(0) | EBB_CEMSTB_MSDD(0)},
	/* Write 0 to CEPINMODE register as per datasheet requirement */
	{EBB_CEPINMODE,
	  0,
	  0},
	/* EBB_CEALE register untouched - assumes FLASHs are not using address latches */
	/* EBB_CEBUSMODE register untouched */
	
	/* FLASH 0 specific stuff starts here */
	{EBB_CERW0,
		EBB_CERW_MASK,
		EBB_CERW_RDAF(1) | EBB_CERW_RDDL(1) |
   EBB_CERW_WRAF(1) | EBB_CERW_WRDL(1) | EBB_CERW_WRAD(1) },
	{EBB_CECS0,
		EBB_CECS_MASK,
		EBB_CECS_RDD(1) | EBB_CECS_RAF(1) | EBB_CECS_RDL(1) |
		EBB_CECS_WDD(1) | EBB_CECS_WAF(1) | EBB_CECS_WDL(1)},
	{EBB_CEWAIT0,
		EBB_CEWAIT_MASK,
		EBB_CEWAIT_WAITF(READ_CLOCKS_0) | EBB_CEWAIT_WAITN(63)},
	{EBB_CEBC0,
		EBB_CEBC_MASK,
		EBB_CEBC_RBC(0)},
  /* For CS0 registers ONLY, the writes to CERW0, CECS0, CEWAIT0 and CEBC0
     are held in shadow registers and are only written when CECFG0 is.
     This means that the CECFG0 write must be the last in the sequence.
  */
	{EBB_CECFG0,
		EBB_CECFG_MASK,
		EBB_CECFG_AEN(1) | EBB_CECFG_BUSHR(1) | EBB_CECFG_BUSHW(1) | 
		EBB_CECFG_AHLDR(1) | EBB_CECFG_SIZE(FLASH_0_WIDTHCODE)},
#endif
#ifdef FLASH_1_DEVICE
	{EBB_CERW1,
		EBB_CERW_MASK,
		EBB_CERW_RDAF(1) | EBB_CERW_RDDL(1) |
    EBB_CERW_WRAF(1) | EBB_CERW_WRDL(1) | EBB_CERW_WRAD(1)},
	{EBB_CECS1,
		EBB_CECS_MASK,
		EBB_CECS_RDD(1) | EBB_CECS_RAF(1) | EBB_CECS_RDL(1) |
		EBB_CECS_WDD(1) | EBB_CECS_WAF(1) | EBB_CECS_WDL(1)},
	{EBB_CECFG1,
		EBB_CECFG_MASK,
		EBB_CECFG_AEN(1) | EBB_CECFG_BUSHR(1) | EBB_CECFG_BUSHW(1) | 
		EBB_CECFG_AHLDR(1) | EBB_CECFG_SIZE(FLASH_1_WIDTHCODE)},
	{EBB_CEWAIT1,
		EBB_CEWAIT_MASK,
		EBB_CEWAIT_WAITF(READ_CLOCKS_1) | EBB_CEWAIT_WAITN(63)},
	{EBB_CEBC1,
		EBB_CEBC_MASK,
		EBB_CEBC_RBC(0)},
#endif
#ifdef FLASH_2_DEVICE
	{EBB_CERW2,
		EBB_CERW_MASK,
		EBB_CERW_RDAF(1) | EBB_CERW_RDDL(1) | 
    EBB_CERW_WRAF(1) | EBB_CERW_WRDL(1) | EBB_CERW_WRAD(1)},
	{EBB_CECS2,
		EBB_CECS_MASK,
		EBB_CECS_RDD(1) | EBB_CECS_RAF(1) | EBB_CECS_RDL(1) |
		EBB_CECS_WDD(1) | EBB_CECS_WAF(1) | EBB_CECS_WDL(1)},
	{EBB_CECFG2,
		EBB_CECFG_MASK,
		EBB_CECFG_AEN(1) | EBB_CECFG_BUSHR(1) | EBB_CECFG_BUSHW(1) | 
		EBB_CECFG_AHLDR(1) | EBB_CECFG_SIZE(FLASH_2_WIDTHCODE)},
	{EBB_CEWAIT2,
		EBB_CEWAIT_MASK,
		EBB_CEWAIT_WAITF(READ_CLOCKS_2) | EBB_CEWAIT_WAITN(63)},
	{EBB_CEBC2,
		EBB_CEBC_MASK,
		EBB_CEBC_RBC(0)},
#endif
#ifdef FLASH_3_DEVICE
	{EBB_CERW3,
		EBB_CERW_MASK,
		EBB_CERW_RDAF(1) | EBB_CERW_RDDL(1) | 
    EBB_CERW_WRAF(1) | EBB_CERW_WRDL(1) | EBB_CERW_WRAD(1)},
	{EBB_CECS3,
		EBB_CECS_MASK,
		EBB_CECS_RDD(1) | EBB_CECS_RAF(1) | EBB_CECS_RDL(1) |
		EBB_CECS_WDD(1) | EBB_CECS_WAF(1) | EBB_CECS_WDL(1)},
	{EBB_CECFG3,
		EBB_CECFG_MASK,
		EBB_CECFG_AEN(1) | EBB_CECFG_BUSHR(1) | EBB_CECFG_BUSHW(1) | 
		EBB_CECFG_AHLDR(1) | EBB_CECFG_SIZE(FLASH_3_WIDTHCODE)},
	{EBB_CEWAIT3,
		EBB_CEWAIT_MASK,
		EBB_CEWAIT_WAITF(READ_CLOCKS_3) | EBB_CEWAIT_WAITN(63)},
	{EBB_CEBC3,
		EBB_CEBC_MASK,
		EBB_CEBC_RBC(0)},
#endif
};

#define RegInitEntries (sizeof(RegInit) / sizeof(struct RegisterInitialization))

/****************************************************************************
 *    END OF FLASH DEFINITION STRUCTURES & VARIABLES - START OF REAL CODE   *
 ****************************************************************************/

/****************************************************************************
 * Function Name  : MaxFlashLinearAddress
 * Arguments      : None
 * Returns        : Largest FLASH linear address that is valid in the system
 ****************************************************************************/
PUBLIC __u32 MaxFlashLinearAddress()
{
	return ((struct FlashInfo*)PtrFixup(FlashChipLayout))[FlashChips - 1].LinearEnd;
}

/****************************************************************************
 * Function Name  : ConfigureEBB
 * Arguments      : None
 * Returns        : None
 * Description    : Performs a list of register read/writes that sets up the
 *					EBB to map in the FLASH chips.
 ****************************************************************************/

PUBLIC void ConfigureEBB(void)
{
	int i;
	const struct RegisterInitialization *pri;

	for (i = 0; i < RegInitEntries; i++)
	{
		pri = &((struct RegisterInitialization *)PtrFixup(RegInit))[i];
		*(pri->pAddress) = ((*(pri->pAddress) & pri->AndMask) | pri->OrMask);
	}
}

/****************************************************************************
 * Function Name  : ReadFlash32
 * Arguments      : FLASH linear address to read
 * Returns        : unsigned 32 bit value at address given, or 0 if the
 *				    address is invalid.
 * Variables      : loop counter, bit field and bit shift counter
 * Modifies       : nothing
 * Description    : Reads a 32 bit value from a FLASH virtual address.
 *					(See flash_virt2phys for an explanation of FLASH virtual
 *					addresses.)
 ****************************************************************************/

PUBLIC __u32 ReadFlash32(__u32 FlashLinearAddress)
{
	int AddressOK;
	__u32 FlashVirtualAddress;
	
	FlashVirtualAddress = FlashLin2Virt(FlashLinearAddress, &AddressOK);
	return (AddressOK != 0) ? *(__u32 *)FlashVirtualAddress : (__u32)0;
}

/****************************************************************************
 * Function Name  : FlashLin2Virt
 * Arguments      : FLASH linear address to convert, address valid flag.
 * Returns        : Virtual address corresponding to linear address given
 *					or 0 if address is invalid.  AddressOK set to true if
 *					returned address is valid, otherwise set to false.
 * Description    : The FLASH linear address assumes that all FLASH chips in
 *					the system are mapped contiguously and that the first
 *					FLASH starts at address zero.  This function uses
 *					the system specific knowledge of how many FLASH chips are
 *					really present and where they live in virtual address
 *					space in order to convert this address into a real
 *					ARM virtual address.
 *					FLASH linear addresses have nothing to do with ARM
 *					linear addresses, they are purely a software concept.
 ****************************************************************************/

PUBLIC __u32 FlashLin2Virt(__u32 FlashLinearAddress, int *pAddressOK)
{
	int	i;
	const struct FlashInfo *pfi;
	
	*pAddressOK = 0;
	for (i = 0; i < FlashChips; i++)
	{
		pfi = &(((struct FlashInfo*)PtrFixup(FlashChipLayout))[i]);
		if (FlashLinearAddress <= pfi->LinearEnd)
		{
			*pAddressOK = 1;
			return (FlashLinearAddress + pfi->Offset);
		}
	}
	return 0;
}

/****************************************************************************
 * Function Name  : FlashVirt2Lin
 * Arguments      : FLASH virtual address to convert, address valid flag.
 * Returns        : FLASH linear address corresponding to the virtual address
 *                  given or 0 if address is invalid.  AddressOK set to true
 *                  if the returned address is valid, otherwise set to false.
 * Description    : The FLASH linear address assumes that all FLASH chips in
 *         					the system are mapped contiguously and that the first
 *				        	FLASH starts at address zero.  This function uses
 *					        the system specific knowledge of how many FLASH chips are
 *					        really present and where they live in virtual address
 *					        space in order to convert this address into a real
 *					        ARM virtual address.
 *					        FLASH linear addresses have nothing to do with ARM
 *					        linear addresses, they are purely a software concept.
 ****************************************************************************/

PUBLIC __u32 FlashVirt2Lin(__u32 FlashVirtualAddress, int *pAddressOK)
{
	int	i;
	const struct FlashInfo *pfi;
	
	if (pAddressOK != 0)
	{
	  *pAddressOK = 0;
	}
	for (i = 0; i < FlashChips; i++)
	{
		pfi = &(((struct FlashInfo*)PtrFixup(FlashChipLayout))[i]);
		
		if ((FlashVirtualAddress >= pfi->VirtStart) &&
		      (FlashVirtualAddress <= (pfi->LinearEnd + pfi->Offset)))
		{
			if (pAddressOK != 0)
			{
			  *pAddressOK = 1;
			}
			return (FlashVirtualAddress - pfi->Offset);
		}
	}
	return 0;
}

/****************************************************************************
 * Function Name  : CopyFlashLin2Virt
 * Arguments      : FLASH linear start address.
 *					Number of bytes to copy.
 *					*ARM* virtual target address.
 * Returns        : Non-zero if address range and length were valid, zero
 *					otherwise.
 * Modifies       : nothing
 * Description    : Copies a block of memory from one or more FLASH chips
 *					to an ARM virtual address.  The source address is
 *					a FLASH linear address (see FlashLin2Virt).
 ****************************************************************************/

PUBLIC __u32 CopyFlashLin2Virt(__u32 LinStartAddr, __u32 Length, __u32 VirtAddress)
{
	int	i;
	const struct FlashInfo *pfi;
	__u32 MaxBytes;
	
	for (i = 0; i < FlashChips; i++)
	{
		pfi = &(((struct FlashInfo*)PtrFixup(FlashChipLayout))[i]);
		/*
			This test fails until we get to the FLASH chip that contains
		   	the starting address - it will succeed therafter.
		*/
		if (LinStartAddr <= pfi->LinearEnd)
		{
			MaxBytes = pfi->LinearEnd - LinStartAddr + 1;
			if (Length <= MaxBytes)
			{
				/* This is the last copy */
				FlashMemcpy(VirtAddress, LinStartAddr + pfi->Offset, Length);
				return 1;
			}
			/* Copy as much out of this one FLASH chip as possible */
			FlashMemcpy(VirtAddress, LinStartAddr + pfi->Offset, MaxBytes);
			VirtAddress += MaxBytes;
			LinStartAddr += MaxBytes;
			Length -= MaxBytes;
		}
	}
	return 0;
}

/****************************************************************************
 * Function Name  : checkFlashCRC32edBlock
 * Arguments      : pointer to byte data (FLASH linear address)
 *					and length of data area to test.
 * Returns        : CRC_RESULT_PASS or CRC_RESULT_FAIL.
 * Description    : Does a CRC check of an area of memory described by
 *					a FLASH linear address and length.  This routine divides
 *					the CRC into chunks of continuous virtual addresses
 *					and performs the CRC in stages, one chunk at a time.
 *					This prevents the overhead of having to lookup the
 *					virtual address of each FLASH linear address for every
 *					byte.
 ****************************************************************************/

PUBLIC __u32 checkFlashCRC32edBlock(__u32 LinStartAddr, __u32 Length)
{
	int	i;
	const struct FlashInfo *pfi;
	__u32 MaxBytes;
	__u32 crc;
	int AddressOK;
	__u32 VirtAddr;
	
	crc = CRC32_INIT;
	for (i = 0; i < FlashChips; i++)
	{
		pfi = &(((struct FlashInfo*)PtrFixup(FlashChipLayout))[i]);
		/*
			This test fails until we get to the FLASH chip that contains
		   	the starting address - it will succeed therafter.
		*/
		if (LinStartAddr <= pfi->LinearEnd)
		{
			MaxBytes = pfi->LinearEnd - LinStartAddr + 1;
			VirtAddr = FlashLin2Virt(LinStartAddr, &AddressOK);
			if (AddressOK == 0)
			{
				/* We've run off the end of valid FLASH linear address space */
				return CRC_RESULT_FAIL;
			}
			if (Length <= MaxBytes)
			{
				/* This is the final block, so see if the CRC value is good */
				crc = checkPartialCRC32edBlock(crc, (__u8 *)VirtAddr, Length);
				return (crc == CRC32_GOOD) ? CRC_RESULT_PASS : CRC_RESULT_FAIL;
			}
			/* Check an intermediate block */
			crc = checkPartialCRC32edBlock(crc, (__u8 *)VirtAddr, MaxBytes);
			VirtAddr += MaxBytes;
			LinStartAddr += MaxBytes;
			Length -= MaxBytes;
		}
	}
	return 0;
}
