/******************************************************************************
 * Copyright(c) 2002 LSI.
 *
 * $ID
 *
 * Description:
 * Soft Console Access Manager (SCAM)
 * ----------------------------------
 * This program controls access to the console functions of other programs
 * in the system. The IPC with other processes is through named pipes.
 *****************************************************************************/

/* Include files */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <string.h>
#include <termios.h>
#include <errno.h>
//#include <libgen.h>

#include "scamcommon.h"

#define SCAM_EXIT_REQUIRED 1
//#define SERIAL_TARGET 1
#define SERIAL "/dev/ttyS0"

typedef struct {
    char *userProgram;
    int rx_fd;
    int tx_fd;
    int echo;
}ProgramData;

ProgramData **programsList;
#define MAX_PROGS 20
#define FIFO_DIR "/fifo"

/* The terms are always in non-canonical mode, which means
   we will get one byte at a time. */
#define SCAM_BUF_SIZE 1
char buf[SCAM_BUF_SIZE];

void setup_term(int fd, int echo);
#define MAXENV 7
char *env[MAXENV];

/******************************************************************************
 * mybasename
 *
 * @param filename
 *
 * @return pointer to filename
 *
 *****************************************************************************/
static char *mybasename(char *filename)
{
  char *p;

  p = strrchr(filename, '/');
  if (p)
  {
    p++;
    return p;
  }
  return filename;
}

/******************************************************************************
 * createProgramsList
 *
 * @param void
 *
 * @return 0 success, -1 failure
 *
 *****************************************************************************/
int createProgramsList(void)
{
  FILE *fp;
  char line[255];
  char *ap;
  int i;

  programsList = (ProgramData **)malloc(sizeof(programsList) * (MAX_PROGS + 1));
  if (!programsList)
  {
    perror("malloc programsList");
    return -1;
  }

  fp = fopen(SCAMCONFIG, "r");
  if (!fp)
  {
    perror(SCAMCONFIG);
    return -1;
  }
  /* Start from 1 - zero is reserved for the shell. */
  for (i = 0; i < MAX_PROGS; i++)
  {
    if (fgets(line, 255, fp) == NULL)
    {
      break;
    }
    if (line[0] == '#')
    {
      continue;
    }
    if ((ap = strchr(line, ':')) == NULL)
    {
      break;
    }
    ap++;
    /* get rid of the newline */
    ap[strlen(ap) - 1] = '\0';
    programsList[i] = (ProgramData *)malloc(sizeof(ProgramData));
    if (!programsList[i])
    {
      perror("malloc ProgramData");
      return -1;
    }
    programsList[i]->userProgram = strdup(ap);
    programsList[i]->tx_fd = -1;
    programsList[i]->rx_fd = -1;
  }
  if (i == 0)
  {
    printf("No entries in scamrc.\n");
    return -1;
  }
  programsList[i] = (ProgramData *)malloc(sizeof(ProgramData));
  if (!programsList[i])
  {
    perror("malloc ProgramData");
    return -1;
  }
  programsList[i]->userProgram = 0;
  programsList[i]->tx_fd = -1;
  programsList[i]->rx_fd = -1;
  return 0;
}

/******************************************************************************
 * initShell
 *
 * @param void
 *
 * @return 0 success, -1 failure
 *
 *****************************************************************************/
#ifdef PIPE_METHOD
int initShell(void)
{
  int parentToChild[2];
  int childToParent[2];

  if (pipe(parentToChild) != 0)
  {
    perror("pipe 1");
    return -1;
  }
  if (pipe(childToParent) != 0)
  {
    perror("pipe 2");
    return -1;
  }

  switch (fork())
  {
    case -1:
      /* it failed */
      perror("fork");
      return -1;
    case 0:
      /* child */
      break;
    default:
      /* parent */
      programsList[0]->tx_fd = childToParent[0];
      programsList[0]->rx_fd = parentToChild[1];      
      return 0;
  }

  close(0);
  if (dup(parentToChild[0]) == -1)
  {
    perror("dup 1");
    exit(1);
  }
  close(1);
  if (dup(childToParent[1]) == -1)
  {
    perror("dup 2");
    exit(1);
  }
  close(2);
  if (dup(childToParent[1]) == -1)
  {
    perror("dup 3");
    exit(1);
  }
  close(parentToChild[0]);
  close(childToParent[1]);

  /* Now exec the shell. */
  if (execl("/bin/sh", "/bin/sh", 0) != 0)
  {
    exit(1);
  }
  /* Never gets here. */
  return 0;
}
#else
int initShell(void)
{
  char masterName[] = "/dev/ptmx";
  char slaveName[40], *p;
  int masterFd, slaveFd;

  /* Open the master pseudo tty */
  masterFd = open(masterName, O_RDWR);
  if (masterFd == -1)
  {
    perror("open master");
    return -1;
  }
  
#if 0
  if (grantpt(masterFd) < 0)
  {
    perror("grantpt");
    close(masterFd);
    return -1;
  }
#endif
	if (unlockpt(masterFd) < 0)
  {
    perror("unlockpt");
		close(masterFd);
		return -1;
	}
  /* Get the name of the associated slave tty */
  p = ptsname(masterFd);
  if (!p)
  {
    perror("ptsname");
    close(masterFd);
    return -1;
  }
  strcpy(slaveName, p);

  /* Open the slave */
  slaveFd = open(slaveName, O_RDWR);
  if (slaveFd == -1)
  {
    perror("open slave");
    close(masterFd);
    return -1;
  }

  switch (fork())
  {
    case -1:
      /* it failed */
      perror("fork");
      return -1;
    case 0:
      /* child */
      break;
    default:
      /* parent */
      programsList[0]->tx_fd = masterFd;
      programsList[0]->rx_fd = masterFd;
      close(slaveFd);
      return 0;
  }

  ioctl(0, TIOCNOTTY, 0);
  close(0);
  close(1);
  close(2);
  close(masterFd);

  signal(SIGUSR1, SIG_DFL);
  signal(SIGUSR2, SIG_DFL);
  signal(SIGINT, SIG_DFL);
  signal(SIGTERM, SIG_DFL);
  signal(SIGHUP, SIG_DFL);

  dup2(slaveFd, 0);
  dup2(slaveFd, 1);
  dup2(slaveFd, 2);

  ioctl(0, TIOCSCTTY, 1);
  tcsetpgrp(0, getpgrp());
  setup_term(0, 0);

  env[0] = "HOME=/";
  env[1] = "PATH=/bin:/sbin:/usr/bin:.";
  env[2] = "SHELL=/bin/sh";
  env[3] = "USER=root";
  env[4] = "TERM=vt102";
  env[5] = "PS1=scam-sh # ";
  env[6] = 0;
  /* Now exec the shell. */
  if (execle("/bin/sh", "/bin/sh", "-", 0, env) != 0)
  {
    perror("execl");
    exit(1);
  }
  /* Never gets here. */
  return 0;
}
#endif

/******************************************************************************
 * initProgramsList
 *
 * @param void
 *
 * @return void
 *
 *****************************************************************************/
void initProgramsList(void)
{
  int i = 0;
  ProgramData **pd;
  char pipe[255];
  
  /* Start from 1 - zero is reserved for the shell. */
  pd = &programsList[0];
  while ((*pd)->userProgram != 0)
  {
    printf("SCAM: Adding %s\n", (*pd)->userProgram);

    if (strcmp((*pd)->userProgram, "shell") == 0)
    {
      (*pd)->echo = 0;
      initShell();
    }
    else
    {
      (*pd)->echo = 1;
      sprintf(pipe, "%s/%s_rx", FIFO_DIR, mybasename((*pd)->userProgram));

      (*pd)->rx_fd = open(pipe, O_RDWR | O_NONBLOCK);
      if ((*pd)->rx_fd == -1)
      {
        perror(pipe);
        printf("Opening for read-only.\n");
      }

      sprintf(pipe, "%s/%s_tx", FIFO_DIR, mybasename((*pd)->userProgram));
      
      (*pd)->tx_fd = open(pipe, O_RDWR | O_NONBLOCK);
      if ((*pd)->tx_fd == -1)
      {
        perror(pipe);
        printf("%s will not be accessible\n", (*pd)->userProgram);
      }
    }
    i++;
    pd++;
  }
}

/******************************************************************************
 * menuSystem
 *
 * @param rx_fd Pointer to receive file descriptor
 * @param tx_fd Pointer to transmit file descriptor
 *
 * @return 0 something selected, 1 nothing selected
 *
 *****************************************************************************/
int menuSystem(int *rx_fd, int *tx_fd)
{
  int i = 0, selection = 0;
  ProgramData **pd;
  char *status = "";
  char selstring[2];

  /* Set term to not echo */
  setup_term(0, 0);

  printf("\n==================== SCAM ======================\n");
  printf("| Build %s %s\n", __DATE__, __TIME__);
  pd = programsList;
  while ((*pd)->userProgram)
  {
    if (((*pd)->tx_fd == -1) && ((*pd)->rx_fd == -1))
    {
      pd++;
      continue;
    }
    if ((*pd)->tx_fd != -1)
    {
      status = "read only";
      if ((*pd)->rx_fd != -1)
	    {
	      status = "read write";
	    }
    }
    printf("|  %d. %s <%s>\n", i + 1, (*pd)->userProgram, status);
    i++;
    pd++;
  }
#ifdef SCAM_EXIT_REQUIRED
  printf("|  %d. exit scam\n", i + 1);
#endif
  printf("|  Note: Type ctrl-d to return to the scam menu.\n");
  printf("Enter selection (1 - %d):", i + 1);

  selstring[1] = '\0';
  while (1)
  {
    selstring[0] = (unsigned char)getchar();
    selection = atoi(selstring);
    if ((selection > 0) && (selection <= i + 1))
    {
      printf("%s\n", selstring);
      break;
    }
  }

  if ((selection > 0) && (selection < i + 1))
  {
    if (programsList[selection - 1]->rx_fd != -1 || programsList[selection - 1]->tx_fd != -1)
    {
      *rx_fd = programsList[selection - 1]->rx_fd;
      *tx_fd = programsList[selection - 1]->tx_fd;
      if (programsList[selection - 1]->echo)
      {
        /* Turn on echo */
        setup_term(0, 1);
      }
      return 0;
    }
  }
#ifdef SCAM_EXIT_REQUIRED
  else
  {
    if (selection == i + 1)
    {
      printf("exiting scam...\n");
      exit(1);
    }
  }
#endif

  printf("error in selection...\n");
  return 1;
}

/******************************************************************************
 * parse_input_for_exit
 *
 * @param b Pointer to string
 *
 * @return 1 exit string matched, 0 no match
 *
 *****************************************************************************/
int parse_input_for_exit(char *b)
{
  /* ctrl-d */
  if (*b == 0x04)
  {
    return 1;
  }
  return 0;
}

/******************************************************************************
 * main
 *
 * @param argc Parameter count
 * @param argv Array of char pointers to parameters
 *
 * @return 
 *
 *****************************************************************************/
int main(int argc, char *argv[])
{
#ifdef SERIAL_TARGET
  int sfd = 0;
#endif
  int rc;
  fd_set rfds;
  int retval;
  int rx_fd, tx_fd;
  int first;

  signal(SIGINT, SIG_IGN);

  if (createProgramsList() == -1)
  {
    return -1;
  }

#ifdef SERIAL_TARGET
  sfd = open(SERIAL, O_RDWR);
  if (sfd == -1)
  {
    perror("open serial");
    return -1;
  }
  /* Copy the serial if file descriptor to stdin, stdout and stderr */
  dup2(sfd, 0);
  dup2(sfd, 1);
  dup2(sfd, 2);
  /* No longer need serial if file descriptor. */
  close(sfd);
#endif

  /* Open the pipes for the programs specified in the list. */
  initProgramsList();

  /* Run until the program is killed. */
  while (1)
  {
    /* Stay in the menu until we get a valid selection. */
    menuSystem(&rx_fd, &tx_fd);

    first = 1;

    /* Stay with the soft console until the "scam" keyword is entered. */
    while (1)
    {
      if (first)
      {
        first = 0;
        if (rx_fd != -1)
        {
          /* write to pipe */
          buf[0]='\n';
          buf[1]='\0';
          rc = write(rx_fd, buf, 2);
          if (rc == -1)
          {
            perror("write to rx pipe");
            return -1;
          }
        }
      }

      FD_ZERO(&rfds);
      FD_SET(0, &rfds);
      FD_SET(tx_fd, &rfds);

      retval = select(tx_fd + 1, &rfds, NULL, NULL, NULL);
      if (retval == -1)
	    {
	      perror("select");
	      return -1;
	    }

      if (FD_ISSET(tx_fd, &rfds))
	    {
	      /* read from pipe */
	      rc = read(tx_fd, buf, SCAM_BUF_SIZE);
	      if (rc == -1)
        {
          perror("read from tx pipe");
          return -1;
        }

	      /* write to serial */
	      rc = write(1, buf, rc);
	      if (rc == -1)
        {
          perror("write to serial");
          return -1;
        }
	    }
      if (FD_ISSET(0, &rfds))
	    {
	      /* read from serial */
	      rc = read(0, buf, SCAM_BUF_SIZE);
	      if (rc == -1)
        {
          perror("read from serial");
          return -1;
        }

	      /* See if we have the "scam" keyword. */
	      if (parse_input_for_exit(buf))
        {
          /* "scam" entered - leave the soft console loop. */
          break;
        }
	      
	      if (rx_fd != -1 && rc > 0)
        {
          /* write to pipe */
          rc = write(rx_fd, buf, rc);
          if (rc == -1)
          {
            perror("write to rx pipe");
            return -1;
          }
        }
	    }
    }
  }

  return 0;
}

/******************************************************************************
 * setup_term
 *
 * @param fd
 *
 * @return void
 *
 *****************************************************************************/
void setup_term(int fd, int echo)
{
	struct termios tty;

	tcgetattr(fd, &tty);

	/* set control chars */
  tty.c_cc[VINTR]  = 3;	/* C-c */
	tty.c_cc[VQUIT]  = 28;	/* C-\ */
	tty.c_cc[VERASE] = 127; /* C-? */
	tty.c_cc[VKILL]  = 21;	/* C-u */
	tty.c_cc[VEOF]   = 4;	/* C-d */
	tty.c_cc[VSTART] = 17;	/* C-q */
	tty.c_cc[VSTOP]  = 19;	/* C-s */
	tty.c_cc[VSUSP]  = 26;	/* C-z */

	/* use line dicipline 0 */
	tty.c_line = 0;

	/* Make it be sane */
	tty.c_cflag &= CBAUD|CBAUDEX|CSIZE|CSTOPB|PARENB|PARODD;
	tty.c_cflag |= CREAD|HUPCL|CLOCAL;


	/* input modes */
	tty.c_iflag = ICRNL | IXON | IXOFF;

	/* output modes */
	tty.c_oflag = OPOST | ONLCR;

	/* local modes */
	tty.c_lflag =
		ISIG | ECHOE | ECHOK | ECHOCTL | ECHOKE | IEXTEN;
  if (echo)
  {
    tty.c_lflag |= ECHO;
  }
//		ISIG | ICANON | ECHO | ECHOE | ECHOK | ECHOCTL | ECHOKE | IEXTEN;

	tcsetattr(fd, TCSANOW, &tty);
}
