/******************************************************************************
 *
 *  (C) Copyright 2004 Trend Micro, Inc.
 *
 *  All Rights Reserved.
 *  This program is an unpublished copyrighted work which is proprietary
 *  to Trend Micro, Inc. and contains confidential information that is not
 *  to be reproduced or disclosed to any other person or entity without
 *  prior written consent from Trend Micro, Inc. in each and every instance.
 *
 *  WARNING:  Unauthorized reproduction of this program as well as
 *  unauthorized preparation of derivative works based upon the
 *  program or distribution of copies by sale, rental, lease or
 *  lending are violations of federal copyright laws and state trade
 *  secret laws, punishable by civil and criminal penalties.
 *
 *****************************************************************************/

/**
 *  @file   tmAUChecker.c
 *  @brief  Sample program for periodically check AcitveUpdate server
 */

/*
 *  Change Log
 *  2004-02-14  Jeremy  initial version
 */
//#define NDEBUG
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <stdarg.h>
#include "TmErrors.h"
#include "tmauIniParser.h"
#include "tmClientInfo.h"
#include "tmPidUtil.h"
#include "tmcfg.h"
#include "tmGetAUInfo.h"
#include "tmShm.h"

#define  TM_FILE_MAX_LEN    256
/*  write a pid file for tmss.sh to "kill -9" to solve unkillable case */
#define  AU_PID_FILE    "/var/run/tmAUChecker.pid"

/* the module name that will show in log file */
static const char ModuleName[] = "tmAUChecker";

static char     gConfigFile[TM_FILE_MAX_LEN] = {0};
static char     gAUServerURL[TM_AU_MAX_STR] = {0};
static int32_t  gAUCheckTime = 0, gAUCheckTimeMax=0;
static uint32_t nShmID = 0;

char strEngineVxDKey[]  = "E.2=";   /* Leading key for the VxD engine */
char strEngineNTKDKey[] = "E.10=";  /* Leading key for the NTKD engine */
char strPatternKey[]    = "P.4";    /* Leading key for the VSAPI pattern */

#define CFG_MODULE_NAME_AU_CHECKER "au_checker"


/**
 * @brief   Reload the settings from the tm.conf, the configuration file  
 *
 * @param   pszINIFile: pointer to the server.ini file
 *
 * @return  1: continue to work, enable key is set
 *          0: stop the work, enable key is not set
 *
 * @remarks This function will be called either the this program receive the 
 *          SIG_HUP or begin at the execution
 */
void run_task(char *cmd, ...);
static int _tmau_LoadSettings(const char *pszFile)
{
    tmcfg_t     cfg;
    const char *ptr;
    int         ret = 0;


    LOGDEBUG("Entering _tmau_LoadSettings"); 

    LOGINFO("Refer config file: [%s]", pszFile);

    if ((cfg = tmcfg_open(pszFile, TMCFG_MODE_READ)) == NULL)
    {
        LOGERROR("Failed to open config file: [%s]", pszFile);
        return ret;
    }

    ptr = tmcfg_get_param_value(cfg, CFG_MODULE_NAME_AU_CHECKER, "au_server_url", NULL);
    if (ptr != NULL)
    {
        strncpy(gAUServerURL, ptr, sizeof(gAUServerURL));
        gAUServerURL[sizeof(gAUServerURL) - 1] = 0; 
    }

    ptr = tmcfg_get_param_value(cfg, CFG_MODULE_NAME_AU_CHECKER, "au_check_time", NULL);
    if (ptr != NULL)
    {
        gAUCheckTime = atoi(ptr);
    }
    
    ptr = tmcfg_get_param_value(cfg, CFG_MODULE_NAME_AU_CHECKER, "enable", NULL);
    if (ptr != NULL)
    {
        ret = atoi(ptr);
    }

    tmcfg_close(cfg);

    LOGDEBUG("Leaving _tmau_LoadSettings");
    return ret;
}


/**
 * @brief   Call the parsing functions in libtmIniParser.so for retrieving
 *          the ActiveUpdate engine and pattern version for PCC.
 *
 * @param   pszINIFile: pointer to the server.ini file
 * @param   pTmuPat: pointer to TmuVersion_t structure
 * @param   pTmuEng: pointer to TmuVersion_t structure
 *
 * @return  zero for success; negative value for error cases
 * 
 * @remarks This function originate from ~/src/utils/au_info/unittest/testINIParser.c
 *          It will store the newest information into input arguments, pTmuPat
 *          and pTmuEng.
 */
static int _tmau_CallIniParser(char *pszINIFile, TmuVersion_t *pTmuPat, TmuVersion_t *pTmuEng)
{
    TmErrorCode     tmErrCode = TM_ERR_ZERO_ERROR;
    char            szPat[TM_AU_MAX_VER];
    char            szEngineVxD[TM_AU_MAX_VER];
    char            szEngineNTKD[TM_AU_MAX_VER];


    LOGDEBUG("Entering _tmau_CallIniParser(), start to call library for parsing");

    if (access(pszINIFile, R_OK) != TM_ERR_ZERO_ERROR)
    {
        LOGERROR("Cannot find the target file, [%s]\n", pszINIFile);
        tmErrCode = AU_INFO_FAILURE(TM_ERR_C_INT_FILE_NOT_FOUND);
    }

    if (tmau_GetPatternVer(pszINIFile, strPatternKey, pTmuPat, &tmErrCode))
    {
        snprintf(szPat, TM_AU_MAX_STR, "%d", pTmuPat->version);        
        LOGDEBUG("The lastest version of pattern[%s] is: %s\n", strPatternKey, szPat);        
    }
    else if (TM_FAILURE(tmErrCode))
    {
        LOGERROR("The error return of parsing pattern version is:\n");
        LOGERROR("\t\"Facility\" is       [%02x]\n", TM_COND_FACILITY(tmErrCode));
        LOGERROR("\t\"Source ID\" is      [%02x]\n", TM_COND_SOURCE(tmErrCode));
        LOGERROR("\t\"Condition Code\" is [%04x]\n", TM_COND_CODE(tmErrCode));
    }    
    
    if (tmau_GetEngineVer(pszINIFile, strEngineVxDKey, pTmuEng, &tmErrCode))
    {
        tmau_MakeVersionBuildString((*pTmuEng), szEngineVxD, TM_AU_MAX_VER);
        LOGDEBUG("The latest version of engine VxD is: %s\n", szEngineVxD);
    }
    else if (TM_FAILURE(tmErrCode))
    {
        LOGERROR("The error return of parsing VxD engine is:\n");
        LOGERROR("\t\"Facility\" is       [%02x]\n", TM_COND_FACILITY(tmErrCode));
        LOGERROR("\t\"Source ID\" is      [%02x]\n", TM_COND_SOURCE(tmErrCode));
        LOGERROR("\t\"Condition Code\" is [%04x]\n", TM_COND_CODE(tmErrCode));
    }

    if (tmau_GetEngineVer(pszINIFile, strEngineNTKDKey, pTmuEng, &tmErrCode))
    {
        tmau_MakeVersionBuildString((*pTmuEng), szEngineNTKD, TM_AU_MAX_VER);
        LOGDEBUG("The latest version of engine NTKD is: %s\n", szEngineNTKD);
    }
    else if (TM_FAILURE(tmErrCode))
    {
        LOGERROR("The error return of parsing NTKD engine are:\n");
        LOGERROR("\t\"Facility\" is       [%02x]\n", TM_COND_FACILITY(tmErrCode));
        LOGERROR("\t\"Source ID\" is      [%02d]\n", TM_COND_SOURCE(tmErrCode));
        LOGERROR("\t\"Condition Code\" is [%04x]\n", TM_COND_CODE(tmErrCode));
    }
  
    if (pTmuPat->version == 0)
    {
        LOGERROR("No pattern number found, treat it as error");
        tmErrCode = AU_INFO_FAILURE(TM_ERR_C_INT_INVALID_FORMAT);
    }
    
    LOGDEBUG("Leaving _tmau_CallIniParser()");
    return tmErrCode;
}


/**
 * @brief   Handle the signals, like SIGTERM, SIGINT, and SIGHUP
 * 
 * @param   nSignal: signal number
 * 
 * @return  void
 */
static void _tmau_SignalHandler(int nSignal)
{
    LOGDEBUG("Entering of _tmau_SignalHandler()");

    switch (nSignal)
    {
        case SIGTERM:
        case SIGINT:
            LOGINFO("Will exit due to signal %d", nSignal);

            LOGDEBUG("Call shm_fini() to delete shared memory.");
            shm_fini(nShmID);

            /*  write a pid file for tmss.sh to "kill -9" to solve unkillable case */
            LOGDEBUG("Remove PID file: %s", AU_PID_FILE);
            remove_pid(AU_PID_FILE);

            exit(0);
            break;
        case SIGHUP:
            LOGINFO("Recevice a reload signal, %d", nSignal);
            _tmau_LoadSettings(gConfigFile);
            break;
        default:
            LOGWARNING("Receive a unknown signal %d", nSignal);
            break;
    }

    LOGDEBUG("Leaving _tmau_SignalHandler()");
    return;
}


int main(int argc, char **argv)
{
  //  TmErrorCode     tmErrCode = TM_ERR_ZERO_ERROR; 
  //  int32_t         i = 0;
    TmuVersion_t    stTmuPatVer, stTmuEngVer;                 
    TmuAUVersion_t  *pAUInfoPool;
    TmuAUVersion_t  TmuVerShm;              /* structure to hold the AU information, from shared memory */
    char            szEngine[TM_AU_MAX_VER];/* buffer to hold the readable engine information */
    char            szWgetCmd[TM_AU_MAX_STR];


    LOGDEBUG("Program started");

    if (argc != 2)
    {
        fprintf(stderr, "Usage: %s <config file> \n", argv[0]);
        exit(1);
    }

    if (access(argv[1], R_OK) == TM_ERR_ZERO_ERROR)
    {
        strncpy(gConfigFile, argv[1], sizeof(gConfigFile));
        gConfigFile[sizeof(gConfigFile) - 1] = 0;
    }
    else
    {
        fprintf(stderr, "Cannot find the config file, [%s]", argv[1]);
        LOGERROR("Cannot find the config file, [%s]", argv[1]);
        exit(1);
    }
    
    LOGDEBUG("Hook SIGTERM");
    if (SIG_ERR == signal(SIGTERM, _tmau_SignalHandler))
    {
        LOGERROR("Register singal SIGTERM fail: %m");
        exit(1);
    }

    LOGDEBUG("Hook SIGINT");
    if (SIG_ERR == signal(SIGINT, _tmau_SignalHandler))
    {
        LOGERROR("Register singal SIGINT fail: %m");
        exit(1);
    }

    /* Monitor for config reload signal, may come from CGI */
    LOGDEBUG("Hook SIGHUP");
    if (SIG_ERR == signal(SIGHUP, _tmau_SignalHandler))
    {
        LOGERROR("Register singal SIGHUP fail: %m");
        exit(1);
    }

    /* Write a pid file for tmss.sh to "kill -9" to solve unkillable case */
    LOGDEBUG("Create PID file: %s", AU_PID_FILE);
    write_pid(AU_PID_FILE);


    memset(&TmuVerShm,   0, sizeof(TmuAUVersion_t));
    memset(&stTmuPatVer, 0, sizeof(TmuVersion_t));
    memset(&stTmuEngVer, 0, sizeof(TmuVersion_t));

    _tmau_LoadSettings(gConfigFile);
	tmci_InitClientTable();

    /** 
     * Wait 15 seconds for other stuff to initialized 
     * This value depends on how long your device need for the initialization
     */
    sleep(15);
    while (1)
    {
        memset(&stTmuPatVer, 0, sizeof(TmuVersion_t));
        memset(&stTmuEngVer, 0, sizeof(TmuVersion_t));

        if ( _tmau_LoadSettings(gConfigFile) == 0)  /* the [au_checker] > [enable] is set as 0 */
        {
            goto Next_Run;
        }

	gAUCheckTimeMax=gAUCheckTime;
	gAUCheckTime=20;// 20 seconds when we do not get the right server.ini
       /**
         * Some arguments that we apply here:
         * -d,  --debug                 print debug output.
         * -a,  --append-output=FILE    append messages to FILE.
         * -O,  --output-document=FILE  write documents to FILE.
         * -N,  --timestamping          don't re-retrieve files unless newer than local.
         * -t,  --tries=number          Set number of retries to number.  
         *                              Specify 0 or inf for infinite retrying.
         * --proxy-user=USER            Specify the username user for authentication on a proxy server. 
         * --proxy-passwd=PASS          Specify the password for authentication on a proxy server.             
         */
       // snprintf(szWgetCmd, TM_AU_MAX_STR - 1, "wget -a /var/log/messages -O server.ini -t 3 -N %s", gAUServerURL );
       //snprintf(szWgetCmd, TM_AU_MAX_STR - 1, "wget -O /var/tmp/server.ini %s", gAUServerURL );
       sprintf(szWgetCmd, "%s", gAUServerURL );
       run_task("/usr/bin/wget","/usr/bin/wget","-O","/var/tmp/server.ini",szWgetCmd,0);      

       //system(szWgetCmd);      
         LOGDEBUG("access server.ini %s",gAUServerURL);
	unsigned long image_size;
	FILE *updateFd;
  	updateFd = fopen("/var/tmp/server.ini","rb");
	if (updateFd){
		fseek(updateFd, 0L, SEEK_END);
		image_size=ftell(updateFd);
	//	DPRINTF("/var/tmp/server.ini size=%lu\n", image_size);
		fclose(updateFd);
		
	}	
	if (image_size>0){
        if (access("/var/tmp/server.ini", R_OK) == TM_ERR_ZERO_ERROR)
        {
            TmErrorCode     tmErrCode = TM_ERR_ZERO_ERROR;

            LOGDEBUG("Start to parse the server.ini");

            /* Parse the newest AU information from newly download server.ini */
            tmErrCode = _tmau_CallIniParser("/var/tmp/server.ini", &stTmuPatVer, &stTmuEngVer);
            
	    unlink("server.ini");
	    

            if (TM_FAILURE(tmErrCode))
            {
                LOGINFO("The return of _tmau_CallIniParser is [%04x]", TM_COND_CODE(tmErrCode));
                LOGINFO("If that is caused by Internet connection problem, you might need to solve it first.");
                goto Next_Run;
            }

            /**
             * Start to read the product AU information from the shared memory.
             * The usage sample here can be leveraged by CGI program or discovery routine.
             */
            LOGDEBUG("Start to read the AU information from shared memory");

            memset(szEngine, 0x00, TM_AU_MAX_VER);

            /**
             * Call the query AU information routine in libtmGetAUInfo.a
             *
             *  The fields declared in shared memory for TmuAUVersion_t:
             * -product= supported product ID, according to AU module's assignment
             * -engine = a TmuVerion_t data type to record VSAPI engine version number
             *           for the supported product.  
             *           For example, 6.910 will store as 0x00069100
             * -pattern= a TmuVersion_t data type to record VSAPI pattern version 
             *           number for the supported product.  
             *           E.g., 1.782.00 will store as 0x0002B818(hex), i.e., 178200(dec)
             */
            memset(&TmuVerShm, 0x00, sizeof(TmuAUVersion_t));
            tmau_GetNewestAUInfo(&TmuVerShm);

            tmau_MakeVersionString(TmuVerShm.engine.version, szEngine, TM_AU_MAX_VER);

            LOGDEBUG("The result from shared memory are:");
            LOGDEBUG("            product ID is: [%d]",   TmuVerShm.product);
            LOGDEBUG("   VASPI Engine number is: [%08x] or [%s]", TmuVerShm.engine.version, szEngine);
            LOGDEBUG("  VASPI Pattern number is: [%d]", TmuVerShm.pattern.version);
  
            if ((stTmuPatVer.version > TmuVerShm.pattern.version) || (stTmuEngVer.version > TmuVerShm.engine.version))
            {
                TmuVerShm.product= 83;
                TmuVerShm.engine.version = stTmuEngVer.version;
                TmuVerShm.pattern.version= stTmuPatVer.version;

                LOGINFO("FOUND NEW AU. Will trigger client information library to update client status");
                tmci_UpdateClientStatus(&TmuVerShm);
            }
            else
            {
                LOGINFO("IDENTICAL AU. The newest AU information from AU server are identical to old ones.");
            }

            /** 
             * Remap the shared memory for recording the "product ID", "VSAPI engine", and 
             * "VSAPI pattern" for the supported products .
             */
            /* todo. add some critical section protection */
            nShmID = shm_init(TMSS_AUINFO_SHMKEY, sizeof(TmuAUVersion_t)*TM_AU_PRODUCT_NUM);
            pAUInfoPool = (TmuAUVersion_t *)shm_attach(nShmID);

            //LOGDEBUG("Remap share memory in tmAUChecker = %X\n", pAUInfoPool);

            /* todo. need to extend for multiple product later */        
            /* Start to store the product AU informaion to the shared memory. */
            LOGDEBUG("Start to store the AU information to shared memory");
            pAUInfoPool[0].product= 83;
            pAUInfoPool[0].engine.version = stTmuEngVer.version;
            pAUInfoPool[0].pattern.version= stTmuPatVer.version;
    
            shm_detach(pAUInfoPool);
	    gAUCheckTime=gAUCheckTimeMax; // now using the configured ones.
        }
	}
        else
        {
            LOGERROR("wget do not successfully download the server.ini");               
        }

Next_Run:
        LOGINFO("Sleep for another %d seconds.", gAUCheckTime);
        sleep(gAUCheckTime);
    }

    shm_fini(nShmID);
    LOGDEBUG("This program will shutdown");

    return 0;
}
void run_task(char *cmd, ...)
{
        va_list args;
        char *param[40];
        char i=0, background=0;
	char *cptr=cmd;
        pid_t  pid;
                
//printf("run_task: %s\n", cmd);  
        va_start(args, cmd);
        //for (;;) {
        for (;*cptr != '\0';cptr++) {
                if (!(param[i++] = va_arg(args, char * )))
                        break;
        }
        va_end(args);
                                                                                                                                               
        if (!strcmp(param[i-2],"&")) {
//printf("run_task: a back ground task\n");  
                background = 1;
                param[i-2] = 0;
        }

        if (!(pid = vfork())) {
		execvp(cmd, param);
                _exit(0);
        }
        else if (!background && pid > 0)
		waitpid(pid, NULL, 0);
}


