/******************************************************************************
 *
 *  (C) Copyright 2004 Trend Micro, Inc.
 *
 *  All Rights Reserved.
 *  This program is an unpublished copyrighted work which is proprietary
 *  to Trend Micro, Inc. and contains confidential information that is not
 *  to be reproduced or disclosed to any other person or entity without
 *  prior written consent from Trend Micro, Inc. in each and every instance.
 *
 *  WARNING:  Unauthorized reproduction of this program as well as
 *  unauthorized preparation of derivative works based upon the
 *  program or distribution of copies by sale, rental, lease or
 *  lending are violations of federal copyright laws and state trade
 *  secret laws, punishable by civil and criminal penalties.
 *
 *****************************************************************************/

/**
 *  @file tmClientInfo.c
 *  @brief TMOG ClientInfo Module
 */

/*
 *  Change Log
 *  2004-02-18 Eric Li Initialization version
 */
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/shm.h>
#include <stdio.h>
#include <assert.h>
#include <errno.h>
#include <arpa/inet.h>
#include "tmClientInfo.h"
#include "tmcfg.h"
#include "tmShm.h"

/* magic key to store on 'next' of allocated cell */
#define MAGIC        0xDEADCAFE
#define CONF_MODULE  "clientinfo"

static const char ModuleName[] = "tmClientInfo";

const char *MsgStatus[] = 
   { "Pending", "Windows no AV", "Unix", "Outdated AV", "Current AV"};

/*
 * MSPATCH      => M
 * TROJAN       => T
 * SPYWARE      => S
 * NETSHARE     => N
 * AVSOFT       => AS
 * AVPATTERN    => AP
 */
const char *MsgVA[] =
    { "M", "T", "S", "N", "AS", "AP"};

const char *MsgLicense[] = { "License Expired", "License OK"};


struct cell
{
    struct cell  *next;
    ClientInfo   data;
} cell;

typedef struct
{
    int shmid;

    void *addr;               /* pointer to the whole memory block */
    ClientTable *pTable;      /* pointer to the ClientTable */ 
    struct cell **freelist;   /* list of free cells */
    struct cell *pEntry;      /* pointer to the cell pool */
} tm_data_info;


/* memory pool management */
tm_data_info gPool;

/* Table pointer */
ClientTable *gpTable = NULL;

/**
  * @brief get a client record memory
  * @return pointer to the newly allocated memory
  */
static void *alloc_client_info(void)
{
     struct cell *pCell = *(gPool.freelist);

     if (pCell == NULL)
         return NULL;

     *(gPool.freelist) = pCell->next;
     pCell->next = (struct cell *)MAGIC;
     return  &(pCell->data);
}

/**
  * @brief free a client record memory
  * @param pointer to the memory being freed
  * @return pointer to the newly allocation memory
  */
static void free_client_info(void *ptr)
{
    struct cell *pCell; 

    pCell = (struct cell *)(ptr - sizeof(struct cell *));
    assert(pCell->next == (struct cell *)MAGIC);
    pCell->next = *(gPool.freelist);
    *(gPool.freelist) = pCell;
}

/**
  * @brief determine if the table list is full
  * @return 0 for not full, 1 for full
  */
static inline int is_full(void)
{
    return (gpTable->size >= CLIENT_MAX_SIZE);
}


/**
  * @brief compute the hash value by mac
  * @param mac MAC address
  * @return hash value
  */
static int mac_hash(const unsigned char *mac)
{
    unsigned long x;

    x = mac[0];
    x = (x << 2) ^ mac[1];
    x = (x << 2) ^ mac[2];
    x = (x << 2) ^ mac[3];
    x = (x << 2) ^ mac[4];
    x = (x << 2) ^ mac[5];

    x ^= x >> 8;

     return (x % CLIENT_HASH_SIZE);
}

/**
  * @brief Get the number of clients in the table
  * @return size of clients
  */
unsigned int tmci_SizeClient(void)
{
    return gpTable->size;
}

/**
  * @brief Get the size of "license-type" clients
  * @param idx license type (licensed or unlicensed)
  * @return size of clients of the specific license type
  */
unsigned int tmci_SizeLicense(uint8_t idx)
{
    if (idx >= LICENSE_SIZE)
    {
        LOGERROR("license out of range.");
        return 0;
    }
    return gpTable->license_size[idx];
}

/**
  * @brief Initialize the database
  * @return NONE
  */
void __tmci_InitClientTable(void)
{
    int i;
    struct list_head *p;

    /* Initialize pointer  */
    for (i = 0; i < CLIENT_HASH_SIZE; i++)
    {
         p = &(gpTable->hash[i]);
         p->prev = p;
         p->next = p;
    }
    p = &(gpTable->fifo);
    p->prev = p;
    p->next = p;

    /* Initialize size */
    gpTable->size = 0;

    for (i = 0; i < LICENSE_SIZE; i++)
         gpTable->license_size[i] = 0;
}

/**
  * @brief Get the pointer of client record
  * @param MAC address of the client
  * @return pointer to the client record
  */
static ClientInfo *__getPtrByMAC(unsigned char *mac)
{
    int hash;
    struct list_head *head, *pos;
    ClientInfo *pNode;

    hash = mac_hash(mac);
    head = &(gpTable->hash[hash]);

    list_for_each(pos, head)
    {
        pNode = list_entry(pos, ClientInfo, hash_head);
        if (!memcmp(CMAC(pNode), mac, MAC_SIZE))
            return pNode;
    }
    return NULL;
}


/**
  * @brief Delete the client record
  * @param mac MAC address of the client
  * @return 0 for success, negative value for failure
  */
static int __tmci_DelClient(unsigned char *mac)
{
    ClientInfo *pNode;

    if ((pNode = __getPtrByMAC(mac)) == NULL)
        return -1;

    gpTable->size--;
    gpTable->license_size[CLICENSE(pNode)]--;
    list_del(&(pNode->hash_head));
    list_del(&(pNode->fifo_head));
    free_client_info(pNode);
    return 0;
}

/**
  * @brief Delete the client record
  * @param mac MAC address of the client
  * @return 0 for success, negative value for failure
  */
int tmci_DelClient(unsigned char *mac)
{
    int ret;
    ret = __tmci_DelClient(mac);
    return ret;
}


/**
  * @brief Cleanup all records in the database
  * @return NONE
  */
void tmci_ResetClientTable(void)
{
    int i;
    ClientInfo *pNode;
    struct list_head *head, *pos;


    for(i=0; i < CLIENT_HASH_SIZE; i++)
    {
        head = &(gpTable->hash[i]);

        for (pos = head->next; pos != (head); )
        {
             pNode = list_entry(pos, ClientInfo, hash_head);
             pos = pos->next;
             list_del(&(pNode->hash_head));
             list_del(&(pNode->fifo_head));
             free_client_info(pNode);
        }
    }

    gpTable->size = 0;

    for (i = 0; i < LICENSE_SIZE; i++)
    {
         gpTable->license_size[i] = 0;
    }
}

/**
  * @brief Add a record to database; Update if same IP already exists
  * @param pClient pointer to the record.
  * @return 0 for success, negative value for failure
  */
static int __tmci_AddClient(ClientInfo *pClient)
{
    int hash;
    struct list_head *head, *fifo;
    ClientInfo *pNode = NULL;

    /* just delete the duplicate if any */
    __tmci_DelClient(CMAC(pClient));

    pNode = (ClientInfo *)alloc_client_info();

    if (!pNode)
    {
        LOGERROR("tmci_AddClient:memory allocation error.\n");
        assert(0);
        return -1;
    }
    memcpy(pNode, pClient, sizeof(ClientInfo));
     gpTable->size++;
   gpTable->license_size[CLICENSE(pNode)]++;

    /* link the node to the table */
    hash = mac_hash(CMAC(pClient));
    head = &(gpTable->hash[hash]);
    fifo = &(gpTable->fifo);
    list_add(&(pNode->hash_head), head);
    list_add(&(pNode->fifo_head), fifo);
 
    return 0;

}

/**
  * @brief Get a 'copy' of client record 
  * @param mac address of the client
  * @param pClient pointer to the buffer
  * @return 0 for success, negative value for failure
  */
int tmci_GetClient(unsigned char *mac, ClientInfo *pClient)
{
    ClientInfo *pNode;

    if ((pNode = __getPtrByMAC(mac)) == NULL)
        return -1;

    memcpy(pClient, pNode, sizeof(ClientInfo));

    return 0;
}


/**
  * @brief Set client VA information
  * @param id client ID info
  * @param va client VA info
  * @return 0 for success, negative value for failure
  */
int tmci_SetVA(struct client_id *id, struct client_va *va)
{
    ClientInfo *pNode;

    LOGINFO("tmci_SetVA(%u.%u.%u.%u) called.\n", NIPQUAD(id->ip));
    if ((pNode = __getPtrByMAC(id->mac)) == NULL)
    {
        ClientInfo info;
        memset(&info, 0, sizeof(ClientInfo));
        memcpy(&(info.idinfo), id, sizeof(struct client_id));
        memcpy(&(info.vainfo), va, sizeof(struct client_va));
        CTVA(&info) = time(NULL);
        return __tmci_AddClient(&info);
    }
    else
    {
        memcpy(&(pNode->idinfo), id, sizeof(struct client_id));
        memcpy(&(pNode->vainfo), va, sizeof(struct client_va));
        CTVA(pNode) = time(NULL);
        return 0;
    }
}


/**
  * @brief Set client AV information
  * @param id client ID info
  * @param av client AV info
  * @return 0 for success, negative value for failure
  */
int tmci_SetAV(struct client_id *id, struct client_av *av)
{
    ClientInfo *pNode;

    LOGINFO("tmci_SetAV(%u.%u.%u.%u) called.\n", NIPQUAD(id->ip));

    if ((pNode = __getPtrByMAC(id->mac)) == NULL)
    {
        ClientInfo info;
        memset(&info, 0, sizeof(ClientInfo));
        memcpy(&(info.idinfo), id, sizeof(struct client_id));
        memcpy(&(info.avinfo), av, sizeof(struct client_av));
        CTAV(&info) = time(NULL);
    LOGINFO("tmci_SetAV add a new Client(time=%lx)\n", CTAV(&info));
        return __tmci_AddClient(&info);
    }
    else
    {
        gpTable->license_size[CLICENSE(pNode)]--;

        CIP(pNode) = id->ip;
        memcpy(CMAC(pNode), id->mac, sizeof(id->mac));
        if (id->hostname[0] != '\0')    /* hostname is only from DRQ */
            memcpy(CHOSTNAME(pNode), id->hostname, sizeof(id->hostname));

        memcpy(&(pNode->avinfo), av, sizeof(struct client_av));
       CTAV(pNode) = time(NULL);
    LOGINFO("tmci_SetAV modify the old client(time=%lx)\n", CTAV(pNode));
        gpTable->license_size[CLICENSE(pNode)]++;
        return 0;
    }
}


/**
  * @brief Set client Exception information
  * @param id client ID info
  * @param exp client Exception info
  * @return 0 for success, negative value for failure
  */
int tmci_SetExp(struct client_id *id, struct client_exp * exp)
{
    ClientInfo *pNode;

    LOGINFO("tmci_SetExp(%u.%u.%u.%u) called.\n", NIPQUAD(id->ip));

    if ((pNode = __getPtrByMAC(id->mac)) == NULL)
    {
        ClientInfo info;
        memset(&info, 0, sizeof(ClientInfo));
        memcpy(&(info.idinfo), id, sizeof(struct client_id));
        memcpy(&(info.expinfo), exp, sizeof(struct client_exp));
        return __tmci_AddClient(&info);
    }
    else
    {
        /* don't touch hostname, since input has no hostname  */
        CIP(pNode) = id->ip;
        memcpy(CMAC(pNode), id->mac, sizeof(id->mac));
        memcpy(&CEXP(pNode), exp, sizeof(struct client_exp));
        return 0;
    }
}


/**
  * @brief Set client Time information
  * @param id client ID info
  * @param timeinfo client Time info
  * @return 0 for success, negative value for failure
  */
int tmci_SetTime(struct client_id *id, struct client_time *timeinfo)
{
    ClientInfo *pNode;

    LOGINFO("tmci_SetTime(%u.%u.%u.%u) called.\n", NIPQUAD(id->ip));

    if ((pNode = __getPtrByMAC(id->mac)) == NULL)
    {
        ClientInfo info;
        memset(&info, 0, sizeof(ClientInfo));
        memcpy(&(info.idinfo), id, sizeof(struct client_id));
        memcpy(&(info.timeinfo), timeinfo, sizeof(struct client_time));
        return __tmci_AddClient(&info);
    }
    else
    {
        memcpy(&(pNode->idinfo), id, sizeof(struct client_id));
        memcpy(&(pNode->timeinfo), timeinfo, sizeof(struct client_time));
        return 0;
    }
}

/**
  * @brief Get multiple clients from a specific index
  * @param pBuf  buffer to store the records
  * @param pSize [IN] unit size of pBuf [OUT] actual size returned
  * @param index starting point to get the records
  * @return 0 for success, negative value for failure
  */
int tmci_GetClientsFrom(ClientInfo *pBuf, int *pSize, int index)
{
    struct list_head *pos;
    ClientInfo *pClient;
    int count = 0, i = -1;

    list_for_each(pos, &gpTable->fifo)
    {
         i++;
         if (i < index)
             continue;

         if (count >= *pSize)
             break;
           
         pClient = list_entry(pos, ClientInfo, fifo_head);
         memcpy(pBuf+count, pClient, sizeof(ClientInfo));
         count++;
    }

    *pSize = count;

    return 0;
}

/**
  * @brief Update the status of all clients 
  * @param pAU  most recent AU versions
  * @return number of client updated
  */
int tmci_UpdateClientStatus(TmuAUVersion_t *pAU)
{
    int count = 0;
    struct list_head *pos;
    ClientInfo *pNode;

    LOGINFO("UpdateStatus() called [engine=%08X pattern=%08X]", 
             pAU->engine.version, pAU->pattern.version);

    list_for_each(pos, &gpTable->fifo)
    {
        pNode = list_entry(pos, ClientInfo, fifo_head);

        /* must have installed AV */
        if (CSTATUS(pNode) != STATUS_AVOLD && CSTATUS(pNode) != STATUS_AVOK)
            continue;

        /* Product ID must match */
        if (CPRODUCT(pNode) == AUPRODUCT(pAU)) 
        {
            /* Either pattern or engine is out of date */
	        LOGERROR("UpdateStatus: the old eng. ver=%08X, pat ver=%08X",CENGINE(pNode),CPATTERN(pNode));
            if (CENGINE(pNode)<AUENGINE(pAU) || CPATTERN(pNode)<AUPATTERN(pAU))
            {
                CSTATUS(pNode) = STATUS_AVOLD;
                CTUPDATE(pNode) = 0;
                count++;
            }
            else
            {
                CSTATUS(pNode) = STATUS_AVOK;
            }
        }
  }

  return count;
}


/*
 * @brief decide whether the client is in URL Filter exception 
 * @param mac MAC address of the client
 * @return 1 for yes, 0 for no
 */
int tmci_IsExpFilter(unsigned char *mac)
{
    ClientInfo info;

    if (tmci_GetClient(mac, &info) < 0)
        return 0;

    return (CEXP(&info).urlfilter);
}

/*
 * @brief decide whether the client is in AV Enforcer exception
 * @param mac MAC address of the client
 * @return 1 for yes, 0 for no
 */
int tmci_IsExpReport(unsigned char *mac)
{
    ClientInfo info;

    if (tmci_GetClient(mac, &info) < 0)
        return 0;

    return (CEXP(&info).report);
}

/**
  * @brief Dump the database distribution
  * @return 0 for success, negative value for failure
  */
int tmci_DumpClientTable(void)
{
    int i;
    ClientInfo *pClient;
    struct list_head *head, *pos;

    printf("[Total Client=%u]\n", gpTable->size);
    head = &gpTable->fifo;
    list_for_each(pos, head)
    {
        pClient = list_entry(pos, ClientInfo, fifo_head);

        printf("\t(%u.%u.%u.%u   ", NIPQUAD(CIP(pClient)));
        for (i = 0; i < MAC_SIZE; i++)
             printf("%02X%c", CMAC(pClient)[i], i==MAC_SIZE-1?')':'-');
        printf("\n");
    }
    printf("\n");

    for (i = 0; i < LICENSE_SIZE; i++)
    {
         printf("[%s, size=%d]\n", MsgLicense[i], gpTable->license_size[i]);
    }
    return 0;
}

/**
  * @brief Setting up the memory pool
  * @return NONE
  */
static void setup_pool(tm_data_info *pData)
{
    int size_table = sizeof(ClientTable);
    int size_free = sizeof(struct cell *);
    int size_entry = CLIENT_MAX_SIZE * sizeof(struct cell);
    int size = size_table + size_free + size_entry;
    int first_time = 0;

    if ((pData->shmid = shmget(TMSS_AV_SHMKEY, size, 0)) == -1)
    {
        LOGINFO("[setup_pool] shmget = -1");
        pData->shmid = shmget(TMSS_AV_SHMKEY, size, TMSS_AV_SHMFLG);
        first_time = 1;
        LOGINFO("[ssetup_pooln] shmget = %d", pData->shmid);
   }

  // pData->addr = (void *)shmat(pData->shmid, TMSS_AV_SHMADDR, 0);
    pData->addr = (void *)shmat(pData->shmid, (void *)0, 0);
    
    LOGDEBUG("[setup_pool] pData->addr = %p", pData->addr);

    if (pData->addr == (void *)-1)
    {
        LOGERROR("shmat aki error. errno=%d", errno);
        return;
    }

    pData->pTable = (ClientTable *)(gPool.addr);
    pData->freelist = (struct cell **)(gPool.addr + size_table);
    pData->pEntry = (struct cell *)(gPool.addr + size_table + size_free);
    gpTable = pData->pTable;
 
    if (first_time)
    {
        int i;
        struct cell *pCell;

        memset(pData->addr, 0, size);

        /* Builup the free cell list */
        *(pData->freelist) = pData->pEntry;
        pCell = pData->pEntry;
        for (i = 0; i < CLIENT_MAX_SIZE-1; i++)
        {
              pCell->next = pCell + 1;
              pCell = pCell->next;
        }
        pCell->next = NULL;

        __tmci_InitClientTable();
    }
}

/**
  * @brief Free up the memory pool
  * @return 0 for success, negative value for failure
  */
static int destroy_pool(tm_data_info *pData)
{

    tmci_ResetClientTable();

    if (shmdt(pData->addr) == -1)
    {
        LOGERROR("clientinfo shmdt error.");
    }
    return shmctl(pData->shmid, IPC_RMID, NULL);

}

/**
  * @brief Delete the client database, (probably never use!)
  * @return 0 for success, negative value for failure
  */
int tmci_DestroyClientTable(void)
{
    return destroy_pool(&gPool);
} 

void tmci_InitClientTable(void)
{
    setup_pool(&gPool);
}
static u_int32_t aspectou32TMSS(char *astr, int dns) {
  u_int32_t addr=0;
  struct in_addr s;

  // So we can input the domain name instead of the IP address.
 if(inet_aton(astr, &s))
    addr = ntohl(s.s_addr);
  else fprintf(stderr,"unable to parse \"%s\"", astr);
  return(addr);
}
/*
 *  Read persistent data from config file
 */
void tmci_ReadClient(const char *file)
{
    tmcfg_t cfg;
    const char *ptr;

    int  i, size = 0;
    char tmp[32];
    struct client_id id;
    struct client_av av;
    struct client_exp exp;
    struct client_time ti;
   
    LOGINFO("tmci_ReadClient() [%s]", file);
    if ((cfg = tmcfg_open(file, TMCFG_MODE_READ)) == NULL)
    {
        LOGERROR("failed to open config file: [%s]", file);
        return;
    }

    ptr = tmcfg_get_param_value(cfg, CONF_MODULE, "count", NULL);
    if (ptr != NULL)
        size = atoi(ptr);

    for (i = 0; i < size; i++)
    {
         memset(&id, 0, sizeof(id));
         memset(&av, 0, sizeof(av));
         memset(&exp, 0, sizeof(exp));
         memset(&ti, 0, sizeof(ti));

         sprintf(tmp, "client%d_mac", i);
         ptr = tmcfg_get_param_value(cfg, CONF_MODULE, tmp, NULL);
         if (ptr == NULL)
             continue;
         StrToMac(ptr, id.mac);

         sprintf(tmp, "client%d_license", i);
         ptr = tmcfg_get_param_value(cfg, CONF_MODULE, tmp, NULL);
         if (ptr != NULL)
             av.license = atoi(ptr);
     	//set the product ID of the new created client to 83
	//av.auversion.product=83;
    	av.status=STATUS_READ_FROM_CONFIG;
 
         sprintf(tmp, "client%d_expreport", i);
         ptr = tmcfg_get_param_value(cfg, CONF_MODULE, tmp, NULL);
         if (ptr != NULL)
             exp.report = atoi(ptr);

         sprintf(tmp, "client%d_expfilter", i);
         ptr = tmcfg_get_param_value(cfg, CONF_MODULE, tmp, NULL);
         if (ptr != NULL)
             exp.urlfilter = atoi(ptr);

         sprintf(tmp, "client%d_ip", i);
         ptr = tmcfg_get_param_value(cfg, CONF_MODULE, tmp, NULL);
         if (ptr != NULL)
         {
		 //id.ip = atol(ptr);
		 id.ip=aspectou32TMSS((char *)ptr,0);
	 }
		 
     LOGDEBUG("tmci_ReadClient: the id.ip=%8X,ip_str=%s",id.ip,ptr);
         tmci_SetAV(&id, &av);
        tmci_SetExp(&id, &exp);
         tmci_SetTime(&id, &ti);   /* just to reset av_time */
     }
    tmcfg_close(cfg);
}


/*
 *  Write client data back to config file
 */
void tmci_WriteClient(const char *file)
{
    tmcfg_t    cfg;
    int        size, i, count = 0;
    ClientInfo *pBuf;
    ClientInfo *ptr;
    char       name[32], value[32];
    

    LOGINFO("tmci_SaveSettings() [%s]", file);

    size = tmci_SizeClient();
    pBuf = (ClientInfo *)malloc(size * sizeof(ClientInfo));
    if (pBuf == NULL)
    {
        LOGERROR("memory allocation error in tmci_WriteClient().");
        return;
    }

    if ((cfg = tmcfg_open(file, TMCFG_MODE_UPDATE)) == NULL)
    {
        LOGERROR("failed to open config file for write: [%s]", file);
        return;
    }

    if (tmci_GetClientsFrom(pBuf, &size, 0) < 0 || size <= 0)
        goto END;

    tmcfg_del_module(cfg, CONF_MODULE);

    for (i = 0; i < size; i++)
    {
        ptr = pBuf + i;

        if (!CLICENSE(ptr) && !CISEXPREPORT(ptr) && !CISEXPFILTER(ptr))
            continue;

        sprintf(name, "client%d_mac", count);
        MacToStr(CMAC(ptr), value);
        tmcfg_set_param_value(cfg, CONF_MODULE, name, value, NULL);

        sprintf(name, "client%d_license", count);
        sprintf(value, "%d", CLICENSE(ptr));
        tmcfg_set_param_value(cfg, CONF_MODULE, name, value, NULL);

        sprintf(name, "client%d_expreport", count);
        sprintf(value, "%d", CISEXPREPORT(ptr));
        tmcfg_set_param_value(cfg, CONF_MODULE, name, value, NULL);

        sprintf(name, "client%d_expfilter", count);
        sprintf(value, "%d", CISEXPFILTER(ptr));
        tmcfg_set_param_value(cfg, CONF_MODULE, name, value, NULL);

	// We also keep the "named" IP. It may be change in hookAction when next MAC accesss the internet with the same/different IPs.
        sprintf(name, "client%d_ip", count);
        //sprintf(value, "%x", CIP(ptr));
	sprintf(value,"%u.%u.%u.%u", NIPQUAD(CIP(ptr)));

        tmcfg_set_param_value(cfg, CONF_MODULE, name, value, NULL);
        count++;
    }
    sprintf(value, "%d", count);
    tmcfg_set_param_value(cfg, CONF_MODULE, "count", value, NULL);

    /* set URL Filter license */
    /* Disabled, the URL Filtering License will check by router directly
    if (tmci_SizeLicense(LICENSE_OK) > 0)
        tmcfg_set_param_value(cfg,"url_filtering", "license_enable", "1", NULL);
    else
        tmcfg_set_param_value(cfg,"url_filtering", "license_enable", "0", NULL);
    */

END:
    tmcfg_flush(cfg);
    tmcfg_close(cfg);
}



