/******************************************************************************
 *
 *  (C) Copyright 2004 Trend Micro, Inc.
 *
 *  All Rights Reserved.
 *  This program is an unpublished copyrighted work which is proprietary
 *  to Trend Micro, Inc. and contains confidential information that is not
 *  to be reproduced or disclosed to any other person or entity without
 *  prior written consent from Trend Micro, Inc. in each and every instance.
 *
 *  WARNING:  Unauthorized reproduction of this program as well as
 *  unauthorized preparation of derivative works based upon the
 *  program or distribution of copies by sale, rental, lease or
 *  lending are violations of federal copyright laws and state trade
 *  secret laws, punishable by civil and criminal penalties.
 *
 *****************************************************************************/

/** 
 *  @file tmCheckOS.c
 *  @brief TMOG Discovery Module - Check OS 
 */

 
/*
 * Module to detect OS version 
 */

#include <netinet/ip_icmp.h>
#include <netdb.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>

#include "tmdpCommonDef.h"  /* common type define */
#include "log.h"

static const char ModuleName[]="ping";

#define MAXPACKET       512   /* max packet size */
#define ALIGNMENT       (4)
#define OS_UNKNOWN      0
#define OS_UNIX         1
#define OS_WINDOWS      2
#define OS_SYSTEM_API_NOT_SUPPORT 3

static u_short in_cksum(u_short *addr, int len);
static int validate(int s, char* buf, int cc);
static int SendData(__u32 ip, int s, int datalen);
static int RcvData(int s, __u32 srcip, char* buf,int* len);

/** function used to return an unique id */
static __u16 UniqueID(void)
{
   /* In Linux RH9, each thread has different PID */
   
   return (getpid() & 0xFFFF);
}

/** function used to send special ICMP packet to destination IP */
static int SendData(__u32 ip, int s, int datalen)
{
    static u_char outpack[MAXPACKET+ALIGNMENT];
    struct icmp *icp = (struct icmp *)(outpack+ALIGNMENT-((long)outpack%ALIGNMENT));
    struct sockaddr_in to;

    LOGINFO("Entering SendData()");
    memset(&to, 0, sizeof(to));
    to.sin_family = AF_INET; 
    to.sin_addr.s_addr = ip;

    icp->icmp_type = ICMP_ECHO;
    icp->icmp_code = 0x13; 
    icp->icmp_cksum = 0;
    icp->icmp_seq = 256;
    icp->icmp_id = UniqueID();

    icp->icmp_cksum = in_cksum((u_short*) icp, datalen + 8);

    LOGINFO("Leaving SendData()");
    return sendto( s, icp, datalen + 8, 0, (SA*)&to, sizeof(struct sockaddr) );
}

/** function used to receive response from special ICMP packet,
    return -1: error, 0: timeout, 1: ok */
static int RcvData(int s, __u32 srcip, char *buf, int *len)
{
    int cc;
    struct icmp *pkt;
    struct sockaddr_in from;
    socklen_t fromlen = sizeof(from);
    
    struct timeval timeout;
    int retval = 0;
    fd_set socketSet;
    
    LOGINFO("Entering RcvData()");

    /* initializes timeout structure */
    timeout.tv_sec = TIME_INTERVAL;
    timeout.tv_usec = 0;

    do
    {    
        FD_ZERO(&socketSet);
        FD_SET(s, &socketSet);
        
        /* waiting for incoming packet */
        retval = select(s+1, &socketSet, NULL, NULL, &timeout);
        
        if(retval == 0)
        {
        	 /* timeout */		
        	 LOGWARNING("Time out, fail to ping the destination machine\n");
        	 return 0;
        }
        else if (retval < 0)
        {
            LOGWARNING("error! select() return < 0");
            return 0;
        }
        
        LOGDEBUG("retval != 0");
        
        if (FD_ISSET(s, &socketSet))
        {
            if ((cc=recvfrom(s, buf, *len, 0, (SA *) &from, &fromlen)) < 0) 
            {
                 LOGERROR("AVE(CAVIT): Ping recvfrom fail.\n");
                 return 0;
            }
            
            LOGINFO("return from  recvfrom");
            
	        if (cc >= 20 + ICMP_MINLEN) 
            {
                struct iphdr *iphdr = (struct iphdr *) buf;
                pkt = (struct icmp *) (buf + (iphdr->ihl << 2)); 
                if (iphdr->saddr != srcip)
                {
                    LOGWARNING("AVE(CAVT): Ping ip not matched.\n");
                    return 0;
                }
                if (pkt->icmp_type != ICMP_ECHOREPLY)
                {
                    LOGWARNING("Error! Ping type(%u) not matched.\n", pkt->icmp_type);
                    return 0;
                }
            }
            else
            {
                LOGWARNING("AVE(CAVIT): Ping invalid packet length.\n");
                return -1;
            }
            
            if(!validate(s, buf, *len)) 
            {
               LOGERROR("validate fail");
               return -1;
            }
            
            LOGINFO("Leaving RcvData");
            
            return 1;
        }
    }
    while(1);
}

/** function used to return the checksum */
static u_short in_cksum(u_short *addr, int len)
{
    register int nleft = len;
    register u_short *w = addr;
    register u_short answer;
    register int sum = 0;

    /*
     *  Our algorithm is simple, using a 32 bit accumulator (sum),
     *  we add sequential 16 bit words to it, and at the end, fold
     *  back all the carry bits from the top 16 bits into the lower
     *  16 bits.
     */
    while( nleft > 1 )  
    {
       sum += *w++;
       nleft -= 2;
    }

    /* mop up an odd byte, if necessary */
    if( nleft == 1 ) 
    {
       u_short   u = 0;
 
       *(u_char *)(&u) = *(u_char *)w ;
       sum += u;
    }

    /*
     * add back carry outs from top 16 bits to low 16 bits
     */
    sum = (sum >> 16) + (sum & 0xffff);   /* add hi 16 to low 16 */
    sum += (sum >> 16);         /* add carry */
    answer = ~sum;            /* truncate to 16 bits */
    return (answer);
}

/**
 * @brief   Check the client's OS type.
 *
 * @param   ip:           client ip 
  * 
 * @see     
 * @remarks 
 * 
 * @return  OS_UNKNOWN, OS_UNIX, OS_WINDOWS, OS_SYSTEM_API_NOT_SUPPORT
 */
int CheckOS(__u32 ip)
{
   int             s, ret;
   struct protoent *proto;
   u_char          packet[MAXPACKET];
   int             size;
   struct          sockaddr_in servaddr;
   int nRec;

   LOGINFO("Entering ping()");
   
   memset(&servaddr, 0, sizeof(servaddr));
   servaddr.sin_family = AF_INET;
   servaddr.sin_addr.s_addr = ip;
 
   size = sizeof(packet);

   if ((proto = (struct protoent *) getprotobyname("icmp")) == NULL) 
   {
      LOGERROR("AVE(CAVIT): getprotobyname() fail. check if /etc/protocols exists.\n");
      return OS_SYSTEM_API_NOT_SUPPORT;
   }

   if ((s = socket(AF_INET, SOCK_RAW, proto->p_proto)) < 0) 
   {
       LOGERROR("AVE(CAVIT): Unable to create ping socket.\n");
       exit(0);
   }

   if (connect(s, (SA*)&servaddr, sizeof(servaddr)) < 0)
   {
       LOGERROR("AVE(CAVIT): Ping unable to connect socket.\n");
       close(s);
       return OS_UNKNOWN;
   }

   if (SendData(ip, s, 1) == -1)
   {
      /* fail to send data */
      LOGERROR("fail to send data in ping");
      
      close(s);
      ret = OS_UNKNOWN;
      return ret;
   }

   nRec = RcvData(s, ip, packet, &size);	 
	 
   if(nRec <= 0) 
      ret = OS_UNKNOWN;
   else if(*(packet+21))
      ret = OS_UNIX;
   else
      ret = OS_WINDOWS;

   close(s);
   
   LOGINFO("Leaving ping()");
   
   return ret;
}

/** function used to validate the response packet */
static int validate(int s, char* buf, int cc)
{
   struct ip *iph;
   struct icmp *icp;
   int hlen;

   iph = (struct ip *) buf;
   hlen = iph->ip_hl << 2;

   if (cc < hlen + ICMP_MINLEN) 
   {
       LOGERROR("AVE(CAVIT): Ping icmp_len not matched.\n");
       return 0;
   }

   cc -= hlen;
   icp = (struct icmp *)(buf + hlen);

   if (icp->icmp_id != UniqueID())
   {
      LOGERROR("AVE(CAVIT): Ping wrong id,id=%u(%u),"
              "type=%d,code=%d,data=%d.", icp->icmp_id, UniqueID(), 
              icp->icmp_type, icp->icmp_code, *(buf+21));
       return 0;               
   }

   return 1;
}

