/******************************************************************************
 *
 *  (C) Copyright 2004 Trend Micro, Inc.
 *
 *  All Rights Reserved.
 *  This program is an unpublished copyrighted work which is proprietary
 *  to Trend Micro, Inc. and contains confidential information that is not
 *  to be reproduced or disclosed to any other person or entity without
 *  prior written consent from Trend Micro, Inc. in each and every instance.
 *
 *  WARNING:  Unauthorized reproduction of this program as well as
 *  unauthorized preparation of derivative works based upon the
 *  program or distribution of copies by sale, rental, lease or
 *  lending are violations of federal copyright laws and state trade
 *  secret laws, punishable by civil and criminal penalties.
 *
 *****************************************************************************/

/** 
 *  @file tmSocketlHandler.c
 *  @brief TMOG Discovery Module - Socket Handler
 */

/*
 *  Change Log
 *  2004-02-17 Jim Wu    Initial Version
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <string.h>
#include <arpa/inet.h>
#include <unistd.h> /* close() */
#include "tmdpCommonDef.h"  /* common type define */
#include "tmrqRequestQueue.h"
#include "tmauIniParser.h"
#include "log.h"
#include "tmGetAUInfo.h"
#include "tmdpClientInfoWrapper.h" 
#include "tmdpMsgComposer.h"        /* header file for message composer.h */

extern void SetPendingAV(uint32_t ip, unsigned char *mac);

static const char ModuleName[]="sockethandler";

/**
 * @brief   Check the client's OS type.
 *
 * @param   ip:           client ip 
  * 
 * @see     
 * @remarks 
 * 
 * @return  OS_UNKNOWN, OS_UNIX, OS_WINDOWS, OS_SYSTEM_API_NOT_SUPPORT
 */ 
extern int CheckOS(__u32 ip); 


/************************************************************************
 * Internal function:
 ***********************************************************************/
static int SendPCCInfo(int nSocket, __u32 nAddress, char *pszMsg, int nSize, int nPort);
static int RecvPCCInfo(int nSocket, char * pszReceived, int nLen);

#define TM_DP_ENGINE_LEN 16
#define TM_DP_PATTERN_LEN 12

int tmdp_CheckClientInfo(__u32 nClientIP, int nCommandType)
{	
    uint16_t nPort = PCC_COMMUNICATION_PORT; 
    int nBuffSend = DQM_DRM_BUFFER_SIZE;           /* size of the szSend */
    char szSend[DQM_DRM_BUFFER_SIZE]; 
         
    int nBuffSize        = 0;    
    int  nReceived       = 0;
    int  nAuLangID       = 0;
    int  nStatus         = STATUS_NOAVWIN;
    int  nLicense        = LICENSE_EXPIRED;
    char szAUProductId   = 0x0;
	char szProductVer[]  = "";
	char szEngineId      = 0x0;	
	char szHostName[]    = "";	
	char *pszReceived    = NULL;	
    char szEngineVer[TM_DP_ENGINE_LEN]     = "";
    char szPatternNum[PATTERN_BUFFER_SIZE] = "";		
	unsigned char szMac[MAC_ADDRESS_SIZE]  = {0, 0, 0, 0, 0, 0};
	
	
    TmuAUVersion_t  TmuVer;  /* structure to hold the AU information, from shared memory */
  
    /* initialize socket */
    int nSocket = socket(PF_INET, SOCK_DGRAM, 0);
    
    if (nSocket < 0)
    {
       /* fail to initialize socket */
       LOGERROR("nSocket init fail %d\n", nSocket);        
  	
  	   /* default is return NOAVWIN */
       return STATUS_NOAVWIN;
    }    
	
    /* according to the command type to prepare DQM/DRM message for PCC */
    if (nCommandType == UPDATE_CLIENT_INFO)
    {         
	nBuffSend = tmdpx_GetDQMSize();

        memset(szSend, 0, DQM_DRM_BUFFER_SIZE);    
        tmdpx_PrepareDQM(szSend);  
    }
    else if (nCommandType == UPDATE_PATTERN_ENGINE)
    {	
       nBuffSend = tmdpx_GetDRMSize(); /* 32 for component info */

       memset(szSend, 0, DQM_DRM_BUFFER_SIZE);

       /* call the query AU information routine in libtmGetAUInfo.a */
       tmau_GetNewestAUInfo(&TmuVer);
       
       tmau_MakeVersionString(TmuVer.engine.version, szEngineVer, TM_DP_ENGINE_LEN);
       
       sprintf(szPatternNum, "%d", TmuVer.pattern.version);
       
       tmdpx_PrepareDRM(szSend, nClientIP, szEngineVer, szPatternNum);  
    }
  
    if (SendPCCInfo(nSocket, nClientIP, szSend, nBuffSend, nPort) < 0)
    {
       /* fail to send information to PCC client */
       LOGERROR("Fail to send packet to PCC client");

       close(nSocket);

       return STATUS_NOAVWIN;
    }
    
    LOGDEBUG("After send PCC messages");
    
    if (nCommandType == UPDATE_PATTERN_ENGINE)
    {
       /* notify PCC to update pattern/engine */
       LOGINFO("Notifying PCC to update pattern/engine...\n\n");      
       
       close(nSocket);    
       /* after the notification, just return, no need to receive response from PCC */  

       return STATUS_AVOLD; 
    }
    
    nBuffSize = tmdpx_GetPCCDRMSize();  /* get message size of PCC DRM */
	
    pszReceived = (char *)malloc(nBuffSize);
    memset(pszReceived, 0, nBuffSize);
	
    nReceived = RecvPCCInfo(nSocket, pszReceived, nBuffSize);
	
    if (nReceived == FAIL)
    {       
       /*LOGDEBUG("tmdp_CheckClientInfo -- Received data fail");*/
       free (pszReceived);
       
       /* NO PCC Client, need to check platform, use ICMP packet to check the OS */
       if (CheckOS(nClientIP) == 1)
       {
          /* found Unix platform */
          LOGINFO("*** OS is Unix/Linux\n\n");	  		  		
          nStatus = STATUS_NOAVUNIX;
       }   
       
       /* other platforms are considered as windows, no av product installed */
  
       /* update the client information table */
       tmdp_UpdateClientInfo(nStatus, nClientIP, szAUProductId, szProductVer, 
   	                         szEngineId, szEngineVer, szPatternNum, 
 	                         nAuLangID, szMac, szHostName, nLicense);  		  

       close(nSocket);
  
       return STATUS_NOAVWIN;
    }
	
    /**
     * Got reply from PCC, 
     * will call tmdp_UpdateClientInfo() to update client information in tmdpx_ParsePCCDRM()
     */       
    if (tmdpx_ParsePCCDRM(pszReceived, nClientIP) == -1)
    {
       /* fail to parse */
       free (pszReceived);

       close(nSocket);

       return STATUS_NOAVWIN;
    }
    
    LOGDEBUG("Parse DRM OK, PCC is installed");
	
    free (pszReceived);
	
    /*LOGDEBUG("tmdp_CheckClientInfo -- Received data size is %d,(%s)\n", nReceived, pszReceived);	*/    
	LOGINFO("Parse DRM OK");        
	
    close(nSocket);
	
    return STATUS_AVOK;
}

static int RecvPCCInfo(int nSocket, char * pszReceived, int nLen)
{
    int nByteRecv, nFromLen = 0;
    struct sockaddr_in fromPCCAddr;    
    struct timeval timeout;
    int retval = 0;
		
    fd_set socketSet;
    FD_ZERO(&socketSet);
    FD_SET(nSocket, &socketSet);
		
    /* initializes timeout structure */
    timeout.tv_sec = TIME_INTERVAL;
    timeout.tv_usec = 0;

    /* waiting for incoming packet */
    retval = select(nSocket+1, &socketSet, NULL, NULL, &timeout);

    LOGDEBUG("After select from PCC messages");

    if(retval == 0)
    {
       /* timeout */		
       LOGERROR("Time out, fail to get contact with PCC\n");
       return FAIL;
    }
    
    if (retval == -1)
    {
        /* fail to receive value */
        LOGERROR("fail to receive a value from select\n");
        return FAIL;
    }
    
    if (!FD_ISSET(nSocket, &socketSet))
    {
    	/* wrong socket id */
    	LOGERROR("got a wrong socket id from select\n");
    	return FAIL;
    }

    /* receive data from PCC client */
    LOGDEBUG("before calling recvfrom");
    
    nFromLen = sizeof(fromPCCAddr);	
    nByteRecv = recvfrom(nSocket, pszReceived, nLen, 0,
            (struct sockaddr *)&fromPCCAddr,&nFromLen);                        
            
    LOGDEBUG("after calling recvfrom");
     
    if (nByteRecv == -1)
    {
    	/* fail to receive data */
    	LOGERROR("fail to Received data from recvfrom\n");
    	
    	return FAIL;
    }
    
    /*LOGDEBUG("Received data size = %d\n", nByteRecv); */
    
    /*LOGDEBUG("Received data %s\n", pszReceived);*/
    
    /*LOGDEBUG("Received datagram from %s\n",inet_ntoa(fromPCCAddr.sin_addr)); */
    
    return nByteRecv;    
}

static int SendPCCInfo(int nSocket, __u32 nAddress, char *pszMsg, int nSize, int nPort)
{
    /* init UDP */	 
    int  nByetSent= 0;
    struct sockaddr_in toPCCAddr;    
 
    /* initialize socket address */    
    memset(&toPCCAddr, 0, sizeof(toPCCAddr));    
    toPCCAddr.sin_family = PF_INET;
    toPCCAddr.sin_port = htons(nPort);
    toPCCAddr.sin_addr.s_addr = nAddress; /*inet_addr(pszAddress);  */
    
    /* use sendto to send data to destination */
    nByetSent = sendto(nSocket, pszMsg, nSize, 0, (struct sockaddr *)&toPCCAddr, sizeof(toPCCAddr));
    
    if (nByetSent == -1)
    {
    	/* fail to send data */
    	LOGERROR("fail to send, byte sent is %d\n", nByetSent);    
    	return FAIL;
    }
    
    LOGDEBUG("After Send out PCC info, byte = %d\n", nByetSent);
    
    return NO_ERROR;
}
