/******************************************************************************
 *
 *  (C) Copyright 2004 Trend Micro, Inc.
 *
 *  All Rights Reserved.
 *  This program is an unpublished copyrighted work which is proprietary
 *  to Trend Micro, Inc. and contains confidential information that is not
 *  to be reproduced or disclosed to any other person or entity without
 *  prior written consent from Trend Micro, Inc. in each and every instance.
 *
 *  WARNING:  Unauthorized reproduction of this program as well as
 *  unauthorized preparation of derivative works based upon the
 *  program or distribution of copies by sale, rental, lease or
 *  lending are violations of federal copyright laws and state trade
 *  secret laws, punishable by civil and criminal penalties.
 *
 *****************************************************************************/

/**
 *  @file tmcfg.h
 *  @brief TMOG Configuration Module. utility to read / update configuration file
 */

/*
 *  Change Log
 *  2001-07-17 Kenneth Kuo  prototype draft
 *  2001-07-19 Kenneth Kuo  configuration file format changed to XML style
 *  2001-07-20 Kenneth Kuo  rename prefix from config_ to tmcfg_
 *                          remove read_string(), read_int()
 *                          add get_param_value()
 *                          remove write_string(), write_int()
 *                          add set_param_value()
 *  2001-08-10 Kenneth Kuo  reformat comment into robodoc style
 *                           #robodoc tmcfg.h ../../doc/tmcfg/tmcfg.html C HTML 
 */

#ifndef TMCFG_HEADER
#define TMCFG_HEADER

#ifdef __cplusplus
extern "C" {
#endif


  /***** tmcfg/sample-configuration-file *************************************
   * NAME
   *     Sample Configuration File
   * EXAMPLE
   *
   * <module name="scan">
   *   <param name="action" value="clean"/>
   *   <param name="clean-fail-action" value="delete"/>
   * </module>
   * <module name="smtp">
   *   <param name="enable" value="yes"/>
   *   <param name="port"   value="10025"/>
   * </module>
   **************************************************************************/

  /***** tmcfg/sample-code ***************************************************
   * NAME
   * 	Sample code using tmcfg library
   * EXAMPLE
   *
   * // reader
   * #include <tmcfg.h>
   * int main() {
   * 	char	cf_path[256] = "/etc/gatelock.conf";
   * 	char	*value, option[256];
   * 	tmcfg_t	handler = tmcfg_open(cf_fpath, TMCFG_MODE_READ);
   * 	value = tmcfg_get_param_value(handler, "scan", "action", option);
   * 	tmcfg_close(handler);
   * }
   *
   * // writer
   * #include <tmcfg.h>
   * int main() {
   * 	char	cf_path[256] = "/etc/gatelock.conf";
   * 	tmcfg_t handler = tmcfg_open(cf_fpath, TMCFG_MODE_UPDATE);
   * 	tmcfg_set_param_value(handler, "scan", "action", "clean", NULL);
   * 	tmcfg_close(handler);
   * }
   **************************************************************************/

  /****s* tmcfg/tmcfg_t ******************************************************
   * NAME
   *     tmcfg_t
   * DESCRIPTION
   *     handle to process configuration file
   * SOURCE
   */
  typedef struct tmcfg_st *tmcfg_t;
  /**************************************************************************/

  /****d* tmcfg/tm_cfg_open_mode_t *******************************************
   * NAME
   *     tmcfg_open_mode_t
   * SOURCE
   */
  typedef enum tmcfg_open_mode_enum{
    TMCFG_MODE_READ,
    TMCFG_MODE_UPDATE
  } tmcfg_open_mode_t;
  /**************************************************************************/

  /****f* tmcfg/tmcfg_open() *************************************************
   * NAME
   *     tmcfg_open()
   * SYNOPSIS
   *     tmcfg_t tmcfg_open(const char *filename, tmcfg_open_mode_t mode);
   * DESCRIPTION
   *     read configuration file
   * PARAMETERS
   *     const char *filename: IN : configuration file name
   *     const char *mode: IN : "+" for update, else for read only
   * RETURN VALUE
   *     tmcfg_t: config handle for later operations
   * ERRORS
   *     NULL: failed to read configuration file
   * NOTES
   *     Remember to call tmcfg_flush() before tmcfg_close()
   * SEE ALSO
   *     tmcfg_get_param_value(), tmcfg_set_param_value(), tmcfg_flush(), tmcfg_close()
   **************************************************************************/
  tmcfg_t tmcfg_open(const char *filename, tmcfg_open_mode_t mode);

  /****f* tmcfg/tmcfg_close() ************************************************
   * NAME
   *     tmcfg_close()
   * SYNOPSIS
   *     int tmcfg_close(tmcfg_t config);
   * DESCRIPTION
   *     release resource hold by config handle
   * PARAMETERS
   *     tmcfg_t config: IN : config handle to close
   * RETURN VALUE
   *     0: success
   * NOTES
   *     It WON'T call tmcfg_flush() automaticly before close.
   * SEE ALSO
   *     tmcfg_open(), tmcfg_flush()
   **************************************************************************/
  int tmcfg_close(tmcfg_t config);

  /****f* tmcfg/tmcfg_get_param_value() **************************************
   * NAME
   *     tmcfg_get_param_value()
   * SYNOPSIS
   *     const char *tmcfg_get_param_value(
   *       tmcfg_t config, const char *module_name, 
   *       const char *param_name, const char **param_option);
   * DESCRIPTION
   *     get parameter value from config handle
   * PARAMETERS
   *     tmcfg_t config: IN : config handle
   *     const char *module_name: IN : module_name
   *     const char *param_name: IN : parameter name
   *     const char **param_option: OUT : parameter option
   * RETURN VALUE
   *     NULL: parameter not found or error occurs
   *     const char*: param value
   * SEE ALSO
   *     tmcfg_open(), tmcfg_close()
   **************************************************************************/
  const char *tmcfg_get_param_value(tmcfg_t config, 
				     const char *module_name, 
				     const char *param_name, 
				     const char **param_option
				     );

  /****f* tmcfg/tmcfg_set_param_value() **************************************
   * NAME
   *     tmcfg_set_param_value()
   * DESCRIPTION
   *     write string value to config handle
   * PARAMETERS
   *     tmcfg_t config: IN : config handle obtained by tmcfg_open() with TMCFG_MODE_UPDATE
   *     const char *module_name: IN : module name
   *     const char *param_name: IN : parameter name
   *     const char *param_value: IN : parameter value
   *     const char *param_option: IN : parameter option
   * RETURN VALUE
   *     0: success
   *    <0: failed
   * NOTES
   *     This operation modifies config handle only, not configuration file.
   * SEE ALSO
   *     tmcfg_open(), tmcfg_flush(), tmcfg_close()
   **************************************************************************/
  int tmcfg_set_param_value(tmcfg_t config,
			     const char *module_name,
			     const char *param_name,
			     const char *param_value,
			     const char *param_option);

  /****f* tmcfg/tmcfg_del_module() ******************************************
   * NAME
   *	tmcfg_del_module()
   * DESCRIPTION
   *    delete/free one module in configure file
   * PARAMETERS
   *	tmcfg_t config: IN : config handle obtained by tmcfg_open() with TMCFG_MODE_UPDATE
   *  	module_name : IN : module name you want to delete
   * RETURN VALUE
   *	0: Cannot find any match module
   *	1: Find and delete match module
   *   -1: Error, Handler is NULL
   * NOTES
   * 	You must call tmcfg_flush() later to write the changes into file
   **************************************************************************/
  int tmcfg_del_module(tmcfg_t config, const char *module_name);



  /****f* tmcfg/tmcfg_del_param() ******************************************
   * NAME
   *	tmcfg_del_param()
   * DESCRIPTION
   *    delete/free one parameter in one module in the configure file
   * PARAMETERS
   *	tmcfg_t config: IN : config handle obtained by tmcfg_open() with TMCFG_MODE_UPDATE
   *  	module_name : IN : module name, which the parameter belongs to
   *	param_name : IN : parameter name, which you want to delete and free
   * RETURN VALUE
   *	0: Cannot find any match module or parameter
   *	1: Find and delete match parameter
   *   -1: Error, Handler is NULL
   * NOTES
   * 	You must call tmcfg_flush() later to write the changes into file
   **************************************************************************/
  int tmcfg_del_param(tmcfg_t config, const char *module_name,
				      const char *param_name);

  /****f* tmcfg/tmcfg_flush() ************************************************
   * NAME
   *     tmcfg_flush()
   * DESCRIPTION
   *     update configuration file from config handle
   * PARAMETERS
   *     tmcfg_t config: IN : config handle obtained by tmcfg_open() with TMCFG_MODE_UPDATE
   * RETURN VALUE
   *     0: success
   * ERRORS
   *    -1: failed
   * NOTES
   *     Don't call tmcfg_flush() more than or equal to twice for the same config handle!!
   * SEE ALSO
   *     tmcfg_open(), tmcfg_set_param_value(), tmcfg_close()
   **************************************************************************/
  int tmcfg_flush(tmcfg_t config);


#ifdef __cplusplus
} /* extern "C" */
#endif

#endif
