/******************************************************************************
 *
 *  (C) Copyright 2004 Trend Micro, Inc.
 *
 *  All Rights Reserved.
 *  This program is an unpublished copyrighted work which is proprietary
 *  to Trend Micro, Inc. and contains confidential information that is not
 *  to be reproduced or disclosed to any other person or entity without
 *  prior written consent from Trend Micro, Inc. in each and every instance.
 *
 *  WARNING:  Unauthorized reproduction of this program as well as
 *  unauthorized preparation of derivative works based upon the
 *  program or distribution of copies by sale, rental, lease or
 *  lending are violations of federal copyright laws and state trade
 *  secret laws, punishable by civil and criminal penalties.
 *
 *****************************************************************************/

/** 
 *  @file tmrqRequestQueue.c
 *  @brief TMOG Request Module
 */

/*
 *  Change Log
 *  2004-02-18 Jim Wu    Initialization Version
 */
#define NDEBUG
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>

#include "log.h"          /* for debug logs */
#include "tmdpCommonDef.h"  /* common type define */
#include "tmrqRequestQueue.h"
#include "tmSemaphore.h"

int *g_pQHead = 0;
int *g_pQTail = 0;
MESSAGE_ITEM *g_pMessageItem = 0;

//#define __DEBUG__
/* function prototype */
static const char ModuleName[]="tmReqQ";

static void tmrq_queue_lock(struct tm_sem_st *pQueueLock)
{
    tm_lock_create(REQUEST_QUEUE_SEMLOCK_KEY, pQueueLock);
    if (tm_lock(pQueueLock) == -1)
    {
        LOGERROR("tmrq_queue_lock(), lock shared memroy error");
    }
}

static void tmrq_queue_unlock(struct tm_sem_st *pQueueLock)
{
    if (tm_unlock(pQueueLock) == -1)
    {
        LOGERROR("tmrq_queue_unlock(), unlock shared memroy error");
    }
    tm_lock_close(pQueueLock);
}

void tmrq_ResetQueue(void)
{
    int nIdx = 0;
    struct tm_sem_st QueueLock;
    
    LOGINFO("(%d) tmrq_ResetQueue", getpid());	
    
    tmrq_queue_lock(&QueueLock);
    *g_pQHead = 0;
    *g_pQTail = 0;    
    
    for (nIdx = 0; nIdx < REQUEST_QUEUE_SIZE; nIdx++)
    {        
        MESSAGE_ITEM * pQueueItem = g_pMessageItem + nIdx;        
        
        LOGDEBUG("(%d) After assign pQueueItem %p", getpid(), pQueueItem);	
    		
        strcpy(pQueueItem->szClientIP, "");
        
        LOGDEBUG("(%d) After assign szClientIP", getpid());	
        
    	pQueueItem->nCommandType = -1;
    	
    	LOGDEBUG("(%d) After assign nCommandType", getpid());	
    }
    
    LOGDEBUG("(%d) Leaving tmrq_ResetQueue", getpid());	
    
    tmrq_queue_unlock(&QueueLock);

}

int tmrq_InsertQueue(MESSAGE_ITEM *pMessageItem)
{
    int nNextPos;
    MESSAGE_ITEM *pNextItem;
    struct tm_sem_st QueueLock;
    
    tmrq_queue_lock(&QueueLock);

    /* insert item in the request queue */
    if (tmrq_IsQueueFull())
    {
       tmrq_queue_unlock(&QueueLock);
       return -1;
    }
	
    LOGINFO("(%d) before insert g_pQTail is %d", getpid(), *g_pQTail);
	
    nNextPos = *g_pQTail;
    
    pNextItem = g_pMessageItem + nNextPos;
    
    strncpy(pNextItem->szClientIP, pMessageItem->szClientIP, IP_SIZE);
    
    pNextItem->nCommandType = pMessageItem->nCommandType;	
    
    *g_pQTail = (*g_pQTail)++;
    *g_pQTail = (*g_pQTail) % REQUEST_QUEUE_SIZE;		
	
    LOGINFO("(%d) after insert g_pQTail now is %d", getpid(), *g_pQTail); 

    tmrq_queue_unlock(&QueueLock);
    return 0;
}

int tmrq_IsQueueFull(void)
{
    int nHeadPos = *g_pQHead;
    int nTailPos = *g_pQTail;
    int nNextPos;
    
    /* don't lock here, already locked in tmrq_InsertQueue() */
	
    LOGINFO("(%d) Check if Queue is full", getpid()); 	
    LOGDEBUG("(%d) head pos is %d, tail pos is %d", getpid(), nHeadPos, nTailPos);

    nNextPos = (nTailPos + 1) % REQUEST_QUEUE_SIZE;
	
    if (nNextPos == nHeadPos)
    {
       /* queue is full */
      LOGINFO("(%d) queue is full", getpid());

      return 1;
    }
	
    LOGINFO("(%d) queue is not full", getpid()); 
	
    return 0;
}

int tmrq_IsQueueEmpty(void)
{
    int nHeadPos = *g_pQHead;
    int nTailPos = *g_pQTail;
    struct tm_sem_st QueueLock;
    
    tmrq_queue_lock(&QueueLock);

    LOGINFO("(%d) Check if it is empty", getpid()); 	
    LOGDEBUG("(%d) head pos is %d, tail pos is %d", getpid(), nHeadPos, nTailPos);

    if (nHeadPos == nTailPos)
    {
       /* queue is empty */
#ifdef __DEBUG__			
       LOGINFO("queue is empty \n");
#endif
	   tmrq_queue_unlock(&QueueLock);
       return 1;
    }
	
	tmrq_queue_unlock(&QueueLock);
    return 0;
}

int tmrq_IsItemExist(MESSAGE_ITEM * pMessageItem)
{
    int nIdx = 0;
    struct tm_sem_st QueueLock;

    LOGINFO("(%d) Check if item exist", getpid());     
	
	tmrq_queue_lock(&QueueLock);
    for (nIdx = 0; nIdx < REQUEST_QUEUE_SIZE; nIdx++)
    {
        MESSAGE_ITEM * pQueueItem = g_pMessageItem + nIdx;
        
        if ((pQueueItem->nCommandType != -1) && (strcmp(pQueueItem->szClientIP, "") != 0))
		{
		    /* item not empty */
    		LOGDEBUG("(%d) Queue item szClientIP is %s", getpid(), pQueueItem->szClientIP);					
    		LOGDEBUG("(%d) pMessageItem item szClientIP is %s", getpid(), pMessageItem->szClientIP);					
    		
    		LOGDEBUG("(%d) Queue item nCommandType is %d", getpid(), pQueueItem->nCommandType);					
    		LOGDEBUG("(%d) pMessageItem item nCommandType is %d", getpid(), pMessageItem->nCommandType);					
    		
            if ((strcmp(pQueueItem->szClientIP, pMessageItem->szClientIP) == 0) &&
    				(pQueueItem->nCommandType == pMessageItem->nCommandType))
            {
               /* item found */					
               LOGINFO("(%d) item found", getpid());					
               tmrq_queue_unlock(&QueueLock);
               return 1;
            }
        }
    }

    LOGINFO("(%d) No item found", getpid());	
	tmrq_queue_unlock(&QueueLock);
    return 0;
}

int tmrq_DeleteLastItem(MESSAGE_ITEM *pMessageItem)
{
    int nIdx = 0;
    struct tm_sem_st QueueLock;

    LOGINFO("(%d) Delete Last Item from Queue", getpid());
	
	tmrq_queue_lock(&QueueLock);
    for (nIdx = 0; nIdx < REQUEST_QUEUE_SIZE; nIdx++)
    {
        MESSAGE_ITEM * pQueueItem = g_pMessageItem + nIdx;
        
        if ((pQueueItem->nCommandType != -1) && (strcmp(pQueueItem->szClientIP, "") != 0))
		{
		    /* item not empty */
    		LOGDEBUG("(%d) Queue item szClientIP is %s", getpid(), pQueueItem->szClientIP);					
    		LOGDEBUG("(%d) pMessageItem item szClientIP is %s", getpid(), pMessageItem->szClientIP);					
    		
    		LOGDEBUG("(%d) Queue item nCommandType is %d", getpid(), pQueueItem->nCommandType);					
    		LOGDEBUG("(%d) pMessageItem item nCommandType is %d", getpid(), pMessageItem->nCommandType);					
    		
            if ((strcmp(pQueueItem->szClientIP, pMessageItem->szClientIP) == 0) &&
    				(pQueueItem->nCommandType == pMessageItem->nCommandType))
            {
               /* item found */					
               LOGDEBUG("(%d) item found, reset it", getpid());					
               strcpy(pQueueItem->szClientIP, "");
               pQueueItem->nCommandType = -1;	
               tmrq_queue_unlock(&QueueLock);
               return 1;
            }
        }
    }

    LOGINFO("(%d) No item found", getpid());	
    tmrq_queue_unlock(&QueueLock);
    return 0;	
}

int tmrq_GetLastItem(MESSAGE_ITEM *pMessageItem)
{
    MESSAGE_ITEM * pQueueItem;
    struct tm_sem_st QueueLock;

    LOGINFO("(%d) Get Last Item from Queue", getpid());
    tmrq_queue_lock(&QueueLock);
	
    pQueueItem = g_pMessageItem + (*g_pQHead);
    
    LOGINFO("(%d) Item from Queue, IP is %s, command is %d", getpid(), pQueueItem->szClientIP, pQueueItem->nCommandType);
	
    strncpy(pMessageItem->szClientIP, pQueueItem->szClientIP, IP_SIZE);
    pMessageItem->nCommandType = pQueueItem->nCommandType;	
    
    *g_pQHead = (*g_pQHead)++;
    *g_pQHead = (*g_pQHead) % REQUEST_QUEUE_SIZE;		

    LOGINFO("(%d) head pos now is %d, tail pos is %d", getpid(), *g_pQHead, *g_pQTail);
	tmrq_queue_unlock(&QueueLock);
    return 0;	
}

MESSAGE_ITEM * RequestQueue_Find(int nItem)
{
    /* return the message item for the nth item */	
    MESSAGE_ITEM * pMessageItem = g_pMessageItem + nItem;

    return pMessageItem;
}

void tmrq_SetMsgItemHead(MESSAGE_ITEM *pMessageItem)
{
    struct tm_sem_st QueueLock;
    
    tmrq_queue_lock(&QueueLock);
    g_pMessageItem = pMessageItem;
    tmrq_queue_unlock(&QueueLock);
}

void tmrq_SetQueueHead(int *pIndex)
{
    struct tm_sem_st QueueLock;
    
    tmrq_queue_lock(&QueueLock);
    
    LOGINFO("(%d) Enter tmrq_SetQueueHead %d", getpid(), *pIndex);
    
    g_pQHead = pIndex;
    
    LOGINFO("(%d) Leave tmrq_SetQueueHead %d", getpid(),*g_pQHead);
    
    tmrq_queue_unlock(&QueueLock);
}

void tmrq_SetQueueTail(int *pIndex)
{
    struct tm_sem_st QueueLock;
    
    tmrq_queue_lock(&QueueLock);
    
    LOGINFO("(%d) Enter tmrq_SetQueueTail %d", getpid(),*pIndex);
    
    g_pQTail = pIndex;    
    
    LOGINFO("(%d) Leave tmrq_SetQueueTail %d", getpid(),*g_pQTail);
    
    tmrq_queue_unlock(&QueueLock);
}
