/******************************************************************************
 *
 *  (C) Copyright 2004 Trend Micro, Inc.
 *
 *  All Rights Reserved.
 *  This program is an unpublished copyrighted work which is proprietary
 *  to Trend Micro, Inc. and contains confidential information that is not
 *  to be reproduced or disclosed to any other person or entity without
 *  prior written consent from Trend Micro, Inc. in each and every instance.
 *
 *  WARNING:  Unauthorized reproduction of this program as well as
 *  unauthorized preparation of derivative works based upon the
 *  program or distribution of copies by sale, rental, lease or
 *  lending are violations of federal copyright laws and state trade
 *  secret laws, punishable by civil and criminal penalties.
 *
 *****************************************************************************/

/**
 *  @file tmURLFilterLicense.c
 *  @brief TMSS URL Filtering Module - License Check
 */

/*
 *  Change Log
 *  2004-05-17 Henry Lin    Initialization Version
 */

#include <stdio.h> /* snprintf() */
#include <string.h> /* strlen() */
#include <unistd.h> /* close() */
#include <stdlib.h> /* atoi() */
#include "tmSocketClient.h"
#include "tmURLFilterConf.h" /* url_filter_config_t */
#include "tmFunction.h" /* tmGethostbyname() */
#include "tmcfg.h"
#include "log.h"
#include "tmShm.h"
#include "tmSemaphore.h"
#include "tmURLFilterLicense.h"
#include "tmGetHardwareInfo.h"
#include <time.h>

#define LIC_CHK_SHORT               300 /* 5 minutes */
#define LIC_CHK_SHORT_TIME          12  /* 12 times */
#define LIC_CHK_SHORT_RETRY         120 /* 2 minutes */

static const char ModuleName[] = "tmURLFilterLicense";

static int GetBoxID(char *strBoxId)
{
    return tmGetBoxID(strBoxId);
}

static int GetVendorID(char *strVendorId)
{
    return tmGetVendorID(strVendorId);
}

/* no license : 0, has license: 1, format error: -1 */
static int tm_entitlement_parse_result(const char *pszReceiveBuf)
{
    char *pStr = NULL;
    char szRetVal[TM_URL_FILTER_ENTITLEMENT_RET_VALUE_LEN + 1];
    int nRetVal;
    int i;
    
   
    pStr = strstr(pszReceiveBuf, TM_URL_FILTER_ENTITLEMENT_RET_MODULE);

    if (pStr == NULL)
    {
        LOGERROR("tm_entitlement_parse_result(), can't find module %s",TM_URL_FILTER_ENTITLEMENT_RET_MODULE);
        return -1;
    }
    
    pStr = strstr(pStr, TM_URL_FILTER_ENTITLEMENT_RET);
    if (pStr == NULL)
    {
        LOGERROR("tm_entitlement_parse_result(), can't find RET");
        return -1;
    }
    
    pStr += sizeof(TM_URL_FILTER_ENTITLEMENT_RET);
    
    for (i = 0; i < TM_URL_FILTER_ENTITLEMENT_RET_VALUE_LEN; i++)
    {
        if (*pStr > '9' || *pStr < '0')
        {
            LOGERROR("tm_entitlement_parse_result(), RET follow by non-digital");
            return -1;
        }
        szRetVal[i] = *pStr++;
    }
    
    if (*pStr != '\r' && *pStr != '\n')
    {
        LOGERROR("tm_entitlement_parse_result(), followed RET=XXX is not a newline");
        return -1;
    }
    
    szRetVal[TM_URL_FILTER_ENTITLEMENT_RET_VALUE_LEN] = '\0';
    
    nRetVal = atoi(szRetVal);
    
    switch (nRetVal)
    {
        case 200:
        case 201:
            return 1;
            break;
        case 210:
        case 211:
        case 212:
        case 214:
        case 404:
        case 410:
        case 411:
        case 500:
        default:
            return 0;
            break;
    }   
}

static int UpdateLicenseConfInMemory(unsigned short nLicenceStatus)
{
    int nUrlFilterConfShmid;
    struct tm_sem_st UrlFilterConfLock;
    url_filter_config_t *pUrlFilterConf;
    
    nUrlFilterConfShmid = shm_init(TMSS_URL_FILTER_CONFSHMKEY, sizeof(url_filter_config_t));
    tm_lock_create(TMSS_URL_FILTER_CONFSEMLOCKKEY, &UrlFilterConfLock);
    if (tm_lock(&UrlFilterConfLock) == -1)
    {
        LOGERROR("UpdateLicenseConfInMemory(), lock shared memroy error");
    }
    pUrlFilterConf = (url_filter_config_t*)shm_attach(nUrlFilterConfShmid);
    
    pUrlFilterConf->nUrlLicenseEnable = nLicenceStatus;
    
    // Yes, we shall also update the nUrlFilterExceptionEnable the same value as nUrlLicenseEnable
    pUrlFilterConf->nUrlFilterExceptionEnable = nLicenceStatus;
    
    shm_detach(pUrlFilterConf);
    if (tm_unlock(&UrlFilterConfLock) == -1)
    {
        LOGERROR("UpdateLicenseConfInMemory(), unlock shared memroy error");
    }
    tm_lock_close(&UrlFilterConfLock);
    
    return 0;
}

static int UpdateLicenseConfInFile(unsigned short nLicenseStatus, const char *pszConfigFilePath)
{
    tmcfg_t cfg = NULL;
    char szStatus[2];
    
    szStatus[1] = '\0';
    szStatus[0] = '0' + nLicenseStatus;
    
    if( NULL == (cfg=tmcfg_open(pszConfigFilePath,TMCFG_MODE_UPDATE)) )
    {
        LOGERROR("UpdateLicenseConfInFile(), open config file fail!");
        return -1;
    }
    if (tmcfg_set_param_value(cfg, CFG_MODULE_URL_FILTER, CFG_PARAM_URL_LICENSE_ENABLE, szStatus,NULL) != 0)
    {
        LOGERROR("UpdateLicenseConfInFile(), write config file for url filtering license fail!");
        tmcfg_close(cfg);
        return -1;
    }
    
   // Yes, we shall also update the nUrlFilterExceptionEnable the same value as nUrlLicenseEnable
     if (tmcfg_set_param_value(cfg, CFG_MODULE_URL_FILTER, CFG_PARAM_URL_FILTER_EXCEPTION_ENABLE, szStatus,NULL) != 0)
    {
        LOGERROR("UpdateLicenseConfInFile(), write config file for url filtering exception enabe fail!");
        tmcfg_close(cfg);
        return -1;
    }
    
    tmcfg_flush(cfg);
    tmcfg_close(cfg);
    
    return 0;
}
extern int sendUrlFilterConfToKernel(url_filter_config_t *pUrlFilterConf);
        
/**                                                                          
  * @brief Send a URL Filtering License Request to Trend Entitlement Server
  * @param pszConfigFilePath configuration file path
  * @return TM_URL_FILTER_ENTITLEMENT_CHECK_ERROR check license fail
  * @return 0 check license ok
  */ 
int tmURLFilterLicenseCheck(char* pszConfigFilePath, int *pnLicStatus)
{
    char szBuf[TM_PACKET_BUF_SIZE];
    char szReceiveBuf[TM_PACKET_BUF_SIZE];
    char str[INET6_ADDRSTRLEN];
    char strBoxId[TMSS_MAX_BUF_BOX_ID];
    char strVendorId[TMSS_MAX_BUF_VENDER_ID];
    url_filter_config_t UrlFilterConfig;
    struct hostent *tm_entitlement_hostent;
    struct in_addr tm_entitlement_ip;
    int t_sd;
    int nRetVal;
    
    CopyUrlConfToLocal(&UrlFilterConfig);
    
    tm_entitlement_hostent = tmGethostbyname(UrlFilterConfig.szEntitleHost);
     
    if (tm_entitlement_hostent == NULL)
    {
       LOGERROR("tmURLFilterLicenseCheck(), tmGethostbyname for  EntitleHost failed");
        return TM_URL_FILTER_ENTITLEMENT_CHECK_ERROR;
   }
   
   
    tm_entitlement_ip.s_addr = inet_addr(inet_ntop(tm_entitlement_hostent->h_addrtype,tm_entitlement_hostent->h_addr_list[0], str, sizeof(str)));
    
    GetBoxID(strBoxId);
    GetVendorID(strVendorId);
    
    snprintf(szBuf, TM_PACKET_BUF_SIZE - 1, TM_URL_FILTER_LICENCE_REQ_PACKET,
             UrlFilterConfig.szEntitlePath,
             strBoxId, strVendorId,
             TM_URL_FILTER_COMPONENT_ID,
             UrlFilterConfig.szEntitleHost, UrlFilterConfig.nEntitlePort);
    
    t_sd = tm_url_filter_tcp_connect(&tm_entitlement_ip, UrlFilterConfig.nEntitlePort);
    
    if(t_sd == TM_URL_FILTER_SOCK_FAILURE)
    {
        LOGERROR("tmURLFilterLicenseCheck(), open connection to entitlement server fail");
        return TM_URL_FILTER_ENTITLEMENT_CHECK_ERROR;
    }

    if(tm_url_filter_sock_srv_write(t_sd, szBuf, strlen(szBuf), 60) == TM_URL_FILTER_SOCK_FAILURE)
    {
        (void)close(t_sd);
        LOGERROR("tmURLFilterLicenseCheck(), send packet to entitlement server fail");
        return TM_URL_FILTER_ENTITLEMENT_CHECK_ERROR;
    }
    LOGDEBUG("tmURLFilterLicenseCheck(), transmit: %s",szBuf);

    if(tm_url_filter_sock_srv_read(t_sd, szReceiveBuf,
                          TM_PACKET_BUF_SIZE, 60) == TM_URL_FILTER_SOCK_FAILURE)
    {
        (void)close(t_sd);
        LOGERROR("tmURLFilterLicenseCheck(), receive packet from entitlement server fail");
        return TM_URL_FILTER_ENTITLEMENT_CHECK_ERROR;
    }
    (void)close(t_sd);
    LOGDEBUG("tmURLFilterLicenseCheck(), receive: %s",szReceiveBuf);
    
    nRetVal = tm_entitlement_parse_result(szReceiveBuf);
    
    if (nRetVal == 1 && UrlFilterConfig.nUrlLicenseEnable == 0)
    {
        UpdateLicenseConfInMemory(1);
        UpdateLicenseConfInFile(1, pszConfigFilePath);
	CopyUrlConfToLocal(&UrlFilterConfig);
	sendUrlFilterConfToKernel(&UrlFilterConfig);
        *pnLicStatus = 1;
        LOGDEBUG("tmURLFilterLicenseCheck(), enable url filtering license");
    }
    else if (nRetVal == 0 && UrlFilterConfig.nUrlLicenseEnable == 1)
    {
        UpdateLicenseConfInMemory(0);
        UpdateLicenseConfInFile(0, pszConfigFilePath);
	CopyUrlConfToLocal(&UrlFilterConfig);
	sendUrlFilterConfToKernel(&UrlFilterConfig);
        *pnLicStatus = 0;
        LOGDEBUG("tmURLFilterLicenseCheck(), disable url filtering license");
    }
    else if (nRetVal == -1)
    {
        *pnLicStatus = UrlFilterConfig.nUrlLicenseEnable;
        LOGERROR("tmURLFilterLicenseCheck(), receive packet from entitlement server fail");
    }
    else
    {
        *pnLicStatus = UrlFilterConfig.nUrlLicenseEnable;
        LOGDEBUG("tmURLFilterLicenseCheck(), license not change");
    }

    return 0;
}

/**                                                                          
  * @brief Do the license check every 5 minutes for 12 times
  * @param pszConfigFilePath configuration file path
  */ 
void *tmURLFilterLicenseCheckForActivate(char* pszConfigFilePath)
{
    int i;
    int nLicStatus = -1;
    int nRetVal;
    
    for (i = 0; i < LIC_CHK_SHORT_TIME; i++)
    {
        LOGERROR("tmURLFilterLicenseCheckForActivate(), Start License Check for Activation: %d",i);
        sleep(LIC_CHK_SHORT);
        
        nRetVal = tmURLFilterLicenseCheck(pszConfigFilePath, &nLicStatus);
        
        LOGERROR("tmURLFilterLicenseCheckForActivate(), nLicStatus: %d",nLicStatus);
        
        if (nRetVal == 0 && nLicStatus == 1)
        {
            LOGERROR("tmURLFilterLicenseCheckForActivate(), License enabled. Stop checking");
            break;
        }
    }
    LOGDEBUG("tmURLFilterLicenseCheckForActivate(), End License Check for Activation");
    return NULL;
}

/**                                                                          
  * @brief Do the license check every 2 minutes until success
  * @param pszConfigFilePath configuration file path
  */ 

// Loop for ever until we get the response from the entitlement server.  The value 
// of nLicStatus shall be 0 or 1 to indicate the lience is enalbed (1) or not (0)
int tmURLFilterLicenseCheckForStartup(char* pszConfigFilePath)
{
    int nLicStatus = -1;
    int nRetVal;
    
    LOGERROR("tmURLFilterLicenseCheckForStartup(), Start License Check for Startup");
   while (1)
    {
  
        nRetVal = tmURLFilterLicenseCheck(pszConfigFilePath, &nLicStatus);
        LOGERROR("tmURLFilterLicenseCheckForStartup(), nLicStatus: %d",nLicStatus);
        
        if (nRetVal == 0)
        {
            LOGERROR("tmURLFilterLicenseCheckForStartup(), License check success. Stop checking");
            return nLicStatus;
        }
        sleep(LIC_CHK_SHORT_RETRY);
    }
    LOGDEBUG("tmURLFilterLicenseCheckForStartup(), End License Check for startup");
	return 0;
}

#include <signal.h>
#include "tmPidUtil.h"
static volatile int GO_HELL=0;
#define MAX_FILE_PATH  128
char gConfigFile[MAX_FILE_PATH] = {0};
#define PIDFILE             "/var/run/tmLicChk.pid"

static int gnURLFilterLicChkInterval = 86400;
static void sig_hup(int signo)
{
   // LoadSettings("by signal SIGHUP");
}

static void sig_usr1(int signo)
{
	tmURLFilterLicenseCheckForActivate(gConfigFile);
} 

unsigned int GetURLFilterLicChkInterval()
{
    url_filter_config_t UrlFilterConfig;
    
    CopyUrlConfToLocal(&UrlFilterConfig);
    
    return UrlFilterConfig.nLicChkInterval;
}

static void go_hell_handler(int signo)
{
    GO_HELL=1;
}
unsigned long last = 0;   /* last time when config reloaded */
unsigned long now;
int nLicStatus = -1;

int main(int argc, char *argv[])
{

/*******************************************************************************
* Porting Required Section Start
*******************************************************************************/
  //  unsigned long last = 0;   /* last time when config reloaded */
  //  unsigned long now;
  //  int nLicStatus = -1;


  //  int nUrlFilterStatisticShmid;
  //  int nUrlFilterConfShmid;


    if (argc != 2)
    {
        fprintf(stderr, "Usage: %s [config file]\n", argv[0]);
        exit(1);
    }

    /* save config file path */
    strncpy(gConfigFile, argv[1], sizeof(gConfigFile));
    gConfigFile[sizeof(gConfigFile)-1] = 0;

/*******************************************************************************
* Porting Required Section End
*******************************************************************************/

    LOGOPEN("FLTLIC");
    LOGINFO("init");
    LOGDEBUG_ON();

    argc -= optind;
    argv += optind;

/*******************************************************************************
* Porting Required Section Start
*******************************************************************************/

   
    
    /* load URL Filtering License Check Interval to a local static variable for later use*/
    gnURLFilterLicChkInterval = GetURLFilterLicChkInterval();

 
    LOGDEBUG("hook SIGHUP");
    { /* monitor for config reload signal */
        if(SIG_ERR==signal(SIGHUP, sig_hup))
        {
            LOGERROR("signal: %m");
            return 1;
        }
    }

   LOGDEBUG("hook SIGUSR1");
    { // monitor for license activation check signal
        if(SIG_ERR==signal(SIGUSR1, sig_usr1))
        {
            LOGERROR("signal: %m");
            return 1;
        }
    }

/*******************************************************************************
* Porting Required Section End
*******************************************************************************/

    LOGDEBUG("hook SIGTERM");
    { /* monitor for shutdown signal */
        if(SIG_ERR==signal(SIGTERM, go_hell_handler))
        {
            LOGERROR("signal: %m");
            return 1;
        }
    }

    LOGDEBUG("hook SIGINT");
    { /* monitor for shutdown signal */
        if(SIG_ERR==signal(SIGINT, go_hell_handler))
        {
            LOGERROR("signal: %m");
            return 1;
        }
    }

    LOGINFO("create pid file: %s", PIDFILE);
    write_pid(PIDFILE);

    tmURLFilterLicenseCheckForStartup(gConfigFile);
    
    // We then perform regular check for filter license.
	
/*******************************************************************************
* Porting Required Section Start
*******************************************************************************/
    for(;GO_HELL==0;)
    {
    time(&now);
    /* Check URL Filtering License from Entitlement server */
    if (now - last >= gnURLFilterLicChkInterval)
    {
	while(1){
	     LOGERROR("tmURLFilterLicense, Time to Check URL Filtering License. now=%lu,last=%lu,interval=%u",now,last,gnURLFilterLicChkInterval);
             if (tmURLFilterLicenseCheck(gConfigFile, &nLicStatus) == 0)
	     {
		     LOGERROR("tmURLFilterLicense, License Check OK. Status=%d",nLicStatus);
		     last = now;
			 break;
	     }
	     else
	     {
		     LOGERROR("tmURLFilterLicense, Check url filtering license fail");
	     }
	sleep(300);
	}
    }
    else
    {
	   // LOGDEBUG("tmURLFilterLicense, NOT Time to Check URL Filtering License. now=%lu,last=%lu,interval=%u",now,last,gnURLFilterLicChkInterval);
    }
    sleep(5);// sleep 5 seconds.
    }
/*******************************************************************************
* Porting Required Section End
*******************************************************************************/

    LOGINFO("remove pid file: %s", PIDFILE);
    remove_pid(PIDFILE);



    LOGINFO("shutdown");
    return 0;
}
