/******************************************************************************
 *
 *  (C) Copyright 2004 Trend Micro, Inc.
 *
 *  All Rights Reserved.
 *  This program is an unpublished copyrighted work which is proprietary
 *  to Trend Micro, Inc. and contains confidential information that is not
 *  to be reproduced or disclosed to any other person or entity without
 *  prior written consent from Trend Micro, Inc. in each and every instance.
 *
 *  WARNING:  Unauthorized reproduction of this program as well as
 *  unauthorized preparation of derivative works based upon the
 *  program or distribution of copies by sale, rental, lease or
 *  lending are violations of federal copyright laws and state trade
 *  secret laws, punishable by civil and criminal penalties.
 *
 *****************************************************************************/

/**
 *  @file tmURLFilterConf.c
 *  @brief TMOG URL Filtering Configuration Loading Module
 */

/*
 *  Change Log
 *  2004-03-03 Henry Lin    Initialization version
 */

//#define NDEBUG
#include <sys/shm.h>
#include <sys/ipc.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h> /* sprintf() */
#include "time.h"
#include "log.h"
#include "tmShm.h"
#include "tmSemaphore.h"
#include "tmURLFilter.h"
#include "tmURLFilterConf.h"
#include "tmcfg.h"
#include <linux/netfilter_ipv4/ipt_webstr.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include <unistd.h>
#include <stdlib.h>


static const char ModuleName[] = "tmURLFilterConf";
unsigned int urlFilterConfigVersion=0;
/**
  * @brief Copy current URL Filtering Config setting from shared memroy to a local variable
  * @param pUrlFilterConfig Output - current URL Filtering Config setting
  * @return none
  */
void CopyUrlConfToLocal(url_filter_config_t *pUrlFilterConfig)
{
    int                     nUrlFilterConfShmid;
    url_filter_config_t     *pUrlFilterConf; // point to shared memroy
    struct tm_sem_st        UrlFilterConfLock;
    
    nUrlFilterConfShmid = shm_init(TMSS_URL_FILTER_CONFSHMKEY, sizeof(url_filter_config_t));
    tm_lock_create(TMSS_URL_FILTER_CONFSEMLOCKKEY, &UrlFilterConfLock);
    if (tm_lock(&UrlFilterConfLock) == -1)
    {
        LOGERROR("CopyUrlConfToLocal(), lock shared memroy error");
    }
    pUrlFilterConf = (url_filter_config_t*)shm_attach(nUrlFilterConfShmid);
    
    memcpy(pUrlFilterConfig,pUrlFilterConf,sizeof(url_filter_config_t));

    shm_detach(pUrlFilterConf);
    if (tm_unlock(&UrlFilterConfLock) == -1)
    {
        LOGERROR("CopyUrlConfToLocal(), unlock shared memroy error");
    }
    tm_lock_close(&UrlFilterConfLock);
}


/**
  * @brief Parse Category from Config file setting string
  * @param pUrlFilterConf Output - Category Config Setting in nCategory array
  * @param pszCategory Input - Category Config setting string (HV Defined)
  * @param nCatLen Input - pszCategory string length
  * @return -1 input error
  * @return 0 parse success
  */
static int ParseCategory(url_filter_config_t *pUrlFilterConf, const char *pszCategory, int nCatLen)
{
    int i;
    int nCatNum;
    char szTemp[3];
    const char *pszTemp;
    int nCategory;

    if ((nCatLen % 2) != 0)
    {
        return -1;
    }
    
    for (i = 0; i < HV_DEFINED_CAT_NUM; i ++)
    {
        pUrlFilterConf->nCategory[i] = 0;
        LOGDEBUG("Initial Category %d in URL Filtering setting to 0\n", i);
    }
    
    nCatNum = nCatLen / 2;
    pszTemp = pszCategory;
    szTemp[2] = '\0';
    for (i = 0; i < nCatNum; i++)
    {
        szTemp[0] = *pszTemp;
        pszTemp++;
        szTemp[1] = *pszTemp;
        pszTemp++;
        nCategory = strtol(szTemp, NULL, 10);
        if ((nCategory < 0) || (nCategory >= (HV_DEFINED_CAT_NUM - 1)))
        {
            LOGERROR("URL Filter Category config error\n");
            continue;
        }
        LOGDEBUG("Enable Cagegory %d in URL Filtering\n", nCategory);
        pUrlFilterConf->nCategory[nCategory] = 1;
    }
    
    return 0;
}

int CheckAndSetStrValue(unsigned char *pDest, const char *pInput, const char *pDefaultValue, int nBufSize, const char *pParaName)
{
    pDest[nBufSize - 1] = '\0';
    
    if (pInput == NULL)
    {
        LOGERROR("failed to read tm.conf %s", pParaName);
        /* Set default value */
        strncpy(pDest, pDefaultValue, nBufSize - 1);
        return -1;
    }
    else
    {
        strncpy(pDest, pInput, nBufSize - 1);
        return 0;
    }
}

int CheckAndSetUIntValue(int *pDest, const char *pInput, const int nMin, const int nMax, const int nDefaultValue, const char *pParaName)
{
    int nTemp;
    
    if (pInput == NULL)
    {
        LOGERROR("failed to read tm.conf %s", pParaName);
        /* Set default value */
        *pDest = nDefaultValue;
        return -1;
    }
    else
    {
        nTemp = atoi(pInput);
        if ((nTemp < nMin) || (nTemp > nMax))
        {
            LOGERROR("failed to read tm.conf : %s = %s", pParaName, pInput);
            /* Set default value */
            *pDest = nDefaultValue;
            return -1;
        }
        else
        {
            *pDest = nTemp;
            return 0;
        }
    }
}

int CheckAndSetUShortValue(unsigned short *pDest, const char *pInput, const unsigned short nMin, const unsigned short nMax, const unsigned short nDefaultValue, const char *pParaName)
{
    unsigned short nTemp;
    
    if (pInput == NULL)
    {
        LOGERROR("failed to read tm.conf %s", pParaName);
        /* Set default value */
        *pDest = nDefaultValue;
        return -1;
    }
    else
    {
        nTemp = atoi(pInput);
        if ((nTemp < nMin) || (nTemp > nMax))
        {
            LOGERROR("failed to read tm.conf : %s = %s", pParaName, pInput);
            /* Set default value */
            *pDest = nDefaultValue;
            return -1;
        }
        else
        {
            *pDest = nTemp;
            return 0;
        }
    }
}

int CheckAndSetStrIntValue(unsigned char *pDest, const char *pInput, const int nMin, const int nMax, const char *pDefaultValue, int nBufSize, const char *pParaName)
{
    int nTemp;
    pDest[nBufSize - 1] = '\0';
    
    if (pInput == NULL)
    {
        LOGERROR("failed to read tm.conf %s", pParaName);
        /* Set default value */
        strncpy(pDest, pDefaultValue, nBufSize - 1);
        return -1;
    }
    else
    {
        nTemp = atoi(pInput);
        if ((nTemp < nMin) || (nTemp > nMax))
        {
            LOGERROR("failed to read tm.conf : %s = %s", pParaName, pInput);
            /* Set default value */
            strncpy(pDest, pDefaultValue, nBufSize - 1);
            return -1;
        }
        else
        {
            strncpy(pDest, pInput, nBufSize - 1);
            return 0;
        }
    }
}

/**
  * @brief Load current URL Filtering Settngs from tm.conf config file to shared memory
  * @param szUrlConfFile Input - tm.conf file path
  * @return -1 - open config file fail or config setting format error
  * @return 0 - ok
  * @remarks called when thpd start or thpd get signal SIGHUP to reload config
  */
int sendUrlFilterConfToKernel(url_filter_config_t *pUrlFilterConf);
int tmLoadUrlFilterConf(const char* szUrlConfFile)
{
    tmcfg_t cfg = NULL;
    int nRetVal = 0;
    int nUrlFilterConfShmid;
    url_filter_config_t *pUrlFilterConf;
    const char *pszVendorId = NULL;
    const char *pszLicenseId = NULL;
    const char *pszUserId = NULL;
    const char *pszOriginProtocol = NULL;
    const char *pszTmUrlFilterHost = NULL;
    const char *pszCategory = NULL;
    const char *pUrlFilterEnable = NULL;
    const char *pTmUrlFilterHostPort = NULL;
    const char *pDrtrMode = NULL;
    const char *pDenyOnError = NULL;
    const char *pUrlCacheSize = NULL;
    const char *pUrlLicenseEnable = NULL;
    const char *pUrlFilterExceptionEnable = NULL;
    const char *pUrlFilterCacheExpireTime = NULL;
    const char *pUrlFilterRatingTimeout = NULL;
#ifdef _PROXY_
    const char *pszProxyEnable = NULL;
    const char *pszProxyHost = NULL;
    const char *pszProxyPort = NULL;
    const char *pszProxyUsername = NULL;
    const char *pszProxyPassword = NULL;
#endif
    const char *pszEntitleHost = NULL;
    const char *pszEntitlePort = NULL;
    const char *pszEntitlePath = NULL;
    const char *pszLicChkInterval = NULL;
    const char *pszAlwayBlock = NULL;
    const char *pszScheduleDay[MAX_WEEK_DAYS];
    const char *pszScheduleFromHour = NULL;
    const char *pszScheduleFromMin = NULL;
    const char *pszScheduleToHour = NULL;
    const char *pszScheduleToMin = NULL;
    struct tm_sem_st UrlFilterConfLock;
    int i;
    
    nUrlFilterConfShmid = shm_init(TMSS_URL_FILTER_CONFSHMKEY, sizeof(url_filter_config_t));
    tm_lock_create(TMSS_URL_FILTER_CONFSEMLOCKKEY, &UrlFilterConfLock);
    if (tm_lock(&UrlFilterConfLock) == -1)
    {
        LOGERROR("tmLoadUrlFilterConf(), lock shared memroy error");
    }
    pUrlFilterConf = (url_filter_config_t*)shm_attach(nUrlFilterConfShmid);

    /* Open Config file for read */
    if( NULL == (cfg=tmcfg_open(szUrlConfFile,TMCFG_MODE_READ)) )
    {
        /* Set default value */
        strncpy(pUrlFilterConf->szVendorId,URL_FILTER_CFG_DEFAULT_VENDOR_ID, URL_FILTER_MAX_VENDORID - 1);
        strncpy(pUrlFilterConf->szLicenseId, URL_FILTER_CFG_DEFAULT_LICENSE_ID, URL_FILTER_MAX_LICENSEID - 1);
        strncpy(pUrlFilterConf->szUserId, URL_FILTER_CFG_DEFAULT_USER_ID, URL_FILTER_MAX_USERID - 1);
        strncpy(pUrlFilterConf->szOriginProtocol, URL_FILTER_CFG_DEFAULT_ORIGIN_PROTOCOL, URL_FILTER_MAX_ORIGINPROTOCOL - 1);
        strncpy(pUrlFilterConf->szTmUrlFilterHost, URL_FILTER_CFG_DEFAULT_TM_URL_FILTER_HOST, URL_FILTER_MAX_HOST - 1);
        strncpy(pUrlFilterConf->szCategory, URL_FILTER_CFG_DEFAULT_CATEGORY, URL_FILTER_MAX_CATEGORY - 1);
            
        ParseCategory(pUrlFilterConf, pUrlFilterConf->szCategory, strlen(pUrlFilterConf->szCategory));
        
        pUrlFilterConf->nUrlFilterEnable = URL_FILTER_CFG_DEFAULT_URL_FILTER_ENABLE;
        pUrlFilterConf->nTmUrlFilterHostPort = URL_FILTER_CFG_DEFAULT_TM_URL_FILTER_HOST_PORT;
        pUrlFilterConf->nDrtrMode = URL_FILTER_CFG_DEFAULT_DRTR_MODE;
        pUrlFilterConf->nDenyOnError = URL_FILTER_CFG_DEFAULT_DENY_ON_ERROR;
        pUrlFilterConf->nUrlCacheSize = URL_FILTER_CFG_DEFAULT_URL_CACHE_SIZE;
        pUrlFilterConf->nUrlLicenseEnable = URL_FILTER_CFG_DEFAULT_URL_LICENSE_ENABLE;
        pUrlFilterConf->nUrlFilterExceptionEnable = URL_FILTER_CFG_DEFAULT_URL_FILTER_EXCEPTION_ENABLE;
        pUrlFilterConf->nUrlFilterCacheExpireTime = URL_FILTER_CFG_DEFAULT_CACHE_EXPIRE_TIME;
        pUrlFilterConf->nUrlFilterRatingTimeout = URL_FILTER_CFG_DEFAULT_RATING_TIMEOUT;
#ifdef _PROXY_
        pUrlFilterConf->nProxyEnable = URL_FILTER_CFG_DEFAULT_PROXY_ENABLE;
        strncpy(pUrlFilterConf->szProxyHost, "", URL_FILTER_MAX_HOST - 1);
        pUrlFilterConf->nProxyPort = URL_FILTER_CFG_DEFAULT_TM_URL_FILTER_PROXY_PORT;
        strncpy(pUrlFilterConf->szProxyUsername, "", URL_FILTER_MAX_PROXY_USERNAME - 1);
        strncpy(pUrlFilterConf->szProxyPassword, "", URL_FILTER_MAX_PROXY_PASSWORD - 1); 
#endif
        strncpy(pUrlFilterConf->szEntitleHost, URL_FILTER_CFG_DEFAULT_TM_ENTITLE_HOST, URL_FILTER_MAX_HOST - 1);
        pUrlFilterConf->nEntitlePort = URL_FILTER_CFG_DEFAULT_TM_URL_FILTER_ENTITLE_PORT;
        strncpy(pUrlFilterConf->szEntitlePath, URL_FILTER_CFG_DEFAULT_TM_ENTITLE_PATH, URL_FILTER_MAX_PATH - 1);
        pUrlFilterConf->nLicChkInterval = URL_FILTER_CFG_DEFAULT_TM_URL_FILTER_LIC_CHK_INTERVAL;
        pUrlFilterConf->alwaysBlock = URL_FILTER_CFG_DEFAULT_ALWAYS_BLOCK;
        pUrlFilterConf->nScheduleDay[0] = URL_FILTER_CFG_DEFAULT_SCHEDULE_DAY;
        pUrlFilterConf->nScheduleDay[1] = URL_FILTER_CFG_DEFAULT_SCHEDULE_DAY;
        pUrlFilterConf->nScheduleDay[2] = URL_FILTER_CFG_DEFAULT_SCHEDULE_DAY;
        pUrlFilterConf->nScheduleDay[3] = URL_FILTER_CFG_DEFAULT_SCHEDULE_DAY;
        pUrlFilterConf->nScheduleDay[4] = URL_FILTER_CFG_DEFAULT_SCHEDULE_DAY;
        pUrlFilterConf->nScheduleDay[5] = URL_FILTER_CFG_DEFAULT_SCHEDULE_DAY;
        pUrlFilterConf->nScheduleDay[6] = URL_FILTER_CFG_DEFAULT_SCHEDULE_DAY;
        pUrlFilterConf->nScheduleFromHour   = URL_FILTER_CFG_DEFAULT_SCHEDULE_FRM_HOUR;
        pUrlFilterConf->nScheduleFromMin     = URL_FILTER_CFG_DEFAULT_SCHEDULE_FRM_MIN;
        pUrlFilterConf->nScheduleToHour = URL_FILTER_CFG_DEFAULT_SCHEDULE_TO_HOUR;
        pUrlFilterConf->nScheduleToMin   = URL_FILTER_CFG_DEFAULT_SCHEDULE_TO_MIN;
        
        LOGERROR("failed to open tm.conf");
        nRetVal = -1;
        goto end;
    }
    LOGDEBUG("Open thpd configuration file %s success", szUrlConfFile);

    pszVendorId = tmcfg_get_param_value(cfg, CFG_MODULE_URL_FILTER, CFG_PARAM_VENDOR_ID, NULL);
    pszLicenseId = tmcfg_get_param_value(cfg, CFG_MODULE_URL_FILTER, CFG_PARAM_LICENSE_ID, NULL);
    pszUserId = tmcfg_get_param_value(cfg, CFG_MODULE_URL_FILTER, CFG_PARAM_USER_ID, NULL);
    pszOriginProtocol = tmcfg_get_param_value(cfg, CFG_MODULE_URL_FILTER, CFG_PARAM_ORIGIN_PROTOCOL, NULL);
    pszTmUrlFilterHost = tmcfg_get_param_value(cfg, CFG_MODULE_URL_FILTER, CFG_PARAM_TM_URL_FILTER_HOST, NULL);
    pszCategory = tmcfg_get_param_value(cfg, CFG_MODULE_URL_FILTER, CFG_PARAM_CATEGORY, NULL);
    pUrlFilterEnable = tmcfg_get_param_value(cfg, CFG_MODULE_URL_FILTER, CFG_PARAM_URL_FILTER_ENABLE, NULL);
    pTmUrlFilterHostPort = tmcfg_get_param_value(cfg, CFG_MODULE_URL_FILTER, CFG_PARAM_TM_URL_FILTER_HOST_PORT, NULL);
    pDrtrMode = tmcfg_get_param_value(cfg, CFG_MODULE_URL_FILTER, CFG_PARAM_DRTR_MODE, NULL);
    pDenyOnError = tmcfg_get_param_value(cfg, CFG_MODULE_URL_FILTER, CFG_PARAM_DENY_ON_ERROR, NULL);
    pUrlCacheSize = tmcfg_get_param_value(cfg, CFG_MODULE_URL_FILTER, CFG_PARAM_URL_CACHE_SIZE, NULL);
    pUrlLicenseEnable = tmcfg_get_param_value(cfg, CFG_MODULE_URL_FILTER, CFG_PARAM_URL_LICENSE_ENABLE, NULL);
    pUrlFilterExceptionEnable = tmcfg_get_param_value(cfg, CFG_MODULE_URL_FILTER, CFG_PARAM_URL_FILTER_EXCEPTION_ENABLE, NULL);
    pUrlFilterCacheExpireTime = tmcfg_get_param_value(cfg, CFG_MODULE_URL_FILTER, CFG_PARAM_CACHE_EXPIRE_TIME, NULL);
    pUrlFilterRatingTimeout = tmcfg_get_param_value(cfg, CFG_MODULE_URL_FILTER, CFG_PARAM_RATING_TIMEOUT, NULL);
#ifdef _PROXY_
    pszProxyEnable = tmcfg_get_param_value(cfg, CFG_MODULE_PROXY, CFG_PARAM_PROXY_ENABLE, NULL);
    pszProxyHost = tmcfg_get_param_value(cfg, CFG_MODULE_PROXY, CFG_PARAM_PROXY_HOST, NULL);
    pszProxyPort = tmcfg_get_param_value(cfg, CFG_MODULE_PROXY, CFG_PARAM_PROXY_PORT, NULL);
    pszProxyUsername = tmcfg_get_param_value(cfg, CFG_MODULE_PROXY, CFG_PARAM_PROXY_USERNAME, NULL);
    pszProxyPassword = tmcfg_get_param_value(cfg, CFG_MODULE_PROXY, CFG_PARAM_PROXY_PASSWORD, NULL);
#endif

    pszEntitleHost = tmcfg_get_param_value(cfg, CFG_MODULE_ENTITLEMENT, CFG_PARAM_ENTITLE_HOST, NULL);
    pszEntitlePort = tmcfg_get_param_value(cfg, CFG_MODULE_ENTITLEMENT, CFG_PARAM_ENTITLE_PORT, NULL);
    pszEntitlePath = tmcfg_get_param_value(cfg, CFG_MODULE_ENTITLEMENT, CFG_PARAM_ENTITLE_PATH, NULL);
    pszLicChkInterval = tmcfg_get_param_value(cfg, CFG_MODULE_URL_FILTER, CFG_PARAM_LIC_CHK_INTERVIAL, NULL);
    pszAlwayBlock = tmcfg_get_param_value(cfg, CFG_MODULE_URL_FILTER, CFG_PARAM_ALWAYS_BLOCK, NULL);
    for (i = 0; i < MAX_WEEK_DAYS; i++)
    {
        char szTemp[sizeof(CFG_PARAM_SCHEDULE_DAY) + 5];
        sprintf(szTemp,"%s%d", CFG_PARAM_SCHEDULE_DAY, i);
        pszScheduleDay[i] = tmcfg_get_param_value(cfg, CFG_MODULE_URL_FILTER, szTemp, NULL);
    }
    pszScheduleFromHour = tmcfg_get_param_value(cfg, CFG_MODULE_URL_FILTER, CFG_PARAM_SCHEDULE_FRM_HOUR, NULL);
    pszScheduleFromMin = tmcfg_get_param_value(cfg, CFG_MODULE_URL_FILTER, CFG_PARAM_SCHEDULE_FRM_MIN, NULL);
    pszScheduleToHour = tmcfg_get_param_value(cfg, CFG_MODULE_URL_FILTER, CFG_PARAM_SCHEDULE_TO_HOUR, NULL);
    pszScheduleToMin = tmcfg_get_param_value(cfg, CFG_MODULE_URL_FILTER, CFG_PARAM_SCHEDULE_TO_MIN, NULL);
    
    nRetVal = CheckAndSetStrValue(pUrlFilterConf->szVendorId, pszVendorId, URL_FILTER_CFG_DEFAULT_VENDOR_ID, URL_FILTER_MAX_VENDORID, "pszVendorId");
    nRetVal = CheckAndSetStrValue(pUrlFilterConf->szLicenseId, pszLicenseId, URL_FILTER_CFG_DEFAULT_LICENSE_ID, URL_FILTER_MAX_LICENSEID, "pszLicenseId");
    nRetVal = CheckAndSetStrValue(pUrlFilterConf->szUserId, pszUserId, URL_FILTER_CFG_DEFAULT_USER_ID, URL_FILTER_MAX_USERID, "pszUserId");
    nRetVal = CheckAndSetStrIntValue(pUrlFilterConf->szOriginProtocol, pszOriginProtocol, URL_FILTER_ORIGIN_PROTOCOL_MIN, URL_FILTER_ORIGIN_PROTOCOL_MAX, URL_FILTER_CFG_DEFAULT_ORIGIN_PROTOCOL, URL_FILTER_MAX_ORIGINPROTOCOL, "pszOriginProtocol");
    nRetVal = CheckAndSetStrValue(pUrlFilterConf->szTmUrlFilterHost, pszTmUrlFilterHost, URL_FILTER_CFG_DEFAULT_TM_URL_FILTER_HOST, URL_FILTER_MAX_HOST, "pszTmUrlFilterHost");
    nRetVal = CheckAndSetStrValue(pUrlFilterConf->szCategory, pszCategory, URL_FILTER_CFG_DEFAULT_CATEGORY, URL_FILTER_MAX_CATEGORY, "pszCategory");
    ParseCategory(pUrlFilterConf, pUrlFilterConf->szCategory, strlen(pUrlFilterConf->szCategory));
    nRetVal = CheckAndSetUShortValue(&(pUrlFilterConf->nUrlFilterEnable), pUrlFilterEnable, 0, 1, URL_FILTER_CFG_DEFAULT_URL_FILTER_ENABLE, "pUrlFilterEnable");
    nRetVal = CheckAndSetUShortValue(&(pUrlFilterConf->nTmUrlFilterHostPort), pTmUrlFilterHostPort, URL_FILTER_PORT_MIN, URL_FILTER_PORT_MAX, URL_FILTER_CFG_DEFAULT_TM_URL_FILTER_HOST_PORT, "pTmUrlFilterHostPort");
    nRetVal = CheckAndSetUShortValue(&(pUrlFilterConf->nDrtrMode), pDrtrMode, 0, 1, URL_FILTER_CFG_DEFAULT_DRTR_MODE, "pDrtrMode");
    nRetVal = CheckAndSetUShortValue(&(pUrlFilterConf->nDenyOnError), pDenyOnError, 0, 1, URL_FILTER_CFG_DEFAULT_DENY_ON_ERROR, "pDenyOnError");
    nRetVal = CheckAndSetUIntValue(&(pUrlFilterConf->nUrlCacheSize), pUrlCacheSize, URL_FILTER_CACHE_MIN, URL_FILTER_CACHE_MAX, URL_FILTER_CFG_DEFAULT_URL_CACHE_SIZE, "pUrlCacheSize");
    nRetVal = CheckAndSetUShortValue(&(pUrlFilterConf->nUrlLicenseEnable), pUrlLicenseEnable, 0, 1, URL_FILTER_CFG_DEFAULT_URL_LICENSE_ENABLE, "pUrlLicenseEnable");
    nRetVal = CheckAndSetUShortValue(&(pUrlFilterConf->nUrlFilterExceptionEnable), pUrlFilterExceptionEnable, 0, 1, URL_FILTER_CFG_DEFAULT_URL_FILTER_EXCEPTION_ENABLE, "pUrlFilterExceptionEnable");
    nRetVal = CheckAndSetUIntValue(&(pUrlFilterConf->nUrlFilterCacheExpireTime), pUrlFilterCacheExpireTime, URL_FILTER_CACHE_EXPIRE_TIME_MIN, URL_FILTER_CACHE_EXPIRE_TIME_MAX, URL_FILTER_CFG_DEFAULT_CACHE_EXPIRE_TIME, "pUrlFilterCacheExpireTime");
    nRetVal = CheckAndSetUShortValue(&(pUrlFilterConf->nUrlFilterRatingTimeout), pUrlFilterRatingTimeout, URL_FILTER_RATING_TIMEOUT_MIN, URL_FILTER_RATING_TIMEOUT_MAX, URL_FILTER_CFG_DEFAULT_RATING_TIMEOUT, "pUrlFilterRatingTimeout");
#ifdef _PROXY_
    nRetVal = CheckAndSetUShortValue(&(pUrlFilterConf->nProxyEnable), pszProxyEnable, 0, 1, URL_FILTER_CFG_DEFAULT_PROXY_ENABLE, "pszProxyEnable");
    nRetVal = CheckAndSetStrValue(pUrlFilterConf->szProxyHost, pszProxyHost, "", URL_FILTER_MAX_HOST, "pszProxyHost");
    nRetVal = CheckAndSetUShortValue(&(pUrlFilterConf->nProxyPort), pszProxyPort, URL_FILTER_PORT_MIN, URL_FILTER_PORT_MAX, URL_FILTER_CFG_DEFAULT_TM_URL_FILTER_PROXY_PORT, "pszProxyPort");
    nRetVal = CheckAndSetStrValue(pUrlFilterConf->szProxyUsername, pszProxyUsername, "", URL_FILTER_MAX_PROXY_USERNAME, "pszProxyUsername");
    nRetVal = CheckAndSetStrValue(pUrlFilterConf->szProxyPassword, pszProxyPassword, "", URL_FILTER_MAX_PROXY_PASSWORD, "pszProxyPassword");
#endif
    nRetVal = CheckAndSetStrValue(pUrlFilterConf->szEntitleHost, pszEntitleHost, "", URL_FILTER_MAX_HOST, "pszEntitleHost");
    nRetVal = CheckAndSetUShortValue(&(pUrlFilterConf->nEntitlePort), pszEntitlePort, URL_FILTER_PORT_MIN, URL_FILTER_PORT_MAX, URL_FILTER_CFG_DEFAULT_TM_URL_FILTER_ENTITLE_PORT, "pszEntitlePort");
    nRetVal = CheckAndSetStrValue(pUrlFilterConf->szEntitlePath, pszEntitlePath, "", URL_FILTER_MAX_PATH, "pszEntitlePath");
    nRetVal = CheckAndSetUIntValue(&(pUrlFilterConf->nLicChkInterval), pszLicChkInterval, URL_FILTER_LIC_CHK_INTERVAL_MIN, URL_FILTER_LIC_CHK_INTERVAL_MAX, URL_FILTER_CFG_DEFAULT_TM_URL_FILTER_LIC_CHK_INTERVAL, "pszLicChkInterval");
    nRetVal = CheckAndSetUShortValue(&(pUrlFilterConf->alwaysBlock), pszAlwayBlock, 0, 1, URL_FILTER_CFG_DEFAULT_ALWAYS_BLOCK, "alwaysBlock");
    nRetVal = CheckAndSetUShortValue(&(pUrlFilterConf->nScheduleDay[0]), pszScheduleDay[0], 0, 1, URL_FILTER_CFG_DEFAULT_SCHEDULE_DAY, "nScheduleDay[0]");
    nRetVal = CheckAndSetUShortValue(&(pUrlFilterConf->nScheduleDay[1]), pszScheduleDay[1], 0, 1, URL_FILTER_CFG_DEFAULT_SCHEDULE_DAY, "nScheduleDay[1]");
    nRetVal = CheckAndSetUShortValue(&(pUrlFilterConf->nScheduleDay[2]), pszScheduleDay[2], 0, 1, URL_FILTER_CFG_DEFAULT_SCHEDULE_DAY, "nScheduleDay[2]");
    nRetVal = CheckAndSetUShortValue(&(pUrlFilterConf->nScheduleDay[3]), pszScheduleDay[3], 0, 1, URL_FILTER_CFG_DEFAULT_SCHEDULE_DAY, "nScheduleDay[3]");
    nRetVal = CheckAndSetUShortValue(&(pUrlFilterConf->nScheduleDay[4]), pszScheduleDay[4], 0, 1, URL_FILTER_CFG_DEFAULT_SCHEDULE_DAY, "nScheduleDay[4]");
    nRetVal = CheckAndSetUShortValue(&(pUrlFilterConf->nScheduleDay[5]), pszScheduleDay[5], 0, 1, URL_FILTER_CFG_DEFAULT_SCHEDULE_DAY, "nScheduleDay[5]");
    nRetVal = CheckAndSetUShortValue(&(pUrlFilterConf->nScheduleDay[6]), pszScheduleDay[6], 0, 1, URL_FILTER_CFG_DEFAULT_SCHEDULE_DAY, "nScheduleDay[6]");

    nRetVal = CheckAndSetUShortValue(&(pUrlFilterConf->nScheduleFromHour), pszScheduleFromHour, URL_FILTER_SCHEDULE_FRM_HOUR_MIN, URL_FILTER_SCHEDULE_FRM_HOUR_MAX, URL_FILTER_CFG_DEFAULT_SCHEDULE_FRM_HOUR, "nScheduleFromHour");
    nRetVal = CheckAndSetUShortValue(&(pUrlFilterConf->nScheduleFromMin), pszScheduleFromMin, URL_FILTER_SCHEDULE_FRM_MIN_MIN, URL_FILTER_SCHEDULE_FRM_MIN_MAX, URL_FILTER_CFG_DEFAULT_SCHEDULE_FRM_MIN, "nScheduleFromMin");
    nRetVal = CheckAndSetUShortValue(&(pUrlFilterConf->nScheduleToHour), pszScheduleToHour, URL_FILTER_SCHEDULE_TO_HOUR_MIN, URL_FILTER_SCHEDULE_TO_HOUR_MAX, URL_FILTER_CFG_DEFAULT_SCHEDULE_TO_HOUR, "nScheduleToHour");
    nRetVal = CheckAndSetUShortValue(&(pUrlFilterConf->nScheduleToMin), pszScheduleToMin, URL_FILTER_SCHEDULE_TO_MIN_MIN, URL_FILTER_SCHEDULE_TO_MIN_MAX, URL_FILTER_CFG_DEFAULT_SCHEDULE_TO_MIN, "nScheduleToMin");
    
    LOGDEBUG("URL Filter Config Setting - Vendor ID: %s",pUrlFilterConf->szVendorId);
    LOGDEBUG("URL Filter Config Setting - License ID: %s",pUrlFilterConf->szLicenseId);
    LOGDEBUG("URL Filter Config Setting - User ID: %s",pUrlFilterConf->szUserId);
    LOGDEBUG("URL Filter Config Setting - Origin Protocol: %s",pUrlFilterConf->szOriginProtocol);
    LOGDEBUG("URL Filter Config Setting - TmUrlFilter Host: %s",pUrlFilterConf->szTmUrlFilterHost);
    LOGDEBUG("URL Filter Config Setting - Category: %s",pUrlFilterConf->szCategory);
    LOGDEBUG("URL Filter Config Setting - Enable: %d",pUrlFilterConf->nUrlFilterEnable);
    LOGDEBUG("URL Filter Config Setting - TmUrlFilter Host Port: %d",pUrlFilterConf->nTmUrlFilterHostPort);
    LOGDEBUG("URL Filter Config Setting - DRTR Mode: %d",pUrlFilterConf->nDrtrMode);
    LOGDEBUG("URL Filter Config Setting - Deny on Error: %d",pUrlFilterConf->nDenyOnError);
    LOGDEBUG("URL Filter Config Setting - Url Cache Size: %d",pUrlFilterConf->nUrlCacheSize);
    LOGDEBUG("URL Filter Config Setting - Url Filter License Enable: %d",pUrlFilterConf->nUrlLicenseEnable);
    LOGDEBUG("URL Filter Config Setting - Url Filter Exception Enable: %d",pUrlFilterConf->nUrlFilterExceptionEnable);
    LOGDEBUG("URL Filter Config Setting - Url Filter Cache Expire Time: %d",pUrlFilterConf->nUrlFilterCacheExpireTime);
    LOGDEBUG("URL Filter Config Setting - Url Filter Rating timeout: %d",pUrlFilterConf->nUrlFilterRatingTimeout);
#ifdef _PROXY_
    LOGDEBUG("URL Filter Config Setting - Proxy Enable: %d",pUrlFilterConf->nProxyEnable);
    LOGDEBUG("URL Filter Config Setting - Proxy Host: %s",pUrlFilterConf->szProxyHost);
    LOGDEBUG("URL Filter Config Setting - Proxy Port: %d",pUrlFilterConf->nProxyPort);
    LOGDEBUG("URL Filter Config Setting - Proxy Username: %s",pUrlFilterConf->szProxyUsername);
    LOGDEBUG("URL Filter Config Setting - Proxy Password: %s",pUrlFilterConf->szProxyPassword);
#endif
    LOGDEBUG("URL Filter Config Setting - Entitlement Host: %s",pUrlFilterConf->szEntitleHost);
    LOGDEBUG("URL Filter Config Setting - Entitlement Port: %d",pUrlFilterConf->nEntitlePort);
    LOGDEBUG("URL Filter Config Setting - Entitlement Path: %s",pUrlFilterConf->szEntitlePath);
    LOGDEBUG("URL Filter Config Setting - URL Filter License Check Interval: %d",pUrlFilterConf->nLicChkInterval);
    LOGDEBUG("URL Filter Config Setting - Schedule AlwaysBlock Enable: %d",pUrlFilterConf->alwaysBlock);
    LOGDEBUG("URL Filter Config Setting - Schedule Day Sun Enable: %d",pUrlFilterConf->nScheduleDay[0]);
    LOGDEBUG("URL Filter Config Setting - Schedule Day Mon Enable: %d",pUrlFilterConf->nScheduleDay[1]);
    LOGDEBUG("URL Filter Config Setting - Schedule Day Tue Enable: %d",pUrlFilterConf->nScheduleDay[2]);
    LOGDEBUG("URL Filter Config Setting - Schedule Day Wed Enable: %d",pUrlFilterConf->nScheduleDay[3]);
    LOGDEBUG("URL Filter Config Setting - Schedule Day Thu Enable: %d",pUrlFilterConf->nScheduleDay[4]);
    LOGDEBUG("URL Filter Config Setting - Schedule Day Fri Enable: %d",pUrlFilterConf->nScheduleDay[5]);
    LOGDEBUG("URL Filter Config Setting - Schedule Day Sat Enable: %d",pUrlFilterConf->nScheduleDay[6]);
    LOGDEBUG("URL Filter Config Setting - Schedule From Hour: %d",pUrlFilterConf->nScheduleFromHour  );
    LOGDEBUG("URL Filter Config Setting - Schedule From Min  : %d",pUrlFilterConf->nScheduleFromMin    );
    LOGDEBUG("URL Filter Config Setting - Schedule To Hour: %d",pUrlFilterConf->nScheduleToHour);
    LOGDEBUG("URL Filter Config Setting - Schedule To Min  : %d",pUrlFilterConf->nScheduleToMin  );
   pUrlFilterConf->version=urlFilterConfigVersion;
   urlFilterConfigVersion++;

end:

    shm_detach(pUrlFilterConf);
    if (tm_unlock(&UrlFilterConfLock) == -1)
    {
        LOGERROR("tmLoadUrlFilterConf(), unlock shared memroy error");
    }
    tm_lock_close(&UrlFilterConfLock);

    if (cfg != NULL)
    {
        tmcfg_close(cfg);
    }
    sendUrlFilterConfToKernel(pUrlFilterConf);

    return nRetVal;
}

int sendUrlFilterConfToKernel(url_filter_config_t *pUrlFilterConf)
{
    int nRetVal = 0;
    size_t len;
    int sockfd = -1; 
    
	// We need to send the AVE configuration to tmss_kernel(ipt_webstr)
	sockfd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
	if (sockfd < 0)
	{
	    fprintf(stderr,"sendUrlFilterConfToKernel:Can not create sock:%s\a\n",strerror(errno)); 
		return -1;
	}

	len=sizeof(url_filter_config_t);
	if (setsockopt(sockfd, IPPROTO_IP, TMSS_SO_SET_URL_FILTER_CONFIG, pUrlFilterConf,len ) < 0)
	{
	    fprintf(stderr,"sendUrlFilterConfToKernel:Can not setsockopt sock:%s\a\n",strerror(errno)); 
	close(sockfd);
	sockfd = -1;
		return -1;
	}
	close(sockfd);
	sockfd = -1;
    return nRetVal;
}

int sendInitClientInfoToKernel(void)
{
    int nRetVal = 0;
    size_t len;
    int sockfd = -1; 
    
	// We need to send the AVE configuration to tmss_kernel(ipt_webstr)
	sockfd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
	if (sockfd < 0)
	{
	    fprintf(stderr,"sendInitClientInfoToKernel:Can not create sock:%s\a\n",strerror(errno)); 
		return -1;
	}

	if (setsockopt(sockfd, IPPROTO_IP, TMSS_SO_INIT_CLIENT_INFO, 0,0 ) < 0)
	{
	    fprintf(stderr,"sendInitClientInfoToKernel:Can not setsockopt sock:%s\a\n",strerror(errno)); 
	close(sockfd);
	sockfd = -1;
		return -1;
	}
	close(sockfd);
	sockfd = -1;
    return nRetVal;
}

