/******************************************************************************
 *
 *  (C) Copyright 2004 Trend Micro, Inc.
 *
 *  All Rights Reserved.
 *  This program is an unpublished copyrighted work which is proprietary
 *  to Trend Micro, Inc. and contains confidential information that is not
 *  to be reproduced or disclosed to any other person or entity without
 *  prior written consent from Trend Micro, Inc. in each and every instance.
 *
 *  WARNING:  Unauthorized reproduction of this program as well as
 *  unauthorized preparation of derivative works based upon the
 *  program or distribution of copies by sale, rental, lease or
 *  lending are violations of federal copyright laws and state trade
 *  secret laws, punishable by civil and criminal penalties.
 *
 *****************************************************************************/

/**
 *  @file tmURLFilterStatistic.c
 *  @brief TMOG URL Filtering Statistic Log Module
 */

/*
 *  Change Log
 *  2004-03-03 Henry Lin    Initialization version
 */

#include <sys/shm.h>
#include <sys/ipc.h>
#include <string.h>
#include <stdlib.h>
#include "time.h"
#include "log.h"
#include "tmShm.h"
#include "tmSemaphore.h"
#include "tmURLFilter.h"
#include "tmURLRater.h" /* TM_URL_FILTER_MAX_CAT_NUMBER */
#include "tmURLFilterConf.h"
#include "tmURLFilterStatistic.h"
#include "tmcfg.h"

static const char ModuleName[] = "tmURLFilterStatistic";


/**
  * @brief Reset the URL Filtering Statistic Log Count
  * @return none
  * @remarks This Function is used by CGI programs to reset Statistic Log
  */
void tmResetURLStatisticCount(void)
{
    int i;
    int nUrlFilterStatisticShmid;
    url_filter_statistic_t *pUrlFilterStatistic;
    time_t *pStartTime;
    struct tm_sem_st UrlFilterStatisticLock;

    nUrlFilterStatisticShmid = shm_init(TMSS_URL_FILTER_LOG_SHMKEY, sizeof(url_filter_statistic_t)*HV_DEFINED_CAT_NUM + sizeof(time_t));
    tm_lock_create(TMSS_URL_FILTER_LOG_SEMLOCKKEY, &UrlFilterStatisticLock);
    if (tm_lock(&UrlFilterStatisticLock) == -1)
    {
        LOGERROR("tmResetURLStatisticCount(), lock shared memroy error");
    }
    pUrlFilterStatistic = (url_filter_statistic_t*)shm_attach(nUrlFilterStatisticShmid);

    for (i = 0; i < HV_DEFINED_CAT_NUM; i++)
    {
        pUrlFilterStatistic[i].nAccessCount = 0;
        pUrlFilterStatistic[i].nBlockedCount = 0;
        pUrlFilterStatistic[i].LastAccessTime = 0;
        pUrlFilterStatistic[i].LastBlockedTime = 0;
    }
    pStartTime = (time_t*)(pUrlFilterStatistic + sizeof(url_filter_statistic_t)*HV_DEFINED_CAT_NUM);
    time(pStartTime);

    shm_detach(pUrlFilterStatistic);
    if (tm_unlock(&UrlFilterStatisticLock) == -1)
    {
        LOGERROR("tmResetURLStatisticCount(), unlock shared memroy error");
    }
    tm_lock_close(&UrlFilterStatisticLock);
    return;
}


/**
  * @brief Add the URL Filtering Log Count. Add access (block) count and record the access (block) time
  * @param BlockCat Input - Blocked category (TrendMicro Defined)
  * @param gTmUrlFilterDefinedCatMap Input - Mapping between TrendMicro Defined Categories and HV Defined Categories
  * @param UrlFilterConfig Input - Current URL Filtering Config settings
  * @return none
  */
void tmAddUrlFilterStatisticCount(unsigned long BlockCat,
                                  const int *gTmUrlFilterDefinedCatMap,
                                  url_filter_config_t *UrlFilterConfig)
{
    url_filter_statistic_t *pUrlFilterStatistic;
    int shmid;
    time_t now;
    struct tm_sem_st UrlFilterStatisticLock;
    
    if (BlockCat >= TM_URL_FILTER_MAX_CAT_NUMBER)
    {
        LOGERROR("tmAddUrlFilterStatisticCount(), Block Category too big: %lu (decimal)",BlockCat);
        return;
    }

    shmid = shm_init(TMSS_URL_FILTER_LOG_SHMKEY, sizeof(url_filter_statistic_t)*HV_DEFINED_CAT_NUM + sizeof(time_t));
    tm_lock_create(TMSS_URL_FILTER_LOG_SEMLOCKKEY, &UrlFilterStatisticLock);
    if (tm_lock(&UrlFilterStatisticLock) == -1)
    {
        LOGERROR("tmAddUrlFilterStatisticCount(), lock shared memroy error");
    }
    pUrlFilterStatistic = (url_filter_statistic_t*)shm_attach(shmid);

    pUrlFilterStatistic[gTmUrlFilterDefinedCatMap[BlockCat]].nAccessCount++;
    time(&now);
    pUrlFilterStatistic[gTmUrlFilterDefinedCatMap[BlockCat]].LastAccessTime = now;
    
    // As we have lots of comments on func tmURLFilter of tmURLFilter.c for Category, then why we still can nCategory
    // checking here, it is because the adding operatio is issued in kerenel.  Actually, we do not need to check here since
    // we have check it in Kernel already.
    if (UrlFilterConfig->nCategory[gTmUrlFilterDefinedCatMap[BlockCat]] == 1)
    {
        pUrlFilterStatistic[gTmUrlFilterDefinedCatMap[BlockCat]].nBlockedCount++;
        pUrlFilterStatistic[gTmUrlFilterDefinedCatMap[BlockCat]].LastBlockedTime = now;
    }

    shm_detach(pUrlFilterStatistic);
    if (tm_unlock(&UrlFilterStatisticLock) == -1)
    {
        LOGERROR("tmAddUrlFilterStatisticCount(), unlock shared memroy error");
    }
    tm_lock_close(&UrlFilterStatisticLock);
    return;
}


/**
  * @brief Write current URL Filtering Statistic Log to Debug Log
  * @return none
  * @remarks This Function is used for debug
  */
void tmDumpURLStatisticCount(void)
{
    int i;
    struct tm *access;
    struct tm *block;
    struct tm *start;
    url_filter_statistic_t *pUrlFilterStatistic;
    int shmid;
    time_t *pStartTime;
    struct tm_sem_st UrlFilterStatisticLock;
    
    shmid = shm_init(TMSS_URL_FILTER_LOG_SHMKEY, sizeof(url_filter_statistic_t)*HV_DEFINED_CAT_NUM + sizeof(time_t));
    tm_lock_create(TMSS_URL_FILTER_LOG_SEMLOCKKEY, &UrlFilterStatisticLock);
    if (tm_lock(&UrlFilterStatisticLock) == -1)
    {
        LOGERROR("tmDumpURLStatisticCount(), lock shared memroy error");
    }
    pUrlFilterStatistic = (url_filter_statistic_t*)shm_attach(shmid);
    
    LOGDEBUG("URL Filtering Statistic Report");

    for (i = 0; i < HV_DEFINED_CAT_NUM; i++)
    {
        LOGDEBUG("Category %d Access %d times",i,pUrlFilterStatistic[i].nAccessCount);
        access = localtime(&pUrlFilterStatistic[i].LastAccessTime);
        LOGDEBUG("Category %d Last Access Time: %04d/%02d/%02d %2d:%02d:%02d",i,access->tm_year + 1900, access->tm_mon+1, access->tm_mday,access->tm_hour, access->tm_min, access->tm_sec);
        LOGDEBUG("Category %d Block %d times",i,pUrlFilterStatistic[i].nBlockedCount);
        block = localtime(&pUrlFilterStatistic[i].LastBlockedTime);
        LOGDEBUG("Category %d Last Block Time: %04d/%02d/%02d %2d:%02d:%02d",i,block->tm_year + 1900, block->tm_mon+1, block->tm_mday,block->tm_hour, block->tm_min, block->tm_sec);
    }
    pStartTime = (time_t*)(pUrlFilterStatistic + sizeof(url_filter_statistic_t)*HV_DEFINED_CAT_NUM);
    start = localtime(pStartTime);
    LOGDEBUG("The records have been recorded since %04d/%02d/%02d %2d:%02d:%02d", start->tm_year + 1900, start->tm_mon+1, start->tm_mday,start->tm_hour, start->tm_min, start->tm_sec);

    shm_detach(pUrlFilterStatistic);
    if (tm_unlock(&UrlFilterStatisticLock) == -1)
    {
        LOGERROR("tmDumpURLStatisticCount(), unlock shared memroy error");
    }
    tm_lock_close(&UrlFilterStatisticLock);
    return;
}

/**
  * @brief Get current URL Filtering Statistic Log
  * @param pUrlFilterStatisticBuf Output - Current URL Filtering Statistic Log
  * @param nBufSize Input - pUrlFilterStatisticBuf buffer size (bytes)
  * @param pStartTime Output - Current URL Filtering Statistic Log Start time
  * @return pUrlFilterStatisticBuf Address of URL Filtering Statistic Log
  * @remarks This Function is used by CGI programs to get URL Filtering Statistic Log
  */
url_filter_statistic_t* tmGetURLStatisticCount(url_filter_statistic_t *pUrlFilterStatisticBuf, int nBufSize, time_t *pStartTime)
{
    url_filter_statistic_t *pUrlFilterStatistic;
    int shmid;
    time_t *pStartTimeRecorded;
    struct tm_sem_st UrlFilterStatisticLock;
   // int size ;
    
    if (nBufSize != sizeof(url_filter_statistic_t)*HV_DEFINED_CAT_NUM)
    {
        return NULL;
    }
    
     //shmid = shm_init(TMSS_URL_FILTER_LOG_SHMKEY, sizeof(url_filter_statistic_t)*HV_DEFINED_CAT_NUM + sizeof(time_t));
    if ((shmid = shmget(TMSS_URL_FILTER_LOG_SHMKEY, sizeof(url_filter_statistic_t)*HV_DEFINED_CAT_NUM + sizeof(time_t), 0)) == -1)
    {// Only for Read Only
	    memset(pUrlFilterStatisticBuf,0,nBufSize);
	    return NULL;// WE do not change the buffer
    }
    tm_lock_create(TMSS_URL_FILTER_LOG_SEMLOCKKEY, &UrlFilterStatisticLock);
    if (tm_lock(&UrlFilterStatisticLock) == -1)
    {
        LOGERROR("tmGetURLStatisticCount(), lock shared memroy error");
    }
    pUrlFilterStatistic = (url_filter_statistic_t*)shm_attach(shmid);
    pStartTimeRecorded = (time_t*)(pUrlFilterStatistic + sizeof(url_filter_statistic_t)*HV_DEFINED_CAT_NUM);

    memcpy(pUrlFilterStatisticBuf,pUrlFilterStatistic,nBufSize);
    *pStartTime = *pStartTimeRecorded;

    shm_detach(pUrlFilterStatistic);
    if (tm_unlock(&UrlFilterStatisticLock) == -1)
    {
        LOGERROR("tmGetURLStatisticCount(), unlock shared memroy error");
    }
    tm_lock_close(&UrlFilterStatisticLock);
    return pUrlFilterStatisticBuf;
}
