/******************************************************************************
 *
 *  (C) Copyright 2004 Trend Micro, Inc.
 *
 *  All Rights Reserved.
 *  This program is an unpublished copyrighted work which is proprietary
 *  to Trend Micro, Inc. and contains confidential information that is not
 *  to be reproduced or disclosed to any other person or entity without
 *  prior written consent from Trend Micro, Inc. in each and every instance.
 *
 *  WARNING:  Unauthorized reproduction of this program as well as
 *  unauthorized preparation of derivative works based upon the
 *  program or distribution of copies by sale, rental, lease or
 *  lending are violations of federal copyright laws and state trade
 *  secret laws, punishable by civil and criminal penalties.
 *
 *****************************************************************************/

/**
 *  @file   tmauIniParser.c
 *  @brief  TMOG ActiveUpdate (Server.ini) Parsing Module 
 */

/*
 *  Change Log
 *  2004-02-14  Jeremy  initial version
 */
//#define NDEBUG
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tmauIniParser.h"

enum SECTION {SEC_PATTERN, SEC_ENGINE, SEC_OTHER};

/* the module name that will show in log file */
static const char ModuleName[] = "libtmIniParser";

static const char COMMON_CHAR = ',';
static const char DOT_CHAR = '.';

char strPattern[]    = "[PATTERN]";
char strEngine[]     = "[ENGINE]";

/**
 * Source codes from AU team - start
 */
static void         Str_TrimAll(char * s);
static char        *Strncpy(char * dst, const char * src, size_t n);
static TmuVersion_t MakeVersionBuild (const char *str);

enum {
    HTAB = '\t',
    LN = '\n',
    FORMFEED = '\f',
    CR = '\r',
    SPACE = ' '
};


static void Str_TrimAll(char *s)
{
    int i;
    int j;
    char *p;
    char *q;

    
    if (s == NULL)
    {
        LOGCRIT("The NULL pointer of input argument!");
    }

    if (strlen(s) == 0) {
        return;
    }

    p = s;
    i = 0;
    while(*p != '\0') {
        switch(*p) {
        case HTAB:
        case LN:
        case FORMFEED:
        case CR:
        case SPACE:
            i ++;
             p ++;
            continue;
        default:
            break;
        }
        break;
    }

    if (*p == '\0') {
        memset(s, 0, strlen(s));
        return;
    }


    q = (char *)malloc(strlen(s) + 1);

    p = s + strlen(s) - 1;
    j = strlen(s) - 1;
    while (p != s) {
        switch(*p) {
        case HTAB:
        case LN:
        case FORMFEED:
        case CR:
        case SPACE:
            j --;
            p --;
            continue;
        default:
            break;
        }
        break;
    }

    memmove(q,s + i,j-i+1);
    memset(s,0,strlen(s));
    memmove(s,q,j-i+1);
    free (q);
}


static char *Strncpy(char *dst, const char *src, size_t n)
{
    if (dst == NULL || src == NULL)
    {
        LOGCRIT("The NULL pointer of input argument!");
    }

    dst[n-1] = '\0';

    if (NULL == (void *)strncpy(dst, src, n))
        return NULL;

    if (dst[n-1] != '\0')
        return NULL;

    return dst;
}


/**
 * @brief   Translate the string type version field to double word format and
 *          return this 32 bits double word.
 *
 * @param   pVer: the field of string type version field to be translated
 *
 * @return  32 bits double word
 *
 * @remarks For example, the input format is 6.801, output format is 0x00068010
 */
uint32_t 
tmau_MakeDWordVersion(const char* pVer)
{
    // major version first !
    uint32_t wMajor = atoi(pVer);
    uint32_t wMinor = 0;

    //  minor !
    char* pMinor =(char *)strchr(pVer,'.');
    char* pStop;
    if ( pMinor != NULL )
    {
        char szVer[5]={ ('0'),('0'),('0'),('0'),('\0') };
        char ch;
        int32_t len;
        for (len=0; (ch=*++pMinor) && len<4; len++ ) {
            if ( !isdigit(ch) ) break;
            szVer[len] = ch;
        }
        wMinor = strtoul(szVer, &pStop, 16);
    }

    return ((uint32_t)wMajor<<16 | ((uint32_t)wMinor));
}


/**
 * @brief   Translate the version field of TmuVersion_t to string format and
 *          put it into a target buffer, strVersion.
 *
 * @param   dwVersion: the field of version information to be translated
 * @param   strVersion: the target buffer to hold the string format
 * @param   str_len: the length of the target buffer
 *
 * @return  void
 */
void 
tmau_MakeVersionString(uint32_t dwVersion, char* strVersion, int32_t nMax)
{
    char        buf[TM_AU_MAX_VER];
    int32_t    dwTmp = dwVersion;
    char        *pTemp;

    snprintf(buf, TM_AU_MAX_VER, "%d.%04x", ((dwVersion >> 16 ) & 0xFFFF), (dwTmp & 0xFFFF));

    // the buf will like 11.90000 format.  we cut the forth byte
    pTemp = strstr(buf, ".");
    *(pTemp + 4) = '\0';
    /*
    end = strlen(buf)-1;
    for ( i=0; i<3 ; i++ ) 
    {
        if ( buf[end] != '0' ) 
            break;
        buf[end--] = '\0';
    }
    */
    Strncpy(strVersion, buf, nMax);
}


/** 
 * @brief   Translate the string format to fit into TmuVersion_t data structure
 *
 * @return  TmuVersion_t structure
 */
static TmuVersion_t MakeVersionBuild (const char *str) {

    TmuVersion_t result;

    memset (&result, 0, sizeof (TmuVersion_t));
    if (!str)
        return result;

    result.version = tmau_MakeDWordVersion (str);

    /* todo. try to find build number and put into TmuVersion_t.build
    // [major].[minor].[build]
    CSV csv (str, '.');
    if (csv.size () >= 3) {
        result.build = atoi (csv.get (2).c_str ());
        return result;
    }

    // [major].[minor]-[build]
    CSV csv2 (str, '-');
    if (csv2.size () >= 2) {
        result.build = atoi (csv2.get (1).c_str ());
        return result;
    }
    */

    return result;
}


/**
 * @brief   Translate the version and build fields of TmuVersion_t to string format and
 *          put them into a target buffer, str.
 *
 * @param   ver: the version structure to be translated
 * @param   str: the target buffer to hold the string format
 * @param   length: the length of the target buffer
 *
 * @return  void
 */
void 
tmau_MakeVersionBuildString (TmuVersion_t ver, char *str, uint32_t length) 
{
    char tmp[TM_AU_MAX_STR];

    if (ver.build) 
    {
         tmau_MakeVersionString (ver.version, tmp, TM_AU_MAX_STR);
         LOGDEBUG("The major and minor version are: [%s]", tmp);
         // todo. check whether we need this and evaluate risk of without it.
         Str_TrimAll (tmp);
         snprintf (str, length, "%s.%u", tmp, ver.build);
         LOGDEBUG("The major and minor version, plus build number are: [%s]", str);
    } 
    else 
    {
         tmau_MakeVersionString (ver.version, str, length);
         LOGDEBUG("The major and minor version are: [%s]", str);
    }
}
/**
 * Source codes from AU team - end
 */



static enum SECTION 
_tmau_GetSectionName(char buffer[]) 
{
    enum SECTION result;

    if (!strncmp(buffer, strPattern, strlen(strPattern)))
    {
        result = SEC_PATTERN;
    }
    else if (!strncmp(buffer, strEngine, strlen(strEngine)))
    {
        result = SEC_ENGINE;
    }
    else
    {
       result = SEC_OTHER;
    }

   return result;
}


static uint32_t 
_tmau_GetPatternVerString(char pszINIInfo[], char *pszPatType, TmErrorCode *pErrCode)
{
    char *s=NULL, *e;

    
    if (strlen(pszINIInfo) < 6) 
    {
        /* Skip the empty line, and keep them in the log file */
        if (strlen(pszINIInfo) > 2)
            LOGDEBUG("The information length is too short to parse, [%s]", pszINIInfo);
        return 0;
    }

    if (!strncmp(pszINIInfo, pszPatType, strlen(pszPatType))) 
    {
        s = pszINIInfo+strlen(pszPatType);

        /* Skip space */
        while (*s == ' ') 
            s++;

        if (*s != '=') 
        {
            /* Exclude the P.4.Merge or P.40 cases, continue on finding correct pattern type */
            if ((*s == '.') || (*s == '0'))
            {
                return 0;
            }
            *pErrCode = AU_INFO_FAILURE(TM_ERR_C_INT_INVALID_FORMAT);
            LOGERR("Invalid format! Cannot find \"=\" after %s in [%s]", pszPatType, pszINIInfo);
            return 0;
        }

        s = (char *)strchr(pszINIInfo, COMMON_CHAR);
        if (s != NULL) 
        {
            s++;

            /* Skip space */
            while (*s == ' ') 
               s++;

            e = (char *) strchr(s, COMMON_CHAR);
            if (e != NULL) 
            {
                *e = 0;
                LOGDEBUG("The pattern version is [%s]", s);
                return atol(s);
            }
            else
            {
                LOGERR("Invalid format! Cannot find second\",\" in line [%s]", pszINIInfo);
            }
        }
        else
        {
            LOGERR("Invalid format! Cannot find first \",\" in line [%s]", pszINIInfo);
        }
    } 

    return 0;
}


static char*
_tmau_GetEngineVerString(char pszINIInfo[], char *pszEngType, TmErrorCode *pErrCode)
{
     char *s, *e;


     if (!strncmp(pszINIInfo, pszEngType, strlen(pszEngType))) 
     {
         s = (char *) strchr(pszINIInfo, ',');
         if (s == NULL)
         {            
             *pErrCode = AU_INFO_FAILURE(TM_ERR_C_INT_INVALID_FORMAT);
             LOGERR("Invalid format! Cannot find first \",\" in line [%s]", pszINIInfo);
             return NULL;
         }
         
         s = (char *) strchr(s+1, COMMON_CHAR);

         if (s != NULL) 
         {
             s++;

             /* Skip space, the format will Major.Minor.Build.  For example, 6.810.1006 */
             while (*s == ' ') 
                 s++;

             e = (char *) strchr(s, COMMON_CHAR);
             if (e != NULL) 
             {
                 char *szTmp;
                 *e = 0;

                 /* todo. need to check the feasibility of "Remove build number" */
                 if ((szTmp= (char *) strchr(s, DOT_CHAR)) != NULL) 
                 {
                      if ((szTmp=(char *) strchr(szTmp+1, DOT_CHAR)) != NULL)
                      {   
                           LOGDEBUG("The engine version is [%s], will remove build number", s);
                           *szTmp = 0;
                      }
                      else
                      {
                           LOGNOTICE("Not well format. Cannot find the build number of engine version, [%s]", s);
                      }
                 }
                 else
                 {
                     LOGWARNING("Invalid format! Cannot find \".\" after engine major version, [%s]", s);
                 }
                 return s;
             }
         }

         *pErrCode = AU_INFO_FAILURE(TM_ERR_C_INT_INVALID_FORMAT);
         LOGERR("Invalid format! Cannot find second\",\" in line [%s]", pszINIInfo);
     }

     return NULL;
}

/**
 * @brief   Parse the SERVER.INI information to extract the version information of requested
 *          pattern type.  The pattern type here is fixed as "P.4".
 *
 * @param   pszINIFile: the location of server.ini
 * @param   pszPatType: pointer to requested pattern type.  For example, the "P.4"
 * @param   pTmuVer: pointer to a TmuVersion_t data
 * @param   pErrCode: address of error code
 *
 * @return  true or false to indicate the execution result
 *
 * @remarks If the IHV do not have the file system, the pINIFile need to point to the precisely
 *          buffer
 */
bool_t
tmau_GetPatternVer(char pszINIFile[], char *pszPatType, TmuVersion_t *pTmuVer, TmErrorCode *pErrCode)
{
    enum SECTION section = SEC_OTHER; 
    FILE        *fp;
    char        szBuffer[TM_AU_MAX_STR];
    uint32_t    dwPattern = 0;


    if (TM_FAILURE(*pErrCode))
    {
        LOGNOTICE("The input error code argument indicate a failure condition, [%08x]!", *pErrCode);
        *pErrCode = AU_INFO_SUCCESS;
    }
    
    if ((fp = fopen(pszINIFile, "r")) == NULL) 
    {
        *pErrCode = AU_INFO_FAILURE(TM_ERR_C_INT_FILE_NOT_FOUND);
        return FALSE;
    }

    /* Parsing server.ini line by line */
    while (fgets(szBuffer, TM_AU_MAX_STR, fp) != NULL) 
    {
        /* Stop parsing if already got the answers we want */
        if (dwPattern != 0) 
            break;

        /* Continue if the line is a comment */
        if (szBuffer[0] == ';')
            continue;

        /* Determine which section we are parsing */
        if (szBuffer[0] == '[') 
        {
            section = _tmau_GetSectionName(szBuffer);
            if (section != SEC_PATTERN)
            {
                section = SEC_OTHER;                
            }
            continue;
        } 

        /* Only care about PATTERN section */
        if (section == SEC_PATTERN) 
        {   
            dwPattern = _tmau_GetPatternVerString(szBuffer, pszPatType, pErrCode);
            pTmuVer->version = dwPattern;
            if (dwPattern > 0)
            {
                break;
            }
        }
    }

    fclose(fp);        
    return TRUE;
}



/**
 * @brief   Parse the SERVER.INI information to extract the version information of requested
 *          engine type
 *
 * @param   pszINIFile: the location of server.ini
 * @param   pszEngType: pointer to requested engine type.  For example, the "E.2="
 * @param   pTmuVer: pointer to a TmuVersion_t data
 * @param   pErrCode: address of error code.
 *
 * @return  true or false to indicate the execution result
 *
 * @remarks If the IHV do not have the file system, the pINIFile need to point to the precisely
 *          buffer
 */
bool_t
tmau_GetEngineVer(char pszINIFile[], char *pszEngType, TmuVersion_t *pTmuVer, TmErrorCode *pErrCode)
{
    enum SECTION section = SEC_OTHER; 
    FILE        *fp;
    char        szBuffer[TM_AU_MAX_STR];
    char        szEngineVersion[TM_AU_MAX_VER];
    char        *pszTmp;  
    

    if (TM_FAILURE(*pErrCode))
    {
        LOGNOTICE("The input error code argument indicate a failure condition, [%08x]!", *pErrCode);
        *pErrCode = AU_INFO_SUCCESS;
    }     
             
    if ((fp = fopen(pszINIFile, "r")) == NULL) 
    {
        *pErrCode = AU_INFO_FAILURE(TM_ERR_C_INT_FILE_NOT_FOUND);
        return FALSE;
    }
   
    memset(szEngineVersion, 0x00, TM_AU_MAX_VER);
 
    /* Parsing server.ini line by line */
    while (fgets(szBuffer, TM_AU_MAX_STR, fp) != NULL) 
    {
        /* Stop parsing if already got the answers we want */
        if (szEngineVersion[0] != 0) 
            break;

        /* Continue if the line is a comment */
        if (szBuffer[0] == ';')
            continue;

        /* Determine which section we are parsing */
        if (szBuffer[0] == '[') 
        {
            section = _tmau_GetSectionName(szBuffer);
            
            if (section != SEC_ENGINE)
            {
                section = SEC_OTHER;                
            }
            continue;
        } 

        /* Only care about ENGINE section */
        if (section == SEC_ENGINE) 
        {
            /*
            if (szEngineVersion[0] != 0) 
                break;
            */

            if ((pszTmp = _tmau_GetEngineVerString(szBuffer, pszEngType, pErrCode)) != NULL)
            {
                *pTmuVer = MakeVersionBuild (pszTmp);
                LOGDEBUG("The engine structure, %08x", pTmuVer->version);
                
                Strncpy(szEngineVersion, pszTmp, TM_AU_MAX_VER);
                
                break;
            } 
        }
    }

    fclose(fp);    
    return TRUE;
}
