%{
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include "tmcfg.h"
#include "tmcfg_impl.h"
#include "log.h"
#include "tmcfg_st.h"
  static const char ModuleName[]="libtmcfg";

#define YYPARSE_PARAM parse_arg

static void yyerror(char *s) {
  fprintf(stderr, "%s\n", s);
}

%}


%start config
%token QS MODULE PARAM NAME VALUE OPTION

%token <charval>  QS

%type <moduleval> config
%type <moduleval> module
%type <charval>   module_s
%type <moduleval> module_list
%type <paramval>  param
%type <paramval>  param_list

%union {
  char                 *charval;
  struct tag_module_st *moduleval;
  struct tag_param_st  *paramval;
}

%%

config:        {
  $$=NULL;
}
|              module_list {
  $$=$1;
  $$->prev->next=NULL;
  $$->prev=NULL;
  {
    struct tag_module_st **ppconfig=(struct tag_module_st **)parse_arg;
    *ppconfig=$$;
  }
}
  ;

param_list:    param {
  $$=$1;
}
|              param_list param {
  $$=$1;
  $2->next=$$;
  $2->prev=$$->prev;
  $$->prev->next=$2;
  $$->prev=$2;
}
  ;

param:         '<' PARAM NAME '=' QS VALUE '=' QS '/' '>' {
  $$=(struct tag_param_st*)malloc(sizeof(struct tag_param_st));
  $$->name=$5;
  $$->value=$8;
  $$->option=NULL;
  $$->prev=$$->next=$$;
}
|              '<' PARAM NAME '=' QS VALUE '=' QS OPTION '=' QS '/' '>' {
  $$=(struct tag_param_st*)malloc(sizeof(struct tag_param_st));
  $$->name=$5;
  $$->value=$8;
  $$->option=$11;
  $$->prev=$$->next=$$;
}
  ;

module_list:   module {
  $$=$1;
}
|              module_list module {
  $$=$1;
  $2->next=$$;
  $2->prev=$$->prev;
  $$->prev->next=$2;
  $$->prev=$2;
}
  ;

module:        module_s module_e {
  $$=(struct tag_module_st*)malloc(sizeof(struct tag_module_st));
  $$->name=$1;
  $$->params=NULL;
  $$->prev=$$->next=$$;
}
|              module_s param_list module_e {
  $$=(struct tag_module_st*)malloc(sizeof(struct tag_module_st));
  $$->name=$1;
  $$->params=$2;
  $2->prev->next=NULL;
  $2->prev=NULL;
  $$->prev=$$->next=$$;
}
  ;

module_s:      '<' MODULE NAME '=' QS '>' { $$=$5; }
  ;

module_e:      '<' '/' MODULE '>'
  ;

%%

int tmcfg_close(tmcfg_t config) {
  if(config==NULL) {
    return -1;
  }

  if(config->filename) free(config->filename);
  if(config->file) {
    /* unlock it first */
    if(config->file_lock!=0) {
      struct flock lock;
      lock.l_whence=SEEK_SET;
      lock.l_start=0;
      lock.l_len=0;
      lock.l_type=F_UNLCK;
      while(-1==fcntl(fileno(config->file), F_SETLKW, &lock) && errno==EINTR);
      config->file_lock=0;
    }
    fclose(config->file);
  }
  if(config->modules) {
    /* free config->modules */
    struct tag_module_st *pm, *pmn;
    struct tag_param_st *pp, *ppn;
    for(pm=config->modules; pm!=NULL; pm=pmn) {
      pmn=pm->next;
      for(pp=pm->params; pp!=NULL; pp=ppn) {
	ppn=pp->next;
	if(pp->name) {
	  free(pp->name);
	}
	if(pp->value) {
	  free(pp->value);
	}
	if(pp->option) {
	  free(pp->option);
	}
	free(pp);
      }
      if(pm->name) {
	free(pm->name);
      }
      free(pm);
    }
  }
  free(config);
  return 0;
}

tmcfg_t tmcfg_open(const char *filename, tmcfg_open_mode_t mode) {
  int parse_ret;
  struct tmcfg_st *config=NULL;
  config=(struct tmcfg_st*)malloc(sizeof(struct tmcfg_st));
  if(config==NULL) {
    LOGERROR("tmcfg_open: malloc: %m");
    goto error;
  }

  memset(config, 0, sizeof(struct tmcfg_st));

  config->mode=mode;
  config->file_lock=0;
  config->modified=0;

  config->filename=strdup(filename);
  if(config->filename==NULL) {
    LOGERROR("tmcfg_open: strdup(%s): %m", filename);
    goto error;
  }

  if(mode==TMCFG_MODE_UPDATE) {
    config->file=fopen(filename, "a+");
    if(config->file==NULL) {
      LOGERROR("tmcfg_open: open %s for \"a+\": %m", filename);
      goto error;
    }
  } else {
    config->file=fopen(filename, "r");
    if(config->file==NULL) {
      LOGERROR("tmcfg_open: open %s for \"r\": %m", filename);
      goto error;
    }
  }

  rewind(config->file);

  {
    struct flock lock;
    lock.l_whence=SEEK_SET;
    lock.l_start=0;
    lock.l_len=0;
    if(mode==TMCFG_MODE_UPDATE) {
      /* get write lock during the whole process */
      lock.l_type=F_WRLCK;
    } else {
      /* get read lock during parsing */
      lock.l_type=F_RDLCK;
    }
    while(-1==fcntl(fileno(config->file), F_SETLKW, &lock) && errno==EINTR);
    config->file_lock=1;
  }
  __cfg_set_yyin(config->file);
  parse_ret=yyparse(&(config->modules));


  /* not TMCFG_MODE_UPDATE */
  if(mode!=TMCFG_MODE_UPDATE) {
    /* release read lock */
    struct flock lock;
    lock.l_whence=SEEK_SET;
    lock.l_start=0;
    lock.l_len=0;
    lock.l_type=F_UNLCK;
    while(-1==fcntl(fileno(config->file), F_SETLKW, &lock) && errno==EINTR);
    config->file_lock=0;

    /* close file */
    fclose(config->file);
    config->file=NULL;
  }

  if(0==parse_ret) {
    return config;
  }

error:

  if(config) tmcfg_close(config);

  return NULL;

};

const char *tmcfg_get_param_value(tmcfg_t config, 
			     const char *module_name, 
			     const char *param_name,
			     const char **param_option) {
  struct tag_module_st *pm;
  struct tag_param_st *pp;

  if(config!=NULL) {
    for(pm=config->modules; pm!=NULL; pm=pm->next) {
      if(strcmp(pm->name, module_name)==0) {
	for(pp=pm->params; pp!=NULL; pp=pp->next) {
	  if(strcmp(pp->name, param_name)==0) {
	    if(param_option!=NULL) {
	      *param_option=pp->option;
	    }
	    return pp->value;
	  }
	}
      }
    }
  }
  return NULL;
}

int tmcfg_set_param_value(tmcfg_t config,
			   const char *module_name,
			   const char *param_name,
			   const char *param_value,
			   const char *param_option) {
  struct tag_module_st *pm, *nm;
  struct tag_param_st *pp, *np;
  char *mn, *pn, *pv, *po;

  nm=NULL;
  np=NULL;
  mn=pn=pv=po=NULL;

  if(config==NULL) {
    return -1;
  }

  config->modified=1;
  for(pm=config->modules; pm!=NULL; pm=pm->next) {
    if(strcmp(pm->name, module_name)==0) {
      for(pp=pm->params; pp!=NULL; pp=pp->next) {
	if(strcmp(pp->name, param_name)==0) {
	  /* modify pp->value, pp->option */
	  if((NULL!=param_value && NULL==(pv=strdup(param_value))) ||
	     (NULL!=param_option && NULL==(po=strdup(param_option)))) {
	    goto error;
	  }
	  
	  if(pp->value) free(pp->value);
	  if(pp->option) free(pp->option);

	  pp->value=pv;
	  pp->option=po;
	  return 0;
	}
      }
      /* insert new param to pm->params */
      if((NULL==(np=(struct tag_param_st*)malloc(sizeof(struct tag_param_st)))) ||
	 (NULL!=param_name && NULL==(pn=strdup(param_name))) ||
	 (NULL!=param_value && NULL==(pv=strdup(param_value))) ||
	 (NULL!=param_option && NULL==(po=strdup(param_option)))) {
	goto error;
	}

	np->name=pn;
	np->value=pv;
	np->option=po;
	
	np->next=pm->params;
	np->prev=NULL;
	if(pm->params != NULL)
		pm->params->prev=np;
	pm->params=np;
	return 0;
      }
  }
  /* insert new module to config->modules, new param to new module*/
  
  if((NULL==(nm=(struct tag_module_st*)malloc(sizeof(struct tag_module_st)))) || 
     (NULL==(np=(struct tag_param_st*)malloc(sizeof(struct tag_param_st)))) ||
     (NULL!=module_name && NULL==(mn=strdup(module_name))) ||
     (NULL!=param_name && NULL==(pn=strdup(param_name))) ||
     (NULL!=param_value && NULL==(pv=strdup(param_value))) ||
     (NULL!=param_option && NULL==(po=strdup(param_option)))) {
    goto error;
  }
  np->name=pn;
  np->value=pv;
  np->option=po;
  np->next=np->prev=NULL;
  nm->name=mn;
  nm->params=np;
  
  nm->next=config->modules;
  nm->prev=NULL;
  if(config->modules!=NULL) {
    config->modules->prev=nm;
  }
  config->modules=nm;
  return 0;

 error:

  LOGERROR("tmcfg_set_param_value: %m");
  if(nm) free(nm);
  if(np) free(np);
  if(mn) free(mn);
  if(pn) free(pn);
  if(pv) free(pv);
  if(po) free(po);
  return -1;
}

int	tmcfg_del_module(tmcfg_t config, const char *module_name) {
	struct tag_module_st	*pm;
	int	mlen;
	if(config == NULL)
		return -1;

	pm = config->modules;
	mlen = strlen(module_name);
	while(pm != NULL) {
		if(pm->name == NULL) {
			pm = pm->next;
			continue;
		}
		if(strncmp(pm->name, module_name, mlen) == 0) {
			struct tag_param_st	*pp, *tp;

			/* If it is the head of chain, chain it up */
			if(pm == config->modules)
				config->modules = pm->next;

			/* re-chain list */
			if(pm->prev != NULL)
				pm->prev->next = pm->next;
			if(pm->next != NULL)
				pm->next->prev = pm->prev;

			/* Free all parameter */
			pp = pm->params;
			if(pp!=NULL) {
				while(pp!=NULL) {
					tp = pp;
					pp = pp->next;
					if(tp->name!=NULL) free(tp->name);
					if(tp->value!=NULL) free(tp->value);
					if(tp->option!=NULL) free(tp->option);
					free(tp);
					config->modified=1;
				}
			}

			/* free memory */
			free(pm->name);
			free(pm);
			return 1;
		}
		pm = pm->next;
	}
	return 0;
}

int	tmcfg_del_param(tmcfg_t config, const char *module_name, const char *param_name) {
	struct tag_module_st    *pm;
	int	mlen, plen;
	if(config == NULL)
		return -1;
	pm = config->modules;
	mlen = strlen(module_name);
	plen = strlen(param_name);
	while(pm!=NULL) {
		if(pm->name == NULL) {
			pm = pm->next;
			continue;
		}
		if(strncmp(pm->name, module_name, mlen) == 0) {
			struct tag_param_st     *pp;
			pp = pm->params;
			if(pp==NULL) 
				return 0;

			while(pp!=NULL) {
				if(pp->name==NULL) {
					pp = pp->next;
					continue;
				}
				
				/* find the parameter, free it */
				if(strncmp(pp->name, param_name, plen) == 0) {
					free(pp->name);
					if(pp->value!=NULL)free(pp->value);
					if(pp->option!=NULL) free(pp->option);

					/* re-chain list */
					if(pp->prev==NULL)
						pm->params = pp->next;
					if(pp->prev!=NULL)
						pp->prev->next = pp->next;
					if(pp->next!=NULL)
						pp->next->prev = pp->prev;
					free(pp);
					config->modified=1;
					return 1;
				} // strncmp param
				pp = pp->next;
			} // while pp
		} // strncmp module
		pm = pm->next;
	}
	return 0;
}

static void __tmcfg_qs_write(FILE *out, const char *p) {
  const char *q;
  for(q=p; *p!='\0'; p++) {
    if(*p=='\"') {
      fwrite(q, 1, p-q+1, out);
      q=p;
    }
  }
  if(p>q) {
    fwrite(q, 1, p-q, out);
  }
}

int tmcfg_flush(tmcfg_t config) {
  struct tag_module_st *pm;
  struct tag_param_st *pp;
  if(config==NULL || config->file==NULL || config->file_lock==0) {
    return -1;
  }

  if(config->modified==0) {
    return 0;
  }

  if(0!=ftruncate(fileno(config->file), 0)) {
    LOGERROR("tmcfg_flush: ftruncate(%d): %m", fileno(config->file));
    return -1;
  }
  rewind(config->file);

  for(pm=config->modules; pm!=NULL; pm=pm->next) {
    fwrite("<module name=\"", 1, 14, config->file);
    if(pm->name) {
      __tmcfg_qs_write(config->file, pm->name);
    }
    fwrite("\">\n", 1, 3, config->file);

    for(pp=pm->params; pp!=NULL; pp=pp->next) {

      fwrite("  <param name=\"", 1, 15, config->file);
      if(pp->name) {
	__tmcfg_qs_write(config->file, pp->name);
      }
      fputc('\"', config->file);

      fwrite(" value=\"", 1, 8, config->file);
      if(pp->value) {
      __tmcfg_qs_write(config->file, pp->value);
      }
      fputc('\"', config->file);

      if(pp->option) {
	fwrite(" option=\"", 1, 9, config->file);
	__tmcfg_qs_write(config->file, pp->option);
	fputc('\"', config->file);
      }

      fwrite("/>\n", 1, 3, config->file);
    }
    fwrite("</module>\n\n", 1, 11, config->file);
  }

  fflush(config->file);
  sync();
  config->modified=0;

  return 0;
}
