/******************************************************************************
 *
 *  (C) Copyright 2004 Trend Micro, Inc.
 *
 *  All Rights Reserved.
 *  This program is an unpublished copyrighted work which is proprietary
 *  to Trend Micro, Inc. and contains confidential information that is not
 *  to be reproduced or disclosed to any other person or entity without
 *  prior written consent from Trend Micro, Inc. in each and every instance.
 *
 *  WARNING:  Unauthorized reproduction of this program as well as
 *  unauthorized preparation of derivative works based upon the
 *  program or distribution of copies by sale, rental, lease or
 *  lending are violations of federal copyright laws and state trade
 *  secret laws, punishable by civil and criminal penalties.
 *
 *****************************************************************************/

/**
 *  @file   tmGetAUInfo.c
 *  @brief  TMOG AcitveUpdate Information Query Module
 */

/*
 *  Change Log
 *  2004-03-19 Jeremy Wu    Initialization version
 */

#include <sys/shm.h>
#include <sys/ipc.h>
#include <string.h>
#include <stdlib.h>
#include "log.h"
#include "tmShm.h"
#include "tmGetAUInfo.h"

static const char ModuleName[] = "tmGetAUInfo";

/**
 * @brief   Query the newest AcitveUpdate information from a global accessible
 *          data area.  In our Linux implementation, this area is a shared
 *          memory; it can be as simple as allocated memory if that is 
 *          accessible among process.
 *
 * @param   pTmuVer: the target structure to hold the AcitveUpdate information,
 *          which is defined in tmauIniParser.h
 *
 * @remarks The fields declared in shared memory for TmuAUVersion_t:
 *          1. product= supported product ID, according to AU module's assignment
 *          2. engine = a TmuVerion_t data type to record VSAPI engine version
 *                    number for the supported product.
 *                    For example, 6.910 will store as 0x00069100
 *          3. pattern= a TmuVersion_t data type to record VSAPI pattern version
 *                    number for the supported product.
 *                    E.g., 1.782.00 will store as 0x0002B818(hex); 178200(dec)
 *
 * @remarks The hardware vendor may need to porting the shm_init(), shm_attach(),
 *          shm_detach(), etc.
 *
 * @return  void
 */
 
int checkIfAUInfoShareMemExist(void)
{
    int shmid = -1;

	if ((shmid = shmget(TMSS_AUINFO_SHMKEY, sizeof(TmuAUVersion_t)*TM_AU_PRODUCT_NUM, 0)) == -1)
		return 0;
	return 1;
}

void tmau_GetNewestAUInfo(TmuAUVersion_t *pTmuVer)
{
    int32_t           nShmID = 0;
    TmuAUVersion_t    *pAUInfoPool;


    nShmID = shm_init(TMSS_AUINFO_SHMKEY, sizeof(TmuAUVersion_t)*TM_AU_PRODUCT_NUM);
    pAUInfoPool = (TmuAUVersion_t *)shm_attach(nShmID);
    LOGDEBUG("tmau_GetNewestAUInfo address = %p\n", pAUInfoPool);


    if (pAUInfoPool == NULL)
    {
        LOGERROR("Attach AU share memory fail");
    }
    else
    {
        LOGDEBUG("The support product ID is: [%d]\n",   pAUInfoPool[0].product);
        LOGDEBUG("   VASPI Engine number is: [%08x]\n", pAUInfoPool[0].engine.version);
        LOGDEBUG("  VASPI Pattern number is: [%08x]\n", pAUInfoPool[0].pattern.version);                        
        pTmuVer->product = pAUInfoPool[0].product;
        pTmuVer->engine.version  = pAUInfoPool[0].engine.version;
        pTmuVer->pattern.version = pAUInfoPool[0].pattern.version;

        shm_detach(pAUInfoPool);
    }

    return;
}


/**
 * @brief   Reset the AcitveUpdate information from a global accessible
 *          data area.  
 *
 * @param   void
 *
 * @remarks The TmuAUVersion_t description and porting notice are same as
 *          tmau_GetNewestAUInfo(). 
 *
 * @remarks The interface is designed for testing purpose.  Only test 
 *          program will call it.
 *
 * @return  void
 */
void tmau_ResetAUInfo(void)
{
    int32_t           nShmID = 0;
    TmuAUVersion_t    *pAUInfoPool;
    
    nShmID = shm_init(TMSS_AUINFO_SHMKEY, sizeof(TmuAUVersion_t)*TM_AU_PRODUCT_NUM);
    pAUInfoPool = (TmuAUVersion_t *)shm_attach(nShmID); 
    LOGDEBUG("tmau_ResetAUInfo address = %p\n", pAUInfoPool);

    if (pAUInfoPool == NULL)
    {
        LOGERROR("Attach AU share memory fail");
    }
    else
    {
        LOGDEBUG("The support product ID is: [%d]\n",   pAUInfoPool[0].product);
        LOGDEBUG("   VASPI Engine number is: [%08x]\n", pAUInfoPool[0].engine.version);
        LOGDEBUG("  VASPI Pattern number is: [%08x]\n", pAUInfoPool[0].pattern.version);                                                                                                              	pAUInfoPool[0].product = 0;
	pAUInfoPool[0].engine.version = 0;
        pAUInfoPool[0].pattern.version = 0;

        shm_detach(pAUInfoPool);
    } 

    return;
}

