/*
 * TrendChip TC3160 ethernet driver
 *
 * TODO:
 * 1. Read from ROM for MAC address.
 * 2. Setup filters.
 * 3. Reliablity check.
 * 4. Multicast handling?
 * 5. Handling ioctl calls?
 * 6. Handling multi-packet frame?
 *
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/spinlock.h>
#include <linux/interrupt.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/module.h>
#include <linux/compatmac.h>
#include <linux/mm.h>
#include <linux/init.h>
#include <linux/timer.h>
#include <asm/virtconvert.h>
#include <asm/tc3160/interrupts.h>
#include <asm/tc3160/eth.h>
//added by Robert
#include <asm/tc3160/gpio.h>

/* Uncomment this for debugging messages */
/* #define ETH_DEBUG */

/* Uncomment this for error messages */
/* #define ETH_ERRMSG */

/* Uncomment this for normal messages */
#define ETH_MSG 

#ifdef ETH_DEBUG
#define DBG_PRINT	printk
#else
static void inline DBG_PRINT(const char *x, ...) { ; }
#endif /* ETH_DEBUG */

#ifdef ETH_ERRMSG
#define ERR_PRINT	printk
#else
static void inline ERR_PRINT(const char *x, ...) { ; }
#endif /* ETH_ERRMSG */

#ifdef ETH_MSG
#define MSG_PRINT	printk
#else
static void inline MSG_PRINT(const char *x, ...) { ; }
#endif /* ETH_ERRMSG */

/* Hack: this is the mac address by default */
#define DEF_MAC_HI	0x00000000
#define DEF_MAC_LO	0xaabbccdd

/* Default maximum link wait */
#define DEF_LINK_LOOP		1000
#define DEF_MAX_LINK_LOOP	0x1000000

/* NI | AI | RS | RU | RI */
#define ETH_IRQ_FLAGS	((1 << 16) | (1 << 15) | (1 << 8) | (1 << 7) | (1 << 6))

/* ST | SR */
#define ETH_RXTX_FLAGS	((1 << 13) | (1 << 1))

/* TE | RE | PR | RA | ASTP | HP | HO | DRTY | DBF */
#define ETH_MAC_FLAGS	((1 << 3) | (1 << 2) | (1 << 18) | (1 << 31) | (1 << 8) | \
		        (1 << 13) | (1 << 15) | (1 << 10) | (1 << 11))

/* Number of descriptors and buffer size */
#define NUM_RDESC	48
#define NUM_TDESC	16
#define R_BUF_SIZE	0x700	/* needs to be < 2KB */
#define T_BUF_SIZE	0x700	/* needs to be < 2KB */

/* Private data for this ethernet device */
typedef struct {
  struct net_device_stats stats; /* stats */
  unsigned long last_rxidx;      /* last rx idx to descriptor */
  unsigned long next_txidx;      /* next tx idx to descriptor */
  unsigned long last_txbytes;    /* last tx transferred bytes */
} TC3160_ETH_PRIV;

/* Descriptors and buffers: notice the alignments */
static struct tc3160_desc rxdsc[NUM_RDESC] __attribute__ ((__aligned__(16)));
static struct tc3160_desc txdsc[NUM_TDESC] __attribute__ ((__aligned__(16)));
static unsigned char rxbuf[NUM_RDESC * R_BUF_SIZE] __attribute__ ((__aligned__(4)));
static unsigned char txbuf[NUM_TDESC * T_BUF_SIZE] __attribute__ ((__aligned__(4)));

/* Reference count for device */
static int dev_count = 0;

/* For storing MAC address */
static unsigned long def_mac_hi = 0L;
static unsigned long def_mac_lo = 0L;

/* Timer */
static struct timer_list eth_timer;

/* Link status flag */
static volatile int link_up = 0;
static volatile int link_configured = 0;
static unsigned long max_link_loop = 0;
static unsigned long ld_count = 0;
static int reset_flag = 0;

/* Device data */
static struct net_device tc3160_eth_dev;

/*  MX29LV160BTC sector architecture for Word Mode(x16)
 *  total = SA0 - SA34 = 35 sectors
 *  SA0  - SA30 32Kwords = 32Kwords x 31 = (F8xxx)
 *  SA31	16Kwords = 16Kwords x  1 = (F8xxx)
 *  SA32 - SA33  4Kwords = 4Kwords  x  2 = (FFxxx)
 *  SA34	 8Kwords = 8Kwords  x  1 = (FExxx)
 */
 
#define FLASH_BASE	0xbc000000
#define FLASH_SIZE	0x200000
#define ADDR_1346	0x555
#define ADDR_25		0x2AA
#define FALSE		-1
#define TRUE		0

/*
 * be care for it will clear all data in chip
 */
/* Function prototypes */
static void tc3160_eth_intr_handler(int irq, void *dev_id, struct pt_regs *regs);
static int tc3160_eth_reset_desc(struct net_device *dev, int *reset);

/* Get the device mac address */
static int tc3160_get_macaddr(unsigned char *addr)
{
  /* TODO: Set up mac address, later on we need to read it from ROM */
  *addr++ = (def_mac_hi & 0x0000ff00) >> 8;
  *addr++ = (def_mac_hi & 0x000000ff);
  *addr++ = (def_mac_lo & 0xff000000) >> 24;
  *addr++ = (def_mac_lo & 0x00ff0000) >> 16;
  *addr++ = (def_mac_lo & 0x0000ff00) >> 8;
  *addr = (def_mac_lo & 0x000000ff);

  return(0);
}

/* Setting up MAC address of ethernet PHY and device data */
static int tc3160_set_mac(struct net_device *dev)
{
  unsigned long hi_mac, low_mac;

  /* Set up device mac address */
  if (tc3160_get_macaddr(&(dev->dev_addr[0])))
    return(-EIO);

  if (!is_valid_ether_addr(&(dev->dev_addr[0])))
    MSG_PRINT("%s: bogus mac address detected.\n", dev->name);

  hi_mac = (dev->dev_addr[0] << 8) | dev->dev_addr[1];
  low_mac = (dev->dev_addr[2] << 24) | (dev->dev_addr[3] << 16) | (dev->dev_addr[4] << 8) | dev->dev_addr[5];

  /* Set up MAC address */
  *(volatile unsigned long *)TC3160_MACAHR_REG = hi_mac;
  *(volatile unsigned long *)TC3160_MACALR_REG = low_mac;

  return(0);
}

/* Setting customized mac address */
static int tc3160_eth_set_macaddr(struct net_device *dev, void *addr)
{
  unsigned long hi_mac, low_mac;

  /* Check if given address is valid ethernet MAC address */
  if (!is_valid_ether_addr(addr))
    return(-EIO);

  /* Turn off IRQ and stop receive/transmit */
  *(volatile unsigned long *)TC3160_CR_REG = 0;
  *(volatile unsigned long *)TC3160_IER_REG = 0;

  netif_carrier_off(dev); /* Shutdown interface */

  /* Save the customize mac address */
  memcpy(&(dev->dev_addr[0]), addr, 6);

  hi_mac = (dev->dev_addr[0] << 8) | dev->dev_addr[1];
  low_mac = (dev->dev_addr[2] << 24) | (dev->dev_addr[3] << 16) | (dev->dev_addr[4] << 8) | dev->dev_addr[5];
  def_mac_hi = hi_mac;
  def_mac_lo = low_mac;

  /* Set up MAC address */
  *(volatile unsigned long *)TC3160_MACAHR_REG = hi_mac;
  *(volatile unsigned long *)TC3160_MACALR_REG = low_mac;

  /* Restart everything */
  link_up = link_configured = 0;

  return(0);
}

/* Writing to transceiver register: Broadcom AC101L */
static void tc3160_eth_mii_write(unsigned long paddr, unsigned long reg, unsigned long data)
{
  volatile unsigned long word;

  /* Wait until not busy */
  while (*(volatile unsigned long *)TC3160_MIIAR_REG & 0x1);
  *(volatile unsigned long *)TC3160_MIIDR_REG = data;

  /* address | register | write mode */
  word = ((paddr << 11) | (reg << 6) | (1 << 1));
    
  /* Wait until not busy */
  while (*(volatile unsigned long *)TC3160_MIIAR_REG & 0x1);
  *(volatile unsigned long *)TC3160_MIIAR_REG = word;
}

/* Reading from transceiver register: Broadcom AC101L */
static unsigned long tc3160_eth_mii_read(unsigned long paddr, unsigned long reg)
{
  volatile unsigned long word;

  /* address | register | write mode */
  word = ((paddr << 11) | (reg << 6));
    
  /* Wait until not busy */
  while (*(volatile unsigned long *)TC3160_MIIAR_REG & 0x1);
  *(volatile unsigned long *)TC3160_MIIAR_REG = word;
    
  /* Wait until not busy */
  while (*(volatile unsigned long *)TC3160_MIIAR_REG & 0x1);
  return(*(volatile unsigned long *)TC3160_MIIDR_REG);
}

/* Initialize the ethernet link status, mac, and flow control */
static int tc3160_link_config(struct net_device *dev, int timeout)
{
  unsigned long word = 0, link = 0;
  unsigned long oldjiffies = 0;
  const unsigned long link_loop = timeout; /* Max. wait for link to established */

  /* If link is up and configured, don't do anything */
  if ((link_up != 0) && (link_configured != 0))
    return(0);
  else
    DBG_PRINT("%s: re-configure link (%ld).\n", dev->name, link_loop);

  /* Turn off IRQ and stop receive/transmit */
  *(volatile unsigned long *)TC3160_CR_REG = 0;
  *(volatile unsigned long *)TC3160_IER_REG = 0;

#if 0 
  /* configure PHY: FDDC/100TXFD/100TXHD/10TXFD/10TXHD capable */
  word = (1 << 0) | (1 << 5) | (1 << 6) | (1 << 7) | (1 << 8) | (1 << 10);
  tc3160_eth_mii_write(0, 4, word);
    
  /* 100Mbps/AutoNego/Restart Nego/FD/no loopback */
  word = (0 << 14) | (1 << 13) | (1 << 12) | (1 << 9) | (1 << 8);
  tc3160_eth_mii_write(0, 0, word);

  /* Wait for link to established */
  for (oldjiffies = jiffies, word = 0; ((oldjiffies + link_loop) > jiffies); word++) { 
    link = tc3160_eth_mii_read(0, 1);
    if (link & (1 << 2)) { /* check link status */
      link_up = 1;
      break;
    } 
  }

  if (word >= link_loop) {
    DBG_PRINT("%s: cannot established link (timeout: %lu).\n", dev->name, link_loop);
    link_up = link_configured = 0;
    return(-EIO);
  } else {
    MSG_PRINT("%s: link established, ", dev->name);
  }
#endif

  link_up = 1;
  MSG_PRINT("%s: link established, ", dev->name);


  *(volatile unsigned long *)TC3160_BMR_REG = 0x00000001; /* DMA reset on */
  *(volatile unsigned long *)TC3160_BMR_REG = 0x00000001; /* Hack: reset it more */
  *(volatile unsigned long *)TC3160_BMR_REG = 0x00000001; /* than once due to HW */
  *(volatile unsigned long *)TC3160_BMR_REG = 0x00000001; /* problem */
  *(volatile unsigned long *)TC3160_BMR_REG = 0x00000001;
  *(volatile unsigned long *)TC3160_BMR_REG = 0x00000000; /* DMA reset off */

  /* Set the descriptor base address */
  *(volatile unsigned long *)TC3160_RLBAR_REG = virt_to_phys(&(rxdsc[0]));
  *(volatile unsigned long *)TC3160_TLBAR_REG = virt_to_phys(&(txdsc[0]));

  word = MiiIfExtPhyMii;
  //link = tc3160_eth_mii_read(0, 1);
//  if (link & (3 << 13)) {                       /* 100Base-Tx */
    word |= MiiIfSpeed100M;
    MSG_PRINT("100TX, ");
//  } else if (word & (3 << 11)) {                /* 10Base-Tx */
//    word |= MiiIfSpeed10M;
//    MSG_PRINT("10TX, ");
//  }

  /* configure PHY if, MII with 100Mbps or 10Mbps */
  *(volatile unsigned long *)TC3160_IFR_REG = word; 

  /* burst length = 32 | BLE | ~DBO */
  word = (32 << 8) | (1 << 7) /* | (1 << 20) */;
  *(volatile unsigned long *)TC3160_BMR_REG = word; /* Bus mode */

  /* FCE */
  word = (1 << 1);
  *(volatile unsigned long *)TC3160_FCR_REG = word; /* enable MAC flow ctrl */
    
  word = ETH_MAC_FLAGS;
  word |= MacLoopbackOff; /* OM = 0 */
//  link = tc3160_eth_mii_read(0, 1);
//  if (link & (5 << 12)) {
    word |= MacFullDuplex;   /* FDM */               
    MSG_PRINT("full duplex");
//  } else {
//    MSG_PRINT("half duplex");
//  }
  MSG_PRINT(".\n");

  *(volatile unsigned long *)TC3160_MACCR_REG = word; /* configure MAC ctrller */

  /* Turn on the IRQ and start receive/transmit */
  *(volatile unsigned long *)TC3160_IER_REG = ETH_IRQ_FLAGS;
  *(volatile unsigned long *)TC3160_CR_REG = ETH_RXTX_FLAGS; /* enable MAC engine */

  link_configured = 1;
  return(0);
}

/* Ethernet hardware initialization */
static int tc3160_eth_hw_init(struct net_device *dev)
{
  if (tc3160_set_mac(dev))
    return(-EIO);

  /* Stop MAC engine */
  *(volatile unsigned long *)TC3160_CR_REG = 0x0; 

  /* Set up multicast hash table to MCHTHR MCHTLR */
  /* Set multicast hash table to accept all */
  *(volatile unsigned long *)TC3160_MCHTHR_REG = 0xffffffff; 
  *(volatile unsigned long *)TC3160_MCHTLR_REG = 0xffffffff; 

  /* Resetting descriptors */
  if (tc3160_eth_reset_desc(&tc3160_eth_dev, &reset_flag))
    return(-EIO);

  /* Configure IFR with PHY and speed, as well as PHY */
  tc3160_link_config(dev, max_link_loop);

  return(0);
}

/* Monitor the status of link, re-do link initialization if necessary. */
static void tc3160_eth_link_monitor(unsigned long dummy)
{
  unsigned long link_status;
  struct net_device *dev = (struct net_device *)dummy;
  int res;

  /* Setup the link/configuration is not done so. */
  if ((link_up == 0) || (link_configured == 0))
    res = tc3160_eth_hw_init(dev);

  /* Get link status and adjust the delay loop accordingly */
//  link_status = ((tc3160_eth_mii_read(0, 1) & (1 << 2)) ? 1 : 0);
  link_status = 1;
  if (link_status) {
    if ((link_configured != 0) && (link_up != 0)) {
      if (!netif_carrier_ok(dev)) {
        netif_carrier_on(dev); /* Up and running, turn netif on */
        max_link_loop = DEF_LINK_LOOP;
        ld_count >>= 2;
        if (ld_count == 0)
	  ld_count = 1;
      }
    } else {
      ld_count <<= 1;
      max_link_loop = ld_count * DEF_LINK_LOOP;
      if (max_link_loop > DEF_MAX_LINK_LOOP)
	max_link_loop = DEF_MAX_LINK_LOOP;
      netif_carrier_off(dev);
      link_configured = 0;
    }
  } else {
    MSG_PRINT("%s: detected link down.\n", dev->name);
    ld_count <<= 1;
    max_link_loop = ld_count * DEF_LINK_LOOP;
    if (max_link_loop > DEF_MAX_LINK_LOOP)
      max_link_loop = DEF_MAX_LINK_LOOP;
    netif_carrier_off(dev);
    link_up = link_configured = 0;
  }

  /* Schedule for the next time */
  mod_timer(&eth_timer, jiffies + (HZ / 2)); /* Check roughly every 0.5 seconds */
}

/* Setting rx/tx descriptors */
static void tc3160_eth_setup_desc(void)
{
  register int i;
  struct tc3160_desc *desc_ptr = NULL;
  unsigned long base_addr = 0;

  /* Setup rx descriptor */
  for (i = 0, desc_ptr = (struct tc3160_desc *)KSEG1ADDR(&rxdsc[0]); 
	i < (NUM_RDESC - 1); i++, desc_ptr++) {
    desc_ptr->desc3 = (unsigned long)virt_to_phys(desc_ptr + 1);
    desc_ptr->desc1 = (0x01000000 | R_BUF_SIZE);
    desc_ptr->desc0 = 0x80000000; /* Owned by DMA */
  }
  desc_ptr->desc3 = (unsigned long)virt_to_phys(&rxdsc[0]);
  desc_ptr->desc1 = (0x03000000 | R_BUF_SIZE);
  desc_ptr->desc0 = 0x80000000; /* Owned by DMA */

  /* Setup tx descriptor */
  for (i = 0, desc_ptr = (struct tc3160_desc *)KSEG1ADDR(&txdsc[0]); 
	i < (NUM_TDESC - 1); i++, desc_ptr++) {
    desc_ptr->desc3 = (unsigned long)virt_to_phys(desc_ptr + 1);
    desc_ptr->desc1 |= (0x61000000 | T_BUF_SIZE);
    desc_ptr->desc0 = 0x0; /* Owned by CPU */
  }
  desc_ptr->desc3 = (unsigned long)virt_to_phys(&txdsc[0]);
  desc_ptr->desc1 = (0x63000000 | T_BUF_SIZE);
  desc_ptr->desc0 = 0x0; /* Owned by CPU */

  /* Point rx descriptor to buffer */
  for (i = 0, desc_ptr = (struct tc3160_desc *)KSEG1ADDR(&rxdsc[0]), 
	base_addr = (unsigned long)virt_to_phys(&rxbuf[0]); 
	i < NUM_RDESC; i++, desc_ptr++, base_addr += R_BUF_SIZE) 
    desc_ptr->desc2 = base_addr;

  /* Point tx descriptor to buffer */
  for (i = 0, desc_ptr = (struct tc3160_desc *)KSEG1ADDR(&txdsc[0]),
	base_addr = (unsigned long)virt_to_phys(&txbuf[0]);
	i < NUM_TDESC; i++, desc_ptr++, base_addr += T_BUF_SIZE) 
    desc_ptr->desc2 = base_addr;
}

/* Starting up the ethernet device */
static int tc3160_eth_open(struct net_device *dev)
{
  TC3160_ETH_PRIV *private = (TC3160_ETH_PRIV *)dev->priv;

  if ((dev == NULL) || (private == NULL))
    return(-EIO);

  MOD_INC_USE_COUNT;
  DBG_PRINT("%s: starting interface.\n", dev->name);

  /* Schedule timer for monitoring link status */
  init_timer(&eth_timer);
  eth_timer.function = tc3160_eth_link_monitor;
  eth_timer.data = (unsigned long)dev;
  mod_timer(&eth_timer, jiffies + (HZ * 2));

  netif_start_queue(dev);

  return(0);
}

/* Stopping the ethernet device */
static int tc3160_eth_close(struct net_device *dev)
{
  if (dev == NULL)
    return(-EIO);

  /* Turn off IRQ and stop receive/transmit */
  *(volatile unsigned long *)TC3160_CR_REG = 0;
  *(volatile unsigned long *)TC3160_IER_REG = 0;

  /* Kill timer */
  del_timer_sync(&eth_timer);

  /* Stop the transmit queue */
  netif_stop_queue(dev);
  link_up = link_configured = 0;

  MOD_DEC_USE_COUNT;

  return(0);
}

/* Setup multicast list */
static void tc3160_eth_set_multicast_list(struct net_device *dev)
{
  return; /* Do nothing */
}

/* Transmit a packet */
static int tc3160_eth_tx(struct sk_buff *skb, struct net_device *dev)
{
  TC3160_ETH_PRIV *private = NULL;
  struct net_device_stats *stats = NULL;
  struct tc3160_desc *desc_ptr = NULL;
  unsigned long length = 0;

  if ((skb == NULL) || (dev == NULL)) {
    return(-EIO);
  } else if ((link_up == 0) || (link_configured == 0)) {
    netif_carrier_off(dev);
    return(-EIO); /* Link is not up yet */
  } else {
    private = (TC3160_ETH_PRIV *)dev->priv;
    stats = &(private->stats);
  }

  if ((length = ((ETH_ZLEN < skb->len) ? skb->len : ETH_ZLEN)) > T_BUF_SIZE)
    return(-ENOMEM); /* Too big */

  /* Send packet to device */
  desc_ptr = (struct tc3160_desc *)KSEG1ADDR(&txdsc[private->next_txidx]);
  if (desc_ptr->desc0 & 0x80000000) {
    stats->tx_dropped++;
    ERR_PRINT("%s: tx error (descriptor not owned by CPU).\n", dev->name);
    return(-EIO);
  } 

  /* Check if previous transmission has error */
  if (desc_ptr->desc0 & 0x00000e82) {
    ERR_PRINT("%s: tx error 0x%08lx\n", dev->name, desc_ptr->desc0);

    /* Update stats */
    stats->tx_errors++;
    stats->tx_packets--;
    stats->tx_bytes -= private->last_txbytes;
  }

  /* Copy packet data to tx buffer */
  memcpy((unsigned char *)KSEG1ADDR(desc_ptr->desc2), skb->data, length);

  /* Setup tx descriptor */
  if (private->next_txidx != (NUM_TDESC - 1))
    desc_ptr->desc1 = (0x61000000 | length);
  else
    desc_ptr->desc1 = (0x63000000 | length);
  desc_ptr->desc0 |= 0x80000000;

  dev->trans_start = jiffies;
  private->next_txidx = ((private->next_txidx + 1) % NUM_TDESC);

  /* Start transmission (should be in suspend mode already) */
  *(volatile unsigned long *)TC3160_TPDR_REG = 0x0;

  /* Update stats */
  stats->tx_packets++;
  stats->tx_bytes += length;
  private->last_txbytes = length;

  /* Free up socket buffer */
  dev_kfree_skb(skb);
  return(0);
}

/* Receiving packet(s) */
static int tc3160_eth_rx(struct net_device *dev)
{
  TC3160_ETH_PRIV *private = NULL;
  struct sk_buff *skb = NULL;
  struct net_device_stats *stats = NULL;
  struct tc3160_desc *desc_ptr = NULL;
  unsigned char *data = NULL;
  unsigned long length = 0, dsc = 0;
  unsigned long count = 0, cnt = 0;
  unsigned long nidx = 0;

  if (dev == NULL)
    return(-EIO);
  else {
    private = (TC3160_ETH_PRIV *)dev->priv;
    stats = &(private->stats);
  }

  /* The one we're starting to read from */
  nidx = ((private->last_rxidx + 1) % NUM_RDESC);

  /* Go thru descriptors list until we ran out or encounterred the one
     being processed by DMA */
  for (cnt = 0, count = nidx; cnt < NUM_RDESC; cnt++, count = ((count + 1) % NUM_RDESC)) {
    desc_ptr = (struct tc3160_desc *)KSEG1ADDR(&rxdsc[count]);
    dsc = desc_ptr->desc0;
    if (dsc & 0x80000000) {
      /* DMA is processing this one, break out the loop */
      break;
    } else if (dsc & 0x4000d8ca) {
      DBG_PRINT("%s: rx error 0x%08lx:%ld\n", dev->name, dsc, count);
      /* We dropped any error packet */
      stats->rx_errors++;
      desc_ptr->desc0 = 0x80000000;
    } else if ((dsc & 0x00000300) != 0x00000300) {
      ERR_PRINT("%s: multi-packet frame detected 0x%08lx\n", dev->name, dsc);
      /* We don't handle multi-packet frame, for now */
      stats->rx_dropped++;
      desc_ptr->desc0 = 0x80000000;
    } else if ((length = ((desc_ptr->desc0 & 0x3fff0000) >> 16)) > R_BUF_SIZE) {
      DBG_PRINT("%s: rx dropped (size too large: %ld)\n", dev->name, length);
      stats->rx_dropped++;
      desc_ptr->desc0 = 0x80000000;
    } else if ((skb = dev_alloc_skb(length + 2)) == NULL) {
      DBG_PRINT("%s: rx dropped (memory unavailable)\n", dev->name);
      stats->rx_dropped++;
      desc_ptr->desc0 = 0x80000000;
    } else {
      data = (unsigned char *)KSEG1ADDR(desc_ptr->desc2);
      skb->dev = dev;
      skb_reserve(skb, 2);
      memcpy(skb_put(skb, length), data, length); 
      desc_ptr->desc0 = 0x80000000;
      skb->protocol = eth_type_trans(skb, dev);
      skb->ip_summed = CHECKSUM_NONE;

      /* Update stats */
      stats->rx_packets++;
      stats->rx_bytes += length;
      dev->last_rx = jiffies;

      /* Send the packet to kernel */
      netif_rx(skb); 
    }
    private->last_rxidx = count; /* This is the last desc we read */
  }

  if (cnt >= NUM_RDESC) 
    ERR_PRINT("%s: rx suspended (no receive buffer available).\n", dev->name);

  /* make sure rx is not suspended */
  *(volatile unsigned long *)TC3160_RPDR_REG = 0x0;

  return(0);
}

/* Resetting rx descriptor: error recovery for "rx buf unavailable" */
static void tc3160_eth_reset_rx_buf(struct net_device *dev)
{
  register int i;
  struct tc3160_desc *desc_ptr = NULL;

  ERR_PRINT("%s: resetting rx buf ..\n", dev->name);

  /* Reset rx desc */
  for (i = 0, desc_ptr = (struct tc3160_desc *)KSEG1ADDR(&rxdsc[0]); 
	i < NUM_RDESC; i++, desc_ptr++) 
    desc_ptr->desc0 = 0x80000000; /* Owned by DMA */

  *(volatile unsigned long *)TC3160_RPDR_REG = 0x0;
  *(volatile unsigned long *)TC3160_CR_REG |= 0x00000002; /* restart process */
}

/* Get the stats information */
static struct net_device_stats *tc3160_eth_stats(struct net_device *dev)
{
  if (dev != NULL) {
    return(&(((TC3160_ETH_PRIV *)dev->priv)->stats));
  } else
    return(NULL);
}

/* Ethernet interrupt handler */
static void tc3160_eth_intr_handler(int irq, void *dev_id, struct pt_regs *regs)
{
  unsigned long status;
  unsigned long mask = 0;
  struct net_device *dev = (struct net_device *)dev_id;

  if (dev_id != (void *)&tc3160_eth_dev) 
    return;

  /* Check status */
  status = *(volatile unsigned long *)TC3160_SR_REG;

  if (status & 0x00010000) { /* NIS */
    mask = 0x00010000;
    if (status & 0x00000040) { /* receive interrupt RI */
      mask |= 0x00000040;
      *(volatile unsigned long *)TC3160_SR_REG = mask;
      tc3160_eth_rx((struct net_device *)dev_id);
    } else {
      ERR_PRINT("%s: unhandled NIS 0x%08lx\n", dev->name, status);
      *(volatile unsigned long *)TC3160_SR_REG = status;
    }
  } else if (status & 0x00008000) { /* AIS */
    mask = 0x00008000;
    if (status & 0x00000080) { /* receive buffer unavaliable RU */
      ERR_PRINT("%s: receive buffer unavailable 0x%08lx\n", dev->name, status);
      mask |= 0x00000080;
      *(volatile unsigned long *)TC3160_SR_REG = mask;
      tc3160_eth_reset_rx_buf((struct net_device *)dev_id); 
    } else if (status & 0x00000100) { /* receive process stopped RPS */
      ERR_PRINT("%s: receive process stopped\n", dev->name);
      mask |= 0x00000100;
      *(volatile unsigned long *)TC3160_SR_REG = mask;
      *(volatile unsigned long *)TC3160_CR_REG |= 0x00000002; /* restart process */
    } else {
      ERR_PRINT("%s: unhandled AIS 0x%08lx\n", dev->name, status);
      *(volatile unsigned long *)TC3160_SR_REG = mask;
    }
  } else {
    ERR_PRINT("%s: Unknown SR 0x%08lx\n", dev->name, status);
    *(volatile unsigned long *)TC3160_SR_REG = status;
  }
}

/* Handling ioctl call */
static int tc3160_eth_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
  /* Not implemented yet */
  return(-EIO);
}

/* Kernel level ethernet initialization */
static int tc3160_eth_init(struct net_device *dev)
{
  if (dev == NULL)
    goto failed;

  /* Ethernet device initialization */
  ether_setup(dev);

  if (tc3160_set_mac(dev))
    goto failed;

  /* Hook up with handlers */
  dev->get_stats = tc3160_eth_stats;
  dev->hard_start_xmit = tc3160_eth_tx;
  dev->open = tc3160_eth_open;
  dev->stop = tc3160_eth_close;
  dev->set_multicast_list = tc3160_eth_set_multicast_list;
  dev->do_ioctl = tc3160_eth_ioctl;
  dev->set_mac_address = tc3160_eth_set_macaddr;

  dev->tx_queue_len = NUM_TDESC; 
  dev->flags &= ~IFF_MULTICAST;
  dev->flags |= IFF_DEBUG;

  return(0);

failed:
  return(-EIO);
}

static int tc3160_eth_reset_desc(struct net_device *dev, int *reset)
{
  TC3160_ETH_PRIV *private;

  if (dev == NULL)
    return(-EIO);
  else
    private = (TC3160_ETH_PRIV *)dev->priv;

  if (*reset)
    memset(dev->priv, 0, sizeof(TC3160_ETH_PRIV));
  *reset = 0;

  /* Clear all tx/rx buffer memory */
  memset((void *)KSEG1ADDR(&rxbuf[0]), 0, NUM_RDESC * R_BUF_SIZE);
  memset((void *)KSEG1ADDR(&txbuf[0]), 0, NUM_TDESC * T_BUF_SIZE);

  /* Initialize the indices */
  private->last_rxidx = NUM_RDESC - 1;
  private->next_txidx = 0;

  /* Intialize the descriptors */
  tc3160_eth_setup_desc();
  return(0);
}

/* Driver installation: this is where the thing starts */
static int __init tc3160_eth_startup(void)
{
  struct net_device *dev = &tc3160_eth_dev;
  TC3160_ETH_PRIV *private = NULL;
  int err = 0, i;

  MSG_PRINT("Ethernet driver for TC3160 version 0.8");
  dev->priv = NULL;

  //flash_Erase();

  MSG_PRINT("Flash Erase complete");

  if (dev_count != 0) {
    err = -EIO;
    goto failed;
  }

  SET_MODULE_OWNER(&tc3160_eth_dev);

  /* Set up default mac address and others */
  def_mac_hi = DEF_MAC_HI;
  def_mac_lo = DEF_MAC_LO;
  max_link_loop = DEF_LINK_LOOP;
  ld_count = 1;
  reset_flag = 1;

  /* Get a device name: normally it'll be eth0 */
  if ((err = dev_alloc_name(&tc3160_eth_dev, "eth%d")) < 0)
    goto failed;

  /* Register ISR */
  if ((err = request_irq(INTR_METH, tc3160_eth_intr_handler, 0, "ethernet", 
		  &tc3160_eth_dev)) != 0) {
    err = -EIO;
    goto failed;
  } else {
    MSG_PRINT(" on %s (IRQ: %d)", dev->name, INTR_METH);
  }

  if (tc3160_get_macaddr(&(dev->dev_addr[0]))) {
    err = -EIO;
    goto failed;
  }

  MSG_PRINT(", (mac %02x", dev->dev_addr[0]);
  for (i = 1; i < 6; i++)
    MSG_PRINT(":%02x", dev->dev_addr[i]);
  MSG_PRINT(")\n");

  /* Allocate memory for private data */
  dev->priv = private = (TC3160_ETH_PRIV *)kmalloc(sizeof(TC3160_ETH_PRIV), GFP_KERNEL);
  if (dev->priv == NULL) {
    err = -ENOMEM;
    goto failed;
  } 
  
  /* Point to driver initialization routine and register the device with kernel */
  dev->init = tc3160_eth_init;
  if (register_netdev(dev) != 0) {
    err = -EIO;
    goto failed;
  }

  MSG_PRINT("%s: driver installation completed.\n", dev->name);
  dev_count++;
  return(0);

failed:
  if (dev->priv != NULL) {
    kfree(dev->priv);
    dev->priv = NULL;
  }
  MSG_PRINT("%s: driver installation failed.\n", dev->name);
  return(err);
}

/* Uninstallation of drive */
static void __exit tc3160_eth_shutdown(void)
{
  if (dev_count != 0) {
    /* Turn off IRQ and stop receive/transmit */
    *(volatile unsigned long *)TC3160_CR_REG = 0;
    *(volatile unsigned long *)TC3160_IER_REG = 0;

    /* Unregister the device and ISR */
    free_irq(INTR_METH, &tc3160_eth_dev);
    unregister_netdev(&tc3160_eth_dev);

    /* Set desc base address registers to 0 */
    *(volatile unsigned long *)TC3160_RLBAR_REG = 0;
    *(volatile unsigned long *)TC3160_TLBAR_REG = 0;

    /* Free up memory */
    if (tc3160_eth_dev.priv != NULL) {
      kfree(tc3160_eth_dev.priv);
      tc3160_eth_dev.priv = NULL;
    }
  } 
}

/* Register startup/shutdown routines */
module_init(tc3160_eth_startup);
module_exit(tc3160_eth_shutdown);

