/*
 * misc.c
 * 
 * This is a collection of several routines from gzip-1.0.3 
 * adapted for Linux.
 *
 * malloc by Hannu Savolainen 1993 and Matthias Urlichs 1994
 * puts by Nick Holloway 1993, better puts by Martin Mares 1995
 * High loaded stuff by Hans Lermen & Werner Almesberger, Feb. 1996
 */

#include <linux/vmalloc.h>
#include <linux/config.h>
#include <asm/io.h>
#include <asm/addrspace.h>
#include <asm/tc3160/sio.h>
#include "address.h"
#include <asm/mipsregs.h>
#include <asm/regdef.h>
#include "testled.h"

/*
 * gzip declarations
 */

#define OF(args)  args
#define STATIC static

#undef memset
#undef memcpy
#define memzero(s, n)     memset ((s), 0, (n))

typedef unsigned char  uch;
typedef unsigned short ush;
typedef unsigned long  ulg;

#define WSIZE 0x8000		/* Window size must be at least 32k, */
				/* and a power of two */

#define INPLACE_MOVE_ROUTINE  0x1000
#define LOW_BUFFER_START      0x2000
#define LOW_BUFFER_MAX       0x90000
#define HEAP_SIZE             0x3000

extern int end;

#define ROMABLE

#ifdef ROMABLE
extern uch __ramdata_start, __ramdata_end;
#endif

#ifdef ROMABLE
#define __ramdata	__attribute__ ((unused, __section__ (".data.ramdata")))
#else
#define __ramdata	
#endif

static const unsigned long free_mem_end_ptr = KSEG1ADDR(0) + CONFIG_TC3162_RAMSIZE - 0x10000; /* Save 64K */

/* variables need to be move to RAM */
static unsigned long free_mem_ptr __ramdata = 0;
static uch *inbuf __ramdata = 0;	     /* input buffer */
static uch window[WSIZE] __ramdata = { 0 };    /* Sliding window buffer */

static unsigned insize __ramdata = 0;  /* valid bytes in inbuf */
static unsigned inptr __ramdata = 0;   /* index of next byte to be processed in inbuf */
static unsigned outcnt __ramdata = 0;  /* bytes in output buffer */

static long bytes_out __ramdata = 0;
static uch *output_data __ramdata = 0;
static unsigned long output_ptr __ramdata = 0;


static unsigned int low_buffer_end __ramdata = 0;
static int high_loaded __ramdata = 0;
static uch *high_buffer_start __ramdata = 0;

static int sio_setup_done __ramdata = 0;

/* gzip flag byte */
#define ASCII_FLAG   0x01 /* bit 0 set: file probably ASCII text */
#define CONTINUATION 0x02 /* bit 1 set: continuation of multi-part gzip file */
#define EXTRA_FIELD  0x04 /* bit 2 set: extra field present */
#define ORIG_NAME    0x08 /* bit 3 set: original file name present */
#define COMMENT      0x10 /* bit 4 set: file comment present */
#define ENCRYPTED    0x20 /* bit 5 set: file is encrypted */
#define RESERVED     0xC0 /* bit 6,7:   reserved */

#define get_byte()  (inptr < insize ? inbuf[inptr++] : fill_inbuf())
		
/* Diagnostic functions */
#ifdef DEBUG
#  define Assert(cond,msg) {if(!(cond)) error(msg);}
#  define Trace(x) fprintf x
#  define Tracev(x) {if (verbose) fprintf x ;}
#  define Tracevv(x) {if (verbose>1) fprintf x ;}
#  define Tracec(c,x) {if (verbose && (c)) fprintf x ;}
#  define Tracecv(c,x) {if (verbose>1 && (c)) fprintf x ;}
#else
#  define Assert(cond,msg)
#  define Trace(x)
#  define Tracev(x)
#  define Tracevv(x)
#  define Tracec(c,x)
#  define Tracecv(c,x)
#endif

static int  fill_inbuf(void);
static void flush_window(void);
static void error(char *m);
static void gzip_mark(void **);
static void gzip_release(void **);
  
/*
 * This is set up by the setup-routine at boot-time
 */
extern char input_data[];
extern int input_len;

static void *malloc(int size);
static void free(void *where);
static void error(char *m);
static void gzip_mark(void **);
static void gzip_release(void **);
 
void puts(const char *);
void Bputs(unsigned char *,int);
  
static const char bars[4] = { '-', '\\', '|', '/' };
static int barcnt __ramdata = 0;
static int cnt __ramdata = 0;
static char pbuf[4] __ramdata = { 0, 0, 0, 0 };
void sio_setup(int flag);

static int  uclk_65000[13]={
			357500,		 
			175500,		 
			119808,		 
			89856,	 
			59904,	 
			44928,	 
			29952,	 
			14976,	 
			7488,	 
			3744,	 
			1872,	 
			936,		 
			343,		 
			};


static void puts_progress(void)
{
  if (!(cnt & 0xf)) {
    pbuf[0] = bars[barcnt];
    pbuf[1] = '\b';
    barcnt = (barcnt + 1) % 4;
    puts(pbuf);
  }
  cnt++;
}

void sio_setup(int flag)
{
	int	div_x,div_y;
	int	word;
	register int i;
	unsigned int baud;

	baud = 0;
	*(unsigned char*)( (0x0c+ 0xBFBF0000 +0x03)  ) = 0x80 ;
	div_y = 65000 ;
	div_x = (unsigned int)(uclk_65000[baud]/ 100 )*2;
	word = (div_x<<16)|div_y;
	*(unsigned int*)( (0x2c+ 0xBFBF0000 )  ) =word;
	*(unsigned char*)( (0x00+ 0xBFBF0000 +0x03)  ) = 0x03 ;
	*(unsigned char*)( (0x04+ 0xBFBF0000 +0x03)  ) = 0x00 ;
	*(unsigned char*)( (0x0c+ 0xBFBF0000 +0x03)  ) = 0x03 ;
	*(unsigned char*)( (0x08+ 0xBFBF0000 +0x03)  ) = 0x0f | 0x0<<6 ;
	*(unsigned char*)( (0x10+ 0xBFBF0000 +0x03)  ) = 0x0 ;
	*(unsigned char*)( (0x24+ 0xBFBF0000 +0x03)  ) = 0x0 ;
	*(unsigned char*)( (0x04+ 0xBFBF0000 +0x03)  ) = 0x05 ;

  if (flag)
    sio_setup_done = 1;
}

void sio_putc(unsigned long c)
{
  while (  !(*(volatile unsigned char*)( (0x14+ 0xBFBF0000 +0x03)  )  & (0x20 ))	)
  {
//lightled(~(LEDa));
//lightled(~(LEDb));	

	}
  *(volatile unsigned char*)( (0x00+ 0xBFBF0000 +0x03)  ) =(unsigned char)c;
}

void ltoa(char *dst, unsigned long l)
{
  register int i;
  unsigned char c;
  unsigned long mask;

  for (mask = 0xf0000000, i = 28; i >= 0; i -= 4, mask >>= 4) {
    c = (unsigned char)((l & mask) >> i);
    if (c <= 9)
      *dst++ = (c + '0');	
    else
      *dst++ = (c - 10 + 'a');
  }
  *dst++ = '\0';
}

void puts(const char *s)
{
  const char *p;
    int j;

  if (!sio_setup_done)
    sio_setup(1);

  for (p = s; *p != '\0'; p++) {
    sio_putc((unsigned long)*p);
    if (*p == '\n')
      sio_putc((unsigned long)'\r');

//	for(j=0;j<0x50;j++);	
  }

}

void Bputs(unsigned char *s,int num)
{
  char *p;
  int i;

  sio_setup(0);

  for(i=0,p=s;i<num;i++,p++) 
    	sio_putc((unsigned long)*p);

}

void force_puts(const char *s)
{
  const char *p;

  sio_setup(0);

  for (p = s; *p != '\0'; p++) {
    sio_putc((unsigned long)*p);
    if (*p == '\n')
      sio_putc((unsigned long)'\r');
  }
}

void puts_tlb_exception(unsigned long cause, unsigned long sr)
{
  char buf[32];

  ltoa(buf, cause);
  force_puts("Encounterred TLB exception: CAUSE=");
  force_puts(buf);
  force_puts(", SR=");
  ltoa(buf, sr);
  force_puts(buf);
  force_puts("\n");
}
#if 0
void interrupt_exception_handler(unsigned long cause, unsigned long sr)
{
#if 1
  __asm__ __volatile__(
	/* GPIO setup */
	"la	$2, 0xb2000500\n"
	"nop\n"
/* GPOEN_REG */
	"la	$4, 0x0 \n"
	"sw	$4, 0x4($2)\n"
	"nop\n"

/* GPVAL_REG */
	"la	$4, 0xffbf\n"
	"sw	$4, 0x0($2)\n"
	"nop\n");
#endif	
}
#endif

//void puts_general_exception(unsigned long cause, unsigned long sr)
void interrupt_exception_handler(unsigned long cause, unsigned long sr)
{
  char buf[32];
  ltoa(buf, cause);
  force_puts("Encounterred general exception: CAUSE=");
  force_puts(buf);
  force_puts(", SR=");
  ltoa(buf, sr);
  force_puts(buf);
  force_puts("\n");
}
void puts_general_exception(unsigned long cause, unsigned long sr)
{
  char buf[32];
  ltoa(buf, cause);
  force_puts("Encounterred general exception: CAUSE=");
  force_puts(buf);
  force_puts(", SR=");
  ltoa(buf, sr);
  force_puts(buf);
  force_puts("\n");
}

#define NO_LENGTH_CHECK
//#define DEBG(x)	puts_progress()
#define DEBG(x)	;
#include "../../../../../lib/inflate.c"

static void *malloc(int size)
{
  void *p;

  if (size < 0) 
    error("Malloc error\n");
  if (free_mem_ptr == 0) 
    error("Memory error\n");

  free_mem_ptr = (free_mem_ptr + 3) & ~3;	/* Align */

  p = (void *)free_mem_ptr;
  free_mem_ptr += size;

  if (free_mem_ptr >= free_mem_end_ptr)
    error("\nOut of memory\n");

  memset(p, 0, size);	/* Clear out the memory */

  return(p);
}

static void free(void *where)
{	/* Don't care */
}

static void gzip_mark(void **ptr)
{
  *ptr = (void *)free_mem_ptr;
}

static void gzip_release(void **ptr)
{
  free_mem_ptr = (long) *ptr;
}
 
void *memset(void* s, int c, size_t n)
{
  int i;
  char *ss = (char*)s;

  for (i = 0; i < n; i++) 
    ss[i] = c;
  return(s);
}

void *memcpy(void* __dest, __const void* __src, size_t __n)
{
  int i;
  char *d = (char *)__dest, *s = (char *)__src;

  for (i = 0; i < __n; i++) 
    d[i] = s[i];
  return(__dest);
}

/* ===========================================================================
 * Fill the input buffer. This is called only when the buffer is empty
 * and at least one byte is really needed.
 */
static int fill_inbuf(void)
{
  if (insize != 0) {
    error("ran out of input data\n");
  }

  inbuf = input_data;
  insize = input_len;
  inptr = 1;
  return(inbuf[0]);
}

/* ===========================================================================
 * Write the output window window[0..outcnt-1] and update crc and bytes_out.
 * (Used for the decompressed data only.)
 */
static void flush_window_low(void)
{
  ulg c = crc;         /* temporary variable */
  unsigned n;
  uch *in, *out, ch;
    
  in = window;
  out = &output_data[output_ptr]; 
  for (n = 0; n < outcnt; n++) {
    ch = *out++ = *in++;
    c = crc_32_tab[((int)c ^ ch) & 0xff] ^ (c >> 8);
  }
  crc = c;
  bytes_out += (ulg)outcnt;
  output_ptr += (ulg)outcnt;
  outcnt = 0;
}

static void flush_window_high(void)
{
  ulg c = crc;         /* temporary variable */
  unsigned n;
  uch *in,  ch;
  in = window;
  for (n = 0; n < outcnt; n++) {
    ch = *output_data++ = *in++;
    if ((ulg)output_data == low_buffer_end) 
      output_data=high_buffer_start;
    c = crc_32_tab[((int)c ^ ch) & 0xff] ^ (c >> 8);
  }
  crc = c;
  bytes_out += (ulg)outcnt;
  outcnt = 0;
}

static void flush_window(void)
{
  if (high_loaded) 
    flush_window_high();
  else
    flush_window_low();
}

static void error(char *x)
{
  puts("\n\n");
  puts(x);
  puts("\n\n -- System halted");

  while(1);	/* Halt */
}

#ifdef ROMABLE
static void clear_ramdata(void)
{
  memset((void *)&__ramdata_start, 0, (uch *)&__ramdata_end - (uch *)&__ramdata_start);
}
#endif

int decompress_kernel(void)
{
  int res;
#ifdef ROMABLE
  clear_ramdata(); // commentted out by KWU
  free_mem_ptr = KSEG1ADDR(0) + CONFIG_TC3162_RAMSIZE - 0x200000; /* Last 2MB */
#else
  free_mem_ptr = (unsigned long)&end;
#endif

  output_data = (uch *)(KERNEL_LOAD_ADDRESS);
  // code starting here, by KWU
#if 0 
  puts("KWU: copy image... \n");
  memcpy(output_data, input_data, *(size_t *)((uch*)input_data-4)); 
  puts("KWU: Ok, booting the kernel.\n");
  return(0);
#endif 
  // code ending here, by KWU

  makecrc();
  puts("Uncompressing Linux... ");
  res = gunzip();

  output_data = NULL;
  //puts("Ok, booting the kernel.\n");

  return(res);
}

