/*  
 * Implementations of Interrupt handlers for lexra 4380. 
 *
 * Written by Robert Wu, Billion Co.
 * Support for TrendChip TC3162
 */
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/kernel_stat.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/timex.h>
#include <linux/malloc.h>
#include <linux/random.h>

#include <asm/bitops.h>
#include <asm/bootinfo.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/mipsregs.h>
#include <asm/system.h>

#ifdef CONFIG_TRENDCHIP_TC3162
#include <asm/tc3162/interrupts.h>
#endif

#define NUM_IRQ_ENTRIES		32

extern asmlinkage void tc3162_handle_int(void);
extern volatile unsigned long *int_mask_reg;
extern irq_desc_t irq_desc[NR_IRQS];

static unsigned int startup_tc3162_irq(unsigned int irq);
static void shutdown_tc3162_irq(unsigned int irq);
static void ack_tc3162_irq(unsigned int irq);
static void end_tc3162_irq(unsigned int irq);
static void tc3162_disable_irq(unsigned int irq_nr);
static void tc3162_enable_irq(unsigned int irq_nr);

static struct hw_interrupt_type tc3162_irq_type = {
  "TC3162",
  startup_tc3162_irq,
  shutdown_tc3162_irq,
  tc3162_enable_irq,
  tc3162_disable_irq,
  ack_tc3162_irq,
  end_tc3162_irq,
  NULL
};

void __init tc3162_irq_init(void)
{

#ifdef CONFIG_TRENDCHIP_TC3162
    /*
     * Setup interrupt structure
     */
    cpu_mask_tbl[INTR_UART] = 0x0;
    cpu_mask_tbl[INTR_TIMER0] = 0x10;
    cpu_mask_tbl[INTR_TIMER1] = 0x20;
    cpu_mask_tbl[INTR_TIMER2] = 0x40;
    cpu_mask_tbl[INTR_TIMER3] = 0x80;
    cpu_mask_tbl[INTR_TIMER4] = 0x100;
    cpu_mask_tbl[INTR_Watchdog] = 0x200;
    cpu_mask_tbl[INTR_GPIO] = 0x400;
    cpu_mask_tbl[INTR_APB_DMA_C0] = 0x4000;
    cpu_mask_tbl[INTR_APB_DMA_C1] = 0x8000;
    cpu_mask_tbl[INTR_AFE_Gasp] = 0x40000;
    cpu_mask_tbl[INTR_DMT] = 0x80000;
    cpu_mask_tbl[INTR_AHB_Bus_Er] = 0x100000;
    cpu_mask_tbl[INTR_EMAC] = 0x200000;
    cpu_mask_tbl[INTR_ATM_SAR] = 0x400000;
    cpu_mask_tbl[INTR_USB] = 0x800000;
    cpu_mask_tbl[INTR_PCI_INTA] = 0x1000000;
    cpu_mask_tbl[INTR_PCI_INTB] = 0x2000000;
    cpu_mask_tbl[INTR_PCI_Bus_Er] = 0x4000000;
    cpu_mask_tbl[INTR_SOFT_Int] = 0x80000000;

#else
    /*
     * Setup interrupt structure
     */
    cpu_mask_tbl[INTR_MTMR1] = 0x1;
    cpu_mask_tbl[INTR_MATM] = 0x2;
    cpu_mask_tbl[INTR_MTMR2] = 0x4;
    cpu_mask_tbl[INTR_METH] = 0x8;
    cpu_mask_tbl[INTR_MTMR3] = 0x10;
    cpu_mask_tbl[INTR_MDMAC] = 0x20;
#ifdef CONFIG_TRENDCHIP_WDL100
    cpu_mask_tbl[INTR_MUART] = 0x40;
    cpu_mask_tbl[INTR_MGPINT1] = 0x80;
    cpu_mask_tbl[INTR_MGPINT2] = 0x80;
    cpu_mask_tbl[INTR_MGPINT3] = 0x400;
    cpu_mask_tbl[INTR_MGPINT4] = 0x400;
    cpu_mask_tbl[INTR_MARB] = 0x800;
#else /* WDL200?? */
    cpu_mask_tbl[INTR_MGPINT1] = 0x80;
    cpu_mask_tbl[INTR_MUART] = 0x40;
    cpu_mask_tbl[INTR_MGPINT2] = 0x100;
    cpu_mask_tbl[INTR_MGPINT3] = 0x200;
    cpu_mask_tbl[INTR_MGPINT4] = 0x400;
    cpu_mask_tbl[INTR_MARB] = 0x800;
    cpu_mask_tbl[INTR_MPCMCIA] = 0x1000;
    cpu_mask_tbl[INTR_MUSB] = 0x2000;
#endif 
#endif /* end of CONFIG_TRENDCHIP_TC3162 */

#ifdef CONFIG_TRENDCHIP_TC3162

    cpu_irq_nr[INTR_UART] = INTR_UART;
    cpu_irq_nr[INTR_TIMER0] = INTR_TIMER0;
    cpu_irq_nr[INTR_TIMER1] = INTR_TIMER1;
    cpu_irq_nr[INTR_TIMER2] = INTR_TIMER2;
    cpu_irq_nr[INTR_TIMER3] = INTR_TIMER3;
    cpu_irq_nr[INTR_TIMER4] = INTR_TIMER4;
    cpu_irq_nr[INTR_Watchdog] = INTR_Watchdog;
    cpu_irq_nr[INTR_GPIO] = INTR_GPIO;
    cpu_irq_nr[INTR_APB_DMA_C0] = INTR_APB_DMA_C0;
    cpu_irq_nr[INTR_APB_DMA_C1] = INTR_APB_DMA_C1;
    cpu_irq_nr[INTR_AFE_Gasp] = INTR_AFE_Gasp;
    cpu_irq_nr[INTR_DMT] = INTR_DMT;
    cpu_irq_nr[INTR_AHB_Bus_Er] = INTR_AHB_Bus_Er;
    cpu_irq_nr[INTR_EMAC] = INTR_EMAC;
    cpu_irq_nr[INTR_ATM_SAR] = INTR_ATM_SAR;
    cpu_irq_nr[INTR_USB] = INTR_USB;
    cpu_irq_nr[INTR_PCI_INTA] = INTR_PCI_INTA;
    cpu_irq_nr[INTR_PCI_INTB] = INTR_PCI_INTB;
    cpu_irq_nr[INTR_PCI_Bus_Er] = INTR_PCI_Bus_Er;
    cpu_irq_nr[INTR_SOFT_Int] = INTR_SOFT_Int;


#else


    cpu_irq_nr[INTR_MTMR1] = INTR_MTMR1;
    cpu_irq_nr[INTR_MATM] = INTR_MATM;
    cpu_irq_nr[INTR_MTMR2] = INTR_MTMR2;
    cpu_irq_nr[INTR_METH] = INTR_METH;
    cpu_irq_nr[INTR_MTMR3] = INTR_MTMR3;
    cpu_irq_nr[INTR_MDMAC] = INTR_MDMAC;

#ifdef CONFIG_TRENDCHIP_WDL100
    cpu_irq_nr[INTR_MUART] = INTR_MUART;
    cpu_irq_nr[INTR_MGPINT1] = INTR_MGPINT1;
    cpu_irq_nr[INTR_MGPINT2] = INTR_MGPINT2;
    cpu_irq_nr[INTR_MGPINT3] = INTR_MGPINT3;
    cpu_irq_nr[INTR_MGPINT4] = INTR_MGPINT4;
    cpu_irq_nr[INTR_MARB] = INTR_MARB;
#else /* WDL200?? */
    cpu_irq_nr[INTR_MGPINT1] = INTR_MGPINT1;
    cpu_irq_nr[INTR_MUART] = INTR_MUART;
    cpu_irq_nr[INTR_MGPINT2] = INTR_MGPINT2;
    cpu_irq_nr[INTR_MGPINT3] = INTR_MGPINT3;
    cpu_irq_nr[INTR_MGPINT4] = INTR_MGPINT4;
    cpu_irq_nr[INTR_MARB] = INTR_MARB;
    cpu_irq_nr[INTR_MPCMCIA] = INTR_MPCMCIA;
    cpu_irq_nr[INTR_MUSB] = INTR_MUSB;
#endif 
#endif /* end of CONFIG_TRENDCHIP_TC3162 */

}

static void __init tc3162_irq_setup(void)
{
   volatile int stat;

   switch (mips_machtype) {
      case MACH_LX4380:
	 tc3162_irq_init();
         /* Setup specific TC3162 interrupt handling bit */
         stat = read_32bit_cp0_register(CP0_STATUS);
	#ifdef CONFIG_TRENDCHIP_TC3160
         stat |= 0x0400; /* Timer */
         stat |= 0x1000; /* Ethernet/UART */
         stat |= 0x3c00; /* 3-6 on */
	#else
		//---------------------------------------------------------------------
		// (1) CU0 <- 1, let CoPocessor 0 to be usable in user mode
		//	  /* however, we should stay in kernel mode, so CP0 is always
		//		 usable regardless of how CU0 is set */
		// (2) BEV <- 0, use regular bootstrap exception vector, which point
		//	  into RAM
		// (3) IntMask ( Int[5:0]Sw[1:0] ) <- 1, turn on all interrupt source
		// (4) KUold/KUprevious/KUcurrent <- 0, kernel Mode
		// (5) IEold/IEprevious/IEcurrent <- 1, enable interrupt
		//---------------------------------------------------------------------

	 stat |=0xf200ff15; 
	#endif
         write_32bit_cp0_register(CP0_STATUS, stat);
	 break;
   }
   set_except_vector(0, tc3162_handle_int);
}

static inline void mask_irq(unsigned int irq_nr)
{
  *int_mask_reg &= ~(cpu_mask_tbl[irq_nr] & 0xffffffff);
}

static inline void unmask_irq(unsigned int irq_nr)
{
  *int_mask_reg |= (cpu_mask_tbl[irq_nr] & 0xffffffff);
}

static void tc3162_disable_irq(unsigned int irq_nr)
{
  unsigned long flags;

  save_and_cli(flags);
  mask_irq(irq_nr);
  restore_flags(flags);
}

static void tc3162_enable_irq(unsigned int irq_nr)
{
  unsigned long flags;

  save_and_cli(flags);
  unmask_irq(irq_nr);
  restore_flags(flags);
}

void __init init_IRQ(void)
{
  register int i;

  tc3162_irq_setup();

  for (i = 0; i < NR_IRQS; i++) {
    irq_desc[i].status = IRQ_DISABLED;
    irq_desc[i].action = NULL;
    irq_desc[i].depth = 1;
    irq_desc[i].handler = &tc3162_irq_type;
  }
}

static unsigned int startup_tc3162_irq(unsigned int irq)
{
  tc3162_enable_irq(irq);
  return(0); /* never anything pending */
}

static void shutdown_tc3162_irq(unsigned int irq)
{
  tc3162_disable_irq(irq);
  return; /* never anything pending */
}

static void ack_tc3162_irq(unsigned int irq)
{
  tc3162_disable_irq(irq);
  return; /* No need to ack here */
}

static void end_tc3162_irq(unsigned int irq)
{
  if (!(irq_desc[irq].status & (IRQ_DISABLED | IRQ_INPROGRESS)))
    tc3162_enable_irq(irq);
}

