/*
 *	Industrial Computer Source WDT500/501 driver for Linux 2.1.x
 *
 *	(c) Copyright 1996-1997 Alan Cox <alan@redhat.com>, All Rights Reserved.
 *				http://www.redhat.com
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *	
 *	Neither Alan Cox nor CymruNet Ltd. admit liability nor provide 
 *	warranty for any of this software. This material is provided 
 *	"AS-IS" and at no charge.	
 *
 *	(c) Copyright 1995    Alan Cox <alan@lxorguk.ukuu.org.uk>
 *
 *	Release 0.08.
 *
 *	Fixes
 *		Dave Gregorich	:	Modularisation and minor bugs
 *		Alan Cox	:	Added the watchdog ioctl() stuff
 *		Alan Cox	:	Fixed the reboot problem (as noted by
 *					Matt Crocker).
 *		Alan Cox	:	Added wdt= boot option
 *		Alan Cox	:	Cleaned up copy/user stuff
 *		Tim Hockin	:	Added insmod parameters, comment cleanup
 *					Parameterized timeout
 *		Tigran Aivazian	:	Restructured wdt_init() to handle failures
 */

#include <linux/errno.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/param.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/interrupt.h>
                                                                                                                                    
#include <asm/cpu.h>
#include <asm/bootinfo.h>
#include <asm/mipsregs.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <linux/miscdevice.h>
                                                                                                                                    
#include <asm/tc3162/tc3162.h>
#include <asm/tc3162/interrupts.h>
#include <asm/tc3162/timer.h>

static int wdt_is_open;

/*
 *	You must set these - there is no sane way to probe for this board.
 *	You can use wdt=x,y to set these now.
 */
 
static int irq=INTR_Watchdog;

#define WD_TIMO (TIMERTICKS_1S*20)		/* 20 second */


/*
 *	Kernel methods.
 */
 
 
/**
 *	wdt_status:
 *	
 *	Extract the status information from a WDT watchdog device. There are
 *	several board variants so we have to know which bits are valid. Some
 *	bits default to one and some to zero in order to be maximally painful.
 *
 *	we then map the bits onto the status ioctl flags.
 */
 
/**
 *	wdt_interrupt:
 *	@irq:		Interrupt number
 *	@dev_id:	Unused as we don't allow multiple devices.
 *	@regs:		Unused.
 *
 *	Handle an interrupt from the board. These are raised when the status
 *	map changes in what the board considers an interesting way. That means
 *	a failure condition occuring.
 */
 
void wdt_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	printk(KERN_CRIT "Would Reboot.\n");
	printk(KERN_CRIT "Initiating system reboot.\n");
	/* all interrupt belong to IRQ */
     	*(volatile unsigned int*)(TC3162_ITR_REG) = 0;                                                                                                                              
        /* disable all IMR */
        *(volatile unsigned int*)(TC3162_IMR_REG) = 0;

	/* reset the system */	
	machine_restart(NULL);
}


/**
 *	wdt_write:
 *	@file: file handle to the watchdog
 *	@buf: buffer to write (unused as data does not matter here 
 *	@count: count of bytes
 *	@ppos: pointer to the position to write. No seeks allowed
 *
 *	A write to a watchdog device is defined as a keepalive signal. Any
 *	write of data will do, as we we don't define content meaning.
 */ 
static ssize_t wdt_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
	tc3162_disable_timer(WD_TIMER);	
	/* reset the timer */
	*(volatile unsigned int*)(TC3162_TIMER5LDR_REG) = (WD_TIMO * SYS_HCLK) * 1000/2;
	tc3162_enable_timer(WD_TIMER);

	return 1;
}

/**
 *	wdt_open:
 *	@inode: inode of device
 *	@file: file handle to device
 *
 *	One of our two misc devices has been opened. The watchdog device is
 *	single open and on opening we load the counters. Counter zero is a 
 *	100Hz cascade, into counter 1 which downcounts to reboot. When the
 *	counter triggers counter 2 downcounts the length of the reset pulse
 *	which set set to be as long as possible. 
 */
 
static int wdt_open(struct inode *inode, struct file *file)
{
	int word,word1;

	if(MINOR(inode->i_rdev)!=WATCHDOG_MINOR)
		return 0;
printk("wdt_open ..\n");
	wdt_is_open=1;
	/* set the timer & enable the timer */
  	word = (WD_TIMO * SYS_HCLK) * 1000/2;
	*(volatile unsigned int*)(TC3162_TIMER5LDR_REG) = word;
                                                                                                                                    
	word = *(volatile unsigned int*)(TC3162_TIMERCTLR_REG);
	word1 = (TIMER_ENABLE << 5) |(TIMER_TOGGLEMODE << (13))+(TIMER_ENABLE<<25);
	word |= word1;
	*(volatile unsigned int*)(TC3162_TIMERCTLR_REG) = word;

	return 0;
}

/**
 *	wdt_close:
 *	@inode: inode to board
 *	@file: file handle to board
 *
 *	The watchdog has a configurable API. There is a religious dispute 
 *	between people who want their watchdog to be able to shut down and 
 *	those who want to be sure if the watchdog manager dies the machine
 *	reboots. In the former case we disable the counters, in the latter
 *	case you have to open it again very soon.
 */
 
static int wdt_release(struct inode *inode, struct file *file)
{
	if(MINOR(inode->i_rdev)!=WATCHDOG_MINOR)
		return 0;
printk("wdt_release..\n");
//	lock_kernel();
	wdt_is_open=0;
	/* disable WDT */
	tc3162_disable_timer(WD_TIMER);
//	unlock_kernel();
	return 0;
}

/*
 *	Kernel Interfaces
 */
 
 
static struct file_operations wdt_fops = {
	owner:		"WatchDog",
	llseek:		NULL,
	read:		NULL,
	write:		wdt_write,
	ioctl:		NULL,
	open:		wdt_open,
	release:	wdt_release,
};

static struct miscdevice wdt_miscdev=
{
	WATCHDOG_MINOR,
	"watchdog",
	&wdt_fops
};

/**
 *	cleanup_module:
 *
 *	Unload the watchdog. You cannot do this with any file handles open.
 *	If your watchdog is set to continue ticking on close and you unload
 *	it, well it keeps ticking. We won't get the interrupt but the board
 *	will not touch PC memory so all is fine. You just have to load a new
 *	module in 60 seconds or reboot.
 */
 
static void __exit wdt_exit(void)
{
	tc3162_disable_timer(WD_TIMER);
	misc_deregister(&wdt_miscdev);
	free_irq(irq, NULL);
}

/**
 * 	wdt_init:
 *
 *	Set up the WDT watchdog board. All we have to do is grab the
 *	resources we require and bitch if anyone beat us to them.
 *	The open() function will actually kick the board off.
 */
 
int __init wdt_init(void)
{
	int ret;

printk("wdt_init ..\n");
	ret = misc_register(&wdt_miscdev);
	if (ret) {
		printk(KERN_ERR "wdt: can't misc_register on minor=%d\n", WATCHDOG_MINOR);
		goto out;
	}
	ret = request_irq(irq, wdt_interrupt, SA_INTERRUPT, "watchdog", NULL);
	if(ret) {
		printk(KERN_ERR "wdt: IRQ %d is not free.\n", irq);
		goto outmisc;
	}

	printk(KERN_INFO "TC3162 watchdog initialize at Interrupt %d\n", irq);
out:
	return ret;

outmisc:
	misc_deregister(&wdt_miscdev);
	return ret;
}

