#ifndef __MEMMAP_H
#define __MEMMAP_H

/*
** Description of channel buffers managed by RISC
**
** VBV	video channel buffer	(encoded)
** ABV	audio channel buffer	(encoded)
** CBV	cddsp channel buffer	(cd-format)
** SBV	sub-picture buffer	(encoded)
** OBV	osd buffer		(encoded)
**
** FORMAT 2000.06.28
**
**          0                 720  736      1024
**          +-------------------+--+--------+
**          |                   |  |        |
**          |                   |  |        |
**          | FRAME BUFFER REF0 |  | FRAME  |
**          |                   |  | BUFFER |
**          |                   |  | B      |
**          +-------------------+  |        |
**          |                   |  |        |
**          |                   |  |        |
**          | FRAME BUFFER REF1 +  +--------+ EVBYA
**          |                   |           |
**          |                   |           |
**   EVBYA2 +-------------------+           |
**          |                               |
**          |       VIDEO BISTSTREAM        |
**          |                               |
**    EABYA +-------------------------------+
**          |        AUDIO BITSTREAM        |
**     CDYA +-------------------------------+
**          |        CDDSP BUFFER           |
**    OSDYA +-------------------------------+
**          |          OSD BUFFER           |
**    SUPYA +-------------------------------+
**          |          SUP BUFFER           |
**  DSP24YA +-------------------------------+
**          |        DSP24 CODE             |
**  IOPYA   +-------------------------------+
**          |          IOP CODE             |
**    AUDYA +-------------------------------+
**          |        AUDIO WORK BUFFER      |
**    PCMYA +-------------------------------+
**          |          PCM BUFFER           |
**          +-------------------------------+
**          |         RISC WORK BUFFER      |
**          +-------------------------------+
**
**
** for special VBV buffer treatment: we use:
**
** 1. (1024-720)=304 in width
** 2. we have to write to strange buffer in some way.
**
**
*/

#define DVD_EXCEPTION_START	0x88200000


#define	SDRAM_BASE_CACHED	0x88000000
#define	SDRAM_BASE_UNCACHED	0xa8000000

#ifdef	TESTCHIP
#define	SDRAM_BASE		SDRAM_BASE_UNCACHED
#else
#define	SDRAM_BASE		SDRAM_BASE_CACHED
#endif

//smwang
#define FRAMEBUF_LINE 1024

#define	SDRAM_WC	1024
#define	SDRAM_W0	720
#define	SDRAM_W1	(SDRAM_WC - SDRAM_W0)

#define	ABV_W	1024			/* AUDIO buf width	*/
#define	SBV_W	1024			/* SUP buf width	*/
#define	OBV_W	1024			/* OSD buf width	*/
#define CBV_W	1024			/* DISC input buf width	*/

/*
** anchors
*/
#define MEM_PCMYA	(dvd_regs->pcmya)
#define	MEM_EABYA	(dvd_regs->eabya)
#define	MEM_CDYA	(dvd_regs->cdya)
#define	MEM_OSDYA	(dvd_regs->osdya)
#define	MEM_SUPYA	(dvd_regs->supya)

#define	MEM_EVBYA	(dvd_regs->evbya)
#define	MEM_EVBYA2	(dvd_regs->evbya2)
#define	MEM_IOPYA	(dvd_regs->iopya)
#define	MEM_DSP24YA	(dvd_regs->dsp24ya)
#define	MEM_AUDYA	(dvd_regs->audya)

/*
** starts
*/
#define	VBV_YS		(0)
#define	ABV_YS		(MEM_EABYA)
#define	CBV_YS		(MEM_CDYA)
#define	OBV_YS		(MEM_OSDYA)
#define	SBV_YS		(MEM_SUPYA)
#define	DSP24_YS	(MEM_DSP24YA)
#define	IOP_YS		(MEM_DSP16YA)
#define	PCM_YS		(MEM_PCMYA)
#define AWV_YS 		(MEM_AUDYA)

/*
** ends
*/
#define	VBV_YE		(MEM_EABYA)
#define	ABV_YE		(MEM_CDYA)
#define	CBV_YE		(MEM_OSDYA)
#define	OBV_YE		(MEM_SUPYA)
#define	SBV_YE		(MEM_DSP24YA)
#define	DSP24_YE	(MEM_DSP16YA)
#define	IOP_YE		(MEM_PCMYA)
#define PCM_YE 		(MEM_AUDYA)

/*
** HEIGHT
*/
#define	VBV_H		(VBV_YE - 0)
#define	ABV_H		(ABV_YE - ABV_YS)
#define	SBV_H		(SBV_YE - SBV_YS)
#define	CBV_H		(CBV_YE - CBV_YS)
#define	OBV_H		(OBV_YE - OBV_YS)



/*
** DEFAULT values
**
** for PAL 888 Y 22-mb-lines: (224k vbs)
**
** 	BIDIR_CHROMA	= 880
**	EVBYA		= 1600
**	EVBYA2 		= 1728
**	EVBYA3 (eabya)	= 1728 + 161 = 1889
**
** for PAL 866 Y 22-mb-lines: (224k vbs)
**
**	BIDIR_CHROMA	= 880
**	EVBYA		= 880 + 540 = 1420
**	EVBYA2 		= 1728
**	EVBYA3 (eabya)	= 1728 + 111 = 1839
**
** for NTSC 866
**	EVBYA		= 1200 + 450 = 1650
**	EVBYA2		= 1440
**	EVBYA3		= 
**	
**
** for PAL 666 Y 22-mb-lines: (224k vbs)
**
**	BIDIR_CHROMA	= 660
**	EVBYA		= 660 + 540 = 1200
**	EVBYA2 		= 1728
**	EVBYA3 (eabya)	= 1728 + 49 = 1777
//	1200*16 + (1728-1200)*288 + 49*1024 = 221440
**
*/


//#define	MEM_411888
//#define	MEM_411866
#define	MEM_411866



#define	MEM_REF0_LUMA_DEF	0
#define	MEM_REF1_LUMA_DEF	576
#define	MEM_REF0_CHROMA_DEF	576+576
#define	MEM_REF1_CHROMA_DEF	576+576+288


//
// in fact this must use FULL when NTSC or SVCD/VCD.
//

#if	defined(MEM_HUGE)

//
// HUGE
//
#define	AGDC_CONFIG		(RF_AGDC_BPIC_LOC_BOTTOM | RF_AGDC_BPIC_SIZE_FULL)
#define	VIDEO_COMPRESS		(RF_COMPRESS_888 \
				| RF_COMPRESS_DITHER_OFF \
				| RF_COMPRESS_CHROMA_FULL) 
#define	MEM_BIDIR_LUMA_DEF	2200
#define	MEM_BIDIR_CHROMA_DEF	2776
#define	MEM_EVBYA_DEF		1420
#define	MEM_EVBYA2_DEF		1728
#define	MEM_EABYA_DEF		1839
#define	MEM_CDYA_DEF		1846
#define	MEM_OSDYA_DEF		1862
#define	MEM_SUPYA_DEF		1880
#define	MEM_DSP24YA_DEF		1940
#define	MEM_IOPYA_DEF		1970
#define	MEM_PCMYA_DEF		2048	// 2120 //1974
#define	MEM_AUDYA_DEF		2096	
#elif	defined(MEM_411888)

//
// 411888
//
#define	AGDC_CONFIG		(RF_AGDC_BPIC_LOC_RIGHT | RF_AGDC_BPIC_SIZE_LINE)
#define	VIDEO_COMPRESS		(RF_COMPRESS_888 \
				| RF_COMPRESS_DITHER_OFF \
				| RF_COMPRESS_CHROMA_FULL)
#define	MEM_BIDIR_LUMA_DEF	0
#define	MEM_BIDIR_CHROMA_DEF	880
#define	MEM_EVBYA_DEF		1600
#define	MEM_EVBYA2_DEF		1728
#define	MEM_EABYA_DEF		1890
#define	MEM_CDYA_DEF		1898
#define	MEM_OSDYA_DEF		1906
#define	MEM_SUPYA_DEF		1914
#define	MEM_DSP24YA_DEF		1954
#define	MEM_IOPYA_DEF		1970
#define	MEM_PCMYA_DEF		1978
#define	MEM_AUDYA_DEF		2002
#elif	defined(MEM_411866)

//
// 411866
//
#define	AGDC_CONFIG		(RF_AGDC_BPIC_LOC_RIGHT | RF_AGDC_BPIC_SIZE_LINE)
#define	VIDEO_COMPRESS		(RF_COMPRESS_866 \
				| RF_COMPRESS_DITHER_ON \
				| RF_COMPRESS_CHROMA_FULL)
#define	MEM_BIDIR_LUMA_DEF	0
#define	MEM_BIDIR_CHROMA_DEF	880
#define	MEM_EVBYA_DEF		1420
#define	MEM_EVBYA2_DEF		1728
#define	MEM_EABYA_DEF		1848	// +0
#define	MEM_CDYA_DEF		1848	// +10
#define	MEM_OSDYA_DEF		1854	// +4
#define	MEM_SUPYA_DEF		1878	// +46
#define	MEM_DSP24YA_DEF		1904	// +24
#define	MEM_IOPYA_DEF		1928	// +2
#define	MEM_PCMYA_DEF		1932	// +42
#define	MEM_AUDYA_DEF		1974	// +20

#elif	defined(MEM_411666)

//
// 411666
//
#define	AGDC_CONFIG		(RF_AGDC_BPIC_LOC_RIGHT | RF_AGDC_BPIC_SIZE_LINE)
#define	VIDEO_COMPRESS		(RF_COMPRESS_666 \
				| RF_COMPRESS_DITHER_ON \
				| RF_COMPRESS_CHROMA_FULL)
#define	MEM_BIDIR_LUMA_DEF	0
#define	MEM_BIDIR_CHROMA_DEF	880
#define	MEM_EVBYA_DEF		1200
#define	MEM_EVBYA2_DEF		1728
#define	MEM_EABYA_DEF		1780	// +0
#define	MEM_CDYA_DEF		1780	// +6
#define	MEM_OSDYA_DEF		1786	// +54
#define	MEM_SUPYA_DEF		1840	// +46
#define	MEM_DSP24YA_DEF		1886	// +24
#define	MEM_IOPYA_DEF		1914	// +2
#define	MEM_PCMYA_DEF		1920	// +42
#define	MEM_AUDYA_DEF		1970	// +20

#endif



#define	PCM_XS		(SDRAM_W1)
#define	ABV_XS		(SDRAM_W1)
#define	SBV_XS		(SDRAM_W1)
#define	cbv_xS		(SDRAM_W1)
#define	OBV_XS		(SDRAM_W1)
#define	VBV_XS		(0)
#define AWV_XS 		(0)


#endif/*__MEMMAP_H*/

