/*
 *	Handle incoming frames
 *	Linux ethernet bridge
 *
 *	Authors:
 *	Lennert Buytenhek		<buytenh@gnu.org>
 *
 *	$Id: br_input.c,v 1.9.2.1 2001/12/24 04:50:05 davem Exp $
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 */

#include <linux/kernel.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/if_bridge.h>
#include <linux/netfilter_bridge.h>
#include <linux/in.h>
#include "br_private.h"

#ifdef IGMP_SNOOP
#include <linux/ip.h>
#include <linux/in.h>
#include <linux/igmp.h>
#include <net/checksum.h>
#endif


unsigned char bridge_ula[6] = { 0x01, 0x80, 0xc2, 0x00, 0x00, 0x00 };
#ifdef IGMP_SNOOP
static char igmp_group_check(struct sk_buff *skb, unsigned char *gmac);
static void update_igmp_snoop_br_fdb(char op, struct net_bridge *br, struct net_bridge_port *p, unsigned char *gmac);
static void update_igmpV3_snoop_br_fdb(struct net_bridge *br, struct net_bridge_port *p, struct sk_buff *skb);
#endif

static int br_pass_frame_up_finish(struct sk_buff *skb)
{
#ifdef CONFIG_NETFILTER_DEBUG
	skb->nf_debug = 0;
#endif
#if 0 //jj; performance/
	netif_rx(skb);
#else
	netif_receive_skb(skb);
#endif

	return 0;
}

#ifdef IGMP_SNOOP
__u8 ip_protocol(struct sk_buff *skb)
{
	__u8 *addr;
	//addr = 	(__u8 *) skb->nh.iph;
	addr = 	(__u8 *) (skb->mac.raw + sizeof(struct ethhdr));
	return *(addr+9);
}

#endif
static void br_pass_frame_up(struct net_bridge *br, struct sk_buff *skb)
{
	struct net_device *indev;
#ifdef IGMP_SNOOP
        unsigned char *dest;
        struct net_bridge_port *p;
        unsigned char gmac[6];
        char op;
                                                                                                                                 
        dest = skb->mac.ethernet->h_dest;
        p = skb->dev->br_port;
        if (/*br->igmpsnoop_enabled && */MULTICAST_MAC(dest) && (skb->mac.ethernet->h_proto == ETH_P_IP))
        {
                /* check whether it is IGMP packet */
                if (ip_protocol(skb) == IPPROTO_IGMP)
		{
	                int pad;
        	        struct sk_buff *pskb;

			//pad = (unsigned long)(skb->nh.raw)&0x3;
			pad = (unsigned long)(skb->mac.raw + sizeof(struct ethhdr)) & 0x3;
	                if (pad)
        	        {
                	        pskb = skb_copy_expand(skb, 4-pad+skb_headroom(skb), skb_tailroom(skb), GFP_ATOMIC);
				memcpy(pskb->mac.raw, skb->mac.raw, sizeof(struct ethhdr));
                        	kfree_skb(skb);
	                        skb = pskb;
        	        }
		
			if ((op=igmp_group_check(skb, gmac)) > 0) 
			{
	                       if (op == 3)
					update_igmpV3_snoop_br_fdb(br,p,skb);
				else
                        	      	update_igmp_snoop_br_fdb(op, br, p, gmac);
			}
		}
        }
#endif

	br->statistics.rx_packets++;
	br->statistics.rx_bytes += skb->len;

	indev = skb->dev;
	skb->dev = &br->dev;
	skb->pkt_type = PACKET_HOST;
	skb_push(skb, ETH_HLEN);
	skb->protocol = eth_type_trans(skb, &br->dev);

	NF_HOOK(PF_BRIDGE, NF_BR_LOCAL_IN, skb, indev, NULL,
			br_pass_frame_up_finish);
}

#if defined(CONFIG_TRENDCHIP_TC3162L2) && defined(TC3162L2_IMEM)
__IMEM int br_handle_frame_finish(struct sk_buff *skb)
#else
int br_handle_frame_finish(struct sk_buff *skb)
#endif
{
	struct net_bridge *br;
	unsigned char *dest;
	struct net_bridge_fdb_entry *dst;
	struct net_bridge_port *p;
	int passedup;

	dest = skb->mac.ethernet->h_dest;

	p = skb->dev->br_port;
	if (p == NULL)
		goto err_nolock;

	br = p->br;
	read_lock(&br->lock);
	if (skb->dev->br_port == NULL)
		goto err;

	passedup = 0;
	if (br->dev.flags & IFF_PROMISC) {
		struct sk_buff *skb2;

		skb2 = skb_clone(skb, GFP_ATOMIC);
		if (skb2 != NULL) {
			passedup = 1;
			br_pass_frame_up(br, skb2);
		}
	}

	if (dest[0] & 1) {
#ifdef IGMP_SNOOP
		if (!(br->dev.flags & IFF_PROMISC) &&
                    MULTICAST_MAC(dest) && (skb->mac.ethernet->h_proto == ETH_P_IP))
	        {
			unsigned char gmac[6];
			char op;

        	        /* check whether it is IGMP packet */
	                if (ip_protocol(skb) == IPPROTO_IGMP)
        	        {
                	        int pad;
                        	struct sk_buff *pskb;
	                        //pad = (unsigned long)(skb->nh.raw)&0x3;
				pad = (unsigned long)(skb->mac.raw + sizeof(struct ethhdr)) & 0x3;
        	                if (pad)
                	        {
                        	        pskb = skb_copy_expand(skb, 4-pad+skb_headroom(skb), skb_tailroom(skb), GFP_ATOMIC);
					memcpy(pskb->mac.raw, skb->mac.raw, sizeof(struct ethhdr));
                                	kfree_skb(skb);
	                                skb = pskb;
        	                }
                                                                                                                                 
                	        if ((op=igmp_group_check(skb, gmac)) > 0)
                        	{
        	                       	if (op == 3)
                	                        update_igmpV3_snoop_br_fdb(br,p,skb);
                        	        else
                                	        update_igmp_snoop_br_fdb(op, br, p, gmac);
	                        }
                	}
	        }

		if (br->igmpsnoop_enabled && MULTICAST_GROUP_MAC(dest) &&
		!((skb->mac.ethernet->h_proto==ETH_P_IP)&&(ip_protocol(skb) == IPPROTO_IGMP))) {	
                        if ((dst = br_fdb_get(br, dest)) != NULL)
                        {
                                br_multicast_forward(br, dst, skb, !passedup);
                                br_fdb_put(dst);
			/* KenWang added 2006/02/07 refer to Mark to solve memory leakage issue -- begin -- */
                        } else {
				if (passedup) goto err;
                        }
			/* KenWang added 2006/02/07 refer to Mark to solve memory leakage issue -- end -- */
                }
                else { // broadcast
                        br_flood_forward(br, skb, !passedup);
                }
#else

		br_flood_forward(br, skb, !passedup);
#endif
		if (!passedup)
			br_pass_frame_up(br, skb);
		goto out;
	}

	dst = br_fdb_get(br, dest);
	if (dst != NULL && dst->is_local) {
		if (!passedup)
			br_pass_frame_up(br, skb);
		else
			kfree_skb(skb);
		br_fdb_put(dst);
		goto out;
	}

	if (dst != NULL) {
		br_forward(dst->dst, skb);
		br_fdb_put(dst);
		goto out;
	}

	br_flood_forward(br, skb, 0);

out:
	read_unlock(&br->lock);
	return 0;

err:
	read_unlock(&br->lock);
err_nolock:
	kfree_skb(skb);
	return 0;
}

#if defined(CONFIG_TRENDCHIP_TC3162L2) && defined(TC3162L2_IMEM)
__IMEM int br_handle_frame(struct sk_buff *skb)
#else
int br_handle_frame(struct sk_buff *skb)
#endif
{
	struct net_bridge *br;
	unsigned char *dest;
	struct net_bridge_port *p;

	dest = skb->mac.ethernet->h_dest;

	p = skb->dev->br_port;
	if (p == NULL)
		goto err_nolock;

	br = p->br;
	read_lock(&br->lock);
	if (skb->dev->br_port == NULL)
		goto err;

	if (!(br->dev.flags & IFF_UP) ||
	    p->state == BR_STATE_DISABLED)
		goto err;

	if (skb->mac.ethernet->h_source[0] & 1)
		goto err;

	if (p->state == BR_STATE_LEARNING ||
	    p->state == BR_STATE_FORWARDING)
		br_fdb_insert(br, p, skb->mac.ethernet->h_source, 0);

	if (br->stp_enabled &&
	    !memcmp(dest, bridge_ula, 5) &&
	    !(dest[5] & 0xF0))
		goto handle_special_frame;

	if (p->state == BR_STATE_FORWARDING) {
		if (br_should_route_hook && br_should_route_hook(&skb)) {
			read_unlock(&br->lock);
			return -1;
		}

		NF_HOOK(PF_BRIDGE, NF_BR_PRE_ROUTING, skb, skb->dev, NULL,
			br_handle_frame_finish);
		read_unlock(&br->lock);
		return 0;
	}

err:
	read_unlock(&br->lock);
err_nolock:
	kfree_skb(skb);
	return 0;

handle_special_frame:
	if (!dest[5]) {
		NF_HOOK(PF_BRIDGE, NF_BR_LOCAL_IN, skb, skb->dev,NULL,
			br_stp_handle_bpdu);
		read_unlock(&br->lock);
		return 0;
	}

	read_unlock(&br->lock);
	kfree_skb(skb);
	return 0;
}
#ifdef IGMP_SNOOP

static void igmp_convertMulticatIPtoMac(__u32 group, unsigned char *gmac)
{
	__u32 u32Temp, temp;
	int i;

	u32Temp = group & 0x007fffff;
	gmac[0]=0x01; gmac[1]=0x00; gmac[2]=0x5e;
	for (i=5; i>=3; --i)
	{
		temp=u32Temp&0xff;
		gmac[i]=temp;
		u32Temp >>= 8;
	}
}
static char igmp_group_check(struct sk_buff *skb, unsigned char *gmac)
{
        struct iphdr *iph;
	__u8 hdrlen;
	struct igmphdr *igmph;

	/* check IP header information */
	//iph = skb->nh.iph;
	iph = skb->mac.raw + sizeof(struct ethhdr);
	hdrlen = iph->ihl << 2;
	if ((iph->version != 4) &&  (hdrlen < 20))
		return -1;
	if (ip_fast_csum((u8 *)iph, iph->ihl) != 0)
		return -1;
	{ /* check the length */
	__u32 len = ntohs(iph->tot_len);
	if (skb->len < len || len < hdrlen)
		return -1; 
	}
	/* parsing the igmp packet */
	igmph = (struct igmphdr *)((u8*)iph+hdrlen);
	igmp_convertMulticatIPtoMac(igmph->group, gmac);
	if ((igmph->type==IGMP_HOST_MEMBERSHIP_REPORT) ||
	    (igmph->type==IGMPV2_HOST_MEMBERSHIP_REPORT)) 
		return 1; /* report and add it */
	else if (igmph->type==IGMP_HOST_LEAVE_MESSAGE)
		return 2; /* leave and delete it */
	else if (igmph->type==IGMPV3_HOST_MEMBERSHIP_REPORT)
		return 3; /* igmp v3 protocol */
	
	return -1;
}

static void update_igmp_snoop_br_fdb(char op, struct net_bridge *br, struct net_bridge_port *p, unsigned char *dest)
{
	struct net_bridge_fdb_entry *dst;
	unsigned char i;
	unsigned char exist;

	/* check whether entry exist */
	dst = br_fdb_get(br, dest);
	if (dst != NULL) 
	{
		for (exist=i=0; i<MAX_SAME_GROUP_PORT; ++i)
		{
			if (dst->igmp_port[i] && dst->igmp_port[i] == p)
			{
				exist = 1;
				break;
			}
		}			
	}
	if (op == 1) /* add */
	{	
		if (dst)
		{
			if (exist == 0)
			{
		                for (exist=i=0; i<MAX_SAME_GROUP_PORT; ++i)
                		{
		                        if (dst->igmp_port[i] == NULL)
                		        {
                                		dst->igmp_port[i] = p;
		                                break;
        	        	        }
	        	        }
			}
			// update the timer to prevent from timeout
			dst->ageing_timer = jiffies;
		}
		else {
			/* insert one fdb entry */
			br_fdb_insert(br, p, dest, 0);
			dst = br_fdb_get(br, dest);
			dst->igmp_port[0] = p;
			for (i=1; i<MAX_SAME_GROUP_PORT; ++i)
				dst->igmp_port[i] = NULL;
		}
	}
	else if (op == 2) /* delete */
	{
		if (dst && exist)
			dst->igmp_port[i] = NULL;
	}
}

static void update_igmpV3_snoop_br_fdb(struct net_bridge *br, struct net_bridge_port *p, struct sk_buff *skb)
{
	struct igmpv3_report *report;
        struct iphdr *iph;
	__u16  numgrps, i;
	unsigned char gmac[6];
                                                                                                                                    
        /* check IP header information */
        //iph = skb->nh.iph;
	iph = skb->mac.raw + sizeof(struct ethhdr);

	report = (struct igmpv3_report *)((u8*)iph+(iph->ihl << 2));
	numgrps = ntohs(report->ngrec);
	for (i=0; i<numgrps; ++i)
	{
		igmp_convertMulticatIPtoMac(report->grec[i].grec_mca, gmac);		
                if ( (report->grec[i].grec_type == IGMPV3_MODE_IS_INCLUDE ||
                      report->grec[i].grec_type == IGMPV3_CHANGE_TO_INCLUDE) &&
                     (report->grec[i].grec_nsrcs == 0)) /* same as leave */
			update_igmp_snoop_br_fdb(2, br, p, gmac);
		else 
			update_igmp_snoop_br_fdb(1, br, p, gmac);
		
	}
}

#endif // IGMP_SNOOP
