/******************************************************************************
 *
 *  (C) Copyright 2004 Trend Micro, Inc.
 *
 *  All Rights Reserved.
 *  This program is an unpublished copyrighted work which is proprietary
 *  to Trend Micro, Inc. and contains confidential information that is not
 *  to be reproduced or disclosed to any other person or entity without
 *  prior written consent from Trend Micro, Inc. in each and every instance.
 *
 *  WARNING:  Unauthorized reproduction of this program as well as
 *  unauthorized preparation of derivative works based upon the
 *  program or distribution of copies by sale, rental, lease or
 *  lending are violations of federal copyright laws and state trade
 *  secret laws, punishable by civil and criminal penalties.
 *
 *****************************************************************************/

/**
 *  @file tmClientInfo.c
 *  @brief TMOG ClientInfo Module
 */

/*
 *  Change Log
 *  2004-02-18 Eric Li Initialization version
 */


#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/shm.h>
#include <net/sock.h>

#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_webstr.h>
#include "../../apps/tmss/include/tmauIniParser.h"
#include "../../apps/tmss/include/tmClientInfo.h"


/* magic key to store on 'next' of allocated cell */
#define MAGIC        0xDEADCAFE
#define CONF_MODULE  "clientinfo"

#define LOGERROR(format, args...) do { if (net_ratelimit()) printk(format, ## args); } while (0)
#if 0
#define LOGDEBUG(format, args...) do { if (net_ratelimit()) printk(format, ## args); } while (0)
#define LOGINFO(format, args...) do { if (net_ratelimit()) printk(format, ## args); } while (0)
#else
#define LOGDEBUG(format, args...)
#define LOGINFO(format, args...)
#endif

//static const char ModuleName[] = "tmClientInfo";

//const char *MsgStatus[] = 
//   { "Pending", "Windows no AV", "Unix", "Outdated AV", "Current AV"};

/*
 * MSPATCH      => M
 * TROJAN       => T
 * SPYWARE      => S
 * NETSHARE     => N
 * AVSOFT       => AS
 * AVPATTERN    => AP
 */
//const char *MsgVA[] =
//    { "M", "T", "S", "N", "AS", "AP"};

//const char *MsgLicense[] = { "License Expired", "License OK"};


struct cell
{
    struct cell  *next;
    ClientInfo   data;
} cell;

typedef struct
{
    int shmid;

    void *addr;               /* pointer to the whole memory block */
    ClientTable *pTable;      /* pointer to the ClientTable */ 
    struct cell **freelist;   /* list of free cells */
    struct cell *pEntry;      /* pointer to the cell pool */
} tm_data_info;


/* memory pool management */
tm_data_info gPool;

/* Table pointer */
ClientTable *gpTable = NULL;

/**
  * @brief get a client record memory
  * @return pointer to the newly allocated memory
  */

/**
  * @brief determine if the table list is full
  * @return 0 for not full, 1 for full
  */
static inline int is_full(void)
{
    return (gpTable->size >= CLIENT_MAX_SIZE);
}


/**
  * @brief compute the hash value by mac
  * @param mac MAC address
  * @return hash value
  */
static int mac_hash(const unsigned char *mac)
{
    unsigned long x;

    x = mac[0];
    x = (x << 2) ^ mac[1];
    x = (x << 2) ^ mac[2];
    x = (x << 2) ^ mac[3];
    x = (x << 2) ^ mac[4];
    x = (x << 2) ^ mac[5];

    x ^= x >> 8;

    return (x % CLIENT_HASH_SIZE);
}


/**
  * @brief Get the number of clients in the table
  * @return size of clients
  */
unsigned int tmci_SizeClient(void)
{
    return gpTable->size;
}

/**
  * @brief Get the size of "license-type" clients
  * @param idx license type (licensed or unlicensed)
  * @return size of clients of the specific license type
  */
unsigned int tmci_SizeLicense(uint8_t idx)
{
    if (idx >= LICENSE_SIZE)
    {
        LOGERROR("license out of range.");
        return 0;
    }
    return gpTable->license_size[idx];
}


/**
  * @brief Get the pointer of client record
  * @param MAC address of the client
  * @return pointer to the client record
  */
static ClientInfo *__getPtrByMAC(unsigned char *mac)
{
    int hash;
    struct list_head *head, *pos;
    ClientInfo *pNode;

    hash = mac_hash(mac);
    head = &(gpTable->hash[hash]);
    list_for_each(pos, head)
    {
        pNode = list_entry(pos, ClientInfo, hash_head);
		if (!memcmp(CMAC(pNode), mac, MAC_SIZE))
            return pNode;
    }
   return NULL;
}

int tmci_SetTime(struct client_id *id, struct client_time *timeinfo)
{
    ClientInfo *pNode;

    LOGINFO("tmci_SetTime(%u.%u.%u.%u) called.\n", NIPQUAD(id->ip));
    if ((pNode = __getPtrByMAC(id->mac)) == NULL)
    {
       LOGERROR("tmci_SetTime: Can not find the Client instance\n"); 
        return 0;//
    }
    else
    {
        memcpy(&(pNode->idinfo), id, sizeof(struct client_id));
        memcpy(&(pNode->timeinfo), timeinfo, sizeof(struct client_time));
        return 0;
    }
        return 0;

}
/*
int tmci_SetIP(uint32_t ip, unsigned char *mac)
{
    ClientInfo *pNode;

    LOGINFO("tmci_SetIP(%u.%u.%u.%u) called.\n", NIPQUAD(ip));
    if ((pNode = __getPtrByMAC(mac)) == NULL)
    {
       LOGERROR("tmci_SetIP: Can not find the Client instance\n"); 
        return 0;//
    }
    else
    {
	    pNode->idinfo.ip=ip;
         return 0;
    }
        return 0;

}
*/
/**
  * @brief Get a 'copy' of client record 
  * @param mac address of the client
  * @param pClient pointer to the buffer
  * @return 0 for success, negative value for failure
  */
int tmci_GetClient(unsigned char *mac, ClientInfo *pClient)
{
    ClientInfo *pNode;

    if ((pNode = __getPtrByMAC(mac)) == NULL)
        return -1;

    memcpy(pClient, pNode, sizeof(ClientInfo));

    return 0;
}


/**
  * @brief Get multiple clients from a specific index
  * @param pBuf  buffer to store the records
  * @param pSize [IN] unit size of pBuf [OUT] actual size returned
  * @param index starting point to get the records
  * @return 0 for success, negative value for failure
  */
int tmci_GetClientsFrom(ClientInfo *pBuf, int *pSize, int index)
{
    struct list_head *pos;
    ClientInfo *pClient;
    int count = 0, i = -1;

    list_for_each(pos, &gpTable->fifo)
    {
         i++;
         if (i < index)
             continue;

         if (count >= *pSize)
             break;
           
         pClient = list_entry(pos, ClientInfo, fifo_head);
         memcpy(pBuf+count, pClient, sizeof(ClientInfo));
         count++;
    }

    *pSize = count;

    return 0;
}



/*
 * @brief decide whether the client is in URL Filter exception 
 * @param mac MAC address of the client
 * @return 1 for yes, 0 for no
 */
int tmci_IsExpFilter(unsigned char *mac)
{
    ClientInfo info;

    if (tmci_GetClient(mac, &info) < 0)
        return 0;

    return (CEXP(&info).urlfilter);
}

/*
 * @brief decide whether the client is in AV Enforcer exception
 * @param mac MAC address of the client
 * @return 1 for yes, 0 for no
 */
int tmci_IsExpReport(unsigned char *mac)
{
    ClientInfo info;

    if (tmci_GetClient(mac, &info) < 0)
        return 0;

    return (CEXP(&info).report);
}

/**
  * @brief Dump the database distribution
  * @return 0 for success, negative value for failure
  */
/*int tmci_DumpClientTable(void)
{
    int i;
    ClientInfo *pClient;
    struct list_head *head, *pos;

    LOGDEBUG("[Total Client=%u]\n", gpTable->size);
    head = &gpTable->fifo;
    list_for_each(pos, head)
    {
        pClient = list_entry(pos, ClientInfo, fifo_head);

        LOGDEBUG("\t(%u.%u.%u.%u   ", NIPQUAD(CIP(pClient)));
        for (i = 0; i < MAC_SIZE; i++)
             LOGDEBUG("%02X%c", CMAC(pClient)[i], i==MAC_SIZE-1?')':'-');
        LOGDEBUG("\n");
    }
    LOGDEBUG("\n");

    for (i = 0; i < LICENSE_SIZE; i++)
    {
         LOGDEBUG("[%s, size=%d]\n", MsgLicense[i], gpTable->license_size[i]);
    }
    return 0;
}
*/
/**
  * @brief Setting up the memory pool
  * @return NONE
  */
static int setup_pool(tm_data_info *pData)
{
    int size_table = sizeof(ClientTable);
    int size_free = sizeof(struct cell *);
    int size_entry = CLIENT_MAX_SIZE * sizeof(struct cell);
    int size = size_table + size_free + size_entry;
    int nRetVal;

    LOGINFO("[Kernel setup_pool] is called ....\n");
    if ((pData->shmid = sys_shmget(TMSS_AV_SHMKEY, size, 0)) == -1)
    {
       LOGERROR("[Kernel:setup_pool] shmget = -1, shall not allowed to get the ClientInfo\n");
       return 0;
    }

    LOGINFO("[Kernel setup_pool] is called ...222222.\n");
   // pData->addr = (void *)shmat(pData->shmid, TMSS_AV_SHMADDR, 0);
    //pData->addr = (void *)sys_shmat(pData->shmid, (void *)0, 0);
     nRetVal = sys_shmat(pData->shmid, (void *)0, 0,(unsigned long *)&pData->addr);
    if(nRetVal)
    {
       LOGERROR("[Kernel:setup_pool] can not sys_shmat, err=%d\n, shall not allowed to get the ClientInfo\n",nRetVal);
	    return 0;
   
    }
   
     LOGINFO("[Kernel setup_pool] is called .333333...\n");
   pData->pTable = (ClientTable *)(gPool.addr);
    pData->freelist = (struct cell **)(gPool.addr + size_table);
    pData->pEntry = (struct cell *)(gPool.addr + size_table + size_free);
    gpTable = pData->pTable;
   LOGINFO("[Kernel setup_pool] is called 444444444....\n");

    return 1;
}


int tmci_InitClientTable(void)
{
    return setup_pool(&gPool);
}




