/*
** $Id: mempool.c 1.23 2004/10/04 17:15:41 cmchang Exp $
*/

/*! \file  mempool.c
   In some embedded system, the memory allocation/free mechanism seems not so robust.
   Therefore, we better allocate memory in the beginning and do not return allocated memory until
   driver is unloaded.
*/

/******************************************************************************
Copyright (c) 2002-2003 Inprocomm, Inc.

All rights reserved. Copying, compilation, modification, distribution
or any other use whatsoever of this material is strictly prohibited
except in accordance with a Software License Agreement with Inprocomm, Inc.
******************************************************************************/

/************************************************************************
*                     C O M P I L E R   F L A G S
*************************************************************************
*/

/************************************************************************
*                E X T E R N A L   R E F E R E N C E S
*************************************************************************
*/
#include "precomp.h"
#include "mempool.h"

/************************************************************************
*                          C O N S T A N T S
*************************************************************************
*/

/************************************************************************
*                         D A T A   T Y P E S
*************************************************************************
*/

/************************************************************************
*                        P U B L I C   D A T A
*************************************************************************
*/

/************************************************************************
*                       P R I V A T E   D A T A
*************************************************************************
*/

/* Now we only support one card in system simulateously. Although multiple
 * cards can still work, but they share the same memory pool.
 */
static ipn_slab_anchor_t ipn_mem_slabs[TOTAL_MEM_TYPE]=
{
        { (sizeof(Frame_t)+3)&~3,    NUM_BLOCK_S(sMaxAId), NULL, NULL, },
        { (sizeof(FragSdu_t)+sizeof(FragMpdu_t)*3+3)&~3,
                                     NUM_BLOCK_M(sMaxAId), NULL, NULL, },
        { (sizeof(sta_info_t)+3)&~3, NUM_BLOCK_L(sMaxAId), NULL, NULL, }
};

/* This variable is used to release physical memory pool */
static int      ipn_mempool_used_count = 0;

/************************************************************************
*                             M A C R O S
*************************************************************************
*/

/************************************************************************
*              F U N C T I O N   D E C L A R A T I O N S
*************************************************************************
*/

/************************************************************************
*                          F U N C T I O N S
*************************************************************************
*/

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
int init_mgt_mem_pool(i2220mgmt_t *priv_p)
{
        ipn_slab_anchor_t       *anchor_p = &ipn_mem_slabs[0];
        ipn_slab_item_t         *item_p;
        int                     pool_size, i, j;

        /* Only one memory pool is created even though multiple cards
         * exist simultaneously. In the future, we should change
         * this mechanism to guarantee enough memory.
         */
        if (ipn_mempool_used_count++ > 0) return 0;


        ASSERT((SLAB_HEAD_SZ &3)==0);
        if (sizeof(Frame_t) > ipn_mem_slabs[LEN_BLOCK_S].block_sz ||
            (sizeof(FragSdu_t)+sizeof(FragMpdu_t)) >
                ipn_mem_slabs[LEN_BLOCK_M].block_sz ||
            sizeof(sta_info_t) > ipn_mem_slabs[LEN_BLOCK_L].block_sz) {

                KDBG_LOG_INIT(KDBG_INFO,"Mempool needs re-partition...\n");
                return -1;
        }

        KDBG_LOG_INIT(KDBG_DEBUG,"priv_p=%d, mib=%d\n",
                sizeof(i2220mgmt_t), sizeof(IEEE_802_11_MIB));
        KDBG_LOG_INIT(KDBG_DEBUG,"IPN_SW_KEY_TABLE_ENTRY=%d, mem=%d\n",
                sizeof(IPN_SW_KEY_TABLE_ENTRY),
                sizeof(IPN_SW_KEY_TABLE_ENTRY) * IPN_KEY_TABLE_SIZE);
        KDBG_LOG_INIT(KDBG_DEBUG,"frame_t=%d, fsdu=%d\n",
                sizeof(Frame_t), sizeof(FragSdu_t));
        KDBG_LOG_INIT(KDBG_DEBUG,"sta_p=%d, mem=%d\n",
                sizeof(sta_info_t), sizeof(sta_info_t)*x_MAX_AID);

        /* Assign block number by x_MAX_AID, which is a module parameter */
        ipn_mem_slabs[LEN_BLOCK_S].num_block = NUM_BLOCK_S(x_MAX_AID);
        ipn_mem_slabs[LEN_BLOCK_M].num_block = NUM_BLOCK_M(x_MAX_AID);
        ipn_mem_slabs[LEN_BLOCK_L].num_block = NUM_BLOCK_L(x_MAX_AID);

        KDBG_LOG_INIT(KDBG_INFO, "total_mempool= %d bytes\n",
                ipn_mem_slabs[LEN_BLOCK_S].num_block *
                        (ipn_mem_slabs[LEN_BLOCK_S].block_sz+SLAB_HEAD_SZ) +
                ipn_mem_slabs[LEN_BLOCK_M].num_block *
                        (ipn_mem_slabs[LEN_BLOCK_M].block_sz+SLAB_HEAD_SZ) +
                ipn_mem_slabs[LEN_BLOCK_L].num_block *
                        (ipn_mem_slabs[LEN_BLOCK_L].block_sz+SLAB_HEAD_SZ));


        for (i = 0; i < TOTAL_MEM_TYPE; i++, anchor_p++) {
                pool_size = ipn_mem_slabs[i].num_block *
                                (ipn_mem_slabs[i].block_sz + SLAB_HEAD_SZ);

                anchor_p->virt_addr = (void*)kal_malloc(pool_size);
                if (anchor_p->virt_addr == NULL) {
                        printk(/*KDBG_LOG_INIT(KDBG_ERR,*/
                                "Memory not enough for slabs!!\n");
                        destroy_mgt_mem_pool(priv_p);
                        return -1;
                }
                KDBG_LOG_INIT(KDBG_DEBUG,"index=%d, virt=0x%08lx, size=%d\n",
                        i, (UINT32)anchor_p->virt_addr, pool_size);

                /* Setup link list of slabs */
                anchor_p->next = NULL;
                item_p = (ipn_slab_item_t *) anchor_p->virt_addr;
                for (j = 0; j < ipn_mem_slabs[i].num_block; j++) {
                        item_p->sz_index = i;   /* Index */
                        item_p->next = anchor_p->next;
                        anchor_p->next = item_p;

                        item_p = (ipn_slab_item_t*) ((UINT8*)item_p +
                                        ipn_mem_slabs[i].block_sz + SLAB_HEAD_SZ);
                }
                anchor_p->max_alloc = 0;
                anchor_p->alloc_count= 0;
        }

        return 0; /* Success */
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
void destroy_mgt_mem_pool(i2220mgmt_t *priv_p)
{
        ipn_slab_anchor_t       *anchor_p;
        ipn_slab_item_t         *item_p;
        int                     i, num;

        /* How many devices use this memory pool simulateously */
        if (--ipn_mempool_used_count > 0) return;

        /* Check if all slabs are free */
        anchor_p = &ipn_mem_slabs[0];
        for (i = 0; i < TOTAL_MEM_TYPE; i++, anchor_p++) {
                item_p = anchor_p->next;
                for (num = 0; item_p; num++, item_p=item_p->next);

                /* This function is also called when fail to create memory pool,
                 * so we should check if the slab is created (virt_addr).
                 * This case should not happen !!
                 */
                if (num != ipn_mem_slabs[i].num_block && anchor_p->virt_addr) {
                        KDBG_LOG_INIT(KDBG_ERR,
                                "Some slabs not freed (index=%d)\n", i);
                        return; /* Don't free memory pool */
                }
        }

        /* Free entire memory pool if all slabs are free */
        anchor_p = &ipn_mem_slabs[0];
        for (i = 0; i < TOTAL_MEM_TYPE; i++, anchor_p++) {
                if (anchor_p->virt_addr) {
                        kal_free(anchor_p->virt_addr);
                }
                anchor_p->virt_addr = NULL;
                anchor_p->next = NULL;
                anchor_p->max_alloc = 0;
                anchor_p->alloc_count= 0;
        }
}

/*----------------------------------------------------------------*/
/*!
* \brief We try to find as small as possible block to accomondate
*        this request.
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
void *slab_mem_alloc(int sz)
{
        ipn_slab_item_t         *item_p;
        int                     i;

        /* If the memory pool is not created successfully, we use system's
         * normal alloc/free mechanism.
         */
        if (ipn_mem_slabs[0].virt_addr == NULL) {
                return (void*)kal_malloc(sz);
        }

        /* First matching sz then stop. If appropriate pool is full,
         * the larger slab should be used.
         */
        for (i = 0; i < TOTAL_MEM_TYPE; i++) {
                if (sz <= ipn_mem_slabs[i].block_sz &&
                    ipn_mem_slabs[i].next) {
                        break;
                }
        }
        if (i >= TOTAL_MEM_TYPE || sz <= 0) {
                KDBG_LOG_INIT(KDBG_WARN, "No slab mamory for size=%d:\n", sz);
		return NULL;
        }

{
//jjunsigned long save_flags;
//jjlocal_irq_save(save_flags);
        item_p = ipn_mem_slabs[i].next;
        ipn_mem_slabs[i].next = item_p->next;
//jjlocal_irq_restore(save_flags);
}
#if 0
        /* Debug */
        ipn_mem_slabs[i].alloc_count++;
        ipn_mem_slabs[i].max_alloc++;
#endif

        return (void*)&item_p->data[0];
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
void slab_mem_free(void *buf)
{
        ipn_slab_item_t *item_p;
        int             index;

        /* This case may occur only when fail to create memory pool */
        if (ipn_mem_slabs[0].virt_addr == NULL) {
                kal_free(buf);
                return;
        }

        item_p = (ipn_slab_item_t*)((UINT8*)buf - SLAB_HEAD_SZ);
        index = item_p->sz_index;
        ASSERT(index>=0 && index<TOTAL_MEM_TYPE);

#if WLAN_INCLUDE_DEBUG
        if (index < 0 || index >= TOTAL_MEM_TYPE) {
                KDBG_LOG_INIT(KDBG_ERR, "Wrong index %d\n", index);
                return;
        }
#endif
        /* Link it to pool */
{
//jjunsigned long save_flags;
//jjlocal_irq_save(save_flags);

        item_p->next = ipn_mem_slabs[index].next;
        ipn_mem_slabs[index].next = item_p;
//jjlocal_irq_restore(save_flags);
}
#if 0
        /* Debug */
        ipn_mem_slabs[index].alloc_count--;
#endif

        return;
}


