/*
** $Id: dot11_if.c 1.88 2004/09/23 19:14:37 cmchang Exp $
*/

/*! \file  dot11_if.c
    \brief Interface functions specific to the dot11 cut interface.

    This file is for the purpose of intergrating our solution with our customer's
    application software, cutting at the 802.11 level. That is, the exchanged data
    packets is already in WLAN format.
*/

/******************************************************************************
Copyright (c) 2002-2003 Integrated Programmable Communications, Inc.

All rights reserved. Copying, compilation, modification, distribution
or any other use whatsoever of this material is strictly prohibited
except in accordance with a Software License Agreement with
Integrated Programmable Communications, Inc.
******************************************************************************/

/************************************************************************
*                     C O M P I L E R   F L A G S
*************************************************************************
*/

/************************************************************************
*                E X T E R N A L   R E F E R E N C E S
*************************************************************************
*/
#include "precomp.h"

/************************************************************************
*                          C O N S T A N T S
*************************************************************************
*/

/************************************************************************
*                         D A T A   T Y P E S
*************************************************************************
*/

/************************************************************************
*                        P U B L I C   D A T A
*************************************************************************
*/

/************************************************************************
*                       P R I V A T E   D A T A
*************************************************************************
*/

/************************************************************************
*                             M A C R O S
*************************************************************************
*/

/************************************************************************
*              F U N C T I O N   D E C L A R A T I O N S
*************************************************************************
*/


/************************************************************************
*                          F U N C T I O N S
*************************************************************************
*/

/*----------------------------------------------------------------*/
/*!
* \brief  PHY_SAP_RX service access point of PHY receiving interface, as
*         defined in 802.11 standard, p.401
*
* This function corresponds to the PHY_SAP_RX channel, see standard p 401.
* Since many functionalities of the SDL BLOCK Reception, Transmission and
* Protocol_Control_AP are handled by IPN2220 MAC hardware, the driver
* software needs to consider only partial of the combined functionality,
* of course.
*
* \param   priv_p  pointer to i2220mgmt_t
* \param   rfb_p   pointer to the structure for RFB of IPN2220 MAC hardware.
* \return  none
*/
/*----------------------------------------------------------------*/
int env_PHY_SAP_RX(i2220mgmt_t *priv_p, rfb_t *rfb_p)
{
        Frame_t         *pdu = (Frame_t *)kal_skb_data(rfb_p->sw.rfb_skb);
        sta_info_t      *sta_p;
        KDBG_ENTER;

        /* Duplicate filter except WDS frames */
        if ((sta_p = task_sta_lookup(priv_p, pdu->addr2)) != NULL){
                sta_p->actTime = kal_jiffies; /* update activity time */
        #if __CPU_ENDIAN__ != __LITTLE_ENDIAN__
                sta_p->bbCr29 = (sta_p->bbCr29*7 + rfb_p->hw.bbCr30) / 8; /*swap*/
        #else
                sta_p->bbCr29 = (sta_p->bbCr29*7 + rfb_p->hw.bbCr29) / 8;
        #endif
                sta_p->sta_statistics.dot11ReceivedFragmentCount++;
                if (sdl_isGroup(pdu->addr1)) {
                        sta_p->sta_statistics.dot11MulticastReceivedFrameCount++;
                }

                if (sdl_retryBit(pdu) &&
                    ieee2host16(pdu->seqCtrl) == sta_p->seqCtrl) {
                        /* update statistics count */
                        mib_import(dot11FrameDuplicateCount)++;
                        sta_p->sta_statistics.dot11FrameDuplicateCount++;
                        KDBG_LOG_RX(KDBG_INFO, "Packet duplicated seqCtl=0x%04x "
                                MACSTR "\n", ieee2host16(pdu->seqCtrl),
                                MAC2STR(pdu->addr2));

                        return RX_DRV_HANDLED_PACKET;
                }
                sta_p->seqCtrl = ieee2host16(pdu->seqCtrl);
        }

        /* Process PS mode of sta_p when aid>0 */
        sdlu_PM_monitor(priv_p, pdu, sta_p);

        /* PS-Poll has no seqCtrl field, so error judgement in sdlu_defragment */
        if (sdl_ftype(pdu) == SYN_ps_poll) {
                if (!priv_p->stop_accept_station) {
                        rx_process_ps_poll(priv_p, pdu, sta_p);
                }
                return RX_DRV_HANDLED_PACKET;
        }

        switch (sdlu_defragment(priv_p, &pdu)){
        case FRAG_STATUS_COMPLETE:
                rfb_p->sw.rfb_skb = pdu->skb; /* important for packet re-use */
                break;
        case FRAG_STATUS_DISCARD:
                KDBG_EXIT;
                return RX_DRV_HANDLED_PACKET;
        case FRAG_STATUS_BEGIN:
                KDBG_EXIT;
                return RX_INDICATE_PACKET_UPWARD;
        }

        if (sdl_basetype(pdu) == SYN_ft_data) {
                /* Process data frame */
                if (pdu->mac_ctl & MACCTL_WDS_FRAME) {
                        return wds_rx_process_data(priv_p, pdu, sta_p, rfb_p);
                }

                /* Discard data frame if AP function is disabled */
                if (priv_p->stop_accept_station) {
                        return RX_DRV_HANDLED_PACKET;
                }
                return rx_process_data(priv_p, pdu, sta_p, rfb_p);
        }

        /* Process management frame */
        switch (sdl_ftype(pdu)) {
        case SYN_beacon:
        case SYN_probe_rsp:
                #if WLAN_INCLUDE_WDSFULL
                if(priv_p->iw_mode == NIC_STATION) {
                        wds_handle_beacon_probersp(priv_p,pdu);
                        return RX_DRV_HANDLED_PACKET;
                }
                #endif

                rx_process_beacon(priv_p, pdu, sta_p);
                break;
        case SYN_probe_req:
                if (priv_p->stop_accept_station) break;
                rx_process_probe_req(priv_p, pdu, sta_p);
                break;
        case SYN_asoc_req:
        case SYN_reasoc_req:
                if (priv_p->stop_accept_station) break;
                rx_process_reasoc_req(priv_p, pdu, sta_p);
                break;
        case SYN_disasoc:
                #if WLAN_INCLUDE_WDSFULL
                if(priv_p->iw_mode == NIC_STATION){
                        wds_disassoc(priv_p, pdu);
                        return RX_DRV_HANDLED_PACKET;
                }
                #endif

                if (priv_p->stop_accept_station) break;
                rx_process_disasoc(priv_p, pdu, sta_p);
                break;
        case SYN_auth:
                #if WLAN_INCLUDE_WDSFULL
                if(priv_p->iw_mode == NIC_STATION){
                        /*Which means from AP to station*/
                        if((ieee2host16(rpdu->authTransSeqNo) %2 ) == 0)
                                wds_auth_rsp(priv_p, pdu);
                        KDBG_EXIT;
                        return RX_DRV_HANDLED_PACKET;
                }
                #endif

                if (priv_p->stop_accept_station) break;
                rx_process_auth(priv_p, pdu, sta_p);
                break;
        case SYN_deauth:
                #if WLAN_INCLUDE_WDSFULL
                /* Because we are not in a mode of AP operation
                 * We should check which AP deauth us. and try to authenticate
                 */
                if(priv_p->iw_mode == NIC_STATION){
                        wds_deauth(priv_p, pdu);
                        return RX_DRV_HANDLED_PACKET;
                }
                #endif

                if (priv_p->stop_accept_station) break;
                rx_process_deauth(priv_p, pdu, sta_p);
                break;
        default:
              KDBG_LOG_RX(KDBG_WARN, "Unknown frameType=0x%04x from "
                      MACSTR "\n", pdu->frameCtrl, MAC2STR(pdu->addr2));
        }

        return RX_DRV_HANDLED_PACKET;
}

/*----------------------------------------------------------------*/
/*!
* \brief  PHY_SAP_TX service access point of PHY receiving interface, as
*         defined in 802.11 standard, p.401
*
* This function corresponds to the PHY_SAP_TX channel, see standard p 401.
* Since many functionalities of the SDL BLOCK Reception, Transmission and
* Protocol_Control_AP are handled by IPN2220 MAC hardware, the driver
* software needs to consider only partial of the combined functionality,
* of course.
*
* \return  none
*/
/*----------------------------------------------------------------*/
int env_PHY_SAP_TX(i2220mgmt_t *priv_p, FragSdu_t *fsdu)
{
        int             result = 0;
        sta_info_t      *sta_p = fsdu->sw.sta_p;

        KDBG_ENTER;

        if (sta_p && sta_p->asoc.adPsm == SYN_power_save &&
            (fsdu->sw.sdu->mac_ctl & MACCTL_CHECK_SLEEP)) {
                /* store pkt and set TIM
                 * Note: here buffered pkt number must be limited
                 */
                KDBG_LOG_TX(KDBG_DEBUG,"Queue packet for " MACSTR "\n",
                        MAC2STR(sta_p->asoc.adAddr));
                sdlQ_Qlast_limit(&sta_p->psQ, fsdu, 30);
                sdl_set_tmap(priv_p,sta_p->aid);
        }
        else if (fsdu->grpa && priv_p->sta_sleeping) {
                /* Note: here buffered pkt number must be limited */
                KDBG_LOG_TX(KDBG_DEBUG,"Queue packet for bcast\n");
                sdlQ_Qlast_limit(&priv_p->bcQ, fsdu, 50);

                /* set bitmap control when dtimCount==0 in preTBTT */
        }
        else {
                result = sdlu_tx_FragSdu(priv_p, fsdu, TXQ_AC0, TRUE);
        }

        KDBG_EXIT;
        return result;
}

/*----------------------------------------------------------------*/
/*!
* \brief SDL utility function, dividing the packet into fragments.
*
* Depends on the MIB value \a dot11FragmentationThreshold, the packet will
* be divided into several fragments and further transmission control will be
* governed under the \a FragSdu_t structure.
*
* \param  hdr       the wlan header
* \param  LLCdata   the LLC data pointer, (wlan or ether frame header excluded)
* \param  LLClength length of the LLC data
* \param  useWep    whether to encrypt this packet or not
* \return the FragSdu_t control structure.
*/
/*----------------------------------------------------------------*/
FragSdu_t *sdlu_fragment(i2220mgmt_t * priv_p, Frame_t *sdu, UINT8 cipherSuite)
{
        FragSdu_t       *fsdu;
        FragMpdu_t      *tpdu;
        UINT8           *LLCdata;
        UINT16          LLClength, maxLLClen;
        UINT16          pduOvhd, n;
        int             f, fTot, grpa;
        KDBG_ENTER;

        LLCdata = sdu->LLCdata;
        LLClength = sdu->LLClength;
        if (cipherSuite == CipherMode_TKIP) {
                LLClength += 8;
        }

        /*-----------------------------------------------------
         * Calculate the total number of fragments
         *-----------------------------------------------------*/
        grpa = sdl_isGroup(sdu->addr1);

        /* Iv and Icv fields not counted in pre-fragment overhead. */
        pduOvhd = sMacHdrLng + sCrcLng;

        if(sdu->mac_ctl & MACCTL_WDS_FRAME)
                pduOvhd += MAC_ADDR_LEN;

        if(sdu->mac_ctl & MACCTL_QOS_FRAME)
                pduOvhd += 2;

        if (grpa) {
            maxLLClen = 2346 - pduOvhd;
            fTot = 1;
        }
        else {
            maxLLClen = mib_import(dot11FragmentationThreshold) - pduOvhd;
            fTot = (LLClength > maxLLClen) ? (LLClength + maxLLClen - 1) / maxLLClen : 1;
        }
        ASSERT(fTot>0);

        /*-----------------------------------------------------
         * Construct FragSdu_t object, Initialize FragSdu structure
         *-----------------------------------------------------*/
        fsdu = (FragSdu_t*)kal_mempool_alloc(sizeof(FragSdu_t) +
                                fTot * sizeof(FragMpdu_t));
        if (!fsdu) {
                KDBG_LOG_SDL(KDBG_CRIT,"Insufficient system memory resource\n");
                goto L_exit;
        }
        ipn_memset(fsdu, 0, sizeof(FragSdu_t));

        fsdu->grpa = grpa;
        fsdu->fTot = fTot;

        /*-----------------------------------------------------
         * Fragmentation loop
         *-----------------------------------------------------*/
        f = 0;
        n = 0;

        do {
                tpdu = fsdu_pdu(fsdu, f);
                tpdu->LLCdata = (UINT8*)LLCdata + n;

                /* only the last frag size is smaller than maxLLClen */
                tpdu->LLClength = LLClength - n;
                if (tpdu->LLClength > maxLLClen){
                        tpdu->LLClength = maxLLClen;
                }

                n += tpdu->LLClength;
                f += 1;
        } while (f < fTot);

#if SUPPORT_TKIP
        /*----------------------------------------------------------------*/
        /* Special case for Tkip Mic: Mic is 8 bytes length and store in  */
        /*         fsdu->TkipMic[40]~[47],                                */
        /*         that mean: we have to check the last                   */
        /*         sdu->LLClength+8 fragment case for pdus assign         */
        /* case 1:                                                        */
        /*                | ......maxLLClen.......|                       */
        /*                   ---------------|---|                         */
        /*                    < maxLLClen-8 | 8 |                         */
        /*                   ---------------|---|                         */
        /* case 2:                                                        */
        /*                | ......maxLLClen.......|.|                     */
        /*                     -----------------|---|                     */
        /*                      > maxLLClen-8   | 8 |                     */
        /*                     -----------------|---|                     */
        /* case 3:                                                        */
        /*                | ......maxLLClen...|...|                       */
        /*                   -----------------|---|                       */
        /*                       maxLLClen    | 8 |                       */
        /*                   -----------------|---|                       */
        /*                                                                */
        /*----------------------------------------------------------------*/
        if (cipherSuite == CipherMode_TKIP) {
                if (grpa) {
                        fsdu->mod4mic = 1;
                } else {
                        fsdu->lastfTot = (LLClength) - (fTot-1)*maxLLClen;
                        if (fsdu->lastfTot > 8) {/* case 1 */
                                fsdu->mod4mic = 1;
                        } else if (fsdu->lastfTot < 8) { /* case 2 */
                                fsdu->mod4mic = 2;
                        } else { /* case 3, lastfTot==8 */
                                fsdu->mod4mic = 3;
                        }
                }
        }
#endif

L_exit:
        KDBG_EXIT;
        return fsdu;
}


/*----------------------------------------------------------------*/
/*!
* \brief SDL utility function
*
* \return
*/
/*----------------------------------------------------------------*/
FragSdu_t *sdlu_sdu2fsdu(i2220mgmt_t *priv_p, Frame_t *sdu, UINT8 cipherSuite)
{
        FragSdu_t       *fsdu;
        KDBG_ENTER;

        if ( (fsdu = sdlu_fragment(priv_p, sdu, cipherSuite)) == NULL) {
                KDBG_EXIT;
                return NULL;
        }
        fsdu->sw.sdu = sdu;
        fsdu->cipherSuite = cipherSuite;

#if SUPPORT_PERF_MEASURE
        if ( fsdu->sw.sdu->skb && (skb_get_Flag(fsdu->sw.sdu->skb) & SKB_TSTAMP_IN_USE)) {
                UINT32 time;
                time = skb_get_TimeStamp(fsdu->sw.sdu->skb);
                // this is the time when upper LLC passed down the packet.
                fsdu->sw.txProcTime1 = time;//skb_get_TimeStamp(fsdu->sw.sdu->skb);
        }
#endif
        KDBG_EXIT;
        return fsdu;
}

/*----------------------------------------------------------------*/
/*!
* \brief SDL utility function
*
* \return
*/
/*----------------------------------------------------------------*/
int sdu_send_request(i2220mgmt_t *priv_p, Frame_t *tsdu, sta_info_t *sta_p,
                        CfPriority_t pri)
{
        FragSdu_t       *fsdu;
        UINT8           *key_p, keyId, cipherSuite;

        KDBG_ENTER;

        /* Check if the traffic should be blocked */
        if (priv_p->rsnBlockingTraffic || (sta_p && sta_p->rsnBlockingTraffic &&
             sdl_basetype(tsdu) == SYN_ft_data &&
             !CheckFor1xFrame(tsdu->LLCdata, tsdu->LLClength))) {
                sdlu_release_sdu(tsdu);
                KDBG_EXIT;
                return 0;
        }

        /* Find out key information if encryption is needed */
        cipherSuite = TxMsduEncrypt(priv_p, tsdu, sta_p, &keyId, &key_p);
        if (cipherSuite != CipherMode_NONE) {
                sdl_setWepBit(tsdu, 1);

                KDBG_LOG_TX(KDBG_DEBUG, "keyId=%d, mode=%d\n", keyId, cipherSuite);
        }

        /* Divide a MSDU into serval fragments (MPDU) */
        if ((fsdu = sdlu_sdu2fsdu(priv_p, tsdu, cipherSuite)) == NULL) {
                sdlu_release_sdu(tsdu);
                KDBG_EXIT;
                return -1; /* Failed: alloc fsdu */
        }
        fsdu->cf = pri;         /* Priority, define as TXQ_NUM in the future */
        fsdu->sw.sta_p = sta_p; /* Important */

        /* Change default value 512 just for safe. Default value
         * is mib_import(dot11MaxTransmitMSDULifetime).
         */
        fsdu->eol = kal_jiffies + TU2JIFFIES(4096);

        if (sdl_wepBit(tsdu)) {
                fsdu->encryptionKeyId = keyId;
                ipn_memcpy(fsdu->encryptKey, key_p, 32);

        #if SUPPORT_TKIP
                /* Proceed to generate TKIP MIC */
                if (cipherSuite == CipherMode_TKIP) {

                #if WP_3210_WORKAROUND
                        /* Waveplus 3210 define software michael */
                        Decide_Sa_Da(tsdu, &fsdu->sa_p, &fsdu->da_p);
                        Run_Michael_Plus(&fsdu->encryptKey[16],
                                fsdu->da_p, fsdu->sa_p, tsdu->LLCdata,
                                tsdu->LLClength,&fsdu->tkipMic[0]);

                #elif SUPPORT_SW_TKIP_MIC
                        Decide_Sa_Da(tsdu, &fsdu->sa_p, &fsdu->da_p);
                        if (tsdu->LLClength < priv_p->HWMICthreshold) {
                                tkipSwGenMic(&fsdu->encryptKey[16],
                                        tsdu->LLCdata, tsdu->LLClength,
                                        fsdu->sa_p, fsdu->da_p,
                                        0, fsdu->tkipMic);
                        }
                        else {
                                tkipHwGenMic(priv_p, fsdu, SECQ_TX, TRUE);
                                KDBG_EXIT;
                                return 0;
                        }
                #else
                        Decide_Sa_Da(tsdu, &fsdu->sa_p, &fsdu->da_p);
                        tkipHwGenMic(priv_p, fsdu, SECQ_TX, TRUE);
                        KDBG_EXIT;
                        return 0;
                #endif
                }
        #endif /* end of SUPPORT_TKIP */
        }

        /* Start to tx out a series of fragments */
        if (env_PHY_SAP_TX(priv_p, fsdu) != 0) {

                KDBG_LOG_TX(KDBG_WARN,"TFCB(%d)/TBB(%d) is not enough.\n",
                        chain_num_elem(&priv_p->tfcb_chain),
                        chain_num_elem(&priv_p->tbb_chain));

                dbg_tx_tdma(priv_p, TXQ_AC0, 0);
                dbg_tx_tdma(priv_p, TXQ_TQ1, 0);

                KDBG_EXIT;
                return TX_TFCB_FAILURE;
        }

        KDBG_EXIT;
        return 0;
}

/*----------------------------------------------------------------*/
/*!
* \brief  SDL utility function, to initialize the Defragment control structures.
*
* This function initializes the \a ap.defrag_buf[] memory to be used to re-assemble
* the received fragments of frames.
*
* \return none
*/
/*----------------------------------------------------------------*/
void sdlu_init_defrag(i2220mgmt_t *priv_p)
{
        int             i;
        PartialSdu_t    *buf_p = priv_p->ap.defrag_buf;

        for (i = 0; i < SYN_defragSize; i++) {
                ipn_memset(buf_p, 0, sizeof(PartialSdu_t));
                buf_p++;
        }
}

/*----------------------------------------------------------------*/
/*!
* \brief  SDL utility function, to reset the Defragment control structures.
*         and release associated partial SDU
*
* This function initializes the \a ap.defrag_buf[] memory to be used to re-assemble
* the received fragments of frames.
*
* \return none
*/
/*----------------------------------------------------------------*/
void sdlu_destroy_defrag(i2220mgmt_t *priv_p)
{
        int             i;
        PartialSdu_t    *buf_p = priv_p->ap.defrag_buf;

        for (i = 0; i < SYN_defragSize; i++) {
                if (buf_p->pdu) {
                        sdlu_release_sdu(buf_p->pdu);
                        buf_p->pdu = NULL;
                }
                buf_p++;
        }
}

/*----------------------------------------------------------------*/
/*!
* \brief  SDL utility function, to perform the defragmentation of the received frames.
*
* This function reassembles a received frame (MPDU) \a pdu into an integrity of MSDU
* frame. Each received frame, before passed up for further processing, will be checked
* here first for its SEQ/FRAG numbers for the reassembling of fragments.
* For fragments of frames, their content will be copied into a corresponding slot in
* \a ap.defrag_buf[], and the frame itself will be released immediately here. After
* the whole MSDU is sucessfully re-assembled, the memory area in \a ap.defrag_buf[] will
* be indicated upwardly.
*
* \param  pdu   pointer of the pointer to the just received frame
* \return TRUE  if the packet can be indicated upward for further dot11 processing \n
*         FALSE if the packet is to be dropped silently
* \post   the \a ap.defrag_buf[] contains the parked MPDU until received the while MSDU
*         or being aged to be purged.
*/
/*----------------------------------------------------------------*/
int sdlu_defragment(i2220mgmt_t *priv_p, Frame_t **pdu_pp)
{
        Frame_t         *pdu = (*pdu_pp);
        PartialSdu_t    *buf_p, *defrag_buf = priv_p->ap.defrag_buf;
        int             k;
        int             result = FRAG_STATUS_DISCARD;

        KDBG_ENTER;

        sdl_ArSearch(defrag_buf, pdu->addr2, sdl_seq(pdu), sdl_frag(pdu), k);
        if (k >= 0){
                buf_p = &defrag_buf[k];

                KDBG_LOG_RX(KDBG_DEBUG,"Match defrag seqCtl=0x%04x len=%d for "
                        MACSTR "\n", pdu->seqCtrl, pdu->LLClength, MAC2STR(pdu->addr2));

                if (pdu->LLClength + buf_p->pdu->LLClength > MAX_RX_PKT_LEN-32){
                        /* Too long, release existed buffer */
                        sdlu_release_sdu(buf_p->pdu);
                        buf_p->pdu = NULL;

                        KDBG_LOG_RX(KDBG_NOTICE,"Too lengthy (len=%d tailroom=%d)\n",
                                pdu->LLClength, MAX_RX_PKT_LEN - buf_p->pdu->LLClength-32);
                        goto L_end;
                }

                /* Append new fragment to buffered Frame_t */
                ipn_memcpy(buf_p->pdu->LLCdata + buf_p->pdu->LLClength,
                                pdu->LLCdata, pdu->LLClength);
                buf_p->pdu->LLClength += pdu->LLClength;

                buf_p->seqCtrl = ieee2host16(pdu->seqCtrl);
                if (sdl_moreFrag(pdu) == 0) {
                        /* Pass it to distribution function */
                        *pdu_pp = buf_p->pdu;
                        buf_p->pdu = NULL;      /* clear cache */

                        sdlu_release_sdu(pdu);
                        result = FRAG_STATUS_COMPLETE;
                }
        }
        else if (sdl_frag(pdu) == 0) {
                if (sdl_moreFrag(pdu) == 0) {
                        /* No defragments (normal case) */
                        result = FRAG_STATUS_COMPLETE;
                        goto L_end;
                }

                sdl_release_ArAge(defrag_buf, kal_jiffies);
                sdl_ArFree(defrag_buf, k);
                if (k >= 0) {
                        /* Initial MPDU of fragmented MSDU */
                        buf_p = &defrag_buf[k];

                        buf_p->pdu = pdu;
                        buf_p->seqCtrl = ieee2host16(pdu->seqCtrl);
                        buf_p->eol = kal_jiffies + HZ / 10; /* temporary 100ms */
                        result = FRAG_STATUS_BEGIN;
                }
        }
L_end:
        KDBG_LOG_RX(KDBG_DEBUG,"Rx seqCtl=0x%04x len=%d (result=%d) for "
                MACSTR "\n", ieee2host16(pdu->seqCtrl), (*pdu_pp)->LLClength,
                result, MAC2STR(pdu->addr2));

        KDBG_EXIT;
        return result;
}


/*-----------------------------------------------------*/
/* The following functions are added by CMC venus 6/12 */
/*-----------------------------------------------------*/
int wlanhdr_to_wlanhdr(i2220mgmt_t *priv_p, Frame_t *tpdu, Frame_t *pdu)
{
        KDBG_ENTER;

        ipn_memcpy(tpdu->addr1, pdu->addr3, MAC_ADDR_LEN);
        ipn_memcpy(tpdu->addr3, pdu->addr2, MAC_ADDR_LEN);
        ipn_memcpy(tpdu->addr2, mib_import(dot11MACAddress), MAC_ADDR_LEN);
        tpdu->frameCtrl = host2ieee16(SYN_data);
        sdl_setFrDs( tpdu, 1 );

        KDBG_EXIT;
        return 0;
}

void sdlu_PM_monitor(i2220mgmt_t *priv_p, Frame_t *pdu, sta_info_t *sta_p)
{
        PwrSave_t       psm;

        KDBG_ENTER;

        /* ignore bcast/multicast probe request frame */
        if (sdl_ftype(pdu) == SYN_probe_req && sdl_isGroup(pdu->addr1)) {
                KDBG_EXIT;
                return;
        }

        psm = sdl_pwrMgt(pdu) ? SYN_power_save : SYN_sta_active;
        if (sta_p && sta_p->asoc.adPsm != psm && sta_p->aid > 0) {
                /* ps mode changed */
                if (psm == SYN_sta_active) {
                        FragSdu_t       *fsdu;

                        KDBG_LOG_MLME(KDBG_INFO, MACSTR " PS mode to Active (qlen=%d)\n",
                                MAC2STR(pdu->addr2), sdlQ_Length(&sta_p->psQ));

                        /* ps mode -> active mode ==> empty buffered packets */
                        while (sdlQ_Length(&sta_p->psQ) != 0) {
                                sdlQ_dequeue_first(&sta_p->psQ, fsdu);
                                if (check_timeout(fsdu->eol)) {
                                        sdlu_release_fsdu(fsdu);
                                        continue;
                                }
                                sdlu_tx_FragSdu(priv_p, fsdu, TXQ_AC0, TRUE);
                        }
                        sdl_reset_tmap(priv_p,sta_p->aid);
                        priv_p->sta_sleeping--;
                }
                else {
                        priv_p->sta_sleeping++;
                        KDBG_LOG_MLME(KDBG_INFO, MACSTR " Active to PS mode (qlen=%d)"
                                "(sta_sleeping=%d)\n", MAC2STR(pdu->addr2),
                                sdlQ_Length(&sta_p->psQ), priv_p->sta_sleeping);

                        /* If STA enters sleeping immediately or leaving,
                         * transmitting pkts will fail. So time margin avoids
                         * rate degrading, especially when scanning.
                         */
                        sta_p->rateUpdateTime = kal_jiffies;
                }
                sta_p->asoc.adPsm = psm;
        }

        KDBG_EXIT;
}

