/*
** $Id: i2220subr.c 1.79 2004/10/04 17:40:23 cmchang Exp $
*/

/*! \file i2220subr.c
    \brief IPN2220 AP/STA common supporting subroutines.

    This is a set of subroutines.
*/

/******************************************************************************
Copyright (c) 2002-2003 Integrated Programmable Communications, Inc.

All rights reserved. Copying, compilation, modification, distribution
or any other use whatsoever of this material is strictly prohibited
except in accordance with a Software License Agreement with
Integrated Programmable Communications, Inc.
******************************************************************************/

/************************************************************************
*                     C O M P I L E R   F L A G S
*************************************************************************
*/

/************************************************************************
*                E X T E R N A L   R E F E R E N C E S
*************************************************************************
*/
#include "precomp.h"

/************************************************************************
*                          C O N S T A N T S
*************************************************************************
*/

/************************************************************************
*                         D A T A   T Y P E S
*************************************************************************
*/


/************************************************************************
*                        P U B L I C   D A T A
*************************************************************************
*/


/************************************************************************
*                       P R I V A T E   D A T A
*************************************************************************
*/

/************************************************************************
*                             M A C R O S
*************************************************************************
*/

/************************************************************************
*              F U N C T I O N   D E C L A R A T I O N S
*************************************************************************
*/

/************************************************************************
*                          F U N C T I O N S
*************************************************************************
*/

/*----------------------------------------------------------------*/
/*!
* \brief  Free the \a sdu.
*
* Deallocates the \a sdu and associated \skb. Note that, for RX frames
* the \a sdu itself is situated in the \a skb->data area and can't be
* deallocated duplicately.
*
* \param  sdu  the frame structure to be released
* \return none
*/
/*----------------------------------------------------------------*/
void sdlu_release_sdu(Frame_t *sdu)
{
        bool    free_sdu;

        ASSERT(sdu);
        free_sdu = (sdu->mac_ctl & MACCTL_SDU_FRAME_IN_SKB) ?
                        FALSE : TRUE;

        if (sdu->skb) kal_free_skb(sdu->skb);
        if (free_sdu)   kal_mempool_free(sdu);

        return;
}

/*----------------------------------------------------------------*/
/*!
* \brief  Free the \a fsdu.
*
* Deallocates the \a fsdu and associated \sdu and \skb. Note that, for RX frames
* the \a sdu itself is situated in the \a skb->data area and can't be
* deallocated duplicately.
*
* \param  fsdu  the fragment control structure to be released
* \return none
*/
/*----------------------------------------------------------------*/
void sdlu_release_fsdu(FragSdu_t *fsdu)
{
        if (fsdu->sw.sdu) {
                sdlu_release_sdu(fsdu->sw.sdu);
        }
        kal_mempool_free(fsdu);
}

/*----------------------------------------------------------------*/
/*!
* \brief  release the TFCB and associated structures, such as \a fsdu,
*         \a sdu and \a skb.
*
* Invoked on the TDMA transmission done situation. Therefore, the TFCB and
* associated structures must be released then.
*
* \param  tfcb_p  pointer to the TFCB structure
* \return none
*/
/*----------------------------------------------------------------*/
void sdlu_release_tfcb(tfcb_t *tfcb_p)
{
        FragSdu_t       *fsdu;

        if ((fsdu = tfcb_p->sw.fsdu) == NULL) {
                KDBG_LOG_UTIL(KDBG_WARN,
                        "TFCB 0x%08x doesn't have an associated fsdu\n",
                        (unsigned)tfcb_p);
                return;
        }

        if (tfcb_p->sw.frag_nr < fsdu->fTot-1) { /* important */

                return;
        }

        /* now we've finished TX all fragments, clean all
         * allocated memories.
         */
        sdlu_release_fsdu(fsdu);
}

/*----------------------------------------------------------------*/
/*!
* \brief  SDL utility function, actually transmit the (fragments of) packets
*         via TFCB to TFIFO.
*
* For each fragment of the packet, the wlan frame header and body is sending
* to \a tx_prepare_TFCB to construct the TFCB / TBB chains. The return value
* from \a tx_prepare_TFCB will be returned after the end of this function.
* The \a tfcb_p and \a fsdu will also be associated here so that when TX-DONE
* interrupt happens we can release all the related memory via \a tfcb_p.
*
* \param  fsdu    pointer to the FragSdu_t control structure
* \param  priv_p  pointer to i2220mgmt_t, the device's private structure
* \param  tdmaChannel  Normal data packets: chan 2, DTIM broadcast packets: chan 0
* \return TX_TFCB_SUCCESS_MORE_BUFF  successful, and more buffer after this operation
*         TX_TFCB_SUCCESS_NO_BUFF  successful, but no more buffer
*         TX_TFCB_FAILURE  failure, the current operation is aborted
*/
/*----------------------------------------------------------------*/
int sdlu_mk_FragSdu_TFCB(i2220mgmt_t *priv_p, FragSdu_t *fsdu, tdma_parameters_t *tdma_p)
{
        tfcb_t          *tfcb_p, *prev_tfcb_p = NULL;
        UINT8           *data[3]; /* TBB: #0 frame-body, #1 NULL (in general) */
        int             len[3], f;
        UINT8           checksum;

        KDBG_ENTER;
        ASSERT(fsdu->fTot > 0);

        if (fsdu->fTot <= 0)
                return -1;

        KDBG_LOG_TX(KDBG_DEBUG,"Prepare %d TFCB/TBB, LLClen=%d\n",
                fsdu->fTot, fsdu->sw.sdu->LLClength);

        priv_p->seqnum += BIT(4); /* seqnum++ */

        tfcb_p = tdma_p->tail_p;
        for (f = 0; f < fsdu->fTot; f++) {
                FragMpdu_t      *tpdu = fsdu_pdu(fsdu, f);

                data[0] = tpdu->LLCdata;
                len[0] = tpdu->LLClength;

                /* Terminator */
                data[1] = NULL; len[1] = 0;

        #if SUPPORT_TKIP
                /*-------------------------------------------------------------
                 * Special case for Tkip Mic: Mic is 8 bytes length and store
                 * in fsdu->TkipMic[0]~[7],
                 *      that mean: we have to prepare the tbb for the Mic field
                 *      see detail at dot11.c sdlu_fragment func
                 */
                /* Note: fsdu->mod4mic=0 means "no TKIP" */
                if (fsdu->mod4mic == 0) {
                        goto L_bypass_tkip;
                }
                ipn_memcpy(tfcb_p->tkipMic, fsdu->tkipMic, 8);

                if (fsdu->mod4mic == 1) {       /* case 1 */
                        if (f == fsdu->fTot - 1) {
                                len[0] -= 8;

                                /* DMA address for pci_alloc_consistent() */
                                data[1] = (UINT8*)(tfcb_p->dmaAddr +
                                                offsetof(tfcb_t, tkipMic[0]));
                                len[1] =  -8; /* Negative */

                                /* Terminator */
                                data[2] = NULL; len[2] = 0;
                        }
                }
                else if (fsdu->mod4mic == 2) {  /* case 2 */
                        if (f == fsdu->fTot - 2) {
                                len[0] -= (8 - fsdu->lastfTot);

                                /* DMA address for pci_alloc_consistent() */
                                data[1] = (UINT8*)(tfcb_p->dmaAddr +
                                                offsetof(tfcb_t, tkipMic[0]));
                                len[1] = -(8 - fsdu->lastfTot); /* Negative */

                                /* Terminator */
                                data[2] = NULL; len[2] = 0;
                        }
                        else if (f == fsdu->fTot - 1) {
                                /* DMA address for pci_alloc_consistent() */
                                data[0] = (UINT8*)(tfcb_p->dmaAddr +
                                                offsetof(tfcb_t, tkipMic[0]) +
                                                (8 - fsdu->lastfTot));
                                len[0] = -(fsdu->lastfTot); /* Negative */
                        }
                }
                else if (fsdu->mod4mic == 3) {  /* case 3 */
                        if (f == fsdu->fTot - 1) {
                                /* DMA address for pci_alloc_consistent() */
                                data[0] = (UINT8*)(tfcb_p->dmaAddr +
                                                offsetof(tfcb_t, tkipMic[0]));
                                len[0] = -8; /* Negative */
                        }
                }
L_bypass_tkip:
        #endif
                /* Important: clear txStatus and tfdCtrlFlag except EOL */
                tfcb_p->hw.tfdCtrlFlag &= TFCB_FLAG_EOL;
                tfcb_p->hw.txStatus = 0;

                if (!len[0] || tx_prepare_TBB(priv_p, data, len, tfcb_p) != 0){
                        /* no TBB are linked to this TFCB */
                        tfcb_p->hw.tfdCtrlFlag |= TFCB_FLAG_NB;
                }

                /* Link the association between fsdu and tfcb_p */
                tfcb_p->sw.fsdu = fsdu;
                tfcb_p->sw.frag_nr = f;
                tx_setup_TFCB_ctrl_info(priv_p, tfcb_p, prev_tfcb_p);

                /* Check if applying TXOP burst, (10 TFCB burst, 5 not) */
                if (priv_p->x_CfbBurst && tdma_p->tdmaChannel == TXQ_AC0 &&
                    fsdu->fTot == 1 && !(tfcb_p->hw.tciCtrlFlag & TFCB_CTRL_RTS)) {
                        /* Work around for chip_venus_0828.bit, which can not
                         * handle RTS/CTS or fragmentation in CFB mode.
                         */
                        if (++priv_p->cfbCount <= 10) {
                                /* continuous 10 TFCB burst */
                                tfcb_p->hw.tfdCtrlFlag |= TFCB_FLAG_CFB;
                                tfcb_p->hw.tciCtrlFlag |= TFCB_CTRL_CFB;
                        }
                        else if (priv_p->cfbCount >= 15) {
                                priv_p->cfbCount = 0;
                        }
                }

                if (prev_tfcb_p) {
                        prev_tfcb_p->nextTfcb = tfcb_p;
                        prev_tfcb_p->hw.nextTfcbPointer =
                                        host2chip_32(tfcb_p->dmaAddr);
                        /* Calculate TFD checksum */
                        prev_tfcb_p->hw.tfdChecksum = 0;
                        nicCalculateChecksumByte((PVOID)&prev_tfcb_p->hw,
                                        TFCB_TFD_SIZE, &checksum);
                        prev_tfcb_p->hw.tfdChecksum = ~checksum;
                }

                prev_tfcb_p = tfcb_p;

                /* Get new one for next usage, including EOL TFCB */
                get_chain_head(&priv_p->tfcb_chain,tfcb_t,tfcb_p);
                ASSERT(tfcb_p);
        }
        /* New EOL TFCB */
        tfcb_p->hw.tfdCtrlFlag = TFCB_FLAG_EOL;

        ASSERT(prev_tfcb_p);
        prev_tfcb_p->nextTfcb = tfcb_p;
        prev_tfcb_p->hw.nextTfcbPointer = host2chip_32(tfcb_p->dmaAddr);
        /* Calculate TFD checksum */
        prev_tfcb_p->hw.tfdChecksum = 0;
        nicCalculateChecksumByte((PVOID)&prev_tfcb_p->hw, TFCB_TFD_SIZE, &checksum);
        prev_tfcb_p->hw.tfdChecksum = ~checksum;

        /* Clear EOL (tfdCtrlFlag -= TFCB_FLAG_EOL) */
        tdma_p->tail_p->hw.tfdChecksum += TFCB_FLAG_EOL;
        tdma_p->tail_p->hw.tfdCtrlFlag &= ~TFCB_FLAG_EOL;

#if WLAN_INCLUDE_DEBUG
        prev_tfcb_p = tdma_p->tail_p;
        for (f = 0; f <= fsdu->fTot; f++) {
                dbg_tx_packet(priv_p, prev_tfcb_p, DBFLG_TX_TFCB);
                prev_tfcb_p = prev_tfcb_p->nextTfcb;
        }
#endif
        tdma_p->count += fsdu->fTot;
        tdma_p->tail_p = tfcb_p;    /* new EOL TFCB */

        KDBG_EXIT;
        return 0;
}

/*----------------------------------------------------------------*/
/*!
* \brief  SDL utility function, actually transmit the (fragments of) packets
*         via TFCB to TFIFO.
*
* Invoking \a sdlu_mk_FragSdu_TFCB to generate the TFCB/TBB chains from /a fsdu.
* Then start up the TDMA device.
*
* \param  fsdu    pointer to the FragSdu_t control structure
* \param  priv_p  pointer to i2220mgmt_t, the device's private structure
* \param  tdmaChannel  Normal data packets: chan 2, DTIM broadcast packets: chan 0
* \return TX_TFCB_SUCCESS_MORE_BUFF  successful, and more buffer after this operation
*         TX_TFCB_SUCCESS_NO_BUFF  successful, but no more buffer
*         TX_TFCB_FAILURE  failure, the current operation is aborted
*/
/*----------------------------------------------------------------*/
int sdlu_tx_FragSdu(i2220mgmt_t *priv_p, FragSdu_t* fsdu, int tdmaChannel, bool chk_res)
{
        tdma_parameters_t       *tdma_p;
        tfcb_t                  *tfcb_eol_p;
        int                     nTFCB, nTBB;
        int                     result;

        KDBG_ENTER;

        tdma_p = &priv_p->tdma[tdmaChannel];
        if ((tfcb_eol_p = tdma_p->tail_p) == NULL) {
                /* it's firt time to startup this TDMA channel */
                get_chain_head(&priv_p->tfcb_chain,tfcb_t,tdma_p->tail_p);
                if (tdma_p->tail_p == NULL) {
                        sdlu_release_fsdu(fsdu);
                        KDBG_EXIT;
                        return TX_TFCB_FAILURE;
                }
                tdma_p->tail_p->hw.tfdCtrlFlag = TFCB_FLAG_EOL;
                tdma_p->head_p = tdma_p->tail_p;
                tdma_p->count++;
        }

        /* calculate the total required TFCBs/TBBs */
        nTFCB = fsdu->fTot;
        nTBB  = fsdu->fTot;

        if (chk_res && (!tx_check_TDMA(&priv_p->tfcb_chain, nTFCB, nTBB) ||
            sdlQ_Length(&tdma_p->waitQ) != 0)) {
                /* Note: the pkt seq may be out of order when fragment
                 *       if we don't check sdlQ_Length(q).
                 * We consider sdlQ_Length(q) here, loop occurs
                 * in function isr_process_tx_done() if chk_res=TRUE
                 */
                sdlQ_Qlast_limit(&tdma_p->waitQ, fsdu, 100);
                KDBG_EXIT;
                return TX_TFCB_FAILURE;
        }

#if !WLAN_INCLUDE_DEBUG
        /* Start tx_timeout if no previous packet is in process */
        if (!priv_p->chk_txTimout && tdma_p->head_p->hw.tfdCtrlFlag & TFCB_FLAG_EOL) {
                priv_p->chk_txTimout = TRUE;
                set_txTimeout(priv_p);
        }
#endif

        /* Queue the fsdu into tdmaChannel now */
        result = sdlu_mk_FragSdu_TFCB(priv_p, fsdu, tdma_p);

#if SUPPORT_PERF_MEASURE
        if (fsdu->sw.txProcTime1) {
                UINT32 t;
                NIC_MCR_READ(priv_p, MCR_LTTR0, &t);
                if (fsdu->sw.txProcTime2) {
                        printk(KERN_ALERT "#TX (%4u)# "
                                "T%4luus P%4luus M%4luus S%4luus\n"
                                , fsdu->sw.sdu->LLClength
                                , t - fsdu->sw.txProcTime1
                                , fsdu->sw.txProcTime2 - fsdu->sw.txProcTime1
                                , fsdu->sw.txProcTime3 - fsdu->sw.txProcTime2
                                , t - fsdu->sw.txProcTime3);
                } else {
                        printk(KERN_ALERT "#Total TX(%4u)# "
                                "TSF(%08lx,%08lx) %4luus\n"
                                , fsdu->sw.sdu->LLClength
                                , fsdu->sw.txProcTime1, t, t - fsdu->sw.txProcTime1);
                }
        }
#endif

        if (tfcb_eol_p == NULL) {
                KDBG_LOG_TX(KDBG_DEBUG,"TDMA%d is started.\n", tdmaChannel);
                dmaStartTxQueue(priv_p, (TX_QUEUE_NUM)tdmaChannel,
                                tdma_p->head_p->dmaAddr);
        }
        else {
                KDBG_LOG_TX(KDBG_DEBUG,"TDMA%d is resumed.\n", tdmaChannel);
                dmaResumeTxQueue(priv_p, (TX_QUEUE_NUM)tdmaChannel);
        }

        KDBG_EXIT;
        return result;
}

/*----------------------------------------------------------------*/
/*!
* \brief  SDL hardware related utility, to send beacon frame.
*
* The already formatted beacon frame \a tdu, will be written into the hardware
* \b BRCR (Beacon Generation Control Register), dword by dword.
* We will do busy-wait till we acquired the \a OWN bit of the BGCR.
* The frame, \a tdu, will be freed at the end of this function.
*
* \param  tdu      the beacon frame to be written to BGCR
* \param  lenInfo  the Information Element length
* \return TRUE if success, FALSE otherwise.
*/
/*----------------------------------------------------------------*/
bool sdlhw_send_beacon(i2220mgmt_t *priv_p, beacon_frame_t *tpdu, int lenInfo)
{
        UINT32          bc_len;
        UINT32          timestampOffset;
        UINT32          rateCode, txPower;
        bool            result = 0;

        KDBG_ENTER;

        /* Exclude time stamp (8 bytes) */
        bc_len = tpdu->LLClength - sizeof(tpdu->infoElem) + lenInfo - 8;
        ASSERT(bc_len>0 && bc_len<=172);

        /* Write bcn content to NIC card */
        if (pauSetBeaconContent(priv_p, 0,
                        (UINT8*)&tpdu->beaconInterval,bc_len) == FALSE) {
                KDBG_LOG_MLME(KDBG_WARN,"Failed to change Bcn content.\n");
                result = 1;
        }

        /* Set the transmission of generated Beacon frames to 1 Mbps or
         * 6 Mbps and the offset that is used to adjust the timestamp.
         * Evaluate offset using function rateComputePpduDuration().
         * For speed, only when fixed rate is used, the generic procedure
         * is used to calculate the timestamp offset.
         */
        /* Beacon also follows fixed data rate */
        if (priv_p->fixedTxRate_p) {
                UINT16  ppduDuration1, ppduDuration2;

                rateComputePpduDuration(24,
                        priv_p->fixedTxRate_p->txRate,
                        priv_p->phyPreambleType, &ppduDuration1);
                rateComputePpduDuration(0,
                        priv_p->fixedTxRate_p->txRate,
                        priv_p->phyPreambleType, &ppduDuration2);

                timestampOffset = ppduDuration1 - ppduDuration2 +
                                        TX_PLCP_DELAY + TX_RF_DELAY;
                rateCode = priv_p->fixedTxRate_p->txRateCode;
                txPower = priv_p->fixedTxRate_p->isERP ?
                        priv_p->txPowerGainForOFDM : priv_p->txPowerGainForCCK;
        }
        else if (priv_p->lowestBRate_p->isERP) {
                timestampOffset = 32 + TX_PLCP_DELAY + TX_RF_DELAY;
                rateCode = TX_RATE_OFDM_6M;
                txPower = priv_p->txPowerGainForOFDM;
        }
        else {
                timestampOffset = 192 + TX_PLCP_DELAY + TX_RF_DELAY;
                rateCode = TX_RATE_LONG_PREAMBLE_1M;
                txPower = priv_p->txPowerGainForCCK;
        }
        if (timestampOffset > 255) {
                timestampOffset = 255;
        }
        NIC_MCR_WRITE(priv_p, MCR_BGCR0, (timestampOffset << 16) |
                (txPower<<8) | rateCode);

        /* Set own_bit to HW with seqnum and BeaconLength */
        priv_p->seqnum += BIT(4); /* seqnum++ */
        NIC_MCR_WRITE(priv_p, MCR_BGCR1,
                BGCR1_BC_OWN | ((ULONG)priv_p->seqnum<<12) |
                ((ULONG)(bc_len+36)<<8) | (ULONG)bc_len );

        KDBG_EXIT;
        return result;
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
void auto_txrate_switch(i2220mgmt_t *priv_p, sta_info_t *sta_p, tfcb_t *tfcb_p)
{
        UINT32          txRateCode = chip2host_16(tfcb_p->hw.txRates) & BITS(0,5);
        UINT8           newTxRate = 0;
        RateSet_t       adSuptRates;

        ASSERT(sta_p);

        if (!sta_p->curTxRate_p || txRateCode != sta_p->curTxRate_p->txRateCode)
                return;

        adSuptRates = sta_p->aid < 0 ?
                      sta_p->asoc.adSuptRates :
                      sta_p->asoc.adSuptRates & priv_p->supportRates;

        sta_p->weightedRetryCount =
                ((10 - RATE_SWITCH_LATEST_COUNT_WEIGHTING) *
                sta_p->weightedRetryCount) / 10 +
                RATE_SWITCH_LATEST_COUNT_WEIGHTING * tfcb_p->hw.mpduRetryCount;

        if (sta_p->weightedRetryCount <= RATE_SWITCH_UPGRADE_THRESHOLD &&
            check_timeout(sta_p->rateUpdateTime + 10*HZ)) {
                rateGetLowestRate(adSuptRates, sta_p->curTxRate_p->txRate + 1,
                                &newTxRate, NULL);

                if (newTxRate == RATE_5_5M && (adSuptRates & RATE_SET_BIT_6M)) {
                        newTxRate = RATE_6M;
                }
                else if (newTxRate >= RATE_6M && newTxRate <= RATE_11M) {
                        if (adSuptRates & RATE_SET_BIT_12M)
                                newTxRate = RATE_12M;
                        else if (adSuptRates & RATE_SET_BIT_11M)
                                newTxRate = RATE_11M;
                }

                sta_p->rateUpdateTime = kal_jiffies - 10*HZ + HZ / 4;

        } else if (sta_p->weightedRetryCount >= RATE_SWITCH_FALLBACK_THRESHOLD &&
                   check_timeout(sta_p->rateUpdateTime + 2*HZ)) {
                rateGetHighestRate(adSuptRates, sta_p->curTxRate_p->txRate - 1,
                                &newTxRate, NULL);

                if (newTxRate <= RATE_11M && newTxRate >= RATE_6M)
                        sta_p->rateUpdateTime = kal_jiffies - 2*HZ + HZ / 10;
                else
                        sta_p->rateUpdateTime = kal_jiffies;

                if (newTxRate == 0) {
                        sta_p->weightedRetryCount = RATE_SWITCH_INITIAL_WEIGHTED_COUNT;
                }
        }

        if (newTxRate) {
                KDBG_LOG_TX(KDBG_DEBUG, MACSTR " Rate Switch: %d --> %d Mbps\n",
                        MAC2STR(sta_p->asoc.adAddr), sta_p->curTxRate_p->txRate>>1,
                        newTxRate>>1);

                sta_p->curTxRate_p = __get_txRate_info(priv_p, newTxRate);
                sta_p->weightedRetryCount = RATE_SWITCH_INITIAL_WEIGHTED_COUNT;
        }

}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
void set_txTimeout(i2220mgmt_t *priv_p)
{
        mgmt_timer_t    *tmr = &priv_p->tx_timer;

        mgmt_start_timer(tmr, TX_TIMEOUT_INTERVAL,
                     (mgmt_timeout_func) i2220hw_reset_mac,
                     (unsigned long) priv_p);
}

#if SUPPORT_LOOPBACK
/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
void sdlu_send_looppacket(i2220mgmt_t* priv_p, int pk_len, int value)
{
        kal_skb_t       *skb;
        Frame_t         *tpdu;
        unsigned long addr;

        skb = kal_alloc_skb(sizeof(Frame_t) + pk_len);
        if (skb == NULL) {
		printk("sdlu_send_looppacket alloc skb fail \n");
                return 0;
	}

        tpdu = (Frame_t *)kal_skb_data(skb);
        tpdu->skb = skb;
        tpdu->mac_ctl = MACCTL_SDU_FRAME_IN_SKB;
        tpdu->LLCdata = (UINT8*)tpdu + sizeof(Frame_t);
        tpdu->LLClength = pk_len;

        ipn_memset(tpdu->addr1, 0xff, MAC_ADDR_LEN); /* bcast */
        ipn_memcpy(tpdu->addr2, mib_import(dot11MACAddress), MAC_ADDR_LEN);
        tpdu->frameCtrl = host2ieee16(SYN_data);
        ipn_memset(tpdu->LLCdata, value, pk_len);

        addr = rx_start_recv_unit(priv_p);

        LOOP_PRINT("RDMA Start Addr:0x%08x payload value:0x%04x\n",addr,value);

        sdu_send_request(priv_p, tpdu, NULL, SYN_contention);
}
#endif

