
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/config.h>
#include <linux/mtd/partitions.h>

#define WINDOW_ADDR 0x1fc00000
#define WINDOW_SIZE 0x4000000
#define BUSWIDTH 2

static int create_mtd_partitions(void);
static void __exit tc3162_mtd_cleanup(void);
	
static struct mtd_info *tc3162_mtd_info;

__u8 tc3162_read8(struct map_info *map, unsigned long ofs)
{
	return __raw_readb(map->map_priv_1 + ofs);
}

__u16 tc3162_read16(struct map_info *map, unsigned long ofs)
{
	return __raw_readw(map->map_priv_1 + ofs);
}

__u32 tc3162_read32(struct map_info *map, unsigned long ofs)
{
	return __raw_readl(map->map_priv_1 + ofs);
}

void tc3162_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	memcpy_fromio(to, map->map_priv_1 + from, len);
}

void tc3162_write8(struct map_info *map, __u8 d, unsigned long adr)
{
	__raw_writeb(d, map->map_priv_1 + adr);
	mb();
}

void tc3162_write16(struct map_info *map, __u16 d, unsigned long adr)
{
	__raw_writew(d, map->map_priv_1 + adr);
	mb();
}

void tc3162_write32(struct map_info *map, __u32 d, unsigned long adr)
{
	__raw_writel(d, map->map_priv_1 + adr);
	mb();
}

void tc3162_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
	memcpy_toio(map->map_priv_1 + to, from, len);
}

struct map_info tc3162_map = {
	name: "Physically mapped flash",
	size: WINDOW_SIZE,
	buswidth: BUSWIDTH,
	read8: tc3162_read8,
	read16: tc3162_read16,
	read32: tc3162_read32,
	copy_from: tc3162_copy_from,
	write8: tc3162_write8,
	write16: tc3162_write16,
	write32: tc3162_write32,
	copy_to: tc3162_copy_to
};

static int num_of_partitions = 0;
static struct mtd_partition tc3162_partition_info[] = {
        {
                //name: "Boot1",
                name: "boot1",
                size: 0x0002000, // 8k bootcode
                offset: 0,
        },
        {
                //name: "Mac",
                name: "mac",
                size: 0x0002000, // 8k Mac/vendor info.
                offset: MTDPART_OFS_APPEND,
        },
        {
                //name: "config",
                name: "factory",
                size: 0x0004000, // 16KB config
                offset: MTDPART_OFS_APPEND,
        },
        {
                //name: "config",
                name: "reserved",
                size: 0x0008000, // 32KB config
                offset: MTDPART_OFS_APPEND,
        },
        {
                //name: "Boot2",
                name: "boot2",
                size: 0x0010000, // 64k bootcode
                offset: MTDPART_OFS_APPEND,
        },
        {                                                                        /* Third partition */
                  name : "kernel",                               /* Kernel section */
                  size    : 0x3c0000, /* MTDPART_SIZ_FULL,*/                 /* Size */
                  offset : MTDPART_OFS_APPEND,   /* Append after bootloader section */
        },
        {                                                                        /* Fourth partition */
                  name : "rootfs",                               /* Root filesystem section */
                  size    :  0x3c0000, /*MTDPART_SIZ_FULL, */  /* Occupy rest of flash */
                  offset :   MTDPART_OFS_APPEND       /* Append after kernel section */
        },
        {
                //name: "config",
                name: "saved config",
                size: 0x0020000, // 128KB config
                offset: 0x3e0000
        },
};

static struct mtd_info *get_mtd_named(char *name)
{
	int i;
	struct mtd_info *mtd;

	for (i = 0; i < MAX_MTD_DEVICES; i++) {
		mtd = get_mtd_device(NULL, i);
		if (mtd) {
			if (strcmp(mtd->name, name) == 0)
				return(mtd);
			put_mtd_device(mtd);
		}
	}
	return(NULL);
}

#define NB_OF(x)  (sizeof(x)/sizeof(x[0]))

int __init tc3162_mtd_init(void)
{
	struct mtd_info *mtd;

	printk(KERN_NOTICE "tc3162 flash device: 0x%lx at 0x%lx.\n", (unsigned long)WINDOW_SIZE, (unsigned long)WINDOW_ADDR);
	tc3162_map.map_priv_1 = (unsigned long)ioremap_nocache(WINDOW_ADDR, WINDOW_SIZE);

	if (!tc3162_map.map_priv_1) {
		printk("Failed to ioremap\n");
		return -EIO;
	}
	tc3162_mtd_info = do_map_probe("cfi_probe", &tc3162_map);
	if (!tc3162_mtd_info)
	{
		printk("tc3162 flash: no probe\n");
		//tc3162_mtd_cleanup();
		return -ENXIO;
	}
	
	tc3162_mtd_info->module = THIS_MODULE;
	num_of_partitions = NB_OF(tc3162_partition_info);

	tc3162_partition_info[6].offset = 0x20000 + (((*(long*)(KSEG1ADDR(WINDOW_ADDR)+0x00020000+76))+128+0x10000) & 0xffff0000);
	tc3162_partition_info[6].size = 0x3e0000 - tc3162_partition_info[6].offset;

	add_mtd_partitions(tc3162_mtd_info, tc3162_partition_info, num_of_partitions);

#ifndef CONFIG_EMBEDDED_RAMDISK
	mtd = get_mtd_named("rootfs");
	if (mtd) {
		ROOT_DEV = MKDEV(MTD_BLOCK_MAJOR, mtd->index);
		put_mtd_device(mtd);
	}
#endif
	return 0;
}

static void __exit tc3162_mtd_cleanup(void)
{
	if (tc3162_mtd_info) {
		del_mtd_partitions(tc3162_mtd_info);
		del_mtd_device(tc3162_mtd_info);
		map_destroy(tc3162_mtd_info);
	}

	if (tc3162_map.map_priv_1) {
		iounmap((void *)tc3162_map.map_priv_1);
		tc3162_map.map_priv_1 = 0;
	}
}

module_init(tc3162_mtd_init);
module_exit(tc3162_mtd_cleanup);

