/*
** $Id: dot3_if.c 1.23 2004/07/23 17:57:27 cmchang Exp $
*/

/*! \file  dot3_if.c
    \brief Interface functions specific to the dot3 cut interface.

    This file is for the purpose of intergrating our solution with our customer's
    application software, cutting at the 802.3 level. That is, the exchanged data
    packets is in Ethernet format.
*/

/******************************************************************************
Copyright (c) 2002-2003 Integrated Programmable Communications, Inc.

All rights reserved. Copying, compilation, modification, distribution
or any other use whatsoever of this material is strictly prohibited
except in accordance with a Software License Agreement with
Integrated Programmable Communications, Inc.
******************************************************************************/

/************************************************************************
*                     C O M P I L E R   F L A G S
*************************************************************************
*/

/************************************************************************
*                E X T E R N A L   R E F E R E N C E S
*************************************************************************
*/
#include "precomp.h"

/************************************************************************
*                          C O N S T A N T S
*************************************************************************
*/
#ifdef ETH_DATA_LEN
#undef ETH_DATA_LEN
#endif
#define ETH_DATA_LEN            1500

#ifdef ETH_HLEN
#undef ETH_HLEN
#endif
#define ETH_HLEN                14

#ifdef ETH_P_IPX
#undef ETH_P_IPX
#endif
#define ETH_P_IPX          0x8137

#ifdef ETH_P_AARP
#undef ETH_P_AARP
#endif
#define ETH_P_AARP       0x80F3

#ifdef ETH_P_8021Q
#undef ETH_P_8021Q
#endif
#define ETH_P_8021Q     0x8100

#ifdef VLAN_HLEN
#undef VLAN_HLEN
#endif
#define VLAN_HLEN       4
/************************************************************************
*                         D A T A   T Y P E S
*************************************************************************
*/


/************************************************************************
*                        P U B L I C   D A T A
*************************************************************************
*/

/************************************************************************
*                       P R I V A T E   D A T A
*************************************************************************
*/

/* See IEEE 802.1H for LLC/SNAP encapsulation/decapsulation */
/* Ethernet-II snap header (RFC1042 for most EtherTypes) */
unsigned char rfc1042_header[] =
{ 0xaa, 0xaa, 0x03, 0x00, 0x00, 0x00 };

/* Bridge-Tunnel header (for EtherTypes ETH_P_AARP and ETH_P_IPX) */
unsigned char bridge_tunnel_header[] =
{ 0xaa, 0xaa, 0x03, 0x00, 0x00, 0xf8 };
/* No encapsulation header if EtherType < 0x600 (=length) */

/************************************************************************
*                             M A C R O S
*************************************************************************
*/

/************************************************************************
*              F U N C T I O N   D E C L A R A T I O N S
*************************************************************************
*/

/************************************************************************
*                          F U N C T I O N S
*************************************************************************
*/

/*----------------------------------------------------------------*/
/*!
* \brief Uses the contents of the ether frame to build the elements
*        of the 802.11 frame.
*
* We don't actually set up the frame header here.  As we have already
* set it properly in the SDL Hypothetical Architecture.  We're only handling
* conversion of DIXII or 802.3+LLC frames to something that works
* with 802.11.
*
* Note -- 802.11 header is NOT part of the skb.  Likewise, the 802.11
*         FCS is also not present and will need to be added elsewhere.
*
* \param   skb  skbuff containing the ether frame
*
* \return  0 on success, non-zero otherwise
*
* \sa      wlanhdr_to_ether, i2220ap_hard_start_xmit
*/
/*----------------------------------------------------------------*/
int wlanhdr_from_ether(kal_skb_t *skb)
{
        UINT16          proto;
        ethhdr_t        *eth_p;

        KDBG_ENTER;

        if (kal_skb_len(skb) < ETH_HLEN) {
                KDBG_LOG_TX(KDBG_WARN, "Too short skb!\n");
                return 1;
        }
        eth_p = (ethhdr_t *) kal_skb_data(skb);

        /* step 1: classify ether frame, DIX or 802.3? */
        proto = kal_ntohs(eth_p->h_proto);
        if ( proto <= ETH_DATA_LEN ) {
                KDBG_LOG_TX(KDBG_DEBUG, "802.3 len: %d\n", kal_skb_len(skb));
                /* codes <= 1500 reserved for 802.3 lengths */
                /* it's 802.3, pass ether payload unchanged,  */

                /* trim off ethernet header */
                kal_skb_pull(skb, ETH_HLEN);

        } else {
                UINT8 *encaps_data;

                KDBG_LOG_TX(KDBG_DEBUG, "DIXII len: %d\n", kal_skb_len(skb));
                /* it's DIXII, time for some conversion */

                /* trim off ethernet header */
                kal_skb_pull(skb, ETH_HLEN - sizeof(eth_p->h_proto));

                if (proto == ETH_P_AARP || proto == ETH_P_IPX) {
                        encaps_data = kal_skb_push(skb, sizeof(bridge_tunnel_header));
                        ipn_memcpy( encaps_data, bridge_tunnel_header, sizeof(bridge_tunnel_header));
                }
                else if (proto >= 0x600) {
                        encaps_data = kal_skb_push(skb, sizeof(rfc1042_header));
                        ipn_memcpy( encaps_data, rfc1042_header, sizeof(rfc1042_header));
                }
        }

        KDBG_EXIT;
        return 0;
}


/*----------------------------------------------------------------*/
/*!
* \brief Uses the contents of a received 802.11 frame to build an ether frame.
*
* This function extracts the src and dest address from the 802.11
* frame to use in the construction of the eth frame.
*
* \param   dev   the net_device object
* \param   rsdu  the 802.11 frame to convert
*
* \return  0 on success, non-zero otherwise
*
* \sa      wlanhdr_from_ether, i2220ap_hard_start_xmit
*/
/*----------------------------------------------------------------*/
int wlanhdr_to_ether(kal_net_device_t *dev, Frame_t *rsdu)
{
        kal_skb_t       *skb = rsdu->skb;
        UINT8           daddr[MAC_ADDR_LEN];
        UINT8           saddr[MAC_ADDR_LEN];
        UINT8           *llc_p;
        ethhdr_t        *eth_p;
        UINT16          type;
        bool            ethernetRepresentation;

        KDBG_ENTER;

        if (sdl_toDs(rsdu) && sdl_frDs(rsdu)==0) {
                ipn_memcpy( daddr, rsdu->addr3, MAC_ADDR_LEN );
                ipn_memcpy( saddr, rsdu->addr2, MAC_ADDR_LEN );
        }
        else if (sdl_toDs(rsdu) && sdl_frDs(rsdu)) {
                ipn_memcpy( daddr, rsdu->addr3, MAC_ADDR_LEN );
                ipn_memcpy( saddr, rsdu->addr4, MAC_ADDR_LEN );
        }
        else if (sdl_toDs(rsdu)==0 && sdl_frDs(rsdu)==0) {
                ipn_memcpy( daddr, rsdu->addr1, MAC_ADDR_LEN );
                ipn_memcpy( saddr, rsdu->addr2, MAC_ADDR_LEN );
        }
        else if (sdl_toDs(rsdu)==0 && sdl_frDs(rsdu)) {
                /* Warning: AP should not be here! */
                ipn_memcpy( daddr, rsdu->addr1, MAC_ADDR_LEN );
                ipn_memcpy( saddr, rsdu->addr3, MAC_ADDR_LEN );
        }

        /* now hdr_len should be fixed size = wlanbody_offset */
        kal_skb_put(skb, wlanbody_offset + rsdu->LLClength);

        /* chop 802.11 header and mac_ctl_descript from skb. */
        kal_skb_pull(skb, wlanbody_offset);
        llc_p = kal_skb_data(skb);

        /* Check the LLC of the received 802.11 packet */
        if (ipn_memcmp(llc_p,rfc1042_header,sizeof(rfc1042_header)) == 0) {
                /* If the LLC is RFC1042 protocol, we need to check the Type field in
                 * the LLC against the Selective Translation Table.
                 * Currently, there are two types in the Selective Translation Table
                 * according to the WiFi's requirements: AppleTalk ARP (0x80F3) and
                 * DIX II IPX (0x8137)
                 */
                type = kal_ntohs( *(UINT16*)(llc_p+sizeof(rfc1042_header)) );

                if (type == ETH_P_AARP || type == ETH_P_IPX) {
                        /* If the type is included in the Selective Translation Table, the
                         * 802.11 MAC is replaced by the ISO/IEC 8802-3 MAC below the LLC.
                         */
                        ethernetRepresentation = FALSE;
                        KDBG_LOG_RX(KDBG_DEBUG,"RFC1042: in the Table\n");
                } else {
                        /* If the type is not included in the Selective Translation Table,
                         * the 802.11 MAC and RFC1042 LLC is replaced by its Ethernet MAC
                         * representation.
                         */
                        ethernetRepresentation = TRUE;
                        KDBG_LOG_RX(KDBG_DEBUG,"RFC1042: not in the Table\n");
                }
        }
        else if (ipn_memcmp(llc_p,bridge_tunnel_header,sizeof(bridge_tunnel_header)) == 0) {
                /* If the LLC is Bridge-Tunnel Encapsulation Protocol,
                 * the 802.11 MAC and Bridge-Tunnel LLC is replaced by
                 * its Ethernet MAC representation.
                 */
                ethernetRepresentation = TRUE;
                KDBG_LOG_RX(KDBG_DEBUG,"Bridge-Tunnel\n");
        }
        else {
                /* If the LLC is neither Bridge-Tunnel Encapsulation Protocol nor
                 * RFC1042 protocol, the 802.11 MAC is replaced by the ISO/IEC 8802-3
                 * MAC below the LLC.
                 */
                ethernetRepresentation = FALSE;

                KDBG_LOG_RX(KDBG_DEBUG,"Other LLC\n");
        }   /* end of check the LLC header */


        /* Replace the 802.11 MAC and the LLC, if needed, by its Ethernet
         * representation or the ISO/IEC 8802-3 MAC.
         */
        if (ethernetRepresentation) {
                kal_skb_push(skb, ETH_HLEN - 8);
                eth_p = (ethhdr_t *) kal_skb_data(skb);
                /* The type field in the Ethernet MAC header is the same as the
                   original Type field in the LLC. */
        } else {
                kal_skb_push(skb, ETH_HLEN);
                eth_p = (ethhdr_t *) kal_skb_data(skb);
                /* Compute the length field in the ISO/IEC 8802-3 MAC */
                eth_p->h_proto = kal_htons(kal_skb_len(skb) - ETH_HLEN);
        }

        /* Copy the destination address and source address */
        ipn_memcpy( eth_p->h_dest,   daddr, MAC_ADDR_LEN );
        ipn_memcpy( eth_p->h_source, saddr, MAC_ADDR_LEN );

        KDBG_EXIT;
        return 0;
}

#if SUPPORT_MSSID
/*----------------------------------------------------------------*/
/*!
* \brief Append VLAN TAG
*
       DEST            SRC        PRO            Payload
|----------|----------|--------|-----......----|
      6                      6           2
       DEST            SRC        PRO           TCI     O_PRO     Payload
|----------|----------|--------|------|-----|---- ......----|
      6                      6       2(0x8100)      2          2
*
*
* \param   skb          the packet
* \param   which_ssid
*
* \return
*/
/*----------------------------------------------------------------*/
void append_vlang_tag(kal_skb_t * skb, ssid_info_t* which_ssid)
{
        unsigned char* packet_p;

        if(!which_ssid->vlan_tag_enabled)
                return;

        /* Do not worry if the header has enough room, because
         * we have sizeof(Frame_t) - 14 byte room
         */
        kal_skb_push(skb,VLAN_HLEN);
        packet_p = (unsigned char*)kal_skb_data(skb);
        ipn_memcpy(packet_p,packet_p + VLAN_HLEN, MAC_ADDR_LEN * 2);

        //VLAN ethernet protocol 0x8100
        packet_p[MAC_ADDR_LEN * 2] = 0x81;
        packet_p[MAC_ADDR_LEN * 2 + 1] = 0x00;

        //TCI, what about priority?
        packet_p[MAC_ADDR_LEN * 2 + 2] = (unsigned char)((which_ssid->vlanid & 0xFF00) >> 8);
        packet_p[MAC_ADDR_LEN * 2 + 3] = (unsigned char)(which_ssid->vlanid & 0x00FF);
}

/*----------------------------------------------------------------*/
/*!
* \brief Remove VLAN TAG and 0x8100.
*
*
* \param   skb          the packet
* \param   tid           return priority
* \param   vlan_id    return vlan_id
*
* \return  0 not vlan packet
*             1 vlan packet
*
*/
/*----------------------------------------------------------------*/
int remove_vlang_tag(kal_skb_t * skb,unsigned int* pri,unsigned int* vlan_id)
{
        unsigned char* eth_prot_p;
        unsigned char  tmp[12];

        eth_prot_p = (unsigned char*)kal_skb_data(skb);

        /*
        **  If the Ethernet protocol type is 0x8100, then this packet
        **  is a VLAN-tagged one.
        */
        if((eth_prot_p[MAC_ADDR_LEN * 2] == 0x81) &&
           (eth_prot_p[MAC_ADDR_LEN * 2 + 1] == 0x00)) {
                *pri = eth_prot_p[MAC_ADDR_LEN * 2 + 2] >> 5;
                *vlan_id = (((unsigned int)eth_prot_p[MAC_ADDR_LEN * 2 + 2]
                             & BITS(0,3)) << 8) +
                            ((unsigned int)eth_prot_p[MAC_ADDR_LEN * 2 + 3]);

                /*
                **         1.Copy Dest and Src MAC address
                **         2.Shorten Frame header by 4 bytes
                **         3.Copy Dest and Src MAC back to ethernet frame
                */
                ipn_memcpy(tmp, kal_skb_data(skb), MAC_ADDR_LEN *2);
                kal_skb_pull(skb, VLAN_HLEN);
                ipn_memcpy(kal_skb_data(skb), tmp, MAC_ADDR_LEN *2);
                return 1;
        }
        return 0;
}
#endif /* end of SUPPORT_MSSID */

