/*
** $Id: wds_os.c 1.28 2004/09/23 19:27:52 cmchang Exp $
*/
/*! \file wds_os.c
    \brief Handling WDS os dependent part

    This file concerns about WDS(Wireless Distribution System), which uses Wireless Medium
    to forward frames among APs. The basic idea is to utilize Linux's bridging modules. Every
    link between 2 APs is represented by a net device with a name like wlan#wds#.

    This file is os depent part. For indepent part, please refer to wds.c
*/

/******************************************************************************
Copyright (c) 2002-2003 Inprocomm, Inc.

All rights reserved. Copying, compilation, modification, distribution
or any other use whatsoever of this material is strictly prohibited
except in accordance with a Software License Agreement with Inprocomm, Inc.
******************************************************************************/
/************************************************************************
*                E X T E R N A L   R E F E R E N C E S
*************************************************************************
*/
/*================================================================*/
/* System Includes */
#include "wlan_os.h"
#include "wlan_kal.h"
#include "wlan_lib.h"
/************************************************************************
*              F U N C T I O N   D E C L A R A T I O N S
*************************************************************************
*/

int wds_open( struct net_device *netdev );
int wds_stop( struct net_device *netdev );
static int set_wdsdevice_name(struct net_device* dev, char* main_device_name);
static int wds_netdevice_structure_create(struct net_device** dev_p,
        struct net_device* main_device);
static int wds_unregister_device(struct net_device* dev);
static int wds_register_device(struct net_device* new_dev,
        struct net_device* main_dev);

extern int wlan_hard_start_xmit(struct sk_buff* skb, struct net_device *netdev);
extern int wlan_do_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd);
extern int wlan_init( struct net_device *netdev);
extern int wlan_set_mac_address(struct net_device *dev, void *addr);

/************************************************************************
*                          F U N C T I O N S
*************************************************************************
*/

/*----------------------------------------------------------------*/
/*!
* \brief interface: wlan0. Used to add a wds link based on its MAC Address.
*
*   \param remote_addr  The remote ap you wish to add.
*   \param name             The net_device's name.
*   \return  Return successful or not.
*/
/*----------------------------------------------------------------*/
int linux_del_wdslink(struct net_device*main_dev,unsigned char *remote_addr)
{
        struct net_device* remote_dev;
        int       rst;

        /* Only when device is valid and is main device, then we can
         * delete wds from it.
         */
        if(!main_dev ||!is_main_device(main_dev))
                return -ENOTTY;

        if(!is_wds_by_mac(main_dev, remote_addr))
                return -ENODEV;

        remote_dev = param_del_wdslink2(main_dev, remote_addr);
        if(!remote_dev)
                return -ENODEV;
        rst = wds_unregister_device(remote_dev);
        if(rst)
                return 0;
        else
                return -EFAULT;
}

/*----------------------------------------------------------------*/
/*!
* \brief interface: wlan0. Used to add a wds link based on its MAC Address.
*
*   \param remote_addr  The remote ap you wish to add.
*   \param name             The net_device's name.
*   \return  Return successful or not.
*/
/*----------------------------------------------------------------*/
int linux_add_wdslink(struct net_device*main_dev, unsigned char *remote_addr)
{
        struct net_device* new_dev;
        int rst;

        /*Only when device is valid and is main device ,then we can add wds to it.*/
        if(!main_dev ||!is_main_device(main_dev))
                return -ENOTTY;

        /*Null address or broadcast address*/
        if(!remote_addr || (remote_addr[0]&0x01))
                return -ENODEV;

        if(is_wds_by_mac(main_dev, remote_addr))
                return -ENODEV;

        rst = wds_netdevice_structure_create(&new_dev,main_dev);
        if(rst)
                return -ENOMEM;

        rst = param_add_wdslink2(main_dev, remote_addr, new_dev,
                0, 0, NULL);

        if(rst){
                kal_free(new_dev);
                return -ENOMEM;
        }

        if(wds_register_device(new_dev, main_dev))
                return 0;
        else{
                param_del_wdslink2(main_dev,remote_addr);
                kal_free(new_dev);
                return -EFAULT;
        }
}

/*----------------------------------------------------------------*/
/*!
* \brief  Utility function to give wds device a suitable name like wla0wds0
*
*   \param dev     The wds net_device structure.
*   \param main_device_name The main device's name.
* \return  The newly given name.
*/
/*----------------------------------------------------------------*/
static int set_wdsdevice_name(struct net_device* dev, char* main_device_name){
        int i = 0;
        do {
                sprintf(dev->name, "%swds%d", main_device_name, i++);
        } while (i < 10000 && dev_get(dev->name));

        return i;
};

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
*
* \return fail:0 success:1
*/
static int wds_unregister_device(struct net_device* dev)
{
        int ret;
        ret = unregister_netdevice(dev);

        if (ret) {
                printk(KERN_ALERT "unregistering WDS device '%s' failed\n",
                        dev->name);
                return 0;
        }
        printk(KERN_ALERT "unregistered WDS netdevice %s\n", dev->name);
        kal_free(dev);
        return 1;
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
*
* \return fail: 0 ,bigger means current wds order
*/
static int wds_register_device(struct net_device* new_dev,struct net_device* main_dev)
{
        int num_wds;
        int ret;
        num_wds = set_wdsdevice_name(new_dev,main_dev->name) + 1;
        ret = register_netdevice(new_dev);

        if (ret) {
                printk(KERN_ALERT "%s: registering WDS device '%s' failed\n",
                        main_dev->name, new_dev->name);
                return 0;
        }

        return num_wds;
}

/*----------------------------------------------------------------*/
/*!
* \brief  Creating the WDS netdevice data structure
*
* Called by wdslink_add to init net device structure.
* This roughly matches the functionality of ether_setup, prepares a set
* of networking call back functions.
*   \param dev     The malloced net_device structure.
*   \param priv_p The private management object.
* \return  success:0 fail:1.
*/
/*----------------------------------------------------------------*/
static int wds_netdevice_structure_create(struct net_device** dev_p,
        struct net_device* main_device)
{
        struct net_device* dev;

        /* verify that there is room for wds# postfix in the interface name */
        if (strlen(main_device->name) > IFNAMSIZ - 5) {
                printk(KERN_ALERT "'%s' too long base device name\n",
                main_device->name);
                return 1;
        }

        (*dev_p) = dev = kal_malloc(sizeof(struct net_device));

        if (dev == NULL) {
		printk("wds_netdevice_structure_create: malloc fail\n");
                return 1;
	}

        /* Setup netdevice's ability to report resources */
        memset(dev, 0, sizeof(struct net_device));

        /* Initialize the network device object. */
        ether_setup(dev);
        dev->hard_start_xmit =    &wlan_hard_start_xmit;
        dev->do_ioctl             =    &wlan_do_ioctl;
        dev->init                    =    &wlan_init;
        dev->open                 =    &wds_open;
        dev->stop                  =    &wds_stop;

        dev->priv = kal_dev_priv(main_device);
        memcpy(dev->dev_addr, main_device->dev_addr, ETH_ALEN);
        dev->base_addr = main_device->base_addr;
        dev->irq = main_device->irq;
        dev->mem_start = main_device->mem_start;
        dev->mem_end = main_device->mem_end;

        return 0;
}

/*----------------------------------------------------------------*/
/*!
* \brief  A method of struct net_device, to open the network interface
*         for normal operation.
* This open procedure is for WDS Device. Therefore it's a virture
* Opens the interface. The interface is opened whenever ifconfig activates it.
* \param   netdev   Linux netdevice.
* \return  zero on success, non-zero otherwise
*/
/*----------------------------------------------------------------*/
int wds_open( struct net_device *netdev )
{
        int             result = 0; /* success */
        MOD_INC_USE_COUNT;
        netif_start_queue(netdev);
        printk(KERN_ALERT "WDS net_device:%s opened\n",netdev->name);
        return result;
}

/*----------------------------------------------------------------*/
/*!
* \brief  A method of struct net_device, to close the network interface.
* This close procedure is for WDS Device. Therefore it's a virture
* Opens the interface. The interface is closed whenever ifconfig deactivates it.
* \param   netdev   Linux netdevice.
* \return  zero on success, non-zero otherwise
*/
/*----------------------------------------------------------------*/
int wds_stop( struct net_device *netdev )
{
        int             result = 0;
        MOD_DEC_USE_COUNT;
        netif_stop_queue(netdev);
        return result;
}

/*----------------------------------------------------------------*/
/*!
* \brief  A method of struct net_device, to close the network interface.
* We should be able to bring down if interface. But I don't how to do it yet.
* \param   netdev   Linux netdevice.
* \return  zero on success, non-zero otherwise
*/
void kal_indicate_wds_tx_down(void *netdev)
{
}

/*----------------------------------------------------------------*/
/*!
* \brief  A method of struct net_device, to close the network interface.
* \param   netdev   Linux netdevice.
* \return  zero on success, non-zero otherwise
*/
void kal_indicate_wds_tx_up(void*netdev)
{
}
